Package: shidashi
Type: Package
Title: A Shiny Dashboard Template Modular System with Chat Bot Support
Version: 0.2.0
Language: en-US
URL: https://dipterix.org/shidashi/,
        https://github.com/dipterix/shidashi
BugReports: https://github.com/dipterix/shidashi/issues
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Authors@R: c(
    person("Zhengjia", "Wang", email="dipterix.wang@gmail.com", 
            role = c("cph", "aut", "cre"), comment = c(ORCID = "0000-0001-5629-1116")),
    person(family = "ColorlibHQ", role = c("cph"), comment = "AdminLTE - Bootstrap 4 Admin Dashboard"),
    person(family = "Bootstrap contributors", role = "ctb", comment = "Bootstrap library"),
    person(family = "Twitter, Inc", role = "cph", comment = "Bootstrap library"),
    person("Ivan", "Sagalaev", role = c("ctb", "cph"), comment = "highlight.js library"),
    person("Rene", "Haas", role = c("ctb", "cph"), comment = "OverlayScrollbars library"),
    person("Zeno", "Rocha", role = c("ctb", "cph"), comment = "Clipboard.js library")
    )
Description: A template dashboard system with AI agent integrated. 
    Comes with default themes that can be customized. 
    Developers can upload modified templates on 'Github', and users can 
    easily download templates with 'RStudio' project wizard.
    The key features of the default template include light and dark theme 
    switcher, resizing graphs, synchronizing inputs across sessions, 
    new notification system, fancy progress bars, and card-like flip 
    panels with back sides, as well as various of 'HTML' tool widgets.
Imports: utils, tools, bslib, digest (>= 0.6.27), ellmer (>= 0.4.0),
        fastmap (>= 1.1.0), formatR (>= 1.11), httr2, S7, shiny (>=
        1.7.0), shinychat, yaml, jsonlite, htmlwidgets
Suggests: cli, coro, logger (>= 0.2.1), processx, promises, later,
        rstudioapi (>= 0.13), htmltools, httpuv, ggplot2, ggExtra,
        testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-10 16:20:02 UTC; dipterix
Author: Zhengjia Wang [cph, aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5629-1116>),
  ColorlibHQ [cph] (AdminLTE - Bootstrap 4 Admin Dashboard),
  Bootstrap contributors [ctb] (Bootstrap library),
  Twitter, Inc [cph] (Bootstrap library),
  Ivan Sagalaev [ctb, cph] (highlight.js library),
  Rene Haas [ctb, cph] (OverlayScrollbars library),
  Zeno Rocha [ctb, cph] (Clipboard.js library)
Maintainer: Zhengjia Wang <dipterix.wang@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-10 17:00:02 UTC
Built: R 4.5.2; ; 2026-04-15 05:10:15 UTC; unix

Package: shidashi
Type: Package
Title: A Shiny Dashboard Template Modular System with Chat Bot Support
Version: 0.2.0
Language: en-US
URL: https://dipterix.org/shidashi/,
        https://github.com/dipterix/shidashi
BugReports: https://github.com/dipterix/shidashi/issues
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Authors@R: c(
    person("Zhengjia", "Wang", email="dipterix.wang@gmail.com", 
            role = c("cph", "aut", "cre"), comment = c(ORCID = "0000-0001-5629-1116")),
    person(family = "ColorlibHQ", role = c("cph"), comment = "AdminLTE - Bootstrap 4 Admin Dashboard"),
    person(family = "Bootstrap contributors", role = "ctb", comment = "Bootstrap library"),
    person(family = "Twitter, Inc", role = "cph", comment = "Bootstrap library"),
    person("Ivan", "Sagalaev", role = c("ctb", "cph"), comment = "highlight.js library"),
    person("Rene", "Haas", role = c("ctb", "cph"), comment = "OverlayScrollbars library"),
    person("Zeno", "Rocha", role = c("ctb", "cph"), comment = "Clipboard.js library")
    )
Description: A template dashboard system with AI agent integrated. 
    Comes with default themes that can be customized. 
    Developers can upload modified templates on 'Github', and users can 
    easily download templates with 'RStudio' project wizard.
    The key features of the default template include light and dark theme 
    switcher, resizing graphs, synchronizing inputs across sessions, 
    new notification system, fancy progress bars, and card-like flip 
    panels with back sides, as well as various of 'HTML' tool widgets.
Imports: utils, tools, bslib, digest (>= 0.6.27), ellmer (>= 0.4.0),
        fastmap (>= 1.1.0), formatR (>= 1.11), httr2, S7, shiny (>=
        1.7.0), shinychat, yaml, jsonlite, htmlwidgets
Suggests: cli, coro, logger (>= 0.2.1), processx, promises, later,
        rstudioapi (>= 0.13), htmltools, httpuv, ggplot2, ggExtra,
        testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-10 16:20:02 UTC; dipterix
Author: Zhengjia Wang [cph, aut, cre] (ORCID:
    <https://orcid.org/0000-0001-5629-1116>),
  ColorlibHQ [cph] (AdminLTE - Bootstrap 4 Admin Dashboard),
  Bootstrap contributors [ctb] (Bootstrap library),
  Twitter, Inc [cph] (Bootstrap library),
  Ivan Sagalaev [ctb, cph] (highlight.js library),
  Rene Haas [ctb, cph] (OverlayScrollbars library),
  Zeno Rocha [ctb, cph] (Clipboard.js library)
Maintainer: Zhengjia Wang <dipterix.wang@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-10 17:00:02 UTC
Built: R 4.6.0; ; 2026-04-26 21:34:38 UTC; unix

# Generated by roxygen2: do not edit by hand

export(accordion)
export(accordion_item)
export(accordion_operate)
export(active_module)
export(add_class)
export(adminlte_sidebar)
export(adminlte_ui)
export(as_badge)
export(as_icon)
export(back_top_button)
export(bslib_dependency)
export(card)
export(card2)
export(card2_close)
export(card2_open)
export(card2_toggle)
export(card_badge)
export(card_operate)
export(card_recalculate_badge)
export(card_tabset)
export(card_tabset_activate)
export(card_tabset_insert)
export(card_tabset_remove)
export(card_tool)
export(clear_notifications)
export(clipboardOutput)
export(combine_html_class)
export(current_module)
export(disable_input_broadcast)
export(disable_input_sync)
export(disable_recalculate_badge)
export(drawer_close)
export(drawer_open)
export(drawer_toggle)
export(enable_input_broadcast)
export(enable_input_sync)
export(enable_recalculate_badge)
export(fire_event)
export(flex_break)
export(flex_container)
export(flex_item)
export(flip)
export(flip_box)
export(format_text_r)
export(get_construct_string)
export(get_event)
export(get_handler)
export(get_theme)
export(guess_body_class)
export(html_asis)
export(html_highlight_code)
export(include_view)
export(info_box)
export(init_app)
export(load_module)
export(mcp_wrapper)
export(module_drawer)
export(module_info)
export(open_url)
export(progressOutput)
export(read_stream_vis)
export(register_input)
export(register_output)
export(register_session)
export(remove_class)
export(remove_html_class)
export(render)
export(renderClipboard)
export(renderProgress)
export(renderStreamViz)
export(reset_output)
export(server_standalone_viewer)
export(set_card_badge)
export(set_handler)
export(shiny_progress)
export(show_notification)
export(show_ui_code)
export(skill_wrapper)
export(streamVizOutput)
export(stream_file_id)
export(stream_init)
export(stream_path)
export(stream_to_js)
export(stream_viz)
export(switch_module)
export(template_root)
export(template_settings)
export(template_settings_get)
export(template_settings_set)
export(unregister_session)
export(updateStreamViz)
export(use_template)
importFrom(digest,digest)
importFrom(ellmer,tool)
importFrom(fastmap,fastmap)
importFrom(formatR,tidy_source)
importFrom(httr2,url_modify)
importFrom(jsonlite,fromJSON)
importFrom(yaml,read_yaml)

# A Shiny Template System

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/shidashi)](https://CRAN.R-project.org/package=shidashi)
[![R-CMD-check](https://github.com/dipterix/shidashi/workflows/R-CMD-check/badge.svg)](https://github.com/dipterix/shidashi/actions)
<!-- badges: end -->

The goal of shidashi is to provide framework for R-shiny templates, especially for dashboard applications.

<div>
<img src="https://raw.githubusercontent.com/dipterix/shidashi/main/inst/screenshots/theme-light.png" width="49%">
<img src="https://raw.githubusercontent.com/dipterix/shidashi/main/inst/screenshots/theme-dark.png" width="49%">
</div>
<small>
*Default template (using [AdminLTE](https://adminlte.io/)) provides two themes: light vs dark*
</small>

## Installation

You can install the released version of shidashi from [CRAN](https://CRAN.R-project.org) with:

```r
install.packages("shidashi")
```

## Demo & Tutorial Application

The demo app requires to install the following extra packages

```r
install.packages(c("ggExtra", "rmarkdown"))
```

Once you have installed these packages, run the following command from R:

```r
library(shidashi)
project <- file.path(tools::R_user_dir('shidashi', which = "data"), "AdminLTE3")

# `use_template` only needs to be called once
use_template(project)
render(project)
```

## Start From Existing Templates

To start a `shidashi` project, open `RStudio` menu from the navigation bar:

> File > New Project... > New Directory > Shidashi Shiny Template (*)

_*You might need to scroll down to find that template option_.

Please enter the project information accordingly. By default, the `Github user` is `dipterix`, and theme is `AdminLTE3`, which lead to [the default template](https://github.com/dipterix/shidashi-templates/tree/master/AdminLTE3).

##### Bare-bone Template

If you want to start from a bare-bone template, change the `theme` option to be `AdminLTE3-bare`.

### File Structure

A typical `shidashi` project has the following file structure:

```
<project root_path>
├─modules/
│ └─<module ID>           - Module folder; folder name is module ID
│   ├─R                   - Module functions shared across UI and server
│   ├─module-ui.html      - Module HTML template
│   └─server.R            - Module-level server function
├─R/                      - Common functions shared across modules
├─views/                  - Small snippets (see `?include_view` function)
├─www/                    - Static files: css, js, img, ...
├─index.html              - Template for homepage
├─modules.yaml            - Module label, order, icon, badge..
└─server.R                - Root server function, usually no modification is required
```



## Contribute

Create your own `Github` repository with name `shidashi-templates`. Add folders named by the themes. Then people can install your themes as templates through `RStudio`.
An easy start is to fork [this repository](https://github.com/dipterix/shidashi-templates/).

library(shiny)
library(shidashi)

ui_notification <- function() {
  column(
    width = 6L,
    actionButton(ns("notif_1"), "Default notification", width = "auto"),
    actionButton(ns("notif_2"), "autohide = FALSE", width = "auto"),
    actionButton(ns("notif_3"), "With title, subtitle, and icon", width = "auto"),
    actionButton(ns("notif_4"), "With different types", width = "auto"),
    actionButton(ns("notif_5"), "Notification that does not close", width = "auto"),
    actionButton(ns("notif_6"), "With shiny components", width = "auto"),
    actionButton(ns("notif_7"), "With progress bar", width = "auto"),
    actionButton(ns("notif_8"), "Clear notifications", width = "auto")
  )
}

server_notification <- function(input, output, session, ...) {
  observeEvent(input$notif_1, {
    show_notification("This is a default notification. It automatically hides itself after 5 seconds, or you can close it via `x` button.")
  })
  observeEvent(input$notif_2, {
    show_notification(
      message = "This notification does not automatically hide itself",
      autohide = FALSE
    )
  })
  observeEvent(input$notif_3, {
    show_notification(
      message = "This notification has title and subtitle",
      autohide = FALSE,
      title = "Hi there!",
      subtitle = "Welcome!",
      icon = "kiwi-bird"
    )
  })
  observeEvent(input$notif_4, {
    show_notification(
      message = "This validation process has finished. You are welcome to proceed.",
      autohide = FALSE,
      title = "Success!",
      subtitle = "type='success'",
      type = "success"
    )
    show_notification(
      message = "Here are some information.",
      autohide = FALSE,
      title = "Information",
      subtitle = "type='info'",
      type = "info"
    )
    show_notification(
      message = "Here are some information.",
      autohide = FALSE,
      title = "Attention!",
      subtitle = "type='warning'",
      type = "warning"
    )
    show_notification(
      message = "Here are some information.",
      autohide = FALSE,
      title = "Error!",
      subtitle = "type='danger'",
      type = "danger"
    )
  })
  observeEvent(input$notif_5, {
    show_notification(
      message = "This notification cannot be closed. It does not automatically disappear. You have to call `clear_notifications` in R to remove it.",
      close = FALSE,
      autohide = FALSE,
      title = "Running...",
      type = "info"
    )
  })
  output$notif_6_out <- renderPlot({
    plot(1:10, col = input$notif_6_inp)
  })
  observeEvent(input$notif_6, {
    show_notification(
      message = div(
        selectInput(ns("notif_6_inp"), "Choose a color?",
                    choices = c("red", "blue", "green", "yellow")),
        plotOutput(ns("notif_6_out"), height = "300px")
      ),
      autohide = FALSE,
      title = "Shiny inputs & outputs",
    )
  })
  observeEvent(input$notif_7, {
    show_notification(
      message = div(
        "The wavelet might take a while to run. Grub a cup of coffee and wait.",
        progressOutput(ns("notif_7_prg"))
      ),
      autohide = FALSE,
      title = "Brewing",
      icon = "coffee",
      close = FALSE,
      class = "notif_7_autoclose"
    )
    on.exit({ clear_notifications(class = "notif_7_autoclose") })
    progress <- shiny_progress(title = "Running", max = 10, outputId = "notif_7_prg")
    for (i in 1:10) {
      Sys.sleep(0.5)
      progress$inc(detail = paste("Channel", i))
    }
  })
  observeEvent(input$notif_8, {
    clear_notifications()
  })
}

verbatim <- function(code) {
  tags$code(
    class = "clipboard-btn shidashi-clipboard-output",
    "data-clipboard-text" = code,
    role = "button", title = "Click to copy!",
    code)
}

ui_card_controls <- function() {

  tagList(
    column(
      width = 4L,
      card(
        inputId = ns("card_control_3"),
        title = "Card Expand",
        class_foot = "display-block-force",
        body_main = div(
          p(
            class = "inline-all",
            "Use ",
            verbatim("card_operate(...)"),
            " to collapse, expand, maximize, or minimize",
            tags$code("card"), ", ",
            tags$code("card2"), ", ",
            tags$code("card_tabset"), ". Even without card tools."
          )
        ),
        footer = tagList(
          fluidRow(
            column(width = 6L, actionButton(ns("card_control_3_collapse"), "Collapse")),
            column(width = 6L, actionButton(ns("card_control_3_expand"), "Expand"))),
          div(class = "space-vertical-5"),
          fluidRow(
            column(width = 6L, actionButton(ns("card_control_3_maximize"), "Maximize")),
            column(width = 6L, actionButton(ns("card_control_3_minimize"), "Minimize"))
          )
        )
      )
    ),
    column(
      width = 4L,
      card_tabset(
        inputId = ns("card_control_1"),
        title = "Cardset",
        active = "Tab B",
        class_body = "padding-20 min-height-100",
        "Tab A" = div(
          p(
            class = "inline-all",
            "To activate a tab, use ",
            verbatim("card_tabset_activate(...)"), ". ",
            "For example, ",
            verbatim('card_tabset_activate(inputId = "card_control_1", title = "Tab A")')
          )
        ),
        "Tab B" = div(
          p(
            class = "inline-all",
            "By specifying ",
            span(tags$code('active="Tab B"')),
            ", this tab is activated by default."
          )
        ),
        footer = fluidRow(
          column(
            width = 6L,
            actionButton(ns("switch_tab_a"), "Switch to Tab A")
          ),
          column(
            width = 6L,
            actionButton(ns("add_tab_a"), "New Tab")
          )
        )
      )
    ),
    column(
      width = 4L,
      card2(
        inputId = ns("card_control_2"),
        title = "Card2",
        class_body = "min-height-100",
        body_main = div(
          class = "padding-20",
          p(
            class = "inline-all",
            "use ",
            verbatim("card2_open(...)"), ", ",
            verbatim("card2_close(...)"), ", ",
            verbatim("card2_toggle(...)"),
            " to control the B-side from within R."
          )
        ),
        body_side = p("Card - side B"),
        footer = fluidRow(
          column(width = 4L, actionButton(ns("card_control_2_open"), "Open B-side")),
          column(width = 4L, actionButton(ns("card_control_2_close"), "Close B-side")),
          column(width = 4L, actionButton(ns("card_control_2_toggle"), "Toggle B-side"))
        )
      )
    )
  )

}

# UI: a single full-size output container.
# The output type (plot vs htmlwidget) is determined at runtime
# by server_standalone_viewer() based on the original render function.
viewer_output <- function() {
  shiny::uiOutput(ns("viewer_content"), class = "fill-width", style = "width:100%;height:100%;")
}

library(shiny)
library(shidashi)

ui_render <- function() {
  column(
    width = 12L,
    h2("Render your first project", class = "shidashi-anchor"),
    p(
      span(
        class = "inline-all",
        "Once you start a new shidashi project, a ",
        tags$code("start.R"), " file will be created with three lines:"
      )
    ),
    fluidRow(
      column(
        6L,
        tags$pre(
          class = "no-padding bg-gray-90 pre-compact",
          tags$code(
            class = "r",
            'library(shidashi)

# Set root path to your project folder
shidashi::template_settings$set(root_path = "<your project folder>")

# Render project
shidashi::render(host = "127.0.0.1", port = 8310L)'
          )
        )
      )
    ),
    p(
      span(
        class = "inline-all",
        "The first line loads the shidashi package into your R session. ",
        "The second line sets the ",
        tags$code("root_path"), ", which should be your project path, ",
        "so that shidashi knows where to look at when rendering websites. ",
        "The last line starts a local R-shiny application from port ",
        tags$code("8310"), "."
      )
    )
  )
}

server_demo <- function(input, output, session, ...) {
}

library(shiny)
library(shidashi)

server_stream_viz <- function(input, output, session, ...) {

  # Set up the stream directory once for this session
  shidashi::stream_init(session)

  # ---- register_output for the stream-viz widget ----------------------------
  shidashi::register_output(
    shidashi::renderStreamViz({
      shidashi::stream_viz("viz_signal", session = session, streaming = TRUE)
    }),
    outputId = "viz_signal",
    description = "Multi-channel signal viewer (stream-viz Three.js widget)",
    download_type = "stream_viz",
    session = session
  )

  # ---- Simulation state -----------------------------------------------------
  simulating <- reactiveVal(FALSE)
  sim_time <- reactiveVal(0)
  init <- FALSE

  observeEvent(input$btn_simulate, {
    currently <- simulating()
    simulating(!currently)
    if (!currently) {
      sim_time(0)
      shiny::updateActionButton(session, "btn_simulate",
                                label = "Stop Simulation",
                                icon = shidashi::as_icon("stop"))
    } else {
      shiny::updateActionButton(session, "btn_simulate",
                                label = "Simulate",
                                icon = shidashi::as_icon("play"))
    }
  })

  # ---- Periodic data generation (50 ms flush rate) --------------------------
  observe({
    if (!simulating()) {
      init <<- FALSE
      return()
    }
    shiny::invalidateLater(100, session)

    n_ch <- isolate(as.integer(input$n_channels))
    sr   <- isolate(as.integer(input$sample_rate))
    buf_sec <- isolate(as.numeric(input$n_seconds))
    n_t  <- as.integer(sr * buf_sec)

    # Advance simulation clock (isolate to avoid self-dependency loop)
    t0 <- isolate(sim_time())
    dt <- 50 / 1000  # 50 ms flush interval
    sim_time(t0 + dt)

    # Generate multi-channel sine waves with noise
    t_seq <- seq(t0, by = 1 / sr, length.out = n_t)
    mat <- matrix(0.0, nrow = n_ch, ncol = n_t)
    for (ch in seq_len(n_ch)) {
      freq      <- 2 + ch * 1.5
      amplitude <- 1 / ch
      mat[ch, ] <- amplitude * sin(2 * pi * freq * t_seq) +
                   rnorm(n_t, sd = 0.1 * amplitude)
    }

    ch_names <- paste0("Ch ", seq_len(n_ch))

    # Write binary stream file
    abspath <- shidashi::stream_path("viz_signal", session)
    shidashi::stream_to_js(
      abspath,
      data = as.numeric(mat),
      type = "float32",
      n_channels   = n_ch,
      n_timepoints = n_t,
      sample_rate  = sr,
      channel_names = ch_names,
      time_start = t0,
      time_end = t0 + buf_sec,
      x_unit = "s"
    )

    # Push update to the browser widget in-place
    if (!init) {
      init <<- TRUE
      shidashi::updateStreamViz("viz_signal", streaming = TRUE)
    }
  })
}

library(shiny)
library(shidashi)

#' During the run-time, the script in the following folders
#' will be loaded:
#' * R/
#' * modules/<ID>/R
#' You can run the following script to debug
if (FALSE) {
  source("inst/template/R/interface.R")
  source("inst/template/modules/test/R/aaa.R")
  session <- shiny::MockShinySession$new()
}

#' Defines the module server
server <- function(input, output, session, ...) {

  output$infobox_progress <- renderProgress({
    val <- input$infobox_make_progress %% 5
    if (val == 2) {
      stop("Click again")
    }
    list(
      value = val / 4 * 100,
      description = sprintf("Progress %d of 4", val)
    )
  })

  observeEvent(input$card_tool_modal, {
    shiny::showModal(shiny::modalDialog(
      "Click to dismiss",
      title = "Alert", easyClose = TRUE
    ))
  })

  output$card_defaulttool_plot <- renderPlot({
    plot(rnorm(20))
  })

  output$card_tool_plot <- renderPlot({
    input$card_tool_rerun
    hist(rnorm(1000))
  })

  output$card2_plot <- renderPlot({
    npoints <- input$card2_plot_npts
    title <- input$card2_plot_title
    if (!length(title) || title == "") {
      title <- "Normal Q-Q Plot"
    }
    qqnorm(rnorm(npoints), main = title)
    abline(a = 0, b = 1, col = "orange3", lty = 2)
  })

  observeEvent(input$add_card, {
    if (input$add_card %% 2) {
      card_tabset_insert(
        inputId = "card_tabset_expand_demo",
        title = "More...", active = TRUE,
        h4("A hidden playground!"),
        hr(),
        p("You can use `card_tabset_insert` to ",
          "insert cards to the cardset. ",
          "However, if you try to insert a card ",
          "whose title has already existed, ",
          "a notification will pop up to warn you."),
        p("Now, try to click on the kiwi-bird (",
          as_icon("kiwi-bird"), ") again.")
      )
    } else {
      card_tabset_remove(
        inputId = "card_tabset_expand_demo",
        title = "More..."
      )
    }
  })

  observeEvent(input$switch_tab_a, {
    card_tabset_activate(
      inputId = "card_control_1",
      title = "Tab A"
    )
  })
  observeEvent(input$add_tab_a, {
    card_tabset_insert(
      inputId = "card_control_1",
      title = sample(LETTERS, 1),
      tags$code("card_tabset_insert(...)")
    )
  })

  observeEvent(input$card_control_2_open, {
    card2_open("card_control_2")
  })
  observeEvent(input$card_control_2_close, {
    card2_close("card_control_2")
  })
  observeEvent(input$card_control_2_toggle, {
    card2_toggle("card_control_2")
  })

  observeEvent(input$card_control_3_collapse, {
    card_operate("card_control_3", method = "collapse")
  })
  observeEvent(input$card_control_3_expand, {
    card_operate("card_control_3", method = "expand")
  })
  observeEvent(input$card_control_3_maximize, {
    card_operate("card_control_3", method = "maximize")
  })
  observeEvent(input$card_control_3_minimize, {
    card_operate("card_control_3", method = "minimize")
  })

  # --- card_badge demo ---

  observeEvent(input$badge_set_refresh, {
    set_card_badge(
      id = "badge_status_demo",
      text = "Refresh needed",
      add_class = "bg-yellow",
      remove_class = "bg-success",
      session = session
    )
  })

  observeEvent(input$badge_set_ready, {
    set_card_badge(
      id = "badge_status_demo",
      text = "Up-to-date",
      add_class = "bg-success",
      remove_class = "bg-yellow",
      session = session
    )
  })

  observeEvent(input$badge_enable_recalculate, {
    enable_recalculate_badge(session = session)
  })

  observeEvent(input$badge_disable_recalculate, {
    disable_recalculate_badge(session = session)
  })

  observeEvent(get_event("run_analysis"), {
    show_notification("Run-analysis flag triggered")
    disable_recalculate_badge(session = session)
  })

}

library(shiny)
library(shidashi)

# ---------------------------------------------------------------------------
# UI helpers
# ---------------------------------------------------------------------------

ui_new_features <- function() {
  tagList(
    # --- Module info ---
    column(
      width = 12L,
      h4("Current Module Info", class = "shidashi-anchor"),
      p(
        class = "inline-all",
        tags$code("current_module()"), " returns a named list ",
        "with ", tags$code("id"), ", ", tags$code("label"), ", ",
        tags$code("group"), ", ", tags$code("icon"), ", ",
        tags$code("badge"), ", and ", tags$code("url"),
        " for the module that is currently running."
      ),
      verbatimTextOutput(ns("current_module_info"))
    ),

    # --- Drawer ---
    column(
      width = 6L,
      card(
        title = "Drawer",
        class_body = "min-height-200",
        body_main = div(
          p(
            class = "inline-all",
            "Open a right-side panel (drawer) from R. ",
            "You can also click the ",
            as_icon("cog"),
            " icon in the navbar. ",
            "Works from within modules (cross-iframe)."
          ),
          html_highlight_code(
            {
              drawer_open()
              drawer_close()
              drawer_toggle()
            },
            width.cutoff = 20L, hover = "overflow-auto"
          )
        ),
        footer = fluidRow(
          column(4L, actionButton(ns("drawer_open"), "Open",
                                  class = "btn-primary btn-sm w-100 mb-2")),
          column(4L, actionButton(ns("drawer_close"), "Close",
                                  class = "btn-danger btn-sm w-100 mb-2")),
          column(4L, actionButton(ns("drawer_toggle"), "Toggle",
                                  class = "btn-secondary btn-sm w-100 mb-2"))
        )
      )
    ),

    # --- open_url ---
    column(
      width = 6L,
      card(
        title = "Open URL",
        class_body = "min-height-200",
        body_main = div(
          p(
            class = "inline-all",
            "Open a URL in the browser from R. ",
            "Accepts any valid URL."
          ),
          html_highlight_code(
            open_url("https://github.com/dipterix/shidashi"),
            width.cutoff = 25L, hover = "overflow-auto"
          )
        ),
        footer = fluidRow(
          column(6L, actionButton(ns("open_url_github"), "GitHub Page",
                                  class = "btn-info btn-sm w-100 mb-2")),
          column(6L, actionButton(ns("open_url_cran"), "CRAN Link",
                                  class = "btn-outline-secondary btn-sm w-100 mb-2"))
        )
      )
    ),

    # --- Resizable card demo ---
    column(
      width = 6L,
      card(
        title = "Resize Handles",
        class_body = "height-300 resize-vertical",
        tools = list(
          card_tool(widget = "maximize")
        ),
        body_main = div(
          p(
            class = "inline-all",
            "Add ", tags$code('class_body = "resize-vertical"'),
            " to make the card body vertically resizable. ",
            "Drag the bottom handle to resize."
          ),
          html_highlight_code(
            card(
              title = "Resizable Card",
              class_body = "height-300 resize-vertical",
              plotOutput(ns("plot"), height = "100%")
            ),
            width.cutoff = 20L, hover = "overflow-auto"
          ),
          plotOutput(ns("resize_demo_plot"), height = "120px")
        )
      )
    ),

    # --- Horizontal resize panel
    column(
      width = 6L,
      card(
        title = "Horizontal Resize (split pane)",
        class_body = "height-300",
        tools = list(
          card_tool(widget = "maximize")
        ),
        body_main = div(
          style = "display:flex;height:100%;",
          div(
            style = "flex:1; padding: 0.5rem; overflow:auto;",
            p("Left pane"),
            p("Drag the divider to resize.")
          ),
          div(class = "resize-horizontal"),
          div(
            style = "flex:1; padding: 0.5rem; overflow:auto;",
            p("Right pane"),
            p("Add class ",
              tags$code("resize-horizontal"),
              " to a divider element."),
            plotOutput(ns("resize_demo_plot2"), height = "120px",
                       width = "100%")
          )
        )
      )
    )
  )
}

# ---------------------------------------------------------------------------
# Server
# ---------------------------------------------------------------------------

server_new_features <- function(input, output, session, ...) {

  # Current module info
  output$current_module_info <- renderPrint({
    info <- current_module()
    utils::str(info)
  })

  # Drawer controls
  observeEvent(input$drawer_open, {
    drawer_open()
  })
  observeEvent(input$drawer_close, {
    drawer_close()
  })
  observeEvent(input$drawer_toggle, {
    drawer_toggle()
  })

  # open_url controls
  observeEvent(input$open_url_github, {
    open_url("https://github.com/dipterix/shidashi")
  })
  observeEvent(input$open_url_cran, {
    open_url("https://cran.r-project.org/package=shidashi")
  })

  # Demo plot for resizable card
  output$resize_demo_plot <- renderPlot({
    par(mar = c(2, 2, 1, 1))
    plot(cars, pch = 19, col = "steelblue", main = "Resize me!")
    abline(lm(dist ~ speed, data = cars), col = "tomato", lwd = 2)
  })

  output$resize_demo_plot2 <- renderPlot({
    par(mar = c(2, 2, 1, 1))
    plot(cars, pch = 19, col = "steelblue", main = "Resize me!")
    abline(lm(dist ~ speed, data = cars), col = "tomato", lwd = 2)
  })
}

library(shiny)
page_title <- function(complete = TRUE) {
  if (complete) {
    "Shiny Dashboard Template - bslib"
  } else {
    "ShiDashi"
  }
}

page_logo <- function(size = c("normal", "small", "large")) {
  "shidashi/img/icon.png"
}
page_loader <- function() {
  NULL
}

body_class <- function() {
  c(
    #--- Start as dark-mode ---
    "dark-mode",

    # drawer has no no-overlay
    "shidashi-drawer-no-overlay"
  )
}

nav_class <- function() {
  c(
    "shidashi-header",
    "navbar",
    "navbar-expand"
  )
}

sidebar_class <- function() {
  c(
    #--- Start as dark-mode ---
    "dark-mode"
  )
}

module_breadcrumb <- function() {}

ui_card_controls <- function() {
  tagList(

    # --- 1. Server-side tabset control ---
    shiny::column(
      width = 4L,
      card(
        title = "Tabset Control",
        class_body = "no-padding",
        shiny::div(
          class = "p-3",
          p(
            "Use ", tags$code("card_tabset_activate()"), " and ",
            tags$code("card_tabset_insert()"), " to control a cardset ",
            "from the server."
          ),
          shiny::fluidRow(
            shiny::column(6L,
              actionButton(ns("switch_tab_a"), "Activate Tab A",
                           class = "btn-primary btn-sm w-100 mb-2")
            ),
            shiny::column(6L,
              actionButton(ns("add_tab_a"), "Insert Random Tab",
                           class = "btn-secondary btn-sm w-100 mb-2")
            )
          ),
          html_highlight_code(
            {
              # Activate a specific tab
              card_tabset_activate("card_control_1", title = "Tab A")

              # Insert / remove a tab dynamically
              card_tabset_insert("card_control_1", title = "Tab X",
                                 p("Dynamic content"))
            },
            width.cutoff = 20L, hover = "overflow-auto"
          )
        )
      ),
      card_tabset(
        inputId = ns("card_control_1"),
        title = "Controlled Cardset",
        class_body = "min-height-150",
        "Tab A" = p("Content of Tab A"),
        "Tab B" = p("Content of Tab B")
      )
    ),

    # --- 2. Server-side card2 control ---
    shiny::column(
      width = 4L,
      card(
        title = "Card2 (B-side) Control",
        class_body = "no-padding",
        shiny::div(
          class = "p-3",
          p(
            "Use ", tags$code("card2_open()"), ", ",
            tags$code("card2_close()"), ", or ",
            tags$code("card2_toggle()"), " to open or close the B-side panel."
          ),
          shiny::fluidRow(
            shiny::column(4L,
              actionButton(ns("card_control_2_open"), "Open",
                           class = "btn-success btn-sm w-100 mb-2")
            ),
            shiny::column(4L,
              actionButton(ns("card_control_2_close"), "Close",
                           class = "btn-danger btn-sm w-100 mb-2")
            ),
            shiny::column(4L,
              actionButton(ns("card_control_2_toggle"), "Toggle",
                           class = "btn-secondary btn-sm w-100 mb-2")
            )
          ),
          html_highlight_code(
            card2_open("card_control_2"),
            width.cutoff = 20L, hover = "overflow-auto"
          )
        )
      ),
      card2(
        inputId = ns("card_control_2"),
        title = "Card with B-side",
        class_body = "min-height-200",
        body_main = tagList(
          p("Main side."),
          p(
            "Click the ", as_icon("ellipsis-v"),
            " button or use the controls above to open/close the B-side."
          )
        ),
        body_side = tagList(
          h5("Configuration Panel"),
          p(tags$code("card2_open()"), " opens this side."),
          p(tags$code("card2_close()"), " closes this side."),
          p(tags$code("card2_toggle()"), " toggles this side.")
        ),
        tools = list(
          card_tool(widget = "card2-switch")
        )
      )
    ),

    # --- 3. Plain card: no btn-tools, purely server-side controlled ---
    shiny::column(
      width = 4L,
      card(
        title = "Server-only Card Control",
        class_body = "no-padding",
        shiny::div(
          class = "p-3",
          p(
            "The card below has ", tags$strong("no tool buttons"),
            " in its header. All collapse / expand / maximize actions are",
            " triggered from R using ", tags$code("card_operate()"), "."
          ),
          shiny::fluidRow(
            shiny::column(6L,
              actionButton(ns("card_control_3_collapse"), "Collapse",
                           class = "btn-warning btn-sm w-100 mb-2")
            ),
            shiny::column(6L,
              actionButton(ns("card_control_3_expand"), "Expand",
                           class = "btn-success btn-sm w-100 mb-2")
            ),
            shiny::column(6L,
              actionButton(ns("card_control_3_maximize"), "Maximize",
                           class = "btn-primary btn-sm w-100")
            ),
            shiny::column(6L,
              actionButton(ns("card_control_3_minimize"), "Minimize",
                           class = "btn-secondary btn-sm w-100")
            )
          ),
          html_highlight_code(
            {
              # No tools needed on the card — server controls it directly
              card_operate("card_control_3", method = "collapse")
              card_operate("card_control_3", method = "expand")
              card_operate("card_control_3", method = "maximize")
              card_operate("card_control_3", method = "minimize")
            },
            width.cutoff = 20L, hover = "overflow-auto"
          )
        )
      ),
      card(
        title = "Controlled Card (No Tools)",
        # inputId enables card_operate() to find this card
        inputId = ns("card_control_3"),
        class_body = "min-height-200",
        # No `tools` argument — this is the key demonstration
        p("This card has no tool buttons."),
        p(
          "Collapse, expand, and maximize are driven entirely from the observer ",
          "in server.R via ", tags$code("card_operate()"), "."
        ),
        p(style = "color: var(--bs-secondary);",
          "Try the buttons in the control card above.")
      )
    )

  )
}

ui_accordion <- function() {

  shiny::fluidRow(
    shiny::column(
      width = 6L,
      shidashi::accordion(
        id = ns("this-accordion"),
        shidashi::accordion_item(
          title = "Step 1",
          collapsed = FALSE,
          footer = shiny::actionLink(
            inputId = ns("accordion_next"),
            label = "Next step"
          ),
          "An accordion is a list of collapsible cards. ",
          "If you expand one of them, the other cards will collapse"
        ),
        shidashi::accordion_item(
          itemId = ns("accordion-step-2"),
          title = list("Step 2", shidashi::as_badge("New|badge-danger")),
          shiny::textInput("input_2", "Input 2"),
          footer = shiny::actionButton(
            inputId = ns("accordion_close"),
            label = "OK, collapse this item"
          ),
          collapsed = TRUE
        )
      )
    ),
    shiny::column(
      width = 6L,
    )
  )

}

server_accordion <- function(input, output, session, ...) {
  shiny::bindEvent(
    shiny::observe({
      shidashi::accordion_operate(
        id = "this-accordion",
        itemId = "accordion-step-2",
        method = "expand"
      )
    }),
    input$accordion_next,
    ignoreInit = TRUE, ignoreNULL = TRUE
  )

  shiny::bindEvent(
    shiny::observe({
      shidashi::accordion_operate(
        id = "this-accordion",
        itemId = "accordion-step-2",
        method = "collapse"
      )
    }),
    input$accordion_close,
    ignoreInit = TRUE, ignoreNULL = TRUE
  )
}

# agents/tools/hello_world.R
#
# Root-level MCP tool: Returns a greeting.
# Used to verify the MCP tunnel works end-to-end.
#
# Enable per module in agents/agent.yaml:
#   tools:
#     root:
#       hello_world: true

hello_world <- shidashi::mcp_wrapper(
  function(session) {
    ellmer::tool(
      fun = function(name = "World") {
        paste0("Hello, ", name, "!")
      },
      name = "hello_world",
      description = "Returns a greeting. Used to verify the MCP tunnel works.",
      arguments = list(
        name = ellmer::type_string(
          "Name to greet (default: 'World')",
          required = FALSE
        )
      )
    )
  }
)

library(shiny)
library(shidashi)

server <- function(input, output, session, ...) {

  server_base_plot(input, output, session, ...)
  server_ggplot_brush(input, output, session, ...)
  server_threebrain(input, output, session, ...)

}

library(shiny)
library(shidashi)

# ===========================================================================
# QUICK START SECTION (What users see first)
# ===========================================================================

ui_quick_start <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Quick Start", class = "shidashi-anchor"),
      tags$div(
        class = "alert alert-info",
        tags$strong("30-Second Setup: "),
        "Click the ", as_icon("robot"), " button in the bottom-right → ",
        "Type a prompt → Done!"
      )
    ),
    column(
      width = 4L,
      card(
        title = "1. Open the Chatbot",
        tags$p(
          "Click ",
          tags$a(
            href = "#",
            class = "btn btn-sm btn-primary",
            `data-shidashi-action` = "drawer-toggle",
            as_icon("robot"), " Open"
          ),
          " or the ", as_icon("robot"), " button at bottom-right."
        )
      )
    ),
    column(
      width = 4L,
      card(
        title = "2. Choose a Mode",
        tags$ul(
          class = "mb-0",
          tags$li(tags$strong("Ask"), " — Q&A only"),
          tags$li(tags$strong("Plan"), " — Can read data"),
          tags$li(tags$strong("Executing"), " — Can change inputs")
        )
      )
    ),
    column(
      width = 4L,
      card(
        title = "3. Try These Prompts",
        tags$ul(
          class = "mb-0",
          tags$li(tags$em("\"What inputs are here?\"")),
          tags$li(tags$em("\"Change color to purple\"")),
          tags$li(tags$em("\"Set points to 300\""))
        )
      )
    )
  )
}


# ===========================================================================
# LIVE DEMO SECTION (Interactive playground)
# ===========================================================================

ui_live_demo <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Try It Now", class = "shidashi-anchor"),
      p("These controls are registered with the AI agent. Open the chatbot and ask it to change them!")
    ),
    column(
      width = 4L,
      card(
        title = "Controls",
        shidashi::register_input(
          inputId = "live_npoints",
          description = "Number of random points to plot (10-500)",
          update = "shiny::updateSliderInput",
          expr = sliderInput(
            inputId = ns("live_npoints"),
            label = "Number of points",
            min = 10, max = 500, value = 100, step = 10
          )
        ),
        shidashi::register_input(
          inputId = "live_color",
          description = "Color of the scatter plot points",
          update = "shiny::updateSelectInput(value=selected)",
          expr = selectInput(
            inputId = ns("live_color"),
            label = "Point color",
            choices = c("steelblue", "tomato", "forestgreen",
                        "orange", "purple"),
            selected = "steelblue"
          )
        ),
        shidashi::register_input(
          inputId = "live_title",
          description = "Title displayed above the scatter plot",
          update = "shiny::updateTextInput",
          expr = textInput(
            inputId = ns("live_title"),
            label = "Plot title",
            value = "Agent-Controlled Scatter Plot"
          )
        )
      )
    ),
    column(
      width = 8L,
      card(
        title = "Output",
        tools = list(card_tool(widget = "maximize")),
        class_body = "min-height-400",
        plotOutput(ns("live_scatter"), height = "350px")
      )
    )
  )
}


# ===========================================================================
# DEVELOPER GUIDE (Collapsible details for those who want to learn more)
# ===========================================================================

ui_developer_guide <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Developer Guide", class = "shidashi-anchor"),
      p("Expand the sections below to learn how to build agent-powered modules.")
    ),

    # --- Section: Register Inputs ---
    column(
      width = 12L,
      accordion(
        id = ns("dev_accordion"),
        accordion_item(
          title = "Register Inputs (shidashi::register_input)",
          tags$p(
            "Wrap any Shiny input with ", tags$code("register_input()"),
            " to make it visible to AI agents:"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "r",
'shidashi::register_input(
  inputId = "my_slider",
  description = "Adjusts the threshold value (0-100)",
  update = "shiny::updateSliderInput",
  expr = sliderInput(ns("my_slider"), "Threshold", 0, 100, 50)
)'
            )
          ),
          tags$p(
            tags$strong("Parameters:"),
            tags$ul(
              tags$li(tags$code("inputId"), " — Input name (without namespace)"),
              tags$li(tags$code("description"), " — What the agent sees"),
              tags$li(tags$code("update"), " — Function to change the value"),
              tags$li(tags$code("expr"), " — The actual Shiny input widget")
            )
          )
        ),

        # --- Section: Register Outputs ---
        accordion_item(
          title = "Register Outputs (shidashi::register_output)",
          tags$p(
            "Wrap any Shiny output with ", tags$code("register_output()"),
            " to help agents understand what's displayed:"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "r",
'shidashi::register_output(
  outputId = "my_plot",
  description = "Histogram showing data distribution",
  expr = plotOutput(ns("my_plot"))
)'
            )
          )
        ),

        # --- Section: agents.yaml ---
        accordion_item(
          title = "Enable Agent Features (agents.yaml)",
          tags$p(
            "Create ", tags$code("agents.yaml"), " in your module folder:"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "yaml",
'# modules/mymodule/agents.yaml
default_mode: "Executing"
system_prompt: |
  You help users analyze data in this module.

modes:
  Ask:
    tools: []
  Plan:
    tools: [shiny_input_info, shiny_output_info]
  Executing:
    tools: [shiny_input_info, shiny_input_update, shiny_output_info]'
            )
          ),
          tags$p(
            tags$strong("Modes control what the agent can do:"),
            tags$ul(
              tags$li(tags$strong("Ask"), " — No tools, just conversation"),
              tags$li(tags$strong("Plan"), " — Can read inputs/outputs"),
              tags$li(tags$strong("Executing"), " — Can also update inputs")
            )
          )
        ),

        # --- Section: Custom Tools ---
        accordion_item(
          title = "Custom MCP Tools (mcp_wrapper)",
          tags$p(
            "Create custom tools with ", tags$code("mcp_wrapper()"), ":"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "r",
'# In module R/ folder
my_tool <- shidashi::mcp_wrapper(function(session) {
  ellmer::tool(
    name = "calculate_stats",
    description = "Calculate statistics for the current data",
    arguments = list(
      column = ellmer::tool_arg("string", "Column name")
    ),
    fun = function(column) {
      data <- session$userData$current_data
      paste("Mean:", mean(data[[column]], na.rm = TRUE))
    }
  )
})'
            )
          ),
          tags$p(
            "Then add to ", tags$code("agents.yaml"), ": ",
            tags$code("tools: [calculate_stats]")
          )
        ),

        # --- Section: Skills ---
        accordion_item(
          title = "Skills (Script-Based Tools)",
          tags$p(
            "For complex operations, create a skill folder with ",
            tags$code("SKILL.md"), " and scripts:"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
"modules/mymodule/skills/analyze/
├── SKILL.md        # Defines tool name, args, descriptions
└── scripts/
    └── run.R       # The actual script"
            )
          ),
          tags$p("Example ", tags$code("SKILL.md"), ":"),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "yaml",
'---
name: analyze_data
description: Run statistical analysis
arguments:
  - name: method
    type: string
    description: "lm" or "glm"
script: scripts/run.R
---'
            )
          )
        )
      )
    )
  )
}


# ===========================================================================
# SERVER
# ===========================================================================

server_aiagent <- function(input, output, session, ...) {
  shidashi::register_session(session)

  shidashi::register_output(
    renderPlot({
      theme <- shidashi::get_theme()
      n <- input$live_npoints %||% 100
      col <- input$live_color %||% "steelblue"
      title <- input$live_title %||% "Scatter Plot"

      set.seed(42)
      x <- rnorm(n)
      y <- x + rnorm(n, sd = 0.5)

      par(
        bg = theme$background, fg = theme$foreground,
        col.main = theme$foreground,
        col.axis = theme$foreground,
        col.lab = theme$foreground
      )
      plot(x, y, pch = 19, col = col, main = title,
           xlab = "X", ylab = "Y")
    }),
    outputId = "live_scatter",
    description = "Scatter plot controlled by live_npoints, live_color, and live_title"
  )
}


# ===========================================================================
# CUSTOM TOOL (for demo)
# ===========================================================================

trigger_refresh <- shidashi::mcp_wrapper(
  function(session) {
    ellmer::tool(
      fun = function() {
        shidashi::fire_event(session$ns("refresh"), Sys.time(), session = session)
        "Refresh triggered."
      },
      name = "trigger_refresh",
      description = "Trigger a refresh action in the AI agent demo module.",
      arguments = list()
    )
  }
)

ui_info_box_basic <- function() {
  tagList(
    column(width = 3L,
           infobox_with_code(
             info_box("Message", icon = NULL)
           )),
    column(width = 3L,
           infobox_with_code(
             info_box(icon = NULL,
                      span(class = "info-box-text", "Likes"),
                      span(class = "info-box-number", "20,331"))
           ))
  )
}

ui_info_box_advanced <- function() {

  tagList(
    column(width = 3L,
           infobox_with_code(
             info_box(icon = "cogs",
                      span(class = "info-box-text", "Configurations"),
                      span(class = "info-box-number", "With icon")))
           ),
    column(width = 3L,
           infobox_with_code(
             info_box(icon = "thumbs-up",
                      class_icon = "bg-green",
                      span(class = "info-box-text", "Likes"),
                      span(class = "info-box-number", "Colored icon")))
           ),
    column(width = 3L,
           infobox_with_code(
             info_box(span(class = "info-box-text", "Calendars"),
                      span(class = "info-box-number", "4 items"),
                      icon = "calendar-alt",
                      class = "bg-yellow", class_icon = NULL)
           )
    ),
    column(width = 3L,
           infobox_with_code(
             info_box(span(class = "info-box-text", "Yes!"),
                      span(class = "info-box-number", "Colored differently"), icon = "star",
                      class = "bg-yellow")
           )
    ),
    column( width = 8L,
            infobox_with_code(
              title = "Info-box (progress bar)",
              info_box(
                span(class = "info-box-text", "Progress | ",
                     actionLink(ns("infobox_make_progress"), "Keep clicking me"),
                     " | ",
                     actionLink(ns("infobox_make_progress_alt"), "Alternative process")),
                progressOutput(ns("infobox_progress")),
                icon = "sync"
              )
            )
    ),
  )

}

server_info_box <- function(input, output, session, ...) {

  output$infobox_progress <- renderProgress({
    val <- input$infobox_make_progress %% 5
    if (val == 2) {
      stop("Click again")
    }
    list(
      value = val / 4 * 100,
      description = sprintf("Progress %d of 4", val)
    )
  })

  observeEvent(input$infobox_make_progress_alt, {
    progress <- shiny_progress(title = "Alternative Procedure", max = 10, outputId = "infobox_progress")
    for (i in 1:10) {
      progress$inc(sprintf("Step %s", i))
      Sys.sleep(1)
    }
  })

}

ui_card_badge <- function() {
  tagList(

    # --- 1. Static badge in card header ---
    shiny::column(
      width = 4L,
      card(
        title = "Static Badge",
        class_body = "no-padding",
        tools = list(
          card_badge("Ready", class = "bg-success shidashi-output-status")
        ),
        shiny::div(
          class = "p-3",
          p(
            tags$code("card_badge()"), " places a badge in the card header. ",
            "It is more flexible than ", tags$code("as_badge()"), " because ",
            "it can be updated from the server at runtime."
          ),
          html_highlight_code(
            card(
              title = "My Card",
              tools = list(
                card_badge("Ready",
                           class = "bg-success shidashi-output-status")
              )
            ),
            width.cutoff = 20L, hover = "overflow-auto"
          )
        )
      )
    ),

    # --- 2. Dynamic badge updated from the server ---
    shiny::column(
      width = 4L,
      card(
        title = "Dynamic Badge",
        class_body = "no-padding",
        tools = list(
          card_badge(
            text = "Up-to-date",
            class = "bg-success shidashi-output-status",
            id = ns("badge_status_demo")
          )
        ),
        shiny::div(
          class = "p-3",
          p(
            "Use ", tags$code("set_card_badge()"), " on the server to change ",
            "the badge text, add, or remove CSS classes dynamically."
          ),
          shiny::fluidRow(
            shiny::column(6L,
              actionButton(
                ns("badge_set_refresh"),
                "Need Refresh",
                class = "btn-warning btn-sm w-100 mb-2"
              )
            ),
            shiny::column(6L,
              actionButton(
                ns("badge_set_ready"),
                "Up-to-date",
                class = "btn-success btn-sm w-100 mb-2"
              )
            )
          ),
          html_highlight_code(
            {
              # Use id to target a single badge by namespace
              set_card_badge(
                id = ns("badge_status_demo"),
                text = "Refresh needed",
                add_class = "bg-yellow",
                remove_class = "bg-success"
              )

              # Use class selector to target all badges with that class
              set_card_badge(
                class = "shidashi-output-status",
                text = "Up-to-date",
                add_class = "bg-success",
                remove_class = "bg-yellow"
              )
            },
            width.cutoff = 20L, hover = "overflow-auto"
          )
        )
      )
    ),

    # --- 3. Recalculate badge convenience helpers ---
    shiny::column(
      width = 4L,
      card(
        title = "Recalculate Badge",
        class_body = "no-padding",
        tools = list(
          card_recalculate_badge()
        ),
        shiny::div(
          class = "p-3",
          p(
            tags$code("card_recalculate_badge()"), " is a pre-styled badge ",
            "that acts as a button to trigger re-run. ",
            "Use ", tags$code("enable_recalculate_badge()"), " / ",
            tags$code("disable_recalculate_badge()"),
            " to flip it on/off from the server."
          ),
          shiny::fluidRow(
            shiny::column(6L,
              actionButton(
                ns("badge_enable_recalculate"),
                "Enable",
                class = "btn-warning btn-sm w-100 mb-2"
              )
            ),
            shiny::column(6L,
              actionButton(
                ns("badge_disable_recalculate"),
                "Disable",
                class = "btn-success btn-sm w-100 mb-2"
              )
            )
          ),
          html_highlight_code(
            {
              # Mark output as stale
              enable_recalculate_badge()

              # Mark output as up-to-date
              disable_recalculate_badge()
            },
            width.cutoff = 20L, hover = "overflow-auto"
          )
        )
      )
    )
  )
}

# mcp_input_update_wrapper <- mcp_wrapper_input_output()
#
# mcp_input_update_generator <- mcp_input_update_wrapper$tool_generator
# register_input <- mcp_input_update_wrapper$input_helpers$register_input_specification


# # Returns two tools:
# #   - shiny_input_info
# #   - shiny_input_update
# unname(sapply(
#   X = mcp_input_update_generator(shiny::MockShinySession$new()),
#   FUN = function(tool) {
#     tool@name
#   }
# ))

options("crayon.enabled" = TRUE)
options("crayon.colors" = 256)

NULL


app <- shidashi:::register_mcp_route(shiny::shinyAppDir("/Users/dipterix/Dropbox (Personal)/projects/shidashi/inst/builtin-templates/bslib-bare"))
shiny::runApp(appDir = app, launch.browser = TRUE, test.mode = FALSE,
    port = 4515L)

library(shiny)
library(shidashi)

# Controls card — simulate button + parameter sliders
ui_stream_viz_controls <- function() {
  shidashi::card(
    title = "Simulate Signal",
    tools = list(
      shidashi::card_tool(widget = "collapse")
    ),
    fluidRow(
      column(
        width = 3L,
        sliderInput(
          inputId = ns("n_channels"),
          label = "Channels",
          min = 1L, max = 64L, value = 10L, step = 1L
        )
      ),
      column(
        width = 3L,
        sliderInput(
          inputId = ns("n_seconds"),
          label = "Duration (s)",
          min = 1L, max = 30L, value = 5L, step = 1L
        )
      ),
      column(
        width = 3L,
        sliderInput(
          inputId = ns("sample_rate"),
          label = "Sample rate (Hz)",
          min = 100L, max = 2000L, value = 500L, step = 100L
        )
      ),
      column(
        width = 3L,
        br(),
        actionButton(
          inputId = ns("btn_simulate"),
          label = "Simulate",
          icon = shidashi::as_icon("play"),
          class = "btn btn-primary w-100"
        )
      )
    )
  )
}

# Viewer card — hosts the streamVizOutput widget
ui_stream_viz_plot <- function() {
  shidashi::card(
    title = "Signal Viewer",
    resizable = TRUE,
    tools = list(
      shidashi::card_tool(widget = "collapse"),
      shidashi::card_tool(widget = "maximize")
    ),

    shidashi::streamVizOutput(ns("viz_signal"), height = "100%")
  )
}

library(shiny)
library(shidashi)

server <- function(input, output, session, ...) {

  server_notification(input, output, session, ...)
  server_progress(input, output, session, ...)
  server_info_box(input, output, session, ...)
  server_flip_box(input, output, session, ...)
  server_accordion(input, output, session, ...)
  server_new_features(input, output, session, ...)

}

#!/usr/bin/env Rscript
# greet.R — Simple greeting script for the demo skill
#
# Usage: Rscript greet.R [name]
# Output: Hello, <name>!

args <- commandArgs(trailingOnly = TRUE)
name <- if (length(args) >= 1L) args[[1L]] else "World"

cat(sprintf("Hello, %s!\n", name))

library(shiny)
library(shidashi)

# ===========================================================================
# FILE STRUCTURE OVERVIEW
# ===========================================================================

ui_app_structure <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Application Structure", class = "shidashi-anchor"),
      tags$p(
        "A shidashi application follows a consistent folder structure. ",
        "The key directories are shown below."
      )
    ),
    column(
      width = 6L,
      card(
        title = "Root Directory",
        tags$pre(
          class = "bg-gray-90 pre-compact",
          tags$code(
"my-shidashi-app/
├── global.R         # Generated by shidashi::render()
├── server.R         # Main server entry point
├── ui.R             # Main UI entry point
├── index.html       # Landing page template
├── modules.yaml     # Module definitions
├── modules/         # Module folder
│   └── <module_id>/
├── R/               # Shared R code
├── views/           # HTML templates
└── www/             # Static assets"
          )
        )
      )
    ),
    column(
      width = 6L,
      card(
        title = "Key Files",
        tags$dl(
          tags$dt(tags$code("modules.yaml")),
          tags$dd("Defines sidebar menu structure, module order, and icons."),

          tags$dt(tags$code("global.R")),
          tags$dd("Auto-generated. Initializes app globals and loads modules."),

          tags$dt(tags$code("R/common.R")),
          tags$dd("Shared functions available to all modules."),

          tags$dt(tags$code("index.html")),
          tags$dd("Landing page template (rendered by whisker).")
        )
      )
    )
  )
}


# ===========================================================================
# MODULE STRUCTURE
# ===========================================================================

ui_module_structure <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Module Structure", class = "shidashi-anchor"),
      tags$p(
        "Each module is a self-contained unit with its own UI, server, and assets."
      )
    ),
    column(
      width = 6L,
      card(
        title = "Module Directory",
        tags$pre(
          class = "bg-gray-90 pre-compact",
          tags$code(
"modules/<module_id>/
├── module-ui.html   # Module UI (whisker template)
├── server.R         # Module server entry point
├── R/               # Module-specific R code
│   └── my-ui.R      # UI functions
├── agents.yaml      # (Optional) AI agent config
└── www/             # (Optional) Module assets"
          )
        )
      )
    ),
    column(
      width = 6L,
      card(
        title = "File Purposes",
        tags$dl(
          tags$dt(tags$code("module-ui.html")),
          tags$dd("HTML template using whisker syntax ({{ ... }})."),

          tags$dt(tags$code("server.R")),
          tags$dd("Defines the module server function. Must export a ",
                  tags$code("server(input, output, session, ...)"), " function."),

          tags$dt(tags$code("R/")),
          tags$dd("Place UI functions and server helpers here. Auto-loaded."),

          tags$dt(tags$code("agents.yaml")),
          tags$dd("Configures AI agent modes and available tools.")
        )
      )
    )
  )
}


# ===========================================================================
# MODULE UI TEMPLATE
# ===========================================================================

ui_module_template <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Module UI Template", class = "shidashi-anchor"),
      tags$p(
        "The ", tags$code("module-ui.html"), " file is a whisker template. ",
        "R expressions inside ", tags$code("{{ }}"), " are evaluated at render time."
      )
    ),
    column(
      width = 12L,
      card(
        title = "Example module-ui.html",
        tags$pre(
          class = "bg-gray-90 pre-compact",
          tags$code(
            class = "html",
'<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <title>{{ module_title() }}</title>
  {{ shidashi::include_view("header.html") }}
</head>
<body class="{{ shidashi::guess_body_class(body_class()) }}">

<nav class="{{ paste(nav_class(), collapse = \' \') }}">
  <ul class="navbar-nav">
    <li class="nav-item">
      <a class="nav-link fw-semibold">{{ module_title() }}</a>
    </li>
  </ul>
</nav>

<div class="container-fluid py-3">
  {{ my_ui_function() }}
</div>

{{ shidashi::module_drawer() }}
{{ shidashi::back_top_button("rocket") }}
{{ shidashi::include_view("footer.html") }}
</body>
</html>'
          )
        )
      )
    )
  )
}


# ===========================================================================
# SERVER PATTERN
# ===========================================================================

ui_server_pattern <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Server Pattern", class = "shidashi-anchor"),
      tags$p(
        "The server defines reactive logic. Place server functions in ",
        tags$code("R/"), " and call them from ", tags$code("server.R"), "."
      )
    ),
    column(
      width = 6L,
      card(
        title = "server.R",
        tags$pre(
          class = "bg-gray-90 pre-compact",
          tags$code(
            class = "r",
"library(shiny)
library(shidashi)

server <- function(input, output, session, ...) {
  # Call server functions defined in R/
  server_mymodule(input, output, session, ...)
}"
          )
        )
      )
    ),
    column(
      width = 6L,
      card(
        title = "R/mymodule-server.R",
        tags$pre(
          class = "bg-gray-90 pre-compact",
          tags$code(
            class = "r",
"server_mymodule <- function(input, output, session, ...) {
  # Register session for event bus and theme changes
  shidashi::register_session(session)

  output$my_plot <- renderPlot({
    theme <- shidashi::get_theme()
    par(bg = theme$background, fg = theme$foreground)
    plot(1:10, main = input$title)
  })
}"
          )
        )
      )
    )
  )
}


# ===========================================================================
# MODULES.YAML
# ===========================================================================

ui_modules_yaml <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("modules.yaml Configuration", class = "shidashi-anchor"),
      tags$p(
        "The ", tags$code("modules.yaml"), " file controls the sidebar menu."
      )
    ),
    column(
      width = 6L,
      card(
        title = "Example Configuration",
        tags$pre(
          class = "bg-gray-90 pre-compact",
          tags$code(
            class = "yaml",
'# Dividers (horizontal lines in sidebar)
divider:
  Components:
    order: 20
  Settings:
    order: 50

# Collapsible groups
groups:
  Visualization:
    icon: "chart-bar"
    order: 25
    open: true

# Module entries
modules:
  dashboard:
    order: 1
    icon: "home"
    label: "Dashboard"
  plots:
    order: 26
    icon: "chart-line"
    label: "Plots"
    group: "Visualization"'
          )
        )
      )
    ),
    column(
      width = 6L,
      card(
        title = "Module Properties",
        tags$dl(
          tags$dt(tags$code("order")),
          tags$dd("Numeric order in sidebar. Lower = higher."),

          tags$dt(tags$code("icon")),
          tags$dd("FontAwesome icon name (without 'fa-' prefix)."),

          tags$dt(tags$code("label")),
          tags$dd("Display text in sidebar."),

          tags$dt(tags$code("group")),
          tags$dd("(Optional) Group name for collapsible section."),

          tags$dt(tags$code("disabled")),
          tags$dd("(Optional) Set to ", tags$code("true"), " to hide module.")
        )
      )
    )
  )
}


# ===========================================================================
# SERVER (empty for documentation module)
# ===========================================================================

server_filestructure <- function(input, output, session, ...) {
  # No server logic needed for this documentation module
}

# ---- shidashi global.R ----
# This file is generated by shidashi::render() and sourced once
# when the Shiny application starts.  It initializes per-application
# state (MCP session registry, module input registries, etc.) that
# is shared across all sessions.
shidashi::init_app()

ui_flip_box <- function() {

  shiny::tagList(
    column(
      width = 12L,
      infobox_with_code(
        flip_box(
          front = info_box(
            icon = "bookmark",
            p("This is the front side. Double-click me")
          ),
          back = info_box(
            icon = NULL,
            p("This is the back side. Double-click me to toggle back.")
          )
        )
      )
    ),
    column(
      width = 12L,
      infobox_with_code(
        flip_box(
          active_on = "manual",
          inputId = ns("flip_demo_3"),
          front = info_box(
            icon = "arrow-right",
            p("This is the front side. Click this button"),
            actionButton(ns("flip_btn_1"), "Flip")
          ),
          back = info_box(
            icon = "arrow-left",
            p("This is the back side. Click this button"),
            actionButton(ns("flip_btn_2"), "Flip again")
          )
        )
      )
    ),
    column(
      width = 12L,
      infobox_with_code(
        flip_box(
          active_on = "click-front",
          inputId = ns("flip_demo_2"),
          front = info_box(
            icon = "th",
            p("This is the front side. Double-click me")
          ),
          back = info_box(
            icon = NULL,
            plotOutput(ns("flip_box_plot")),
            p(
              "Click this button and a progress will run. ",
              "Once the progress finishes, the box will flip back."
            ),
            actionButton(ns("show_progress"), "Show Progress")
          )
        )
      )
    )
  )

}

ui_flip_card <- function() {
  tagList(
    column(
      width = 12L,
      card(
        title = "Card with flip box",
        class_body = "no-padding",
        tools = list(
          card_tool(widget = "flip")
        ),
        flip_box(
          front = div(
            p(
              class = "padding-20",
              "While you can still flip the box by double-clicking, ",
              "alternatively, clicking the 'flip' icon in the tool bar also flips the box."
            ),
            plotOutput(ns("flip_card_plot"))
          ),
          back = div(
            class = "padding-20",
            p("The data used to generate the figure"),
            tableOutput(ns("flip_card_table"))
          )
        )
      )
    )
  )
}

server_flip_box <- function(input, output, session, ...) {

  output$flip_box_plot <- renderPlot({
    plot(rnorm(100, 10 * sin(seq(0, 2, length.out = 100))), pch = 16,
         ylab = "Response", las = 1)
  })

  observeEvent(input$show_progress, {
    progress <- shiny_progress(title = "Running algorithms", max = 10, shiny_auto_close = TRUE)
    for (i in 1:10) {
      progress$inc(sprintf("Running part %s", i))
      Sys.sleep(0.1)
    }
    flip("flip_demo_2")
  })

  observeEvent(input$flip_btn_1, {
    flip("flip_demo_3")
  })
  observeEvent(input$flip_btn_2, {
    flip("flip_demo_3")
  })

  output$flip_card_plot <- renderPlot({
    data(iris)
    with(iris, {
      plot(Sepal.Length, Sepal.Width, col = Species, pch = 20)
    })
  })

  output$flip_card_table <- renderTable({
    data(iris)
    iris
  })

}

library(shiny)
library(shidashi)
library(ggplot2)
library(ggExtra)
library(plyr)

ui_demo_summary <- function() {
  fluidRow(
    column(
      width = 3L,
      info_box(
        icon = "cog",
        span(class = "info-box-text", "Memory Usage"),
        span(class = "info-box-number", "324 MB")
      )
    ),
    column(
      width = 3L,
      info_box(
        icon = "thumbs-up",
        span(class = "info-box-text", "Likes"),
        span(class = "info-box-number", "12,320"),
        class_icon = "bg-red"
      )
    ),
    column(
      width = 3L,
      info_box(
        icon = "shopping-cart",
        span(class = "info-box-text", "Sales"),
        span(class = "info-box-number", "1,829"),
        class_icon = "bg-success"
      )
    ),
    column(
      width = 3L,
      info_box(
        icon = "users",
        span(class = "info-box-text", "Users"),
        span(class = "info-box-number", "1,290"),
        class_icon = "bg-yellow"
      )
    )
  )
}

ui_demo_monthly <- function() {
  fluidRow(
    column(
      width = 9L,
      div(class = "shidashi-anchor max-height-0 visibility-none", "Sales Report"),
      card(
        title = "Sales Report",
        class_body = "height-300",
        resizable = TRUE,
        tools = list(
          as_badge(sprintf("%s|bg-primary", Sys.Date())),
          card_tool(widget = "flip", title = "See details"),
          card_tool(widget = "collapse"),
          card_tool(widget = "link", href = "https://github.com/dipterix/shidashi"),
          card_tool(widget = "maximize")
        ),
        flip_box(
          class = "fill",
          front = plotOutput(ns("sales_report"), height = "100%"),
          back = shiny::tableOutput(ns("sales_table"))
        ),
        footer = fluidRow(
          column(
            width = 4L,
            div(
              class = "description-block border-right",
              span(
                class = "description-percentage text-success",
                as_icon("caret-up"), "17%"
              ),
              h5(
                class = "description-header",
                "$35,210.43"
              ),
              span(
                class = "description-text",
                "TOTAL REVENUE"
              )
            )
          ),
          column(
            width = 4L,
            div(
              class = "description-block border-right",
              span(
                class = "description-percentage text-warning",
                as_icon("caret-left"), "0%"
              ),
              h5(
                class = "description-header",
                "$10,390.90"
              ),
              span(
                class = "description-text",
                "TOTAL COST"
              )
            )
          ),
          column(
            width = 4L,
            div(
              class = "description-block",
              span(
                class = "description-percentage text-success",
                as_icon("caret-up"), "10%"
              ),
              h5(
                class = "description-header",
                "$14,123.90"
              ),
              span(
                class = "description-text",
                "TOTAL PROFIT"
              )
            )
          )
        )
      )
    ),
    column(
      width = 3L,
      card(
        title = "Goal Completion",
        class_body = "height-300",
        tools = list(
          card_tool(widget = "collapse")
        ),
        footer = fluidRow(
          column(
            width = 12L,
            div(
              class = "description-block",
              span(
                class = "description-percentage text-danger",
                as_icon("caret-down"), "30%"
              ),
              h5(
                class = "description-header",
                "1163"
              ),
              span(
                class = "description-text",
                "GOAL COMPLETIONS"
              )
            )
          )
        ),
        column(
          width = 12L,
          h6(class = "text-center", "Goal Completion"),
          progressOutput(ns("sales_report_prog1"), description = "",
                         "Add Products to Cart",
                         span(class = "float-end", "123/150"),
                         value = 123 / 150 * 100),
          progressOutput(ns("sales_report_prog2"), description = "",
                         "Complete Purchase",
                         class = "bg-red",
                         span(class = "float-end", "310/400"),
                         value = 310 / 400 * 100),
          progressOutput(ns("sales_report_prog3"), description = "",
                         "Visit Premium Page",
                         class = "bg-success",
                         span(class = "float-end", "480/800"),
                         value = 480 / 800 * 100),
          progressOutput(ns("sales_report_prog4"), description = "",
                         "Inquiries",
                         class = "bg-yellow",
                         span(class = "float-end", "250/500"),
                         value = 250 / 500 * 100)
        )
      )
    )
  )
}

ui_demo_details <- function() {
  fluidRow(
    column(
      width = 4L,
      card2(
        title = "Scatter plot",
        class_body = "no-padding",
        body_main = flip_box(
          front = div(
            class = "fill-width height-450 min-height-450 resize-vertical",
            plotOutput(ns("iris_plot"), height = "100%")
          ),
          back = tableOutput(ns("iris_plot_data"))
        ),
        body_side = div(
          class = "padding-top-50",
          shidashi::register_input(
            inputId = "iris_threshold",
            description = "Threshold Petal.Width and visualize the iris data with only selected data",
            update = "shiny::updateSliderInput",
            expr = sliderInput(inputId = ns("iris_threshold"),
                               label = "Threshold by Petal.Width",
                               min = 0, max = 3, value = 0, step = 0.1)
          )
        )
      )
    ),
    column(
      width = 8L,
      div(class = "shidashi-anchor max-height-0 visibility-none", "Analysis"),
      card_tabset(
        title = "Analysis",
        class_body = "fill flex-container no-padding min-height-450",
        tools = list(
          card_tool(widget = "custom", icon = "sync", inputId = ns("refresh"), title = "Generate analysis"),
          card_tool(widget = "maximize")
        ),
        "Histogram" = div(
          class = "fill position-absolute",
          plotOutput(ns("distibution_plot"), height = "100%")
        ),
        "Summary" = div(
          class = "fill position-absolute overflow-auto",
          tableOutput(ns("summary_table"))
        )
      )
    )
  )
}

server_demo <- function(input, output, session, ...) {

  shidashi::register_session(session)
  local_data <- reactiveValues()

  output$sales_report <- renderPlot({
    theme <- shidashi::get_theme()
    par(bg = theme$background, fg = theme$foreground,
        col.lab = theme$foreground, col.main = theme$foreground,
        col.axis = theme$foreground,
        las = 1)
    data("AirPassengers")
    y <- as.vector(t(AirPassengers))
    x <- seq(1, 12, length.out = length(y))
    plot(x, y, axes = FALSE, type = "l",
         xlab = "Time", ylab = "", main = "Sales (1 year)")
    axis(1, at = 1:12,
         labels = c("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
                    "Oct", "Nov", "Dec"))
    axis(2, seq(100, 600, 250))
  })

  output$sales_table <- renderTable({
    data("airquality")
    airquality <- head(airquality)
    names(airquality) <- c("Department", "# Transactions", "Revenue (x$1000)",
                           "Returned", "Month", "Day")
    airquality
  }, hover = TRUE, spacing = "s", width = "100%")

  output$iris_plot_data <- renderTable({
    data(iris)
    iris[iris$Petal.Width > input$iris_threshold,
         c("Petal.Length", "Petal.Width", "Species")]
  }, striped = TRUE, spacing = "s", width = "100%")

  generate_ggtheme <- function(
    theme,
    panel.background = element_rect(
      fill = theme$background, color = theme$background),
    plot.background = element_rect(
      fill = theme$background, color = theme$background),
    panel.grid.major = element_blank(),
    panel.grid.minor = element_blank(),
    axis.line.x.bottom = element_line(color = theme$foreground),
    axis.line.y.left = element_line(color = theme$foreground),
    legend.key = element_rect(fill = theme$background, colour = theme$background),
    rect = element_rect(fill = theme$background, colour = theme$foreground),
    title = element_text(color = theme$foreground),
    text = element_text(color = theme$foreground),
    line = element_line(color = theme$foreground),
    ...) {
    ggplot2::theme(
      panel.background = panel.background,
      plot.background = plot.background,
      panel.grid.major = panel.grid.major,
      panel.grid.minor = panel.grid.minor,
      axis.line.x.bottom = axis.line.x.bottom,
      axis.line.y.left = axis.line.y.left,
      legend.key = legend.key,
      rect = rect,
      title = title,
      text = text,
      line = line,
      ...
    )
  }

  shidashi::register_output(
    renderPlot({
      data(iris)
      theme <- shidashi::get_theme()
      ggtheme <- generate_ggtheme(theme)

      iris <- iris[iris$Petal.Width > input$iris_threshold, ]

      validate(
        need(nrow(iris) > 0, "No data point selected")
      )

      ggplot(data = iris) +
        aes(x = Sepal.Length, y = Petal.Length, color = Species) +
        geom_point() +
        geom_rug(col = "steelblue", alpha = 0.1, linewidth = 1.5) + ggtheme
    }),
    outputId = "iris_plot",
    description = "Scatter plot of iris, thresholded by input `iris_threshold`",
    download_type = "image"
  )

  run_analysis <- function() {
    show_notification(
      title = "Generating analysis...",
      subtitle = "This might take a while",
      class = "bg-primary",
      close = FALSE,
      autohide = FALSE,
      progressOutput(ns("data_gen_pro"), description = "Loading data...",
                     size = "xs", class = "bg-yellow")
    )
    on.exit({ clear_notifications() })

    progress <- shiny_progress("", max = 10, outputId = "data_gen_pro")
    for (i in 1:10) {
      progress$inc(sprintf("step %s", i), message = ifelse(
        i > 5, "Analyze data", "Loading data"
      ))
      Sys.sleep(0.2)
    }

    local_data$data <- data.frame(
      name = c(rep("A", 500), rep("B", 500), rep("B", 500), rep("C", 20),
               rep("D", 100),
               sample(LETTERS, 20000, replace = TRUE)),
      value = c(rnorm(500, 10, 5), rnorm(500, 13, 1), rnorm(500, 18, 1),
                rnorm(20, 25, 4), rnorm(100, 12, 1), rnorm(20000, 15, 30) )
    )
  }

  observeEvent(input$refresh, {
    run_analysis()
  })

  observeEvent(shidashi::get_event(session$ns("refresh"), session = session), {
    run_analysis()
  })

  output$distibution_plot <- renderPlot({
    validate(
      need(is.data.frame(local_data$data), "Please press the refresh button on the top-right tool bar")
    )
    theme <- shidashi::get_theme()

    data <- local_data$data

    sample_size <- do.call("rbind", lapply(split(data, data$name), function(x) {
      data.frame(
        name = x$name[[1]],
        num = nrow(x)
      )
    }))
    merged <- merge(data, sample_size, by = "name", all.x = TRUE, all.y = FALSE)
    merged$myaxis <- factor(paste0(merged$name, "\n", "n=", merged$num))

    ggtheme <- generate_ggtheme(
      theme,
      legend.position = "none",
      axis.line.y.left = element_blank(),
      axis.text = element_text(color = theme$foreground)
    )
    ggplot(merged) +
      aes(myaxis, value, fill = name) +
      geom_violin(width = 1) +
      geom_boxplot(width = 0.1,
                   color = "grey",
                   alpha = 0.2) +
      geom_jitter(
        height = 0,
        width = 0.1,
        size = 0.1,
        alpha = 0.2
      ) +
      ggtheme +
      xlab("")
  })

  output$summary_table <- renderTable({
    validate(
      need(is.data.frame(local_data$data), "Please press the refresh button on the top-right tool bar")
    )
    data <- local_data$data

    sample_size <- do.call("rbind", lapply(split(data, data$name), function(x) {
      data.frame(
        name = x$name[[1]],
        num = nrow(x)
      )
    }))
    sample_size
  })
}


# Module-level MCP tool: Triggers the "refresh" button in the demo module.
# This fires the same reactive chain as clicking the sync icon on the
# Analysis card (inputId = ns("refresh")).
#
# Module-level tools are auto-enabled — no agent.yaml entry needed.

trigger_refresh <- shidashi::mcp_wrapper(
  function(session) {

    ellmer::tool(
      fun = function() {
        shidashi::fire_event(session$ns("refresh"), Sys.time(), session = session)
        "Refresh triggered successfully. The analysis is being regenerated."
      },
      name = "trigger_refresh",
      description = "Trigger the refresh/regenerate analysis action in the demo module. This generates random data and updates the histogram and summary table.",
      arguments = list()
    )
  }
)

library(shiny)
library(shidashi)

ui_use_template <- function() {
  column(
    width = 12,
    h2("Start with a default template", class = "shidashi-anchor"),
    div(
      tags$ul(
        tags$li(
          span(
            class = "inline-all",
            "Start your ", tags$code("RStudio"), ", click on the ", tags$code("File"),
            " menu, choose ", tags$code("New Project...")
          )
        ),
        tags$li(
          span(
            class = "inline-all",
            "In the project wizard, ",
            "choose ", tags$code("New Directory")
          )
        ),
        tags$li(
          span(
            class = "inline-all",
            "Find and select ",
            tags$code("Shidashi Shiny Template")
          )
        ),
        tags$li(
          span(
            class = "inline-all",
            "Enter your project name, ",
            "as well as the project directory. ",
            "Please accept the default user & theme"
          )
        ),
        tags$li(
          span(
            class = "inline-all",
            "Click on ",
            tags$code("Create Project"), ", ",
            tags$code("RStudio"),
            " will create a ", tags$code("shidashi"),
            " project containing this demo."
          )
        )
      )
    )
  )
}

ui_card2 <- function() {

  card_ui <- card2(
    title = "Card2 Example",
    class_body = "height-300",
    tools = list(
      card_tool(widget = "link", href = "https://github.com/dipterix"),
      card_tool(widget = "refresh"),
      card_tool(widget = "collapse"),
      card_tool(widget = "maximize")
    ),
    body_main = plotOutput(
      outputId = ns("card2_plot"),
      height = "100%"
    ),
    body_side = fluidRow(
      column( 6L,
             textInput(ns("card2_plot_title"), "Plot title")
      ),
      column( 6L,
             sliderInput(ns("card2_plot_npts"), "# of points",
                         min = 1, max = 100, value = 10, step = 1, round = TRUE)
      )
    )
  )

  shiny::tagList(
    shiny::column( width = 4L, card_ui ),
    shiny::column(
      width = 8L,
      card_tabset(
        class_body = "height-300 overflow-y-auto",
        inputId = ns("card2_code_cardset"),
        "UI code" = fluidRow(
          column(
            width = 12L,
            html_highlight_code(
              get_construct_string(card_ui), quoted = TRUE,
              width.cutoff = 80L, indent = 2, wrap = TRUE,
              args.newline = TRUE, blank = FALSE, copy_on_click = TRUE,
              hover = "overflow-auto"
            )
          )
        ),
        "server.R" = fluidRow(
          column(
            width = 12L,
            html_highlight_code(
              output$card2_plot <- renderPlot({
                npoints <- input$card2_plot_npts
                title <- input$card2_plot_title
                if (!length(title) || title == "") {
                  title <- "Normal Q-Q Plot"
                }
                qqnorm(rnorm(npoints), main = title)
                abline(a = 0, b = 1, col = "orange3", lty = 2)
              }),
              hover = "overflow-auto"
            )
          )
        )
      )
    )
  )
}

library(shiny)
library(shidashi)

# ---------------------------------------------------------------------------
# Section 1 — Session Info
# ---------------------------------------------------------------------------

ui_session_info <- function() {
  tagList(
    h3("Session Registry", class = "shidashi-anchor mt-0"),
    p(
      class = "inline-all",
      tags$code("register_session(session)"), " stores each Shiny module session ",
      "in a global registry keyed by ", tags$code("session$token"), ". ",
      "The registry entry holds the ", tags$code("shared_id"), " (resolved from ",
      "the URL query string or auto-generated), the session object itself, ",
      "a reactive event bus, and MCP tool bindings."
    ),
    fluidRow(
      column(
        width = 6L,
        card(
          title = "Registry Entry for This Session",
          class_body = "min-height-200",
          tools = list(
            card_tool(widget = "collapse")
          ),
          body_main = div(
            p(
              class = "inline-all",
              "The entry below is read directly from ",
              tags$code("globals_session_registry()"), " via ",
              tags$code("shidashi:::get_session_entry(session$token)"), "."
            ),
            verbatimTextOutput(ns("session_entry_info"))
          ),
          footer = actionButton(
            ns("refresh_entry"),
            label = "Refresh",
            class = "btn-sm btn-outline-primary"
          )
        )
      ),
      column(
        width = 6L,
        card(
          title = "register_session() — shared_id resolution",
          class_body = "min-height-200",
          tools = list(
            card_tool(widget = "collapse")
          ),
          body_main = div(
            p(
              class = "inline-all",
              tags$code("register_session(session)"), " reads ",
              tags$code("shared_id"), " directly from the URL query string ",
              "(\u0060?shared_id=my_id\u0060) or generates a random string. ",
              "It is stored in the session registry entry and accessible via ",
              tags$code("shidashi:::get_session_entry(session$token)$shared_id"), "."
            ),
            html_highlight_code(
              {
                # In module server:
                shidashi::register_session(session)
                # shared_id is resolved from:
                #   1. ?shared_id= query param (URL)
                #   2. auto-generated random string
                # It is stored in the session registry entry.
                shidashi::fire_event("demo.ping", list(ts = Sys.time()))
              },
              width.cutoff = 30L, hover = "overflow-auto"
            )
          )
        )
      )
    )
  )
}

# ---------------------------------------------------------------------------
# Section 2 — Local Events
# ---------------------------------------------------------------------------

ui_local_events <- function() {
  tagList(
    hr(),
    h3("Local Events", class = "shidashi-anchor"),
    p(
      class = "inline-all",
      tags$code("fire_event(key, value)"), " stores a reactive value on the ",
      "current session's event bus. ",
      tags$code("get_event(key)"), " reads it back within a reactive context. ",
      "Events are scoped to the session by default (", tags$code("global = FALSE"), ")."
    ),
    fluidRow(
      column(
        width = 6L,
        card(
          title = "fire_event / get_event",
          class_body = "min-height-250",
          tools = list(
            card_tool(widget = "collapse")
          ),
          body_main = div(
            html_highlight_code(
              {
                shidashi::fire_event("my.key", list(value = 42, ts = Sys.time()))
                evt <- shidashi::get_event("my.key")
                evt$value  # 42
              },
              width.cutoff = 30L, hover = "overflow-auto"
            ),
            hr(),
            fluidRow(
              column(6L, numericInput(ns("local_value"), "Value to send", value = 1, min = 1)),
              column(6L, textInput(ns("local_key"), "Event key", value = "demo.local"))
            )
          ),
          footer = fluidRow(
            column(
              6L,
              actionButton(ns("fire_local"), "fire_event()",
                           class = "btn-sm btn-primary w-100")
            ),
            column(
              6L,
              actionButton(ns("read_local"), "get_event()",
                           class = "btn-sm btn-outline-secondary w-100")
            )
          )
        )
      ),
      column(
        width = 6L,
        card(
          title = "Local Event Log",
          class_body = "height-300 overflow-auto",
          tools = list(
            card_tool(widget = "collapse")
          ),
          body_main = verbatimTextOutput(ns("local_event_log"))
        )
      )
    )
  )
}

# ---------------------------------------------------------------------------
# Section 3 — Global (Cross-Tab) Events
# ---------------------------------------------------------------------------

ui_global_events <- function() {
  tagList(
    hr(),
    h3("Global Events (Cross-Tab Broadcasting)", class = "shidashi-anchor"),
    p(
      class = "inline-all",
      "When ", tags$code("global = TRUE"), " is passed to ",
      tags$code("fire_event()"), ", the event is propagated to ",
      tags$em("all"), " live sessions that share the same ",
      tags$code("shared_id"), ". Open this module in a second browser tab ",
      "and click the button — both tabs will receive the event."
    ),
    p(
      class = "inline-all text-muted",
      "Tabs share a ", tags$code("shared_id"), " when they open the same ",
      "shidashi dashboard URL. You can also pin a specific ",
      tags$code("shared_id"), " via the URL query string: ",
      tags$code("?shared_id=your_id"), "."
    ),
    fluidRow(
      column(
        width = 6L,
        card(
          title = "Broadcast a Global Event",
          class_body = "min-height-250",
          tools = list(
            card_tool(widget = "collapse")
          ),
          body_main = div(
            html_highlight_code(
              {
                shidashi::fire_event(
                  key    = "global.ping",
                  value  = list(message = "Hello from another tab!", ts = Sys.time()),
                  global = TRUE     # broadcast to all tabs with same shared_id
                )
              },
              width.cutoff = 30L, hover = "overflow-auto"
            ),
            hr(),
            textInput(ns("global_message"), "Message to broadcast",
                      value = "Hello from this tab!"),
            div(
              class = "text-muted small mt-2",
              "Shared ID: ", textOutput(ns("shared_id_display"), inline = TRUE)
            )
          ),
          footer = actionButton(ns("fire_global"), "Broadcast: fire_event(global=TRUE)",
                                class = "btn-sm btn-warning w-100")
        )
      ),
      column(
        width = 6L,
        card(
          title = "Received Global Events",
          class_body = "height-300 overflow-auto",
          tools = list(
            card_tool(widget = "collapse")
          ),
          body_main = div(
            p(
              class = "text-muted small inline-all",
              "This panel updates reactively whenever a ", tags$code("global.ping"),
              " event arrives — on any tab."
            ),
            verbatimTextOutput(ns("global_event_log"))
          )
        )
      )
    )
  )
}

# ---------------------------------------------------------------------------
# Server
# ---------------------------------------------------------------------------

server_session_events <- function(input, output, session, ...) {

  shidashi::register_session(session)

  # Keep a local log of events (not reactive — just a character vector in a
  # reactiveVal so we can append to it)
  local_log  <- reactiveVal(character(0))
  global_log <- reactiveVal(character(0))

  # ---- Section 1: session registry info ----------------------------------

  entry_info <- reactiveVal(NULL)

  get_entry <- function() shidashi:::get_session_entry(session$token)

  # Populate on load
  observe({
    entry <- get_entry()
    token <- session$token
    if (is.null(entry)) {
      entry_info("(session not yet in registry)")
      return()
    }
    entry_info(paste(
      paste0("token:      ", token),
      paste0("shared_id:  ", entry$shared_id),
      paste0("namespace:  ", entry$namespace),
      paste0("registered: ", format(entry$registered_at, "%Y-%m-%d %H:%M:%S")),
      sep = "\n"
    ))
  })

  observeEvent(input$refresh_entry, {
    entry <- get_entry()
    token <- session$token
    if (is.null(entry)) {
      entry_info("(session entry not found)")
      return()
    }
    entry_info(paste(
      paste0("token:      ", token),
      paste0("shared_id:  ", entry$shared_id),
      paste0("namespace:  ", entry$namespace),
      paste0("registered: ", format(entry$registered_at, "%Y-%m-%d %H:%M:%S")),
      paste0("mcp_tools:  ", entry$tools$size()),
      sep = "\n"
    ))
  })

  output$session_entry_info <- renderText({
    entry_info() %||% "(loading...)"
  })

  # ---- Section 2: local events -------------------------------------------

  output$shared_id_display <- renderText({
    entry <- shidashi:::get_session_entry(session$token)
    if (is.null(entry)) "(unknown)" else entry$shared_id
  })

  observeEvent(input$fire_local, {
    key   <- trimws(input$local_key)
    value <- as.integer(input$local_value)
    if (!nzchar(key)) {
      show_notification("Event key must not be empty.", type = "warning")
      return()
    }
    fire_event(key = key, value = list(value = value, ts = Sys.time()), session = session)
    line <- sprintf("[%s] fired '%s' = %s", format(Sys.time(), "%H:%M:%S"), key, value)
    local_log(c(local_log(), line))
  })

  observeEvent(input$read_local, {
    key <- trimws(input$local_key)
    if (!nzchar(key)) {
      show_notification("Event key must not be empty.", type = "warning")
      return()
    }
    evt <- get_event(key = key, session = session)
    if (is.null(evt)) {
      line <- sprintf("[%s] get_event('%s') -> NULL (not yet fired)", format(Sys.time(), "%H:%M:%S"), key)
    } else {
      line <- sprintf("[%s] get_event('%s') -> value=%s ts=%s",
                      format(Sys.time(), "%H:%M:%S"), key,
                      evt$value %||% "?",
                      format(evt$ts, "%H:%M:%S"))
    }
    local_log(c(local_log(), line))
  })

  output$local_event_log <- renderText({
    log <- local_log()
    if (!length(log)) return("(no events yet — fire one above)")
    paste(rev(log), collapse = "\n")
  })

  # ---- Section 3: global events ------------------------------------------

  observeEvent(input$fire_global, {
    msg <- trimws(input$global_message)
    if (!nzchar(msg)) msg <- "(empty message)"
    fire_event(
      key    = "global.ping",
      value  = list(message = msg, ts = Sys.time(),
                    from_tab = session$token),
      global = TRUE,
      session = session
    )
  })

  # React to the global.ping event arriving on this session
  shidashi::register_session(session)

  observe({
    evt <- get_event("global.ping", session = session)
    if (is.null(evt)) return()
    line <- sprintf("[%s] from_tab=%s | \"%s\"",
                    format(evt$ts %||% Sys.time(), "%H:%M:%S"),
                    substr(evt$from_tab %||% "?", 1, 8),
                    evt$message %||% "(no message)")
    # isolate() prevents global_log from being a reactive dependency of this
    # observer — without it, writing global_log() re-triggers this observe,
    # causing an infinite loop.
    global_log(c(isolate(global_log()), line))
  })

  output$global_event_log <- renderText({
    log <- global_log()
    if (!length(log)) return("(no global events yet — click broadcast above, or open a second tab)")
    paste(rev(log), collapse = "\n")
  })
}

ui_progress <- function() {
  tagList(
    column(
      width = 12L,
      card(
        title = "Different styles of progress",
        fluidRow(
          column(
            width = 6L,
            progressOutput(ns("prog_1"), description = 'class = "bg-primary" (default)', value = 40),
            progressOutput(ns("prog_2"), description = 'class = "bg-info"', class = "bg-info", value = 100),
            progressOutput(ns("prog_3"), description = 'class = "bg-success"', class = "bg-success", value = 80),
            progressOutput(ns("prog_4"), description = 'class = "bg-warning"', class = "bg-warning", value = 70),
            progressOutput(ns("prog_5"), description = 'class = "bg-danger"', class = "bg-danger", value = 50)
          ),
          column(
            width = 6L,
            p(
              span(
                class = "inline-all",
                "There are two ways of changing the progress. ",
                "The first approach is via ",
                tags$code("renderProgress({...})"),
                " method:",
                actionLink(ns("prog_btn1"), "click me!")
              )
            ),
            p(
              span(
                class = "inline-all",
                "The second choice is to use ",
                tags$code("shiny_progress()"),
                " function with input id:",
                actionLink(ns("prog_btn2"), "click me!")
              )
            ),
            p(
              "Here's an example of embedding progress bar in the notification.",
              actionButton(ns("prog_btn3"), "Click me!")
            )
          )
        )
      )
    )
  )
}

server_progress <- function(input, output, session, ...) {

  data <- fastmap::fastmap()
  data$set("progress", 0)

  output$prog_1 <- renderProgress({
    print(input$prog_btn1)
    progress <- data$get("progress") + 10
    if (progress > 100) { progress <- 0 }
    data$set("progress", progress)
    list(
      value = progress,
      description = sprintf("Progress: %.0f%%", progress)
    )
  })

  observeEvent(input$prog_btn2, {
    progress <- shiny_progress(title = "Method 2", max = 10, outputId = "prog_1")
    for (i in 1:10) {
      progress$inc(sprintf("step - %d", i))
      Sys.sleep(0.3)
    }
  }, ignoreInit = TRUE, ignoreNULL = TRUE)

  observeEvent(input$prog_btn3, {
    show_notification(
      message = div(
        "The wavelet might take a while to run. Grub a cup of coffee and wait.",
        progressOutput(ns("notif_7_prg"), class = "bg-success")
      ),
      autohide = FALSE,
      title = "Brewing",
      icon = "coffee",
      close = FALSE,
      class = "notif_7_autoclose"
    )
    on.exit({ clear_notifications(class = "notif_7_autoclose") })
    progress <- shiny_progress(title = "Running", max = 10, outputId = "notif_7_prg")
    for (i in 1:10) {
      Sys.sleep(0.5)
      progress$inc(detail = paste("Channel", i))
    }
  })
}

ui_card_basic <- function() {

  shiny::tagList(
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Basic Card Example",
        "Card body"
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Primary Card Example 1",
        class = "card-outline card-primary",
        'class = "card-outline card-primary"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Primary Card Example 2",
        class = "card-primary",
        'class = "card-primary"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Primary Card Example 3",
        class = "bg-primary",
        'class = "bg-primary"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Secondary Card",
        class = "card-secondary",
        'class = "card-secondary"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Success Card",
        class = "card-success",
        'class = "card-success"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Info Card",
        class = "card-info",
        'class = "card-info"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Warning Card",
        class = "card-warning",
        'class = "card-warning"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Danger Card",
        class = "card-danger",
        'class = "card-danger"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Dark Card",
        class = "card-dark",
        'class = "card-dark"'
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Light Card",
        class = "card-light",
        'class = "card-light"'
      )
    ))
  )

}

library(shiny)
library(shidashi)

ui_prerequisite <- function() {
  column(
    width = 7L,
    h2("Pre-requisite for this tutorial", class = "shidashi-anchor"),
    p(
      span(
        class = "inline-all",
        "I believe you have had `shidashi` installed. ",
        "However, this tutorial requires some extra packages. ",
        "Please run the following R command:"
      )
    ),
    html_highlight_code(
      install.packages(c("ggExtra", "rmarkdown")),
      hover = "overflow-auto"
    ),
    tags$small(
      "* You can click the code to copy the code."
    )
  )
}

#' @title Generates small icon widgets
#' @description The icons cane be displayed at header line within
#' \code{\link{accordion}}, \code{\link{card}}, \code{\link{card2}},
#' \code{\link{card_tabset}}. See their examples.
#' @param inputId the button id, only necessary when \code{widget}
#' is \code{"custom"}
#' @param title the tip message to show when the mouse cursor hovers
#' on the icon
#' @param widget the icon widget type; choices are \code{"maximize"},
#' \code{"collapse"}, \code{"remove"}, \code{"flip"},
#' \code{"refresh"}, \code{"link"}, and \code{"custom"}; see 'Details'
#' @param icon icon to use if you are unsatisfied with the default ones
#' @param class additional class for the tool icons
#' @param href,target used when \code{widget} is \code{"link"}, will
#' open an external website; default is open a new tab
#' @param start_collapsed used when \code{widget} is \code{"collapse"},
#' whether the card should start collapsed
#' @param ... passed to the tag as attributes
#' @return 'HTML' tags to be included in \code{tools} parameter in
#' \code{\link{accordion}}, \code{\link{card}}, \code{\link{card2}},
#' \code{\link{card_tabset}}
#' @details There are 7 \code{widget} types:
#' \describe{
#' \item{\code{"maximize"}}{allow the elements to maximize
#' themselves to full-screen}
#' \item{\code{"collapse"}}{allow the elements to collapse}
#' \item{\code{"remove"}}{remove a \code{\link{card}} or
#' \code{\link{card2}}}
#' \item{\code{"flip"}}{used together with \code{\link{flip_box}},
#' to allow card body to flip over}
#' \item{\code{"refresh"}}{refresh all shiny outputs}
#' \item{\code{"link"}}{open a hyper-link pointing to external
#' websites}
#' \item{\code{"custom"}}{turn the icon into a \code{actionButton}.
#' in this case, \code{inputId} must be specified.}
#' }
#' @export
card_tool <- function(inputId = NULL, title = NULL, widget = c("maximize", "collapse", "remove", "flip", "refresh", "link", "custom"), icon, class = "", href = "#", target = "_blank", start_collapsed = FALSE, ...) {
  widget <- match.arg(widget)

  if (missing(icon)) {
    icon <- switch(
      widget,
      maximize = as_icon("expand"),
      collapse = as_icon(ifelse(start_collapsed, "plus", "minus")),
      remove = as_icon("times"),
      refresh = as_icon("sync-alt"),
      link = as_icon("external-link-alt"),
      flip = as_icon("adjust"),
      {
        stop("Custom widget must provide a valid icon; see ?shiny::icon")
      }
    )
  } else {
    icon <- as_icon(icon)
  }

  if (length(inputId) == 1) {
    class <- combine_class("btn btn-tool action-button", class)
  } else {
    class <- combine_class("btn btn-tool", class)
  }

  if ( widget == "custom" ) {
    widget <- NULL
  }

  if (startsWith(href, "#")) {
    target <- "_self"
  }

  return(
    shiny::a(
      id = inputId,
      type = "button",
      href = href,
      target = target,
      class = class,
      # AdminLTE3
      `data-card-widget` = widget,
      title = title,
      icon,
      ...
    )
  )

}

library(shiny)
page_title <- function(complete = TRUE) {
  if (complete) {
    "Shiny Dashboard Template - Barebone"
  } else {
    "ShiDashi"
  }
}
page_logo <- function(size = c("normal", "small", "large")) {
  # Relative path to your logo icon in www/
  "shidashi/img/icon.png"
}
page_loader <- function() {
  # if no loader is needed, then return NULL
  shiny::div(
    class = "preloader flex-column justify-content-center align-items-center",
    shiny::img(
      class = "animation__shake",
      src = page_logo("large"),
      alt = "Logo", height = "60", width = "60"
    )
  )
}
body_class <- function() {
  c(
    #--- Fix the navigation banner ---
    #"layout-navbar-fixed",

    #--- Collapse the sidebar at the beginning ---
    # "sidebar-collapse",

    #--- Let control sidebar open at the beginning ---
    # "control-sidebar-slide-open",

    #--- Fix the sidebar position ---
    "layout-fixed",

    #--- Default behavior when collapsing sidebar
    # "sidebar-mini", "sidebar-mini-md", "sidebar-mini-xs"

    #--- Hide the navbar-nav-iframe
    "navbar-iframe-hidden",

    #--- Start as dark-mode ---
    "dark-mode"

    #--- Make scrollbar thinner ---
    # "fancy-scroll-y"

  )
}
nav_class <- function() {
  c(
    "main-header",
    "navbar",
    "navbar-expand",
    "navbar-dark",
    "navbar-primary"
  )
}

module_breadcrumb <- function() {}

ui_card_tools <- function() {
  shiny::tagList(
    shiny::column( width = 2L, card_with_code(
      card(
        class_body = "height-300",
        resizable = TRUE,
        title = "Badges",
        tools = list(
          as_badge("New|badge-info"),
          as_badge("3|badge-warning")
        ),
        div(
          class = "padding-20",
          p(
            "Add badges to the top-right corner. ",
            "Use \"|\" to indicate the badge classes; ",
            "for example: \"badge-info\", \"badge-warning\"..."),
          hr(),
          p("Use `resizable = TRUE` to make card resizable.")
        )
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Default Tools",
        class_body = "height-300",
        resizable = TRUE,
        tools = list(
          card_tool(widget = "link", href = "https://github.com/dipterix"),
          card_tool(widget = "collapse"),
          card_tool(widget = "maximize")
        ),
        plotOutput(ns("card_defaulttool_plot"), height = "100%")
      )
    )),
    shiny::column( width = 4L, card_with_code(
      card(
        title = "Custom Tools",
        class_body = "min-height-300",
        tools = list(
          card_tool(inputId = ns("card_tool_modal"), widget = "custom",
                    title = "Show alert", icon = "bell")
        ),
        "Click the bell icon (",
        as_icon("bell"),
        ") at the top-right corner to show a pop-up alert", br(),
        "The reaction requires server-side shiny observer. ",
        "Go to \"server.R\", add:",
        html_highlight_code(
          observeEvent(input$card_tool_modal, {
            shiny::showModal(shiny::modalDialog(
              "Click to dismiss",
              title = "Alert", easyClose = TRUE
            ))
          }), width.cutoff = 15L, args.newline = FALSE,
          hover = "overflow-auto"
        )
      )
    )),
    shiny::column( width = 3L, card_with_code(
      card(
        title = "Collapsed",
        start_collapsed = TRUE,
        class_body = "height-300 no-padding",
        tools = list(
          as_badge("30 messages |badge-primary"),
          card_tool(inputId = ns("card_tool_rerun"), widget = "custom",
                    title = "Re-run", icon = "random"),
          card_tool(widget = "collapse", start_collapsed = TRUE)
        ),
        p(
          style = "padding: 10px; margin: 0;",
          "Click the shuffle (", as_icon("random"), ") button."
        ),
        plotOutput(ns("card_tool_plot"), height = "254px")
      )
    )),
  )
}

#' Wrap an \verb{MCP} Tool Generator Function
#'
#' @description
#' Creates a wrapper around a generator function to ensure it returns valid
#' Model Context Protocol (\verb{MCP}) tool definitions. The wrapper validates
#' input and filters output to contain only \code{'ellmer::ToolDef'} objects.
#'
#' @param generator A function that accepts a `session` parameter and returns
#'   either a single tool object or a list/vector of such objects; see
#'   \code{\link[ellmer]{tool}}.
#'
#' @return A wrapped function with class \code{'shidashi_mcp_wrapper'} that:
#'   - Accepts a `session` parameter
#'   - Calls the generator function with the session
#'   - Normalizes the output to a list
#'   - Filters to keep only valid tool objects
#'   - Returns a list of tool objects (possibly empty)
#'
#' @details
#'   The wrapper performs the following validations:
#'   - Ensures `generator` is a function
#'   - Checks that `generator` accepts a `session` parameter
#'
#'   The returned function automatically handles both single tool definitions
#'   and lists of tools, providing a consistent interface for \verb{MCP} tool
#'   registration.
#'
#' @examples
#' # Define a generator function that returns tool definitions
#' my_tool_generator <- function(session) {
#'   # Define MCP tools using ellmer package
#'
#'   tool_rnorm <- tool(
#'     function(n, mean = 0, sd = 1) {
#'       shiny::updateNumericInput(session, "rnorm", value = rnorm)
#'     },
#'     description = "Draw numbers from a random normal distribution",
#'     arguments = list(
#'       n = type_integer("The number of observations. Must be positive"),
#'       mean = type_number("The mean value of the distribution."),
#'       sd = type_number("The standard deviation of the distribution.")
#'     )
#'   )
#'
#'   # or `list(tool_rnorm)`
#'   tool_rnorm
#' }
#'
#' # Wrap the generator
#' wrapped_generator <- mcp_wrapper(my_tool_generator)
#'
#' @export
mcp_wrapper <- function(generator) {
  stopifnot(
    "generator must be a function" = is.function(generator),
    "generator must accept an arguments: session" =
      "session" %in% names(formals(generator))
  )
  structure(
    function(session) {
      # TODO: should we consider wraping with try-catch warning
      res <- generator(session = session)
      if (inherits(res, "ellmer::ToolDef")) {
        res <- list(res)
      } else {
        res <- as.list(res)
      }
      res <- res[vapply(res, function(tool) { inherits(tool, "ellmer::ToolDef") }, FALSE)]
      res
    },
    class = c("shidashi_mcp_wrapper", "function")
  )
}

setup_mcp_proxy <- function(port = NULL, overwrite = TRUE, verbose = TRUE) {
  src <- system.file("mcp-proxy", "shidashi-proxy.mjs", package = "shidashi")
  if (!nzchar(src)) {
    return(invisible(NULL))
  }

  mcp_server_dir <- file.path(tools::R_user_dir("shidashi", "cache"), "mcp_server")
  ports_dir <- file.path(mcp_server_dir, "ports")
  dir.create(ports_dir, recursive = TRUE, showWarnings = FALSE)

  # Write port record when a port is supplied.
  if (!is.null(port)) {
    port_file <- file.path(
      ports_dir,
      paste0(format(as.numeric(Sys.time()) * 1000, scientific = FALSE, digits = 15), ".json")
    )
    writeLines(
      paste0('{"port":', as.integer(port), ',"created":"',
             format(Sys.time(), "%Y-%m-%dT%H:%M:%SZ", tz = "UTC"), '"}'),
      port_file
    )
    # Prune: keep only the 10 most-recent port records.
    port_files <- sort(list.files(ports_dir, pattern = "\\.json$", full.names = TRUE))
    if (length(port_files) > 10L) {
      file.remove(port_files[seq_len(length(port_files) - 10L)])
    }
  }

  # Copy proxy script to user cache.
  dest <- file.path(mcp_server_dir, "mcp-proxy.mjs")
  if (!file.exists(dest) || isTRUE(overwrite)) {
    file.copy(src, dest, overwrite = TRUE)
    if (verbose) message("Installed MCP proxy to:\n  ", dest)
  } else {
    if (verbose) message("MCP proxy already exists (overwrite = FALSE):\n  ", dest)
  }

  if (verbose) {
    snippet <- paste0(
      '{\n',
      '  "servers": {\n',
      '    "shidashi": {\n',
      '      "type": "stdio",\n',
      '      "command": "node",\n',
      '      "args": ["', dest, '"]\n',
      '    }\n',
      '  }\n',
      '}'
    )
    message(
      "\nPaste the following into your .vscode/mcp.json",
      " (or equivalent MCP settings):\n\n",
      snippet,
      "\n\nTo target a specific shidashi session, append its port as an extra arg:\n",
      '  "args": ["', dest, '", "<port>"]\n'
    )
  }

  invisible(dest)
}

#' Create \verb{MCP} Tools for Shiny Input Management
#'
#' @description
#' Builds a \code{\link{mcp_wrapper}} that exposes two \verb{MCP} tools:
#' \code{shiny_input_info} (query registered inputs) and
#' \code{shiny_input_update} (set input values). Inputs must first be
#' registered via the returned helper functions before they become visible
#' to the \verb{MCP} tools.
#'
#' @return A list with two elements:
#' \describe{
#'   \item{\code{input_helpers}}{A list of helper functions for managing
#'     input specifications:
#'     \describe{
#'       \item{\code{register_input_specification(expr, inputId, description, update, writable, quoted, env)}}{Registers
#'         a shiny input for \verb{MCP} access and returns the evaluated
#'         UI element. The \code{expr} argument is a call expression that
#'         creates a shiny input widget, e.g.
#'         \code{shiny::textInput(inputId = "x", label = "X")}.
#'         All metadata (\code{inputId}, \code{description}, \code{update})
#'         must be provided explicitly by the module writer.
#'         Returns the evaluated UI element (e.g. an HTML tag object),
#'         so the call can be used inline in UI definitions.}
#'       \item{\code{update_input_specification(inputId, description, type, update, writable)}}{Modifies
#'         the spec of an already-registered input. All arguments except
#'         \code{inputId} are optional; only supplied values are changed.
#'         Returns a list with \code{item} (the updated \code{data.frame} row)
#'         and \code{changed} (logical).}
#'       \item{\code{get_input_specification()}}{Returns a \code{data.frame}
#'         of all registered input specs (columns: \code{inputId},
#'         \code{description}, \code{type}, \code{update}, \code{writable}).
#'         Returns an empty \code{data.frame} with the same columns when
#'         no inputs are registered.}
#'     }
#'   }
#'   \item{\code{tool_generator}}{A \code{shidashi_mcp_wrapper} that, given a
#'     \code{session}, returns a named list of \code{ellmer::ToolDef} objects:
#'     \code{shiny_input_info} and \code{shiny_input_update}.}
#' }
#'
#' @details
#' The \code{update} specification string follows the pattern
#' \code{"pkg::fun"} or \code{"pkg::fun(key=formal, ...)"}.
#' The key-value pairs override the
#' default argument names passed to the update function:
#' \itemize{
#'   \item \code{id} — the formal argument name for the input ID
#'     (default \code{"inputId"})
#'   \item \code{value} — the formal argument name for the new value
#'     (default \code{"value"})
#'   \item \code{session} — the formal argument name for the session
#'     (default \code{"session"})
#' }
#'
#' For example, \code{"shiny::updateSelectInput(id=inputId,value=select)"}
#' means the update call will use \code{inputId} for the ID argument and
#' \code{select} (not \code{value}) for the value argument.
#'
#' Values received from \verb{MCP} are JSON-encoded strings. The update tool
#' attempts to decode them with \code{jsonlite::fromJSON()} before passing
#' them to the update function, falling back to the raw string on failure.
#'
#' @examples
#' wrapper <- mcp_wrapper_input_output()
#'
#' # Register inputs inline — returns the UI element for use in UI code
#' text_ui <- wrapper$input_helpers$register_input_specification(
#'   expr = shiny::textInput(inputId = "my_text", label = "User name"),
#'   inputId = "my_text",
#'   description = "User name",
#'   update = "shiny::updateTextInput"
#' )
#'
#' select_ui <- wrapper$input_helpers$register_input_specification(
#'   expr = shiny::selectInput("my_select", "Choose a dataset",
#'                             choices = c("iris", "mtcars")),
#'   inputId = "my_select",
#'   description = "Choose a dataset to visualise",
#'   update = "shiny::updateSelectInput(value=selected)"
#' )
#'
#' # Inspect all registered specs
#' wrapper$input_helpers$get_input_specification()
#'
#' # The MCP tool generator (pass to your MCP server registration)
#' shiny_input_wrapper <- wrapper$tool_generator
#'
#' # Initialization with a mock session
#' tools <- shiny_input_wrapper(shiny::MockShinySession$new())
#'
#' @param input_specs An optional \code{fastmap::fastmap()} object to use as
#'   the backing store for input specifications.  When \code{NULL} (the
#'   default) a fresh \code{fastmap} is created.  Passing an existing
#'   \code{fastmap} allows multiple wrapper instances (e.g. one created
#'   during UI rendering and another during server initialization) to share
#'   the same input registry.
#'
#' @noRd
mcp_wrapper_input_output <- function(input_specs = fastmap::fastmap(), output_specs = fastmap::fastmap()) {

  # stores the input ID, description, type, update function, writable for a
  # session inputId should be relative to session, meaning
  # "btn" not session$ns("btn")

  normalize_update_fun <- function(update) {
    # update <- "updateSelectInput(id=inputId,value=select)"
    if (!grepl(":", update)) {
      update <- sprintf("shiny::%s", update)
    }
    # record the original spec string before stripping the call signature
    spec  <- update
    parts <- strsplit(update, "[:]+", perl = TRUE)[[1]]
    fun_part <- parts[[2]]
    pkg      <- parts[[1]]
    # fun_part might be updateTextInput, or
    # updateSelectInput(id=inputId,value=select)
    fun_name    <- fun_part
    fields <- list(
      id      = "inputId",
      value   = "value",
      session = "session"
    )
    if (endsWith(fun_part, ")")) {
      fun_name   <- sub("^([^(]+)\\(.*\\)$", "\\1", fun_part, perl = TRUE)
      args_inner <- sub("^[^(]+\\((.*)\\)$", "\\1", fun_part, perl = TRUE)
      # parse key=value pairs; skip quoted-string values like session="session"
      for (pair in strsplit(args_inner, ",")[[1]]) {
        kv <- strsplit(trimws(pair), "\\s*=\\s*", perl = TRUE)[[1]]
        if (length(kv) != 2L) next
        key <- trimws(kv[[1]])
        val <- trimws(kv[[2]])
        fields[[key]] <- val
      }
    }
    fun_impl <- asNamespace(pkg)[[fun_name]]
    if (!is.function(fun_impl)) {
      stop("Unable to find update function `", pkg, "::", fun_name, "`")
    }

    list(
      update       = spec,
      fun_impl    = fun_impl,
      pkg         = pkg,
      fun         = fun_name,
      fields      = fields
    )
  }

  register_input_spec <- function(
    expr,
    inputId,
    update,
    description = "",
    writable = TRUE,
    quoted = FALSE,
    env = parent.frame()
  ) {
    "
    Register a shiny input for MCP tool access and return the UI element.

    The `expr` argument should be a call expression that creates a shiny
    input widget, e.g. `shiny::textInput(inputId = 'x', label = 'X')`.
    The expression is evaluated and its result (the UI element) is
    returned, so this function can be used inline in place of the
    original input constructor.

    Usage:
      register_input_spec(
        expr = shiny::textInput(inputId = ns('my_text'), label = 'Name'),
        inputId = 'my_text',
        description = 'Name',
        update = 'shiny::updateTextInput',
        writable = TRUE
      )

    @param expr        A call expression that creates a shiny input widget,
      e.g. `shiny::textInput(inputId = 'x', label = 'X')` or
      `selectInput('sel', 'Choose', choices = c('a','b'))`.
    @param inputId     Character scalar. The shiny input ID.
    @param description Character scalar. A human-readable description
      of the input's purpose, shown to LLM agents via the MCP info tool.
    @param update      Character scalar. The update function spec,
      e.g. 'shiny::updateTextInput' or
      'shiny::updateSelectInput(value=selected)'.
      Field mappings (e.g. value=selected) override the default
      argument names passed to the update function.
    @param writable    Logical scalar (default TRUE). Whether the MCP
      update tool is allowed to change this input.

    @return The evaluated UI element produced by the `expr` expression.
      The input specification is registered as a side effect.
    "
    if (!quoted) {
      expr <- substitute(expr)
    }

    # Normalise and validate the update spec
    update_info <- normalize_update_fun(update)

    item <- data.frame(
      inputId     = inputId,
      description = paste(description, collapse = " "),
      type        = truc_string(deparse1(expr), max_char = 100),
      update      = update_info$update,
      writable    = as.logical(writable)[[1]]
    )

    input_specs$set(inputId, item)

    # Evaluate the expression and return the UI element
    eval(expr, envir = env)
  }

  class(register_input_spec) <- c("register_input_impl", "function")

  register_output_spec <- function(expr, outputId, description = "", quoted = FALSE, env = parent.frame()) {
    if (!quoted) {
      expr <- substitute(expr)
    }

    description <- trimws(paste(description, collapse = ""))
    if (!nzchar(description)) {
      description <- deparse1(expr)
    }

    item <- data.frame(
      outputId    = outputId,
      description = paste(description, collapse = " "),
      type        = truc_string(deparse1(expr), max_char = 150, side = "both")
    )

    output_specs$set(outputId, item)

    return(invisible(item))
  }
  class(register_output_spec) <- c("register_output_impl", "function")

  update_input_spec <- function(
    inputId,
    description = NULL,
    type = NULL,
    update = NULL,
    writable = NULL
  ) {
    "
    Update the specification of an already-registered shiny input.

    Usage:
      update_input_spec(inputId, description, type, update, writable)

    @param inputId     Character scalar. Must match a previously registered
      input ID; an error is raised otherwise.
    @param description Character or NULL. New description (replaces old).
    @param type        Character or NULL. New widget type.
    @param update      Character or NULL. New update function spec string.
    @param writable    Logical or NULL. New writable flag.

    @return A list (invisible) with:
      - item:    the updated 1-row data.frame.
      - changed: logical, TRUE if any field was modified.
    "
    if (!input_specs$has(inputId)) {
      stop("Input `", inputId, "` has not been registered.")
    }
    item <- input_specs$get(inputId)
    changed <- FALSE
    if (!is.null(description)) {
      item$description <- paste(description, collapse = " ")
      changed <- TRUE
    }
    if (!is.null(type)) {
      item$type <- type
      changed <- TRUE
    }
    if (!is.null(update)) {
      update_info <- normalize_update_fun(update)
      item$update <- update_info$update
      changed <- TRUE
    }
    if (!is.null(writable)) {
      item$writable <- as.logical(writable)[[1]]
      changed <- TRUE
    }
    if (changed) {
      input_specs$set(inputId, item)
    }

    invisible(list(
      item = item,
      changed = changed
    ))
  }

  get_input_spec <- function() {
    "
    Retrieve all registered input specifications as a data.frame.

    Usage:
      get_input_spec()

    @return A data.frame with columns: inputId, description, type,
      update, writable. Returns an empty data.frame with the same
      columns when no inputs have been registered.
    "
    if (input_specs$size() == 0) {
      return(data.frame(
        inputId     = character(),
        description = character(),
        type        = character(),
        update      = character(),
        writable    = logical()
      ))
    }
    items <- input_specs$as_list()
    re <- do.call("rbind", items)
    row.names(re) <- NULL
    re
  }

  wrapper <- mcp_wrapper(function(
    session = shiny::getDefaultReactiveDomain()
  ) {

    query_requests <- fastmap::fastmap()

    # Observer: when the browser responds via setInputValue, store the
    # result in query_requests so shiny_query_ui_result can fetch it.
    local({
      shiny::bindEvent(
        shiny::observe({
          res <- session$input[["@shiny_query_ui_result@"]]
          res <- as.list(res)
          rid <- res$request_id
          if (length(rid) != 1 || !query_requests$has(rid)) {
            return()
          }
          entry <- query_requests$get(rid)
          entry$result <- res
          # str(res)
          # message(Sys.time() - entry$request_timestamp)
          query_requests$set(rid, entry)
        }, domain = session, priority = 101),
        session$input[["@shiny_query_ui_result@"]],
        ignoreNULL = TRUE, ignoreInit = FALSE
      )
    })

    shiny_input_info <- ellmer::tool(
      name = "shiny_input_info",
      description = paste(
        "Query registered shiny input specifications.",
        "Returns input IDs, descriptions, types, update functions,",
        "whether each is writable, and (when a session is active)",
        "whether each currently exists and its current value."
      ),
      arguments = list(
        inputIds = ellmer::type_array(
          ellmer::type_string(
            description = "Shiny input ID"
          ),
          description = "Optional: specific input IDs to query. Omit to list all registered inputs.",
          required = FALSE
        )
      ),
      fun = function(inputIds = character()) {
        inputIds <- unlist(inputIds)
        inputIds <- inputIds[!is.na(inputIds) & nzchar(inputIds)]
        if (length(inputIds) > 0) {
          items <- input_specs$mget(inputIds)
        } else {
          items <- input_specs$as_list()
        }
        # split each row into list

        if (!is.null(session)) {
          input <- shiny::isolate(shiny::reactiveValuesToList(session$input))
          items <- lapply(items, function(item) {
            if (is.null(item)) { return(NULL) }
            item <- as.list(item)
            item$exists <- item$inputId %in% names(input)
            item$current_value <- input[[item$inputId]]
            item
          })
        } else {
          items <- lapply(items, function(item) {
            as.list(item)
          })
        }

        items
      }
    )

    shiny_input_update <- ellmer::tool(
      name = "shiny_input_update",
      description = paste(
        "Update a shiny input value by its ID.",
        "The value will be sent to the corresponding shiny update function",
        "(e.g. updateTextInput, updateSelectInput, updateNumericInput).",
        "Call `shiny_input_info()` first to discover available input IDs,",
        "their types, current values, and whether they are writable.",
        "After changing the inputs, always call `shiny_input_info` again",
        "to verify the changes."
      ),
      arguments = list(
        inputId = ellmer::type_string(
          description = "Shiny input ID of which the value is to be changed",
          required = TRUE
        ),
        value = ellmer::type_string(
          description = "The new value for the input. Use JSON encoding for non-string values (e.g. 123, [1,2,3], {\"a\":1})."
        )
      ),
      fun = function(inputId, value) {
        # TODO: add a mode for tentative updating the input (highlight the
        # inputs and mark the values instead of changing them)
        if (!input_specs$has(inputId)) {
          stop(
            "There is no input ID: `",
            inputId,
            "`. Available IDs are: ",
            paste(input_specs$keys(), collapse = ", "),
            ". Call `shiny_input_info()` to get their information."
          )
        }

        item <- input_specs$get(inputId)
        if (!item$writable) {
          stop("Input ID: `", inputId, "` is read-only.")
        }

        active_inputIds <- shiny::isolate(names(session$input))
        if (!item$inputId %in% active_inputIds) {
          stop(
            "Input ID: `", inputId,
            "` is inactive or missing from this session."
          )
        }

        # Decode JSON-encoded value
        value <- tryCatch(
          jsonlite::fromJSON(value),
          error = function(e) value
        )

        update_info <- normalize_update_fun(item$update)

        expr <- as.call(structure(
          list(
            quote(update_info$fun_impl),
            session,
            inputId,
            value
          ),
          names = c(
            "",
            update_info$fields$session %||% "session",
            update_info$fields$id %||% "inputId",
            update_info$fields$value %||% "value"
          )
        ))

        eval(expr)

        # Wait for Shiny to update
        Sys.sleep(0.1)

        return(invisible(list(
          updated = TRUE,
          shiny_namespace = session$ns(NULL),
          inputId = inputId,
          value = value
        )))
      }
    )


    shiny_query_ui <- ellmer::tool(
      name = "shiny_query_ui",
      description = paste(
        "Request the HTML content of a UI element by CSS selector.",
        "This sends a query to the browser and returns a request_id.",
        "The browser response is asynchronous; call `tool__shiny_query_ui_result`",
        "with the returned request_id to retrieve the actual content.",
        "Wait briefly (1-2 seconds) before fetching the result."
      ),
      arguments = list(
        css_selector = ellmer::type_string(
          description = "A CSS selector to query (e.g. '#my_output', '.card-body', 'div[data-id=\"plot\"]').",
          required = TRUE
        )
      ),
      fun = function(css_selector) {
        request_id <- rand_string()

        query_requests$set(request_id, list(
          request_id = request_id,
          request_timestamp = Sys.time(),
          selector = css_selector,
          result = NULL
        ))

        session$sendCustomMessage("shidashi.query_ui", list(
          selector = css_selector,
          request_id = request_id,
          input_id = session$ns("@shiny_query_ui_result@")
        ))

        # Wait for the browser to store the result in the map — up to 10 x 1s.
        # Doing the wait in the sender avoids the race where the client calls
        # shiny_query_ui_result before Shiny has flushed the reactive observer.
        coro::async(function() {
          for (i in seq_len(10)) {
            coro::async_sleep(1)
            if (query_requests$has(request_id) &&
                !is.null(query_requests$get(request_id)$result)) {
              break
            }
          }
          paste0(
            "Request registered (id: ", request_id, "). ",
            "Call `tool__shiny_query_ui_result(request_id = \"", request_id, "\")` ",
            "to retrieve the result."
          )
        })()
      }
    )

    shiny_query_ui_result <- ellmer::tool(
      name = "shiny_query_ui_result",
      description = paste(
        "Fetch the result of a previous `shiny_query_ui` request.",
        "Returns the innerHTML of the matched element, or an inline image",
        "if the element contains <img> or <canvas>.",
        "An optional text note may be appended with additional context",
        "(e.g. outerHTML, not-found message, or hidden-element content).",
        "If the result is not yet ready, wait ~0.5 s and call again."
      ),
      arguments = list(
        request_id = ellmer::type_string(
          description = "The request_id returned by a prior `shiny_query_ui` call.",
          required = TRUE
        )
      ),
      fun = function(request_id) {
        # 1. Unknown request_id
        if (!query_requests$has(request_id)) {
          stop(
            "Unknown request_id: '", request_id, "'. ",
            "It may have already been consumed or was never created. ",
            "Call `shiny_query_ui` first to register a new request."
          )
        }

        entry <- query_requests$get(request_id)
        res <- entry$result

        # 2. Result available — consume and return
        if (!is.null(res)) {
          query_requests$remove(request_id)
          # note is an optional free-text annotation added by newer JS versions
          # (e.g. outerHTML context, not-found message, hidden-element innerHTML).
          # Older JS omits it entirely; R does not branch on it — agents interpret it.
          note <- res$note %||% ""

          # Primary dispatch: image_data present → inline image
          if (length(res$image_data) == 1 && nzchar(res$image_data)) {
            mime <- res$image_type %||% "image/png"
            img <- ellmer::ContentImageInline(type = mime, data = res$image_data)
            if (nzchar(note)) {
              return(list(img, ellmer::ContentText(note)))
            }
            return(img)
          }

          # Default: return html string, append note when present
          html <- res$html %||% ""
          if (nzchar(note)) {
            html <- paste(
              c(html, "\n\n<!-- NOTE: ", note, "-->"),
              collapse = "\n"
            )
          }
          return(html)
        }

        # 3. Not yet available — check age
        elapsed <- as.double(
          difftime(Sys.time(), entry$request_timestamp, units = "secs")
        )
        if (elapsed > 30) {
          stop(
            "No response from browser for selector '",
            entry$selector, "' within 30 seconds. ",
            "The selector is most likely invalid (no matching results)."
          )
        }
        # Ask the client to wait briefly and retry
        coro::async(function() {
          coro::async_sleep(1)
          paste0(
            "Result not yet available (", round(elapsed, 1), "s elapsed). ",
            "Wait ~0.5 s and call `tool__shiny_query_ui_result(request_id = \"",
            request_id, "\")` again."
          )
        })()
      }
    )

    shiny_output_info <- ellmer::tool(
      name = "shiny_output_info",
      description = paste(
        "List registered Shiny output elements and optionally retrieve",
        "their rendered HTML content. When outputIds is omitted, returns",
        "all registered outputs with their descriptions. You can get the",
        "HTML content of output via `shiny_query_ui(selector)`"
      ),
      arguments = list(
        outputIds = ellmer::type_array(
          ellmer::type_string(description = "Shiny output ID"),
          description = "Optional: specific output IDs to query. Omit to list all registered outputs.",
          required = FALSE
        )
      ),
      fun = function(outputIds = character()) {
        outputIds <- unlist(outputIds)
        outputIds <- outputIds[!is.na(outputIds)]
        if (length(outputIds) > 0) {
          items <- output_specs$mget(outputIds)
        } else {
          items <- output_specs$as_list()
        }

        results <- lapply(items, function(item) {
          if (is.null(item)) return(NULL)
          as.list(item)
        })

        if (is.null(session)) {
          return(results)
        }

        results <- lapply(results, function(item) {
          item$css_selector <- sprintf("#%s", session$ns(item$outputId))
          item
        })

        return(results)
      }
    )

    list(
      shiny_input_info = shiny_input_info,
      shiny_input_update = shiny_input_update,
      shiny_query_ui = shiny_query_ui,
      shiny_query_ui_result = shiny_query_ui_result,
      shiny_output_info = shiny_output_info
    )
  })



  list(
    input_helpers = list(
      register_input_specification = register_input_spec,
      register_output_specification = register_output_spec,
      update_input_specification = update_input_spec,
      get_input_specification = get_input_spec
    ),
    tool_generator = wrapper
  )

}

find_expr <- function(call, env) {
  if (!is.call(call)) {
    stop("find_expr needs a function call, got:\n", deparse1(call))
    return(NULL)
  }
  fn_symbol <- call[[1L]]
  # Resolve the actual function object
  fn <- NULL

  if (
    identical(fn_symbol, quote(shiny::bindEvent)) ||
    identical(fn_symbol, quote(bindEvent))
  ) {
    call <- match.call(definition = shiny::bindEvent, call = call)
    call <- call$x
    fn_symbol <- call[[1L]]
  }

  if (is.call(fn_symbol) && identical(fn_symbol[[1L]], quote(`::`))) {
    # pkg::fun(...) form
    pkg <- as.character(fn_symbol[[2L]])
    fun_name <- as.character(fn_symbol[[3L]])
    ns <- asNamespace(pkg)
    fn <- ns[[fun_name]]
  } else {
    # try to get the function
    tryCatch(
      {
        fn <- eval(fn_symbol, envir = new.env(parent = env))
      },
      error = function(e) {}
    )
  }

  if (is.function(fn)) {
    call <- match.call(fn, call)
  }

  call <- as.list(call)

  expr <- call[["expr"]]
  if (is.null(expr)) {
    # First positional argument
    expr <- call[[2]]
  }

  expr
}


# Internal helper: set up download handler, popout handler, and send
# output metadata to JS. Called by register_output().
register_output_widgets <- function(
  render_expr,
  render_env,
  outputId,
  download_type = "image",
  download_function = NULL,
  output_opts = list(),
  extension = NULL,
  description = "",
  session = shiny::getDefaultReactiveDomain()
) {

  if (is.null(session)) { return(invisible()) }

  ns <- session$ns
  ns_outputId <- ns(outputId)
  input <- session$input

  # Determine enabled widgets
  widgets <- "popout"
  if (!identical(download_type, "no-download")) {
    widgets <- c("download", widgets)
  }

  # --- Download handler setup ---
  if ("download" %in% widgets) {
    download_ns_id <- ns(paste0(outputId, "__download"))
    trigger_ns_id <- paste0(ns_outputId, "__download_trigger")
    modal_prefix <- paste0(outputId, "__dlmodal_")

    shiny::bindEvent(
      shiny::observe({
        # Build modal UI based on download_type
        modal_ui <- switch(
          download_type,
          "image" = shiny::tagList(
            shiny::numericInput(
              inputId = ns(paste0(modal_prefix, "width")),
              label = "Width (cm)",
              value = 30, min = 1, max = 200, step = 0.1
            ),
            shiny::numericInput(
              inputId = ns(paste0(modal_prefix, "height")),
              label = "Height (cm)",
              value = 20, min = 1, max = 200, step = 0.1
            ),
            shiny::textInput(
              inputId = ns(paste0(modal_prefix, "filename")),
              label = "Filename",
              value = paste0(outputId, "_", format(Sys.time(), "%Y%m%d_%H%M%S"))
            ),
            shiny::downloadButton(
              outputId = download_ns_id,
              label = "Download"
            )
          ),
          "threeBrain" = shiny::tagList(
            shiny::textInput(
              inputId = ns(paste0(modal_prefix, "filename")),
              label = "Filename",
              value = paste0(outputId, "_", format(Sys.time(), "%Y%m%d_%H%M%S"))
            ),
            shiny::textInput(
              inputId = ns(paste0(modal_prefix, "title")),
              label = "Title",
              value = "RAVE Viewer"
            ),
            shiny::downloadButton(
              outputId = download_ns_id,
              label = "Download"
            )
          ),
          "data" = shiny::tagList(
            shiny::textInput(
              inputId = ns(paste0(modal_prefix, "filename")),
              label = "Filename",
              value = paste0(outputId, "_", format(Sys.time(), "%Y%m%d_%H%M%S"))
            ),
            shiny::downloadButton(
              outputId = download_ns_id,
              label = "Download"
            )
          ),
          "htmlwidget" = shiny::tagList(
            shiny::textInput(
              inputId = ns(paste0(modal_prefix, "filename")),
              label = "Filename",
              value = paste0(outputId, "_", format(Sys.time(), "%Y%m%d_%H%M%S"))
            ),
            shiny::checkboxInput(
              inputId = ns(paste0(modal_prefix, "self_contained")),
              label = "Self-contained",
              value = TRUE
            ),
            shiny::downloadButton(
              outputId = download_ns_id,
              label = "Download"
            )
          ),
          "stream_viz" = shiny::tagList(
            shiny::textInput(
              inputId = ns(paste0(modal_prefix, "filename")),
              label = "Filename",
              value = paste0(outputId, "_", format(Sys.time(), "%Y%m%d_%H%M%S"))
            ),
            shiny::downloadButton(
              outputId = download_ns_id,
              label = "Download"
            )
          ),
          # fallback
          shiny::tagList(
            shiny::downloadButton(
              outputId = download_ns_id,
              label = "Download"
            )
          )
        )

        shiny::showModal(shiny::modalDialog(
          title = if (nzchar(description)) description else paste("Download", outputId),
          modal_ui,
          easyClose = TRUE,
          footer = shiny::modalButton("Cancel")
        ), session = session)
      }, domain = session),
      input[[paste0(outputId, "__download_trigger")]],
      ignoreNULL = TRUE, ignoreInit = TRUE
    )

    # Download handler
    session$output[[paste0(outputId, "__download")]] <- shiny::downloadHandler(
      filename = function() {
        fname <- input[[paste0(modal_prefix, "filename")]]
        if (!length(fname) || !nzchar(fname)) {
          fname <- paste0(outputId, "_", format(Sys.time(), "%Y%m%d_%H%M%S"))
        }
        switch(
          download_type,
          "image" = paste0(fname, ".pdf"),
          "threeBrain" = paste0(fname, ".html"),
          "htmlwidget" = paste0(fname, ".html"),
          "stream_viz" = paste0(fname, ".bin"),
          {
            if (length(extension)) {
              extension <- gsub("^[\\.]{0,}", ".", extension)
            } else {
              extension <- ""
            }
            paste0(fname, extension)
          }

        )
      },
      content = function(file) {
        switch(
          download_type,
          "image" = {
            width_cm <- input[[paste0(modal_prefix, "width")]]
            height_cm <- input[[paste0(modal_prefix, "height")]]
            if (!length(width_cm) || is.na(width_cm)) {
              width_cm <- 30
            }
            if (!length(height_cm) || is.na(height_cm)) {
              height_cm <- 20
            }
            width_in <- width_cm / 2.54
            height_in <- height_cm / 2.54

            grDevices::pdf(
              file,
              onefile = TRUE,
              useDingbats = FALSE,
              width = width_in,
              height = height_in
            )

            tryCatch(
              {
                res <- eval(render_expr, envir = new.env(parent = render_env))
                if (inherits(res, "ggplot")) {
                  # render ggplot
                  print(res)
                }
              },
              finally = {
                grDevices::dev.off()
              }
            )
          },
          "threeBrain" = {
            tb <- asNamespace("threeBrain")
            title <- input[[paste0(modal_prefix, "title")]]
            if (!length(title) || !nzchar(title)) {
              title <- "RAVE Viewer"
            }
            widget <- eval(render_expr, envir = new.env(parent = render_env))
            tb$save_brain(widget, file, title = title)
          },
          "htmlwidget" = {
            widget <- eval(render_expr, envir = new.env(parent = render_env))
            htmlwidgets::saveWidget(widget, file, selfcontained = TRUE)
          },
          "data" = {
            if (is.function(download_function)) {
              download_function(file)
            }
          },
          "stream_viz" = {
            bin_path <- stream_path(outputId, session)
            if (file.exists(bin_path)) {
              file.copy(bin_path, file, overwrite = TRUE)
            }
          }
        )

        # download finishes, dismiss the modal
        shiny::removeModal()
      }
    )
  }

  # --- Send output metadata to JS ---
  session$sendCustomMessage("shidashi.register_output_widgets", list(
    outputId = ns_outputId,
    widgets = as.list(widgets),
    download_type = download_type,
    token = session$token
  ))

  # --- Store render info in session registry ---
  entry <- get_session_entry(session$token)
  if (!is.null(entry)) {
    entry$output_renderers$set(outputId, list(
      render_expr = render_expr,
      render_env = render_env,
      output_opts = output_opts,
      extension = extension,
      download_type = download_type
    ))
  }

  invisible()
}

#' @name register_io
#' @title Register Shiny Inputs and Outputs for \verb{MCP} Access
#' @description
#' Register \code{shiny} inputs and outputs for \verb{MCP} (Model Context
#' Protocol) agent access.
#'
#' \code{register_input()} wraps a \code{shiny} input constructor to
#' register metadata.  It evaluates \code{expr} and returns the UI element.
#'
#' \code{register_output()} is a server-side function that registers a
#' render function call (e.g. \code{renderPlot(\{...\})}), assigns it to
#' \code{session$output}, registers the \verb{MCP} output spec, and sets
#' up download-widget handlers.  The UI overlay icons are injected
#' entirely by \verb{JS}.
#'
#' @param expr For \code{register_input}: a call expression that creates
#'   a \code{shiny} input widget.
#'   For \code{register_output}: a render function call such as
#'   \code{renderPlot(\{...\})}.
#' @param inputId character string.  The \code{shiny} input ID
#'   (without the module namespace prefix).
#' @param outputId character string.  The \code{shiny} output ID
#'   (without the module namespace prefix).
#' @param update character string.  The fully qualified update function,
#'   e.g. \code{"shiny::updateTextInput"}.  Field mappings such as
#'   \code{"shiny::updateSelectInput(value=selected)"} override the
#'   default argument names passed to the update function.
#' @param description character string.  A human-readable description of
#'   the input or output purpose, exposed to \verb{LLM} agents via \verb{MCP} tools.
#' @param writable logical (default \code{TRUE}).  Whether the \verb{MCP}
#'   update tool is allowed to change this input.
#' @param quoted logical (default \code{FALSE}).  If \code{TRUE},
#'   \code{expr} is treated as already quoted; otherwise it is captured
#'   with \code{substitute()}.
#' @param env the environment in which to evaluate \code{expr}.
#' @param ... reserved for future use.
#' @param output_opts a named list of extra options for the output
#'   (e.g. width, height defaults).
#' @param download_function a custom download handler function.  When
#'   \code{download_type = "data"}, this function receives the file path
#'   and writes the download content.
#' @param download_type character string.  One of \code{"image"},
#'   \code{"threeBrain"}, \code{"data"}, or \code{"no-download"}.
#' @param extension character vector of allowed file extension for
#'   download, or \code{NULL}.
#' @param session the \code{shiny} session object.  For
#'   \code{register_output}, defaults to
#'   \code{shiny::getDefaultReactiveDomain()}.
#' @return \code{register_input} returns the evaluated UI element.
#'   \code{register_output} is called for its side effects (assigning
#'   the render function and registering widgets) and returns \code{NULL}
#'   invisibly.
#' @seealso \code{\link{init_app}}, \code{\link{mcp_wrapper}}
#' @examples
#' \dontrun{
#' # inside a shidashi module UI function:
#' ns <- shiny::NS("demo")
#'
#' register_input(
#'   expr = shiny::sliderInput(
#'     inputId = ns("threshold"),
#'     label = "Threshold",
#'     min = 0, max = 1, value = 0.5
#'   ),
#'   inputId = "threshold",
#'   update = "shiny::updateSliderInput",
#'   description = "Filter threshold for the plot"
#' )
#'
#' # inside a shidashi module server function:
#' register_output(
#'   expr = renderPlot(\{ plot(iris) \}),
#'   outputId = "my_plot",
#'   description = "Scatter plot of iris data",
#'   download_type = "image"
#' )
#' }
#' @export
register_input <- function(expr,
                           inputId,
                           update,
                           description = "",
                           writable = TRUE,
                           quoted = FALSE,
                           env = parent.frame()) {
  if (!quoted) {
    expr <- substitute(expr)
  }

  register_input_impl <- get0(
    x = ".register_input",
    envir = env,
    mode = "function",
    inherits = TRUE
  )

  if (isTRUE(inherits(register_input_impl, "register_input_impl"))) {
    register_input_impl(
      expr = expr,
      inputId = inputId,
      description = description,
      update = update,
      writable = writable,
      quoted = TRUE,
      env = env
    )
  } else {
    eval(expr, envir = env)
  }

}

#' @rdname register_io
#' @export
register_output <- function(
  expr,
  outputId,
  description = "",
  quoted = FALSE,
  env = parent.frame(),
  ...,
  output_opts = list(),
  download_function = NULL,
  download_type = c("image", "htmlwidget", "threeBrain", "no-download", "data", "stream_viz"),
  extension = NULL,
  session = shiny::getDefaultReactiveDomain()
) {

  if (is.null(session)) {
    stop("shidashi::register_output must run in a shiny server function.")
  }

  download_type <- match.arg(download_type)

  if (!quoted) {
    expr <- substitute(expr)
  }

  # Parse the render call to extract the inner expr
  parsed_expr <- find_expr(expr, env)

  # Register MCP output spec (records the call, does NOT eval)
  register_output_impl <- get0(
    x = ".register_output",
    envir = env,
    mode = "function",
    inherits = TRUE
  )

  if (isTRUE(inherits(register_output_impl, "register_output_impl"))) {
    register_output_impl(
      expr = expr,
      outputId = outputId,
      description = description,
      quoted = TRUE,
      env = env
    )
  }

  # Set up download/popout widgets
  register_output_widgets(
    render_expr = parsed_expr,
    render_env = env,
    outputId = outputId,
    download_type = download_type,
    download_function = download_function,
    output_opts = output_opts,
    extension = extension,
    description = description,
    session = session
  )

  # Evaluate the full render call and assign to session output
  render_function <- eval(expr, envir = env)
  if (!is.null(session)) {
    session$output[[outputId]] <- render_function
  }

  invisible(NULL)
}

test_that("mcp_wrapper validates generator is a function", {
  expect_error(mcp_wrapper("not a function"), "generator must be a function")
})

test_that("mcp_wrapper validates generator accepts session param", {
  expect_error(
    mcp_wrapper(function(x) NULL),
    "generator must accept"
  )
})

test_that("mcp_wrapper returns a function with correct class", {
  gen <- function(session) {
    ellmer::tool(
      fun = function() "hello",
      name = "test_tool",
      description = "A test tool"
    )
  }
  wrapped <- mcp_wrapper(gen)
  expect_s3_class(wrapped, "shidashi_mcp_wrapper")
  expect_true(is.function(wrapped))
})

test_that("mcp_wrapper normalises single ToolDef to list", {
  gen <- function(session) {
    ellmer::tool(
      fun = function() "hello",
      name = "single_tool",
      description = "Returns a single tool"
    )
  }
  wrapped <- mcp_wrapper(gen)
  result <- wrapped(session = NULL)
  expect_type(result, "list")
  expect_length(result, 1)
  expect_true(inherits(result[[1]], "ellmer::ToolDef"))
})

test_that("mcp_wrapper filters out non-ToolDef objects", {
  gen <- function(session) {
    list(
      ellmer::tool(
        fun = function() "valid",
        name = "valid_tool",
        description = "A valid tool"
      ),
      "not a tool",
      42,
      NULL
    )
  }
  wrapped <- mcp_wrapper(gen)
  result <- wrapped(session = NULL)
  expect_length(result, 1)
  expect_equal(result[[1]]@name, "valid_tool")
})

test_that("mcp_wrapper returns empty list when generator returns no tools", {
  gen <- function(session) {
    list("no tools here", 123)
  }
  wrapped <- mcp_wrapper(gen)
  result <- wrapped(session = NULL)
  expect_type(result, "list")
  expect_length(result, 0)
})

library(shiny)
library(shidashi)

# ---- Base R plot ----

ui_base_plot <- function() {
  column(
    width = 6L,
    card(
      title = "Base R Plot",
      class_body = "min-height-450",
      tools = list(
        card_tool(widget = "maximize")
      ),
      plotOutput(ns("base_plot"), height = "400px")
    )
  )
}

server_base_plot <- function(input, output, session, ...) {
  shidashi::register_session(session)

  register_output(
    renderPlot({
      theme <- shidashi::get_theme()
      par(
        bg = theme$background, fg = theme$foreground,
        col.lab = theme$foreground, col.main = theme$foreground,
        col.axis = theme$foreground
      )
      x <- seq(0, 4 * pi, length.out = 200)
      plot(x, sin(x), type = "l", lwd = 2,
           col = "steelblue", main = "Sine Wave",
           xlab = "x", ylab = "sin(x)")
      lines(x, cos(x), col = "coral", lwd = 2)
      legend("topright",
             legend = c("sin", "cos"),
             col = c("steelblue", "coral"),
             lwd = 2, bg = theme$background,
             text.col = theme$foreground)
    }),
    outputId = "base_plot",
    description = "Sine and cosine wave plot (base R graphics)",
    download_type = "image"
  )
}

# ---- ggplot with brush ----

ui_ggplot_brush <- function() {
  column(
    width = 6L,
    card(
      title = "ggplot2 with Brush",
      class_body = "min-height-450",
      tools = list(
        card_tool(widget = "maximize")
      ),
      plotOutput(ns("gg_brush_plot"), height = "350px",
                 brush = brushOpts(id = ns("gg_brush"))),
      hr(),
      h5("Brushed Points"),
      verbatimTextOutput(ns("brush_info"))
    )
  )
}

server_ggplot_brush <- function(input, output, session, ...) {
  shidashi::register_session(session)

  register_output(
    renderPlot({
      theme <- shidashi::get_theme()
      ggtheme <- ggplot2::theme(
        panel.background = ggplot2::element_rect(
          fill = theme$background, color = theme$background),
        plot.background = ggplot2::element_rect(
          fill = theme$background, color = theme$background),
        panel.grid.major = ggplot2::element_blank(),
        panel.grid.minor = ggplot2::element_blank(),
        axis.line.x.bottom = ggplot2::element_line(color = theme$foreground),
        axis.line.y.left = ggplot2::element_line(color = theme$foreground),
        legend.key = ggplot2::element_rect(
          fill = theme$background, colour = theme$background),
        rect = ggplot2::element_rect(
          fill = theme$background, colour = theme$foreground),
        title = ggplot2::element_text(color = theme$foreground),
        text = ggplot2::element_text(color = theme$foreground),
        line = ggplot2::element_line(color = theme$foreground)
      )

      ggplot2::ggplot(datasets::mtcars,
                      ggplot2::aes(x = wt, y = mpg, color = factor(cyl))) +
        ggplot2::geom_point(size = 3) +
        ggplot2::labs(
          title = "Motor Trend Cars",
          x = "Weight (1000 lbs)",
          y = "Miles per Gallon",
          color = "Cylinders"
        ) + ggtheme
    }),
    outputId = "gg_brush_plot",
    description = "ggplot scatter of mtcars (wt vs mpg) with brush selection",
    output_opts = list(brush = brushOpts(id = ns("gg_brush"))),
    download_type = "image"
  )

  output$brush_info <- renderPrint({
    brush <- input$gg_brush
    if (is.null(brush)) {
      cat("Drag to select points on the plot above.\n")
    } else {
      pts <- brushedPoints(datasets::mtcars, brush,
                           xvar = "wt", yvar = "mpg")
      if (nrow(pts) == 0L) {
        cat("No points selected.\n")
      } else {
        print(pts[, c("mpg", "cyl", "wt", "hp")])
      }
    }
  })
}

# ---- threeBrain viewer ----

ui_threebrain <- function() {
  column(
    width = 12L,
    card(
      title = "threeBrain 3D Viewer",
      class_body = "min-height-450",
      tools = list(
        card_tool(widget = "maximize")
      ),
      threebrain_output_ui(ns("brain_viewer"), height = "500px"),
      hr(),
      h5("Double-click Event"),
      verbatimTextOutput(ns("brain_click_info"))
    )
  )
}

threebrain_output_ui <- function(outputId, height = "500px") {
  if (!requireNamespace("threeBrain", quietly = TRUE)) {
    return(shiny::div(
      class = "alert alert-warning",
      shiny::h4("threeBrain not installed"),
      shiny::p("Install with: ",
               shiny::code('install.packages("threeBrain")'))
    ))
  }
  threeBrain::threejsBrainOutput(outputId, height = height)
}

server_threebrain <- function(input, output, session, ...) {

  if (!requireNamespace("threeBrain", quietly = TRUE)) {
    output$brain_click_info <- renderPrint({
      cat("Package 'threeBrain' is required but not installed.\n")
    })
    return(invisible())
  }

  b <- threeBrain::merge_brain()
  b$template_object$set_electrodes(data.frame(
    Electrode = 1:10,
    x = stats::rnorm(10) * 20,
    y = stats::rnorm(10) * 20,
    z = stats::rnorm(10) * 20,
    Label = LETTERS[1:10]
  ))

  register_output(
    threeBrain::renderBrain({
      b$plot()
    }),
    outputId = "brain_viewer",
    description = "3D brain viewer with 10 random electrodes",
    download_type = "threeBrain"
  )

  proxy <- threeBrain::brain_proxy("brain_viewer")

  output$brain_click_info <- renderPrint({
    event <- proxy$mouse_event_double_click
    if (is.null(event)) {
      cat("Double-click an electrode in the 3D viewer.\n")
    } else {
      str(event)
    }
  })
}

library(shiny)
library(shidashi)

card_with_code <- function(expr, env = parent.frame(),
                           class = "height-50", width.cutoff = 25L) {
  expr <- substitute(expr)
  x <- eval(expr, envir = env)

  code <- clipboardOutput(
    clip_text = format_text_r(
      get_construct_string(x), quoted = TRUE,
      width.cutoff = width.cutoff
    ), class = "float-end", as_card_tool = TRUE
  )
  expr[["footer"]] <- code
  eval(expr, envir = env)
}

library(shiny)
library(shidashi)

# ===========================================================================
# QUICK START SECTION
# ===========================================================================

ui_quick_start <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Quick Start", class = "shidashi-anchor"),
      tags$div(
        class = "alert alert-info",
        tags$strong("2-Minute Setup: "),
        "Run one command in R, paste config into VS Code, done!"
      )
    ),
    column(
      width = 6L,
      card(
        title = "1. Run in R Console",
        tags$pre(
          class = "bg-gray-90 pre-compact",
          tags$code(
            class = "r",
            "shidashi:::setup_mcp_proxy()"
          )
        ),
        tags$p(
          class = "mb-0",
          "This prints a JSON config — copy it!"
        )
      )
    ),
    column(
      width = 6L,
      card(
        title = "2. Paste into VS Code",
        tags$p(
          "Create ", tags$code(".vscode/mcp.json"), " and paste the config."
        ),
        tags$p(
          class = "mb-0",
          "Restart VS Code. The shidashi MCP server is now available!"
        )
      )
    ),
    column(
      width = 12L,
      card(
        title = "3. Test It",
        tags$ol(
          class = "mb-0",
          tags$li("Start your shidashi app: ", tags$code("shidashi::render()")),
          tags$li("Open a module page in your browser"),
          tags$li(
            "In VS Code Copilot Chat, ask: ",
            tags$em("\"Use the hello_world tool to greet me\"")
          )
        )
      )
    )
  )
}


# ===========================================================================
# DETAILS SECTION (For those who want more info)
# ===========================================================================

ui_details <- function() {
  fluidRow(
    column(
      width = 12L,
      h3("Details", class = "shidashi-anchor"),
      p("Expand sections below for more configuration options.")
    ),
    column(
      width = 12L,
      accordion(
        id = ns("details_accordion"),

        # --- Architecture ---
        accordion_item(
          title = "How It Works",
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
"VS Code/Claude Code    (MCP Client)
       |
       | stdio (JSON-RPC)
       v
  mcp-proxy.mjs        (Node.js proxy)
       |
       | HTTP
       v
  Shiny /mcp           (Your app)"
            )
          ),
          tags$p(
            "The proxy translates stdio into HTTP. ",
            "It auto-discovers the active Shiny port from cached records."
          )
        ),

        # --- VS Code config ---
        accordion_item(
          title = "VS Code Configuration",
          tags$p(
            "Create ", tags$code(".vscode/mcp.json"), " with the path from ",
            tags$code("setup_mcp_proxy()"), ":"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "json",
'// .vscode/mcp.json
{
  "servers": {
    "shidashi": {
      "type": "stdio",
      "command": "node",
      "args": ["<PROXY_PATH>"]
    }
  }
}'
            )
          ),
          tags$p(
            "To target a specific port: ",
            tags$code('"args": ["<PROXY_PATH>", "8310"]')
          )
        ),

        # --- Claude Code config ---
        accordion_item(
          title = "Claude Code Configuration",
          tags$p(
            "Create ", tags$code(".mcp.json"), " in your project root:"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "json",
'// .mcp.json
{
  "mcpServers": {
    "shidashi": {
      "command": "node",
      "args": ["<PROXY_PATH>"]
    }
  }
}'
            )
          ),
          tags$p("Or use ", tags$code("~/.claude/mcp.json"), " for global config.")
        ),

        # --- Find proxy path ---
        accordion_item(
          title = "Finding the Proxy Path",
          tags$p(
            "The path is platform-specific. Use R to find it:"
          ),
          tags$pre(
            class = "bg-gray-90 pre-compact",
            tags$code(
              class = "r",
'# Method 1: setup_mcp_proxy() prints and returns the path
proxy_path <- shidashi:::setup_mcp_proxy()

# Method 2: compute it directly
proxy_path <- file.path(
  tools::R_user_dir("shidashi", "cache"),
  "mcp_server", "mcp-proxy.mjs"
)'
            )
          )
        ),

        # --- Troubleshooting ---
        accordion_item(
          title = "Troubleshooting",
          tags$dl(
            tags$dt("\"No tools available\""),
            tags$dd("Open a module page first. The Shiny session must be active."),

            tags$dt("\"Connection refused\""),
            tags$dd(
              "The Shiny app isn't running. Start it with ",
              tags$code("shidashi::render()"), "."
            ),

            tags$dt("Tools appear but calls fail"),
            tags$dd(
              "Make sure the agent is in ", tags$strong("Executing"), " mode ",
              "and bound to a session."
            ),

            tags$dt("Proxy not found"),
            tags$dd(
              "Run ", tags$code("shidashi:::setup_mcp_proxy()"),
              " again in R."
            )
          )
        )
      )
    )
  )
}


# ===========================================================================
# SERVER
# ===========================================================================

server_mcpsetup <- function(input, output, session, ...) {
  # No server logic needed for this documentation module
}

infobox_with_code <- function(x, title = "",
                              class = "height-100",
                              class_body = "padding-5 bg-gray-70") {
  shiny::div(
    class = "fill-width position-relative",
    shiny::span(
      class = "position-absolute padding-bottom-5 bg-gray",
      style = "right: 0; z-index: 100;",
      clipboardOutput(clip_text = get_construct_string(x),
                      as_card_tool = TRUE, message = "Copy to clipboard")
    ),
    x
  )
}

#' Render a 'shidashi' project
#' @param root_path the project path, default is the demo folder from
#' \code{template_root()}
#' @param ... additional parameters passed to \code{\link[shiny]{runApp}},
#' such as \code{host}, \code{port}
#' @param launch_browser whether to launch browser; default is \code{TRUE}
#' @param as_job whether to run as 'RStudio' jobs; this options is only
#' available when 'RStudio' is available
#' @param test_mode whether to test the project; this options is helpful when
#' you want to debug the project without relaunching shiny applications
#' @param prelaunch expression to execute before launching the session; the
#' expression will execute in a brand new session
#' @param prelaunch_quoted whether the expression is quoted; default is false
#' @return This functions runs a 'shiny' application, and returns the job id
#' if 'RStudio' is available.
#'
#' @examples
#'
#' template_root()
#'
#' if(interactive()) {
#'   render()
#' }
#'
#' @export
render <- function(
  root_path = template_root(),
  ...,
  prelaunch = NULL,
  prelaunch_quoted = FALSE,
  launch_browser = TRUE,
  as_job = TRUE,
  test_mode = getOption("shiny.testmode", FALSE)
) {
  if (!dir.exists(root_path)) {
    stop("`root_path` cannot be found: ", root_path)
  }
  root_path <- normalizePath(root_path, mustWork = TRUE, winslash = "/")

  if (!prelaunch_quoted) {
    prelaunch <- substitute(prelaunch)
  }

  # Resolve port: use caller-supplied port= or pick a random one.
  dots <- list(...)
  mcp_port <- dots[["port"]]
  if (is.null(mcp_port)) {
    mcp_port <- httpuv::randomPort()
  }
  dots[["port"]] <- mcp_port

  # Write port record and keep proxy up-to-date in user cache.
  setup_mcp_proxy(port = mcp_port, overwrite = TRUE, verbose = FALSE)

  # Copy global.R from inst/ so that shinyAppDir sources it at startup.
  # global.R calls shidashi::init_app() to create per-application state.
  global_src <- system.file("global.R", package = "shidashi")
  if (nzchar(global_src)) {
    file.copy(global_src, file.path(root_path, "global.R"), overwrite = TRUE)
  }

  # Write template_settings$set into ui.R so shinyAppDir picks up the correct
  # root_path regardless of working directory or how the app is launched.
  writeLines(
    c(
      sprintf('shidashi::template_settings$set("root_path" = "%s")', root_path),
      "shidashi::adminlte_ui()"
    ),
    file.path(root_path, "ui.R")
  )

  if (!as_job || system.file(package = "rstudioapi") == "" ||
     !rstudioapi::isAvailable(version_needed = "1.4.1717",
                              child_ok = FALSE)) {

    shidashi::template_settings$set("root_path" = root_path)
    eval(prelaunch, envir = new.env(parent = globalenv()))

    # Use shinyAppDir so that ui.R / server.R are loaded normally, then
    # chain the MCP handler in front of Shiny's built-in httpHandler.
    app <- register_mcp_route(shiny::shinyAppDir(root_path))
    do.call(shiny::runApp, c(
      list(appDir = app, launch.browser = launch_browser, test.mode = test_mode),
      dots
    ))
  } else {
    script <- file.path(root_path, "_rs_job.R")

    # Build app object in the job script so httpHandler is attached
    run_call <- as.call(c(
      list(quote(shiny::runApp)),
      list(
        appDir = quote(app),
        launch.browser = launch_browser,
        test.mode = test_mode
      ),
      dots
    ))
    s <- c(
      'options("crayon.enabled" = TRUE)',
      'options("crayon.colors" = 256)\n',
      deparse(prelaunch),
      "\n",
      sprintf("app <- shidashi:::register_mcp_route(shiny::shinyAppDir(\"%s\"))", root_path),
      deparse(run_call)
    )
    writeLines(
      con = script,
      s
    )
    rstudioapi::jobRunScript(
      path = script, workingDir = root_path,
      name = basename(root_path)
    )
  }

}

#' Template function to include 'snippets' in the view folder
#' @description Store the reusing 'HTML' segments in the
#' \code{views} folder. This function should be used in the
#' \code{'index.html'} template
#' @param file files in the template \code{views} folder
#' @param ... ignored
#' @param .env,.root_path internally used
#' @return rendered 'HTML' segments
#' @examples
#' \dontrun{
#' # in your 'index.html' file
#' <html>
#' <header>
#' {{ shidashi::include_view("header.html") }}
#' </header>
#' <body>
#'
#' </body>
#' <!-- Before closing html tag -->
#' {{ shidashi::include_view("footer.html") }}
#' </html>
#' }
#' @export
include_view <- function(file, ..., .env = parent.frame(),
                         .root_path = template_root()) {
  tryCatch({
    file <- normalizePath(file.path(.root_path, "views", file),
                          mustWork = TRUE)
  }, error = function(e) {
    stop(call. = NULL, "Cannot find views/", file)
  })
  list2env(list(.env = .env), envir = .GlobalEnv)
  args <- NULL
  if (is.environment(.env$env)) {
    args <- get0("@args", envir = .env$env, ifnotfound = NULL)
  }
  if (!is.list(args)) {
    args <- as.list(.env)
  }
  # more_args <- list(...)
  # for(nm in names(more_args)) {
  #   if (nm != "") {
  #     args[[nm]] <- more_args[[nm]]
  #   }
  # }
  argnames <- names(args)
  args <- args[!argnames %in% c("headContent", "suppressDependencies", "filename", "document_", "text_")]
  call <- as.call(c(
    list(quote(shiny::htmlTemplate),
         filename = file,
         document_ = FALSE),
    args
  ))
  return(eval(call, envir = .env))
  # shiny::htmlTemplate(file, ..., document_ = FALSE)
}

#' Reset shiny outputs with messages
#' @description Forces outdated output to reset and show a silent message.
#' @param outputId output ID
#' @param message output message
#' @param session shiny reactive domain
#' @return No value
#' @export
reset_output <- function(outputId, message = "This output has been reset",
                         session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.reset_output", list(
    outputId = session$ns(outputId),
    message = message
  ))
}

{
    library(shiny)
    server <- function(input, output, session) {
        shiny::observeEvent(session$clientData$url_search, {
            req <- list(QUERY_STRING = session$clientData$url_search)
            resource <- shidashi::load_module(request = req)
            if (resource$has_module) {
                module_table <- shidashi::module_info()
                module_table <- module_table[module_table$id %in% 
                  resource$module$id, ]
                if (nrow(module_table)) {
                  group_name <- as.character(module_table$group[[1]])
                  if (is.na(group_name)) {
                    group_name <- "<no group>"
                  }
                  if (system.file(package = "logger") != "") {
                    logger::log_info("Loading - { module_table$label[1] } ({group_name}/{ module_table$id })")
                  }
                  shiny::moduleServer(resource$module$id, resource$module$server, 
                    session = session)
                }
            }
        })
    }
}

#' Output placeholder for a streaming visualization widget
#'
#' Use in a Shiny UI to reserve a slot for a multi-channel signal viewer
#' that is driven by binary stream data.
#'
#' @param outputId Character scalar. Output ID that matches the corresponding
#'   \code{\link{renderStreamViz}} call.
#' @param width,height CSS width and height of the widget container.
#' @return An HTML output element suitable for inclusion in a Shiny UI.
#' @seealso \code{\link{renderStreamViz}}, \code{\link{updateStreamViz}}
#' @export
streamVizOutput <- function(outputId, width = "100%", height = "400px") {
  htmlwidgets::shinyWidgetOutput(
    outputId, "stream_viz",
    width = width, height = height,
    package = "shidashi"
  )
}

#' Render a streaming widget
#'
#' Server-side render function for \code{\link{streamVizOutput}}.
#' The expression should evaluate to a \code{\link{stream_viz}} object.
#'
#' @param expr An R expression that returns a \code{\link{stream_viz}} widget.
#' @param env Environment in which to evaluate \code{expr}.
#' @param quoted Logical. Whether \code{expr} is already quoted.
#' @return A server-side render function for use with Shiny.
#' @seealso \code{\link{streamVizOutput}}, \code{\link{updateStreamViz}}
#' @export
renderStreamViz <- function(expr, env = parent.frame(), quoted = FALSE) {
  if (!quoted) expr <- substitute(expr)
  htmlwidgets::shinyRenderWidget(expr, streamVizOutput, env, quoted = TRUE)
}

#' Create a streaming widget
#'
#' Creates a multi-channel signal viewer widget that fetches its data from
#' the binary stream file produced by \code{\link{stream_to_js}}.
#'
#' The \code{outputId} together with \code{session} is used to auto-compute
#' the token-qualified stream file identifier via
#' \code{\link{stream_file_id}}.  You can override this by passing an
#' explicit \code{stream_id}.
#'
#' Use \code{\link{updateStreamViz}} for in-place updates without recreating
#' the widget.
#'
#' @param outputId Character scalar.  The output ID matching the
#'   corresponding \code{\link{streamVizOutput}} call.  Used to compute
#'   \code{stream_id} automatically when \code{stream_id} is \code{NULL}.
#' @param session Shiny session object.  Defaults to the active reactive
#'   domain.  Together with \code{outputId} determines the token-qualified
#'   stream file identifier.
#' @param stream_id Character scalar or \code{NULL}.  When \code{NULL}
#'   (default) the stream identifier is computed as
#'   \code{stream_file_id(outputId, session, token = token)}.  Provide an
#'   explicit value to override for non-Shiny or advanced use.
#' @param width,height Widget dimensions (passed to \code{\link[htmlwidgets]{createWidget}}).
#' @param refresh_rate Numeric scalar (milliseconds). Polling interval used
#'   when the viewer is in \emph{active streaming} mode (play button pressed).
#'   Lower values give smoother animation but increase CPU/network load.
#'   Default \code{33} (approximately 30 Hz).
#' @param show_controls Logical. Whether to display the hover toolbar
#'   (play/pause, zoom, reset, export).  Default \code{TRUE}.
#' @param streaming Logical.  Whether to start active streaming immediately
#'   when the widget is rendered.  Default \code{FALSE}; set to \code{TRUE}
#'   to begin polling for new data automatically.  Can also be toggled later
#'   via \code{\link{updateStreamViz}}.
#' @param token Character scalar or \code{NULL}.  When \code{NULL} (default)
#'   the session's own token is used.  Override with a parent session token
#'   when running inside a standalone viewer (pop-out window) so that the
#'   child session fetches the parent's binary stream file.
#' @return An \pkg{htmlwidgets} object.
#' @seealso \code{\link{streamVizOutput}}, \code{\link{renderStreamViz}},
#'   \code{\link{updateStreamViz}}, \code{\link{stream_file_id}}
#' @export
stream_viz <- function(outputId, session = shiny::getDefaultReactiveDomain(),
                       stream_id = NULL, width = NULL, height = NULL,
                       refresh_rate = 33, show_controls = TRUE,
                       streaming = FALSE, token = NULL) {
  if (is.null(stream_id)) {
    stream_id <- stream_file_id(outputId, session, token = token)
  }
  htmlwidgets::createWidget(
    name = "stream_viz",
    x = list(
      stream_id = stream_id,
      refresh_rate = refresh_rate,
      show_controls = show_controls,
      streaming = isTRUE(streaming)
    ),
    width = width,
    height = height,
    package = "shidashi"
  )
}

#' Build the token-qualified stream file identifier
#'
#' Combines \code{session$token} with \code{id} to produce the string used
#' both by \code{\link{stream_path}} (as the filename stem) and by the browser
#' as the URL path component under \code{stream/}.
#'
#' @param id Character scalar. The base stream identifier (no path separators).
#' @param session Shiny session object. Defaults to the active reactive domain.
#' @param token Character scalar or \code{NULL}.  When \code{NULL} (default)
#'   the session's own token is used.  Override with a parent session token
#'   so that a standalone viewer (child session) can address the parent's
#'   binary stream file.
#' @return Character scalar: \code{"{token}_{id}"}.
#' @seealso \code{\link{stream_path}}, \code{\link{stream_viz}},
#'   \code{\link{updateStreamViz}}
#' @export
stream_file_id <- function(id, session = shiny::getDefaultReactiveDomain(),
                           token = NULL) {
  if (is.null(token)) {
    token <- if (!is.null(session)) session$token else "static"
  }
  nsid <- if (!is.null(session)) session$ns(id) else id
  paste0(token, "_", nsid)
}

#' Trigger an in-place update of a streaming widget
#'
#' Sends a custom Shiny message that causes the browser to re-fetch the binary
#' stream file and re-render the signal viewer without tearing down and
#' recreating the widget.  Call this after writing new data with
#' \code{\link{stream_to_js}}.  The \code{outputId} is used both
#' to locate the widget as the stream file identifier — it must
#' match the \code{id} passed to \code{\link{stream_path}}.
#'
#' @param session Shiny session object. Defaults to the active reactive domain.
#' @param outputId Character scalar. The output ID passed to
#'   \code{\link{streamVizOutput}} (and to \code{\link{stream_path}}).
#' @param streaming Logical or \code{NULL}.  \code{TRUE} starts active
#'   streaming (the widget polls for new data via \code{requestAnimationFrame});
#'   \code{FALSE} pauses it; \code{NULL} (default) leaves the current
#'   streaming state unchanged.
#' @param token Character scalar or \code{NULL}.  When \code{NULL} (default)
#'   the session's own token is used.  Override with a parent session token
#'   when running inside a standalone viewer (pop-out window) so that the
#'   child session fetches the parent's binary stream file.
#' @return Invisibly \code{NULL}.
#' @seealso \code{\link{streamVizOutput}}, \code{\link{stream_to_js}},
#'   \code{\link{stream_file_id}}
#' @export
updateStreamViz <- function(outputId, streaming = NULL, token = NULL,
                            session = shiny::getDefaultReactiveDomain()) {
  msg <- list(
    id = session$ns(outputId),
    stream_id = stream_file_id(outputId, session, token = token)
  )
  if (is.logical(streaming) && length(streaming) == 1L && !is.na(streaming)) {
    msg$streaming <- streaming
  }
  session$sendCustomMessage("stream_viz.render", msg)
}

ui_card_set <- function() {
  tagList(
    shiny::column(
      width = 4L,
      card_tabset(
        class_body = "height-450",
        inputId = ns("card_tabset_demo"),
        title = "House of cards",
        "Code" = flex_container(
          direction = "column",
          flex_item(
            class = "fill-height",
            p("Use key-value pairs to create tab-panels. The keys will be the panel name"),
            html_highlight_code(
              card_tabset(
                title = "House of cards",
                "Tab 1" = p("Tab content 1"),
                "Tab 2" = p("Tab content 2")
              ),
              class = "padding-5 margin-5",
              width.cutoff = 15L,
              hover = "overflow-auto"
            )
          ),
          flex_item(
            class = "fill-height",
            p("Use `names` argument. The length of `names` must be consistent with `...`"),
            html_highlight_code(
              card_tabset(
                title = "House of cards",
                p("Tab content 1"),
                p("Tab content 2"),
                names = c("Tab 1", "Tab 2")
              ),
              class = "padding-5 margin-5 fill",
              width.cutoff = 15L,
              hover = "overflow-auto"
            )
          )
        ),
        "Tab 1" = p("Tab content 1"),
        "Tab 2" = p("Tab content 2")
      )
    ),
    shiny::column(
      width = 4L,
      card_tabset(
        class_body = "height-450",
        inputId = ns("card_tabset_tools_demo"),
        title = "With Tools",
        tools = list(
          as_badge("New|badge-success"),
          card_tool(widget = "collapse"),
          card_tool(widget = "maximize")
        ),
        "Code" = flex_container(
          direction = "column",
          flex_item(
            class = "fill-height",
            html_highlight_code(
              card_tabset(
                title = "Cardset with Tools",
                "Tab 1" = p("Tab content 1"),
                class_body = "height-500",
                tools = list(
                  as_badge("New|badge-success"),
                  card_tool(widget = "collapse"),
                  card_tool(widget = "maximize")
                )
              ),
              class = "padding-5 margin-5",
              width.cutoff = 15L,
              hover = "overflow-auto"
            )
          )
        ),
        "Tab 1" = p("Tab content 1")
      )
    ),
    shiny::column(
      width = 4L,
      card_tabset(
        class_body = "height-450",
        inputId = ns("card_tabset_expand_demo"),
        title = "Self Expand",
        tools = list(
          card_tool(widget = "custom",
                    icon = shiny::icon("kiwi-bird"),
                    inputId = ns("add_card"))
        ),
        "Code" = flex_container(
          direction = "column",
          flex_item(
            class = "fill-height",
            p("Click on the button to add a new tab!"),
            html_highlight_code(
              card_tabset(
                inputId = ns("card_tabset_expand_demo"),
                title = "Cardset with Tools",
                "server.R" = p("..."),
                class_body = "height-500",
                tools = list(
                  card_tool(widget = "custom",
                            icon = shiny::icon("kiwi-bird"),
                            inputId = ns("add_card"))
                )
              ),
              class = "padding-5 margin-5",
              width.cutoff = 15L,
              hover = "overflow-auto"
            )
          )
        ),
        "server.R" = flex_container(
          direction = "column",
          flex_item(
            class = "fill-height",
            p("An observer listens to click event on `add_card`."),
            html_highlight_code(
              observeEvent(input$add_card, {
                if (input$add_card %% 2) {
                  card_tabset_insert(
                    inputId = "card_tabset_expand_demo",
                    title = "More...",
                    p("...")
                  )
                } else {
                  card_tabset_remove(
                    inputId = "card_tabset_expand_demo",
                    title = "More..."
                  )
                }
              }),
              class = "padding-5 margin-5",
              width.cutoff = 15L,
              hover = "overflow-auto"
            )
          )
        )
      )
    )
  )
}

# ---- MCP (Model Context Protocol) JSON-RPC 2.0 handler for shidashi ----
#
# This module implements a Streamable HTTP MCP endpoint that lives inside
# the same httpuv process as the Shiny app.  No sidecar process needed.
#
# Entry point: `mcp_http_handler(req)` is called from `mcp_app_handler()`
# when `req$PATH_INFO == "/mcp"` and `req$REQUEST_METHOD == "POST"`.
#
# Phase 3: Dynamic per-session tools with stateful MCP↔Shiny binding.
#
# Session registry: session$token -> entry
# Entry structure:
#   list(
#     shiny_session      = <ShinySession>,
#     shidashi_module_id = <string or NULL>,
#     mcp_session_ids    = c(<array of MCP session IDs>)
#     namespace          = "",
#     url                = "",
#     registered_at      = <POSIXct>,
#     tools              = <named list of ToolDef objects or NULL>
#   )

# ---------- MCP-specific session binding helpers --------------------------
# Generic helpers (register_session, unregister_session, sweep_closed_sessions,
# get_session_entry) are in globals.R.

# Given a MCP session ID, find shiny session tokens
mcp_tool_bound_shinysessions <- function(mcp_session_id) {
  if (length(mcp_session_id) != 1 || is.na(mcp_session_id) || !nzchar(mcp_session_id)) {
    return(character(0L))
  }
  registry <- globals_session_registry()
  tokens <- registry$keys()
  tokens[
    vapply(tokens, function(token) {
      entry <- registry$get(token)
      isTRUE(mcp_session_id %in% entry$mcp_session_ids)
    }, FUN.VALUE = FALSE)
  ]
}

# Given a MCP session ID, unregister the mcp session from shiny sessions
mcp_tool_unregister_shinysession <- function(mcp_session_id) {
  tokens <- mcp_tool_bound_shinysessions(mcp_session_id)
  if (!length(tokens)) { return() }

  registry <- globals_session_registry()
  lapply(tokens, function(token) {
    entry <- registry$get(token)
    entry$mcp_session_ids <- entry$mcp_session_ids[!entry$mcp_session_ids %in% mcp_session_id]
    registry$set(token, entry)
  })

  invisible(TRUE)
}

# ---------- app-level HTTP handler ----------------------------------------

# Create an HTTP handler for the MCP endpoint
register_mcp_route <- function(app) {
  default_handlers <- app$httpHandler
  mcp_handler <- function(req) {
    if (!identical(req$PATH_INFO, "/mcp")) return(NULL)
    if (identical(req$REQUEST_METHOD, "POST")) {
      return(mcp_http_handler(req))
    }
    if (identical(req$REQUEST_METHOD, "DELETE")) {
      # Clean up MCP session binding on teardown
      sid <- req$HTTP_MCP_SESSION_ID
      mcp_tool_unregister_shinysession(sid)
      return(shiny::httpResponse(200L, "application/json", "{}"))
    }
    if (identical(req$REQUEST_METHOD, "GET")) {
      # Informational response for browser / debugging
      info <- jsonlite::toJSON(
        list(
          status  = "ok",
          message = "shidashi MCP endpoint active. Use POST with JSON-RPC 2.0 to interact."
        ),
        auto_unbox = TRUE, null = "null"
      )
      return(shiny::httpResponse(200L, "application/json", info))
    }
    # Other methods
    shiny::httpResponse(405L, "", "")
  }

  app$httpHandler <- function(req) {
    response <- mcp_handler(req)
    if (!is.null(response)) {
      return(response)
    }
    return(default_handlers(req))
  }

  # Exclude /mcp from httpuv static-path handling so POST/DELETE
  # requests reach the R handler instead of being rejected with 400.
  app$staticPaths <- c(app$staticPaths, list(mcp = httpuv::excludeStaticPath()))

  return(app)
}

# ---------- public entry point -------------------------------------------

#' Handle an incoming MCP HTTP request
#'
#' Called from \code{adminlte_ui()} when \code{req$PATH_INFO == "/mcp"}.
#' Reads the JSON-RPC 2.0 body, dispatches to the appropriate handler,
#' and returns a \code{shiny::httpResponse()}.
#'
#' @param req The Rook request environment
#' @return A \code{shiny::httpResponse} object
#' @noRd
mcp_http_handler <- function(req) {

  # --- sweep stale Shiny sessions (defensive) ---------------------------
  sweep_closed_sessions()

  # --- read body --------------------------------------------------------
  mcp_session_id <- req$HTTP_MCP_SESSION_ID  # may be NULL on initialize
  body_raw <- tryCatch(
    req$rook.input$read(),
    error = function(e) raw(0)
  )
  if (length(body_raw) == 0L) {
    return(mcp_json_error(
      id = NULL,
      code = -32700L,
      message = "Parse error: empty body",
      mcp_session_id = mcp_session_id
    ))
  }

  body_text <- rawToChar(body_raw)
  msg <- tryCatch(
    jsonlite::fromJSON(body_text, simplifyVector = TRUE,
                       simplifyDataFrame = FALSE,
                       simplifyMatrix = FALSE),
    error = function(e) NULL
  )
  if (is.null(msg)) {
    return(mcp_json_error(
      id = NULL,
      code = -32700L,
      message = "Parse error: invalid JSON",
      mcp_session_id = mcp_session_id
    ))
  }

  # --- validate JSON-RPC envelope ---------------------------------------
  jsonrpc <- msg$jsonrpc
  method  <- msg$method
  id      <- msg$id
  params  <- msg$params

  # TODO: what if this is a parallel call?
  if (!identical(jsonrpc, "2.0") || !is.character(method) ||
      length(method) != 1L) {
    return(mcp_json_error(
      id = id,
      code = -32600L,
      message = "Invalid Request: missing jsonrpc or method",
      mcp_session_id = mcp_session_id
    ))
  }

  # --- notifications (no `id`) => 202 Accepted --------------------------
  is_notification <- is.null(id)

  if (is_notification) {
    # notifications/initialized is the only one we expect in Phase 1
    # TODO: should we notify in shiny?
    return(shiny::httpResponse(
      status = 202L,
      content_type = "",
      content = ""
    ))
  }

  # --- dispatch ----------------------------------------------------------


  result <- tryCatch(
    switch(
      method,
      "initialize"  = mcp_handle_initialize(id, params),
      "tools/list"  = mcp_handle_tools_list(id, params, mcp_session_id),
      "tools/call"  = mcp_handle_tools_call(id, params, mcp_session_id),
      "ping"        = mcp_handle_ping(id, mcp_session_id),
      # "resources/list" = mcp_handle_resources_list(id, params, mcp_session_id),
      # "resources/read" = mcp_...,
      # default: method not found
      mcp_json_error(
        id = id,
        code = -32601L,
        message = paste0("Method not found: ", method),
        mcp_session_id = mcp_session_id
      )
    ),
    error = function(e) {
      mcp_json_error(
        id = id,
        code = -32603L,
        message = paste0("Internal error: ", conditionMessage(e)),
        mcp_session_id = mcp_session_id
      )
    }
  )

  result
}

# ---------- method handlers ----------------------------------------------

#' Handle MCP `initialize` request
#' @keywords internal
#' @noRd
mcp_handle_initialize <- function(id, params) {

  # Generate a unique MCP session ID
  mcp_session_id <- digest::digest(
    list(Sys.time(), rand_string(), id, params,
         Sys.getpid(), sample.int(.Machine$integer.max, 1L)),
    algo = "sha256"
  )

  pkg_version <- tryCatch(
    as.character(utils::packageVersion("shidashi")),
    error = function(e) "0.0.0"
  )

  body <- jsonlite::toJSON(
    list(
      jsonrpc = "2.0",
      id      = id,
      result  = list(
        protocolVersion = "2025-03-26",
        capabilities    = list(
          tools   = list(listChanged = TRUE)
        ),
        serverInfo      = list(
          name    = "shidashi",
          version = pkg_version
        )
      )
    ),
    auto_unbox = TRUE,
    null = "null"
  )

  shiny::httpResponse(
    status       = 200L,
    content_type = "application/json",
    content      = body,
    headers      = list(`Mcp-Session-Id` = mcp_session_id)
  )
}

#' Handle MCP `ping` request
#' @keywords internal
#' @noRd
mcp_handle_ping <- function(id, mcp_session_id = NULL) {
  mcp_json_result(id, structure(list(), names = character(0L)), mcp_session_id)
}

# Pre-list shared tool & skill schemas (available before session binding).
# Reads agents/tool-schema.yaml (for tools), discovers skills from
# agents/skills/ via skill_wrapper(), and injects builtin shiny_* tools.
# No caching — always re-reads so changes are picked up immediately.
mcp_prelisted_tool_schemas <- function() {

  schemas <- list()

  # ---- Builtin shiny_* tool schemas (always available) ----
  schemas[["tool__shiny_input_info"]] <- list(
    name        = "tool__shiny_input_info",
    description = paste(
      "Query registered shiny input specifications.",
      "Returns input IDs, descriptions, types, update functions,",
      "whether each is writable, and (when a session is active)",
      "whether each currently exists and its current value."
    ),
    inputSchema = list(
      type       = "object",
      properties = list(
        inputIds = list(
          type        = "array",
          items       = list(type = "string", description = "Shiny input ID"),
          description = "Optional: specific input IDs to query. Omit to list all registered inputs."
        )
      )
    )
  )

  schemas[["tool__shiny_input_update"]] <- list(
    name        = "tool__shiny_input_update",
    description = paste(
      "Update a shiny input value by its ID.",
      "The value will be sent to the corresponding shiny update function",
      "(e.g. updateTextInput, updateSelectInput, updateNumericInput).",
      "Call `tool__shiny_input_info` first to discover available input IDs,",
      "their types, current values, and whether they are writable."
    ),
    inputSchema = list(
      type       = "object",
      properties = list(
        inputId = list(
          type        = "string",
          description = "Shiny input ID of which the value is to be changed"
        ),
        value = list(
          type        = "string",
          description = "The new value for the input. Use JSON encoding for non-string values (e.g. 123, [1,2,3], {\"a\":1})."
        )
      ),
      required = list("inputId")
    )
  )

  schemas[["tool__shiny_query_ui"]] <- list(
    name        = "tool__shiny_query_ui",
    description = paste(
      "Request the HTML content of a UI element by CSS selector.",
      "This sends a query to the browser and returns a request_id.",
      "The browser response is asynchronous; call `tool__shiny_query_ui_result`",
      "with the returned request_id to retrieve the actual content.",
      "Wait briefly (1-2 seconds) before fetching the result."
    ),
    inputSchema = list(
      type       = "object",
      properties = list(
        css_selector = list(
          type        = "string",
          description = "A CSS selector to query (e.g. '#my_output', '.card-body', 'div[data-id=\"plot\"]')."
        )
      ),
      required = list("css_selector")
    )
  )

  schemas[["tool__shiny_query_ui_result"]] <- list(
    name        = "tool__shiny_query_ui_result",
    description = paste(
      "Fetch the result of a previous `tool__shiny_query_ui` request.",
      "Returns the innerHTML of the matched element, or an inline image",
      "if the element is a canvas or contains only an <img> tag.",
      "If the result is not yet available and the request has not timed out,",
      "wait a moment and try again."
    ),
    inputSchema = list(
      type       = "object",
      properties = list(
        request_id = list(
          type        = "string",
          description = "The request_id returned by a prior `tool__shiny_query_ui` call."
        )
      ),
      required = list("request_id")
    )
  )

  schemas[["tool__shiny_output_info"]] <- list(
    name        = "tool__shiny_output_info",
    description = paste(
      "List registered Shiny output elements and optionally retrieve",
      "their rendered HTML content. When outputIds is omitted, returns",
      "all registered outputs with their descriptions. You can get the",
      "HTML content of output via `tool__shiny_query_ui(selector)`"
    ),
    inputSchema = list(
      type       = "object",
      properties = list(
        outputIds = list(
          type        = "array",
          items       = list(type = "string", description = "Shiny output ID"),
          description = "Optional: specific output IDs to query. Omit to list all registered outputs."
        )
      )
    )
  )

  # ---- tool-schema.yaml: user-defined pre-listed tool schemas ----
  root_path <- tryCatch(template_root(), error = function(e) NULL)
  if (!is.null(root_path) && dir.exists(root_path)) {

    schema_path <- file.path(root_path, "agents", "tool-schema.yaml")
    if (file.exists(schema_path)) {
      tool_schema_conf <- tryCatch(
        yaml::read_yaml(schema_path),
        error = function(e) NULL
      )
      if (is.list(tool_schema_conf$tools)) {
        for (ts in tool_schema_conf$tools) {
          if (!length(ts$name) || !nzchar(ts$name)) next
          full_name <- sprintf("tool__%s", ts$name)
          schemas[[full_name]] <- list(
            name        = full_name,
            description = ts$description %||% "",
            inputSchema = ts$inputSchema %||% list(
              type = "object",
              properties = structure(list(), names = character(0))
            )
          )
        }
      }
    }

    # ---- Skill schemas: auto-discovered from agents/skills/ ----
    skills_dir <- file.path(root_path, "agents", "skills")
    if (dir.exists(skills_dir)) {
      skill_dirs <- list.dirs(skills_dir, recursive = FALSE, full.names = TRUE)
      for (sdir in skill_dirs) {
        sname <- basename(sdir)
        if (!file.exists(file.path(sdir, "SKILL.md"))) next
        skill_tool <- tryCatch({
          wrapper <- skill_wrapper(sdir)
          wrapper()
        }, error = function(e) NULL)
        if (!inherits(skill_tool, "ellmer::ToolDef")) next
        skill_tool@name <- sprintf("skill__%s", sname)
        schemas[[skill_tool@name]] <- ellmer_tool_schema(skill_tool)
      }
    }
  }

  schemas
}

#' Handle MCP `tools/list` request
#' @keywords internal
#' @noRd
mcp_handle_tools_list <- function(id, params, mcp_session_id) {

  # Built-in tools (always available)
  tools <- list(
    list(
      name        = "list_shinysessions",
      description = "List all active Shiny sessions. Each session represents a browser tab or iframe. Returns token, module_id, available tool names, and registration time. Call this first to find a session to bind to, then call `register_shinysession` to bind the session.", # nolint: line_length_linter.
      inputSchema = list(
        type       = "object",
        properties = structure(list(), names = character(0))
      )
    ),
    list(
      name        = "register_shinysession",
      description = "Bind this MCP session to a Shiny session by token. After binding, per-session tools become available. Call `list_shinysessions` first to obtain the list of available shiny session token. You can switch sessions by calling this again with a different token.", # nolint: line_length_linter.
      inputSchema = list(
        type       = "object",
        properties = list(
          token = list(
            type        = "string",
            description = "The Shiny session token from list_shinysessions"
          )
        ),
        required = list("token")
      )
    ),
    list(
      name        = "get_session_info",
      description = "Show the current MCP-to-Shiny binding status: bound token, module_id, and available tools with their descriptions.", # nolint: line_length_linter.
      inputSchema = list(
        type       = "object",
        properties = structure(list(), names = character(0))
      )
    )
  )

  # Collect names of bound-session tools for deduplication
  bound_tool_names <- character(0)

  # If bound to a Shiny session, add per-session tools
  bound_token <- mcp_tool_bound_shinysessions(mcp_session_id = mcp_session_id)
  if (length(bound_token)) {
    # Only use the first session: can only bind one at a time
    bound_token <- bound_token[[1]]
    entry <- get_session_entry(bound_token)
    if (length(entry) && entry$tools$size() > 0) {
      # Filter tools by current agent mode
      module_id <- entry$namespace
      current_mode <- globals_get_agent_mode(module_id = module_id)
      enabled_tools <- Filter(function(t) {
        is_tool_enabled_for_mode(t, current_mode)
      }, entry$tools$as_list())
      if (length(enabled_tools)) {
        schema <- lapply(enabled_tools, ellmer_tool_schema)
        tools <- c(tools, unname(schema))
        bound_tool_names <- vapply(schema, `[[`, "name",
                                   FUN.VALUE = character(1))
      }
    }
  }

  # Pre-listed tools & skills (visible even without a bound session).
  # When a session IS bound, skip any pre-listed schema whose name
  # already appeared in bound-session tools (live version wins).
  prelisted <- mcp_prelisted_tool_schemas()
  if (length(prelisted)) {
    for (pschema in prelisted) {
      if (!pschema$name %in% bound_tool_names) {
        tools <- c(tools, list(pschema))
      }
    }
  }

  # ask_user built-in (always available; adapts to Shiny / console / reject)
  tools <- c(tools, list(list(
    name        = "ask_user",
    description = paste(
      "Ask the user a question via a modal dialog (when a Shiny",
      "session is bound) or the R console. Use this when you need",
      "the user to make a choice, confirm an action, or provide",
      "free-form input. Returns the user's response or indicates",
      "cancellation."
    ),
    inputSchema = list(
      type       = "object",
      properties = list(
        message = list(
          type        = "string",
          description = "The question or message to show the user."
        ),
        choices = list(
          type        = "array",
          items       = list(type = "string"),
          description = "Optional predefined choices as buttons (e.g. ['Yes', 'No'])."
        ),
        allow_freeform = list(
          type        = "boolean",
          description = "Whether to show a text area for free-form input. Default: true."
        )
      ),
      required = list("message")
    )
  )))

  mcp_json_result(id, list(tools = tools), mcp_session_id)
}


#' Handle MCP `tools/call` request
#' @keywords internal
#' @noRd
mcp_handle_tools_call <- function(id, params, mcp_session_id) {

  tool_name <- params$name
  arguments <- params$arguments

  if (!is.character(tool_name) || length(tool_name) != 1L) {
    return(mcp_json_error(
      id = id,
      code = -32602L,
      message = "Invalid params: missing or invalid tool name",
      mcp_session_id = mcp_session_id
    ))
  }

  # ---- built-in tools (always available) --------------------------------
  result <- switch(
    tool_name,
    "list_shinysessions" = mcp_tool_list_shinysessions(arguments),
    "register_shinysession" = mcp_tool_register_shinysession(
      arguments, mcp_session_id),
    "get_session_info" = mcp_tool_get_session_info(mcp_session_id),
    "ask_user" = {
      # In MCP mode (Mode 2), asking users via browser modals doesn't work
      # because users are interacting in VSCode/IDE, not in the browser.
      # Return an MCP error instructing the AI to ask questions in its own chat.
      list(
        content = list(list(
          type = "text",
          text = paste0(
            "The ask_user tool is not available in MCP mode. ",
            "Please ask the user directly in your chat interface instead."
          )
        )),
        isError = TRUE
      )
    },
    NULL # not a built-in
  )

  if (!is.null(result)) {
    # After a successful register_shinysession, the tool list changes;
    # notify the client via SSE so it refreshes its tool catalogue.
    if (identical(tool_name, "register_shinysession") &&
        isFALSE(result$isError)) {
      return(mcp_json_result(
        id, result, mcp_session_id,
        content_type  = "text/event-stream",
        notifications = list(
          list(jsonrpc = "2.0", method = "notifications/tools/list_changed")
        )
      ))
    }
    return(mcp_json_result(id, result, mcp_session_id))
  }

  # ---- per-session tools (require binding) ------------------------------
  bound_token <- mcp_tool_bound_shinysessions(mcp_session_id = mcp_session_id)
  if (!length(bound_token)) {
    result <- list(
      content = list(list(
        type = "text",
        text = paste0(
          "No Shiny session bound, or the session has been ended/closed. ",
          "Call `list_shinysessions` first to list available sessions, and ",
          "then `register_shinysession` to bind to a shiny session. ",
          "To un-register or re-register, call this tool again with other ",
          "tokens."
        )
      )),
      isError = TRUE
    )
    return(mcp_json_result(id, result, mcp_session_id))
  }

  bound_token <- bound_token[[1]]
  entry <- get_session_entry(bound_token)

  if (is.null(entry)) {
    result <- list(
      content = list(list(
        type = "text",
        text = "Bound Shiny session has ended/closed. Call `list_shinysessions` first to list available sessions, and then `register_shinysession` to bind to a shiny session"
      )),
      isError = TRUE
    )
    return(mcp_json_result(id, result, mcp_session_id))
  }

  # Look up tool in session's registered tools
  tool_obj <- entry$tools$get(tool_name, missing = NULL)

  if (is.null(tool_obj)) {
    result <- list(
      content = list(list(
        type = "text",
        text = paste0(
          "Tool '", tool_name, "' not found on bound session. ",
          "Available tools: ", paste(entry$tools$keys(), collapse = ", ")
        )
      )),
      isError = TRUE
    )
    return(mcp_json_result(id, result, mcp_session_id))
  }

  # ---- Mode guard: reject if mode is "None" ----
  current_mode <- globals_get_agent_mode(module_id = entry$namespace)
  if (identical(current_mode, "None")) {
    result <- list(
      content = list(list(
        type = "text",
        text = paste0(
          "Agent mode is 'None'. All tools & skills are disabled. ",
          "Change the mode in the dashboard to enable tool calls."
        )
      )),
      isError = TRUE
    )
    return(mcp_json_result(id, result, mcp_session_id))
  }

  # ---- Mode guard: reject if tool is not enabled for current mode ----
  if (!is_tool_enabled_for_mode(tool_obj, current_mode)) {
    result <- list(
      content = list(list(
        type = "text",
        text = paste0(
          "Tool '", tool_name, "' is not enabled for mode '",
          current_mode %||% "(none)", "'. ",
          "Switch to an appropriate mode before calling this tool."
        )
      )),
      isError = TRUE
    )
    return(mcp_json_result(id, result, mcp_session_id))
  }

  # ---- Skill script mode guard: per-script overrides ----
  skill_scripts <- tool_obj@annotations$shidashi_skill_scripts
  if (length(skill_scripts) && identical(arguments$action, "script") &&
      length(arguments$file_name) && nzchar(arguments$file_name)) {
    if (!is_script_enabled_for_mode(skill_scripts, arguments$file_name,
                                    current_mode)) {
      result <- list(
        content = list(list(
          type = "text",
          text = paste0(
            "Script '", arguments$file_name, "' in skill '", tool_name,
            "' is not enabled for mode '", current_mode %||% "(none)", "'."
          )
        )),
        isError = TRUE
      )
      return(mcp_json_result(id, result, mcp_session_id))
    }
  }

  # ---- Destructive/needs_confirmation category guard ----
  tool_category <- tool_obj@annotations$shidashi_category
  needs_confirm <- is.character(tool_category) &&
    any(c("destructive", "needs_confirmation") %in% tool_category)

  # Also check per-script category for skills
  if (!needs_confirm && length(skill_scripts) &&
      identical(arguments$action, "script") &&
      length(arguments$file_name) && nzchar(arguments$file_name)) {
    script_cat <- get_script_category(skill_scripts, arguments$file_name)
    needs_confirm <- any(c("destructive", "needs_confirmation") %in% script_cat)
  }

  if (needs_confirm) {
    # Check confirmation policy
    policy <- globals_get_confirmation_policy(
      module_id = entry$namespace, missing = "auto_allow"
    )

    if (identical(policy, "auto_reject")) {
      result <- list(
        content = list(list(
          type = "text",
          text = paste0(
            "Tool '", tool_name, "' requires confirmation but policy is set to 'Auto-reject'. ",
            "Change the confirmation policy in the dashboard to 'Auto-allow' to enable."
          )
        )),
        isError = TRUE
      )
      return(mcp_json_result(id, result, mcp_session_id))
    }

    if (identical(policy, "ask")) {
      # In MCP mode (Mode 2), we can't show browser modals for confirmation.
      # Auto-reject and instruct user to change policy if they want to allow.
      result <- list(
        content = list(list(
          type = "text",
          text = paste0(
            "Tool '", tool_name, "' requires confirmation. ",
            "In MCP mode, interactive confirmation is not available. ",
            "Change the confirmation policy in the dashboard to 'Auto-allow' to enable this tool."
          )
        )),
        isError = TRUE
      )
      return(mcp_json_result(id, result, mcp_session_id))
    }
    # policy == "auto_allow": proceed without confirmation
  }

  # Call the ToolDef with the provided arguments
  provider <- get_mcp_provider()
  result <- ellmer_tool_call(tool_obj, arguments, provider = provider)

  # Handle async tool results (promises)
  if (promises::is.promise(result)) {
    return(promises::then(result, function(res) {
      mcp_json_result(id, res, mcp_session_id)
    }))
  }

  mcp_json_result(id, result, mcp_session_id)
}

# ---------- built-in tool implementations ----------------------------------

#' List active Shiny sessions (enhanced with module_id and tool names)
#' @keywords internal
#' @noRd
mcp_tool_list_shinysessions <- function(arguments) {
  registry <- globals_session_registry()
  tokens <- registry$keys()

  sessions_info <- lapply(tokens, function(tk) {
    entry <- registry$get(tk)
    if (is.null(entry) || is.null(entry$shiny_session)) return(NULL)
    closed <- tryCatch(entry$shiny_session$isClosed(), error = function(e) TRUE)
    if (isTRUE(closed)) return(NULL)

    tool_names <- character(0)
    if (entry$tools$size() > 0L) {
      tool_names <- entry$tools$keys()
    }

    list(
      token         = tk,
      module_id     = entry$shidashi_module_id %||% "",
      namespace     = entry$namespace,
      # url           = entry$url,
      tool_names    = tool_names,
      registered_at = format(entry$registered_at, "%Y-%m-%dT%H:%M:%S")
    )
  })
  sessions_info <- Filter(Negate(is.null), sessions_info)

  text <- as.character(
    jsonlite::toJSON(unname(as.list(sessions_info)),
                     auto_unbox = TRUE, null = "null")
  )
  list(
    content = list(list(type = "text", text = text)),
    isError = FALSE
  )
}

#' Bind MCP session to a Shiny session
#' @keywords internal
#' @noRd
mcp_tool_register_shinysession <- function(arguments, mcp_session_id) {

  if (missing(mcp_session_id) || length(mcp_session_id) != 1 ||
      is.na(mcp_session_id) || !is.character(mcp_session_id) ||
      !nzchar(mcp_session_id)) {
    return(list(
      content = list(list(
        type = "text",
        text = "Error: No mcp_session_id. Call initialize first."
      )),
      isError = TRUE
    ))
  }

  token <- arguments$token
  if (!is.character(token) || length(token) != 1L || !nzchar(token)) {
    return(list(
      content = list(list(
        type = "text",
        text = "Error: 'token' (string) is required. Call `list_shinysessions` to obtain a valid token."
      )),
      isError = TRUE
    ))
  }


  # Validate Shiny session exists and is open
  entry <- get_session_entry(token)

  # Unregister existing sessions
  mcp_tool_unregister_shinysession(mcp_session_id = mcp_session_id)

  if (is.null(entry)) {
    return(list(
      content = list(list(
        type = "text",
        text = paste0(
          "Error: No active session for token '", token,
          "'. Call list_shinysessions to list available session tokens. ",
          "In addition, your previous sessions have been unregistered."
        )
      )),
      isError = TRUE
    ))
  }

  # re-fetch
  entry <- get_session_entry(token)

  # Bind MCP session to Shiny token
  entry$mcp_session_ids <- unique(c(entry$mcp_session_ids, mcp_session_id))
  registry <- globals_session_registry()
  registry$set(token, entry)

  # Build response with info about the bound session
  tool_info <- lapply(entry$tools$as_list(), function(tool_obj) {
    if (!inherits(tool_obj, "ellmer::ToolDef")) { return() }
    list(name = tool_obj@name,
         description = paste(tool_obj@description, collapse = "\n"))
  })
  tool_info <- drop_null(tool_info)

  info <- list(
    status    = "bound",
    token     = token,
    module_id = entry$shidashi_module_id %||% "",
    tools     = unname(as.list(tool_info))
  )

  # Build a message that explicitly tells the AI the tools are callable
  tool_names <- vapply(tool_info, `[[`, "name", FUN.VALUE = character(1L))
  if (length(tool_names)) {
    msg <- paste0(
      "Successfully bound to Shiny session '", token, "'.",
      " Please NOW search for MCP tools before invoking deferred tools.",
      # " The following tools will be available: ",
      # paste(tool_names, collapse = ", "), ".",
      " * Call `tools/list` (e.g. tool_search_tool_regex) to see their ",
      "full schemas. You MUST search the tools again to enable ",
      "the new deferred tools."
    )
  } else {
    msg <- paste0(
      "Successfully bound to Shiny session '", token, "'.",
      " No per-session tools are registered on this session."
    )
  }

  # detail <- as.character(jsonlite::toJSON(info, auto_unbox = TRUE, null = "null"))
  # text <- paste0(msg, "\n\nDetails:\n", detail)
  text <- msg
  list(
    content = list(list(type = "text", text = text)),
    isError = FALSE
  )
}

#' Show current MCP session binding status
#' @keywords internal
#' @noRd
mcp_tool_get_session_info <- function(mcp_session_id) {
  if (
    missing(mcp_session_id) ||
      length(mcp_session_id) != 1 ||
      is.na(mcp_session_id) ||
      !is.character(mcp_session_id) ||
      !nzchar(mcp_session_id)
  ) {
    return(list(
      content = list(list(
        type = "text",
        text = as.character(jsonlite::toJSON(
          list(status = "no_mcp_session", message = "Missing mcp_session_id. Call initialize first."),
          auto_unbox = TRUE,
          null = "null"
        ))
      )),
      isError = FALSE
    ))
  }

  bound_token <- mcp_tool_bound_shinysessions(mcp_session_id)
  if (!length(bound_token)) {
    return(list(
      content = list(list(
        type = "text",
        text = as.character(jsonlite::toJSON(
          list(status = "unbound",
               message = "Not bound to any active Shiny session or session is ended. Call register_shinysession to bind one."),
          auto_unbox = TRUE, null = "null"
        ))
      )),
      isError = FALSE
    ))
  }

  entry <- get_session_entry(bound_token)
  if (is.null(entry)) {
    return(list(
      content = list(list(
        type = "text",
        text = as.character(jsonlite::toJSON(
          list(status = "session_closed",
               message = "Bound session has closed. Call register_shinysession with a new token."),
          auto_unbox = TRUE, null = "null"
        ))
      )),
      isError = FALSE
    ))
  }

  tool_info <- lapply(entry$tools$as_list(), function(tool_obj) {
    if (!inherits(tool_obj, "ellmer::ToolDef")) { return() }
    list(name = tool_obj@name,
         description = paste(tool_obj@description, collapse = "\n"))
  })
  tool_info <- drop_null(tool_info)

  info <- list(
    status    = "bound",
    token     = bound_token,
    module_id = entry$shidashi_module_id %||% "",
    # url       = entry$url,
    tools     = unname(as.list(tool_info)) # array of objects
  )

  text <- as.character(jsonlite::toJSON(info, auto_unbox = TRUE, null = "null"))
  list(
    content = list(list(type = "text", text = text)),
    isError = FALSE
  )
}


#' Ask the user a question
#'
#' Tries three strategies in order:
#' \enumerate{
#'   \item If a live Shiny session is provided, ask via a browser modal.
#'   \item If \code{interactive()}, ask via the R console.
#'   \item Otherwise reject with an error.
#' }
#' @param arguments list with \code{message}, optional \code{choices},
#'   optional \code{allow_freeform}.
#' @param shiny_session A Shiny session or \code{NULL}.
#' @return A \code{promises::promise} (Shiny path) or a plain list with
#'   \code{content} and \code{isError}.
#' @keywords internal
#' @noRd
mcp_tool_ask_user <- function(arguments, shiny_session = NULL) {

  message_text <- arguments$message
  if (!is.character(message_text) || !nzchar(message_text)) {
    return(list(
      content = list(list(
        type = "text",
        text = "Invalid params: 'message' is required."
      )),
      isError = TRUE
    ))
  }

  choices <- as.character(unlist(arguments$choices))
  allow_freeform <- !identical(arguments$allow_freeform, FALSE)

  # --- Strategy 1: Shiny browser modal -----------------------------------
  session_ok <- !is.null(shiny_session) &&
    is.environment(shiny_session) &&
    !isTRUE(tryCatch(shiny_session$isClosed(), error = function(e) TRUE))

  if (session_ok) {
    return(mcp_tool_ask_user_shiny(
      message_text, choices, allow_freeform, shiny_session,
      tool_name = arguments$tool_name,
      intent = arguments$intent
    ))
  }

  # --- Strategy 2: interactive R console ---------------------------------
  if (interactive()) {
    return(mcp_tool_ask_user_console(
      message_text, choices, allow_freeform
    ))
  }

  # --- Strategy 3: reject ------------------------------------------------
  list(
    content = list(list(
      type = "text",
      text = "Cannot ask user: no Shiny session available and R is not interactive."
    )),
    isError = TRUE
  )
}

# Ask user via Shiny browser modal (returns a promise)
mcp_tool_ask_user_shiny <- function(message_text, choices, allow_freeform,
                                    shiny_session,
                                    tool_name = NULL, intent = NULL) {
  request_id <- rand_string(prefix = "ask_user_")
  input_id <- shiny_session$ns("@shidashi_ask_user_result@")

  payload <- list(
    request_id = request_id,
    input_id = input_id,
    message = message_text,
    choices = choices,
    allow_freeform = allow_freeform
  )
  if (length(tool_name) == 1 && nzchar(tool_name)) {
    payload$tool_name <- tool_name
  }
  if (length(intent) == 1 && nzchar(intent)) {
    payload$intent <- intent
  }
  shiny_session$sendCustomMessage("shidashi.ask_user", payload)

  check_fn <- coro::async(function() {
    remaining <- 120L  # 120 x 500ms = 60 seconds timeout

    while (remaining >= 0) {
      remaining <- remaining - 1L
      res <- shiny::isolate(
        shiny_session$input[["@shidashi_ask_user_result@"]]
      )
      if (!is.null(res) && identical(res$request_id, request_id)) {
        if (isTRUE(res$cancelled)) {
          return(list(
            content = list(list(type = "text",
                                text = "User cancelled the request.")),
            isError = FALSE
          ))
        } else {
          return(list(
            content = list(list(type = "text",
                                text = res$value %||% "")),
            isError = FALSE
          ))
        }
      } else {
        coro::async_sleep(0.5)
      }
    }

    return(list(
      content = list(list(type = "text",
                          text = "Timeout: no response from user within 60 seconds.")),
      isError = FALSE
    ))

  })
  check_fn()
}

# Ask user via the R console (synchronous, returns a plain list)
mcp_tool_ask_user_console <- function(message_text, choices, allow_freeform) {
  cat("\n", message_text, "\n", sep = "")
  answer <- NULL

  if (length(choices)) {
    sel <- utils::menu(choices, title = "Select an option:")
    if (sel == 0L) {
      return(list(
        content = list(list(type = "text",
                           text = "User cancelled the request.")),
        isError = FALSE
      ))
    }
    answer <- choices[[sel]]
  }

  if (allow_freeform) {
    prompt <- if (is.null(answer)) "Your response: " else "Additional input (or Enter to skip): "
    freeform <- readline(prompt)
    if (nzchar(freeform)) {
      answer <- if (is.null(answer)) freeform else paste0(answer, "\n", freeform)
    }
  }

  if (is.null(answer) || !nzchar(answer)) {
    answer <- "(no response)"
  }

  list(
    content = list(list(type = "text", text = answer)),
    isError = FALSE
  )
}

# ---------- internal helpers ----------------------------------------------

ellmer_tool_schema <- function(tool_obj) {
  # Do NOT remove
  # DIPSAUS DEBUG START
  # tool_obj <- ellmer::tool(
  #   rnorm,
  #   description = "Draw numbers from a random normal distribution",
  #   arguments = list(
  #     n = ellmer::type_integer("The number of observations. Must be a positive integer."),
  #     mean = ellmer::type_number("The mean value of the distribution."),
  #     sd = ellmer::type_number("The standard deviation of the distribution. Must be a non-negative number.")
  #   )
  # )

  ellmer <- asNamespace("ellmer")

  dummy_provider <- ellmer::Provider(name = "dummy",
                                     model = "dummy",
                                     base_url = "https://dummy")
  schema <- ellmer$as_json(dummy_provider, tool_obj@arguments)
  # Remove OpenAI-specific quirks if any
  schema$additionalProperties <- NULL

  list(
    name = tool_obj@name,
    description = tool_obj@description,
    inputSchema = schema
  )
}

#' Call a ToolDef with MCP arguments
#'
#' Splices the JSON arguments into the ToolDef's callable interface.
#' Returns MCP result format.
#' @keywords internal
#' @noRd
ellmer_tool_call <- function(tool_obj, arguments, provider = NULL) {
  if (is.null(arguments)) arguments <- list()

  # ToolDef inherits from class_function — it's directly callable
  # Call with the arguments from JSON

  tool_error <- function(e) {
    list(
      content = list(list(
        type = "text",
        text = paste0("Error executing tool '", tool_obj@name, "': ",
                      conditionMessage(e))
      )),
      isError = TRUE
    )
  }

  tryCatch({
    ret <- do.call(tool_obj, arguments)

    # Handle promises (async tools like shiny_query_ui)
    if (promises::is.promise(ret)) {
      return(promises::then(
        ret,
        onFulfilled = function(value) {
          content_to_mcp(value, provider)
        },
        onRejected = tool_error
      ))
    }

    content_to_mcp(ret, provider)
  }, error = tool_error)

}

# ---------- JSON helpers -------------------------------------------------

#' Build a JSON-RPC 2.0 error response
#' @keywords internal
#' @noRd
mcp_json_error <- function(id, code, message, mcp_session_id = NULL) {

  headers <- list()
  if (length(mcp_session_id) == 1 && nzchar(mcp_session_id)) {
    headers[["Mcp-Session-Id"]] <- mcp_session_id
  }

  body <- jsonlite::toJSON(
    list(
      jsonrpc = "2.0",
      id      = id,
      error   = list(
        code    = code,
        message = message
      )
    ),
    auto_unbox = TRUE,
    null = "null"
  )

  shiny::httpResponse(
    status       = 200L,
    content_type = "application/json",
    content      = body,
    headers      = headers
  )
}


#' Build a JSON-RPC 2.0 success response
#'
#' @param id JSON-RPC request id.
#' @param result The result payload (list).
#' @param mcp_session_id Optional MCP session ID for the response header.
#' @param content_type Either \code{"application/json"} (default) or
#'   \code{"text/event-stream"}.  When SSE, each JSON-RPC message is
#'   emitted as an \code{event: message} frame.
#' @param notifications A list of additional JSON-RPC notification objects
#'   (each a named list with \code{jsonrpc}, \code{method}, and optionally
#'   \code{params}) to append after the result.  Only used when
#'   \code{content_type} is \code{"text/event-stream"}.
#' @keywords internal
#' @noRd
mcp_json_result <- function(id, result, mcp_session_id = NULL,
                            content_type = c("application/json",
                                             "text/event-stream"),
                            notifications = NULL) {
  content_type <- match.arg(content_type)
  headers <- list()
  if (length(mcp_session_id) == 1 && nzchar(mcp_session_id)) {
    headers[["Mcp-Session-Id"]] <- mcp_session_id
  }

  result_msg <- list(jsonrpc = "2.0", id = id, result = result)

  if (identical(content_type, "text/event-stream")) {
    messages <- list(result_msg)
    if (length(notifications)) {
      messages <- c(messages, notifications)
    }
    body <- paste0(
      vapply(messages, function(msg) {
        json <- jsonlite::toJSON(msg, auto_unbox = TRUE, null = "null")
        paste0("event: message\ndata: ", json, "\n\n")
      }, FUN.VALUE = character(1L)),
      collapse = ""
    )
  } else {
    body <- jsonlite::toJSON(result_msg, auto_unbox = TRUE, null = "null")
  }

  shiny::httpResponse(
    status       = 200L,
    content_type = content_type,
    content      = body,
    headers      = headers
  )
}

library(shiny)

# Debug
if (FALSE) {
  template_settings$set(
    "root_path" = "inst/builtin-templates/bslib-bare/"
  )
}

server <- function(input, output, session) {

  shidashi::stream_init(session)
  shidashi::enable_input_broadcast(session)
  shidashi::enable_input_sync(session)

  # Load and dispatch module server on navigation (register first)
  shiny::observeEvent(session$clientData$url_search, {
    req <- list(QUERY_STRING = session$clientData$url_search)
    resource <- shidashi::load_module(request = req)
    if (resource$has_module) {

      module_table <- shidashi::module_info()
      module_table <- module_table[module_table$id %in% resource$module$id, ]
      if (nrow(module_table)) {
        group_name <- as.character(module_table$group[[1]])
        if (is.na(group_name)) {
          group_name <- "<no group>"
        }
        if (system.file(package = "logger") != "") {
          logger::log_info("Loading - { module_table$label[1] } ({group_name}/{ module_table$id })")
        }
        shiny::moduleServer(resource$module$id, resource$module$server, session = session)
      }
    }
  })

  output$drawer_output <- shiny::renderPrint({
    module_data <- shidashi::active_module()
    if (is.null(module_data)) {
      "No module"
    } else {
      utils::str(module_data)
    }
  })
}

#' Initialize the shidashi stream directory for a Shiny session
#'
#' Call once in the server function to set up the directory that
#' \code{\link{stream_path}} and \code{\link{stream_to_js}} will write to.
#' When running inside a shidashi template the directory is automatically
#' resolved from the template root; when running in plain Shiny a temporary
#' directory is created and registered with
#' \code{\link[shiny]{addResourcePath}} so that the browser can fetch
#' \code{stream/{token}_{id}.bin}.  An optional cleanup hook is registered
#' to remove this session's files when the session ends.
#'
#' @param session Shiny session object.  Defaults to the currently active
#'   reactive domain.
#' @return Invisibly returns the absolute path to the stream directory.
#' @seealso \code{\link{stream_path}}, \code{\link{stream_to_js}}
#' @export
stream_init <- function(session = shiny::getDefaultReactiveDomain()) {
  in_shidashi <- !is.null(template_settings$get("root_path", NULL))
  if (in_shidashi) {
    path <- file.path(template_root(), "www", "stream")
    dir.create(path, recursive = TRUE, showWarnings = FALSE)
  } else {
    path <- file.path(tempdir(), "shidashi-stream")
    dir.create(path, recursive = TRUE, showWarnings = FALSE)
    shiny::addResourcePath("stream", path)
  }
  Sys.setenv(SHIDASHI_STREAM_PATH = path)
  # Clean up this session's files when the session ends
  if (!is.null(session)) {
    token <- session$token
    session$onSessionEnded(function() {
      files <- list.files(path, pattern = paste0("^", token, "_"),
                          full.names = TRUE)
      unlink(files)
    })
  }
  invisible(path)
}

#' Get the absolute path for a shidashi stream file
#'
#' Returns the full path to the binary file that will be served at
#' \code{stream/{token}_{id}.bin}.  The filename embeds the session token so
#' that simultaneous Shiny sessions never overwrite each other's data.
#' Call \code{\link{stream_init}} at least once per session before relying on
#' this function.
#'
#' @param id Character scalar. Stream identifier.  Must be a valid file-name
#'   component (no path separators).
#' @param session Shiny session object.  Defaults to the currently active
#'   reactive domain.
#' @return Invisible character scalar: absolute path to the \code{.bin} file.
#' @seealso \code{\link{stream_init}}, \code{\link{stream_to_js}}
#' @export
stream_path <- function(id, session = shiny::getDefaultReactiveDomain()) {
  stopifnot(is.character(id), length(id) == 1L, nzchar(id))
  env_path <- Sys.getenv("SHIDASHI_STREAM_PATH", unset = "")
  stream_dir <- if (nzchar(env_path)) {
    env_path
  } else {
    file.path(template_root(), "www", "stream")
  }
  dir.create(stream_dir, showWarnings = FALSE, recursive = TRUE)
  token <- if (!is.null(session)) session$token else "static"
  nsid  <- if (!is.null(session)) session$ns(id) else id
  invisible(file.path(stream_dir, paste0(token, "_", nsid, ".bin")))
}

#' Write data to a \pkg{shidashi} stream binary file
#'
#' Serializes \code{data} together with a JSON header into the binary
#' envelope expected by the JavaScript \code{window.shidashi.fetchStreamData(id)}
#' method.
#'
#' \strong{Wire format}
#' \preformatted{
#'   [endianFlag: 1 byte]  [headerLen: uint32 LE]  [header: UTF-8 JSON]  [body]
#' }
#' \code{endianFlag} is always \code{0x01} (little-endian).
#' \code{header} is a JSON object containing at minimum \code{data_type}
#' plus any extra fields passed via \code{...}.
#'
#' \strong{Body encoding by type}
#' \describe{
#'   \item{\code{"raw"}}{Passed through verbatim; \code{data} must be a
#'     \code{raw} vector or will be coerced via \code{as.raw()}.}
#'   \item{\code{"json"}}{\code{data} is serialized with
#'     \code{jsonlite::toJSON(auto_unbox = TRUE)}.}
#'   \item{\code{"int32"}}{\code{data} is coerced to integer and written as
#'     4-byte little-endian signed integers.}
#'   \item{\code{"float32"}}{\code{data} is coerced to double and written as
#'     4-byte little-endian IEEE 754 single-precision floats.}
#'   \item{\code{"float64"}}{\code{data} is coerced to double and written as
#'     8-byte little-endian IEEE 754 double-precision floats.}
#' }
#'
#' @param abspath Character scalar. Absolute path to the target \code{.bin}
#'   file.  Use \code{\link{stream_path}(id)} to obtain a path under the
#'   app's \code{www/stream/} directory.
#' @param data The data to serialize.  See Details for how each \code{type}
#'   interprets this argument.
#' @param type Character scalar. One of \code{"raw"}, \code{"json"},
#'   \code{"int32"}, \code{"float32"}, or \code{"float64"}.
#' @param ... Additional named scalar values to embed in the JSON header and
#'   expose to the JavaScript caller via \code{result.header}.
#' @return Invisibly returns \code{abspath}.
#' @seealso \code{\link{stream_path}}
#' @export
stream_to_js <- function(abspath, data,
                         type = c("raw", "json", "int32", "float32", "float64"),
                         ...) {
  type <- match.arg(type)

  # Build header JSON
  extra <- list(...)
  header_list <- c(list(data_type = type), extra)
  # Compute a signature so the JS side can skip redundant renders
  signature <- digest::digest(header_list, algo = "xxhash32")
  header_list$timestamp <- as.numeric(Sys.time()) * 1000  # epoch ms
  header_list$signature <- signature
  header_json <- jsonlite::toJSON(header_list, auto_unbox = TRUE)
  header_bytes <- charToRaw(header_json)

  # Encode body
  body_bytes <- switch(
    type,
    raw = {
      if (is.raw(data)) data else as.raw(data)
    },
    json = {
      charToRaw(jsonlite::toJSON(data, auto_unbox = TRUE))
    },
    int32 = {
      writeBin(as.integer(data), raw(), size = 4L, endian = "little")
    },
    float32 = {
      writeBin(as.double(data), raw(), size = 4L, endian = "little")
    },
    float64 = {
      writeBin(as.double(data), raw(), size = 8L, endian = "little")
    }
  )

  # Write binary envelope
  con <- file(abspath, "wb")
  on.exit(close(con), add = TRUE)

  # endianFlag: 0x01 = little-endian
  writeBin(as.raw(0x01L), con)
  # headerLen: 4-byte uint32 LE
  writeBin(as.integer(length(header_bytes)), con, size = 4L, endian = "little")
  # header JSON bytes
  writeBin(header_bytes, con)
  # body bytes
  writeBin(body_bytes, con)

  invisible(abspath)
}


#' Read a shidashi stream binary file
#'
#' Reads the binary envelope written by \code{\link{stream_to_js}} and
#' returns the header and decoded body as a list.
#'
#' @param path Character scalar. Absolute path to a \code{.bin} file
#'   produced by \code{\link{stream_to_js}}.
#' @return A list with components:
#'   \describe{
#'     \item{\code{header}}{Named list parsed from the JSON header
#'       (contains \code{data_type}, \code{signature}, \code{timestamp},
#'       and any extra fields).}
#'     \item{\code{data}}{Decoded body: a \code{raw} vector for
#'       \code{"raw"}, an R object for \code{"json"}, or a numeric/integer
#'       vector for \code{"int32"}, \code{"float32"}, \code{"float64"}.}
#'   }
#' @seealso \code{\link{stream_to_js}}, \code{\link{stream_path}}
#' @export
read_stream_vis <- function(path) {
  con <- file(path, "rb")
  on.exit(close(con), add = TRUE)

  # endianFlag (1 byte)
  endian_flag <- readBin(con, what = "raw", n = 1L)
  endian <- if (identical(endian_flag, as.raw(0x01L))) "little" else "big"

  # headerLen (uint32)
  header_len <- readBin(con, what = "integer", n = 1L, size = 4L, endian = endian)

  # header JSON bytes
  header_raw <- readBin(con, what = "raw", n = header_len)
  header <- jsonlite::fromJSON(rawToChar(header_raw), simplifyVector = TRUE)

  # body: remaining bytes
  body_raw <- readBin(con, what = "raw", n = file.info(path)$size)

  data_type <- header$data_type %||% "raw"
  data <- switch(
    data_type,
    raw = body_raw,
    json = jsonlite::fromJSON(rawToChar(body_raw), simplifyVector = TRUE),
    int32 = readBin(body_raw, what = "integer", n = length(body_raw) %/% 4L,
                    size = 4L, endian = endian),
    float32 = readBin(body_raw, what = "double", n = length(body_raw) %/% 4L,
                      size = 4L, endian = endian),
    float64 = readBin(body_raw, what = "double", n = length(body_raw) %/% 8L,
                      size = 8L, endian = endian),
    body_raw
  )

  list(header = header, data = data)
}

#' @title Download 'shidashi' templates from 'Github'
#' @param path the path to create 'shidashi' project
#' @param user 'Github' user name
#' @param theme the theme to download
#' @param repo repository if the name is other than \code{'shidashi-templates'}
#' @param branch branch name if other than \code{'main'} or \code{'master'}
#' @param ... ignored
#' @return the target project path
#' @details To publish a 'shidashi' template, create a 'Github' repository
#' called \code{'shidashi-templates'}, or fork the
#' \href{https://github.com/dipterix/shidashi-templates}{built-in templates}.
#' The \code{theme} is the sub-folder of the template repository.
#'
#' An easy way to use a template in your project is through the 'RStudio'
#' project widget. In the 'RStudio' navigation bar, go to "File" menu,
#' click on the "New Project..." button, select the "Create a new project"
#' option, and find the item that creates 'shidashi' templates. Use the
#' widget to set up template directory.
#'
#' @export
use_template <- function(
  path, user = "dipterix", theme = "bslib",
  repo = "shidashi-templates", branch = "main", ...
) {

  # ensure path exists
  dir.create(path, recursive = TRUE, showWarnings = FALSE)

  # Download template
  temppath <- tempfile()
  tempzip <- paste0(temppath, ".zip")
  if (file.exists(tempzip)) {
    unlink(tempzip)
  }
  if (dir.exists(temppath)) {
    unlink(temppath, recursive = TRUE, force = TRUE)
  }
  old <- options()
  on.exit({
    options(old)
    unlink(tempzip)
    unlink(temppath, recursive = TRUE, force = TRUE)
  })

  options(timeout = 10000)

  url <- sprintf("https://github.com/%s/%s/archive/%s.zip", user, repo, branch)
  if (branch %in% c("main", "master")) {
    tryCatch({
      suppressWarnings({
        utils::download.file(url, destfile = tempzip, cacheOK = FALSE)
      })
    }, error = function(e) {
      old_branch <- branch
      if (branch == "main") {
        branch <<- "master"
      } else {
        branch <<- "main"
      }
      message("Branch `", old_branch, "` does not have zip file, switching to `", branch, "`.", appendLF = TRUE)
      url <- sprintf("https://github.com/%s/%s/archive/%s.zip", user, repo, branch)
      utils::download.file(url, destfile = tempzip, cacheOK = FALSE)
    })

  } else {
    utils::download.file(url, destfile = tempzip, cacheOK = FALSE)
  }


  # url <- "https://github.com/dipterix/shidashi-templates/archive/refs/heads/master.zip"
  # utils::download.file(url, destfile = tempzip, cacheOK = FALSE)
  utils::unzip(tempzip, exdir = temppath)

  folder_name <- sprintf("%s-%s", repo, branch)
  root <- file.path(temppath, folder_name)
  if (!dir.exists(root)) {
    stop("Cannot find branch folder `", folder_name, "`. Please report this issue to \n\thttps://github.com/dipterix/shidashi/issues")
  }
  project_dir <- normalizePath(file.path(root, theme), mustWork = TRUE)

  fs <- list.files(project_dir, all.files = FALSE, recursive = FALSE, full.names = TRUE, include.dirs = TRUE, no.. = TRUE)

  if (!length(fs)) {
    if (theme == "") {
      stop("Empty sub-module repository. Your theme contains no file. Abort.")
    }
    # This could be a sub-module, parse .gitmodules
    submodule_spath <- file.path(root, ".gitmodules")
    if (!file.exists(submodule_spath)) {
      stop("Theme `", theme, "` exists but it's empty. Abort.")
    }
    settings <- readLines(submodule_spath)
    # Find submodule
    sel <- startsWith(settings, sprintf("[submodule \"%s\"]", theme))
    if (!any(sel)) {
      stop("Theme `", theme, "` is empty. I guess it's a git submodule. However, I cannot locate it in the `.gitmodules` file.")
    }
    settings <- settings[seq(which(sel), length(settings))]
    sel <- which(grepl("\\[submodule", settings))
    if (length(sel) > 1) {
      settings <- settings[seq(sel[[1]], sel[[2]] - 1)]
    }
    sel <- grepl("^[^a-zA-Z0-9]url[ =]", settings)
    if (!any(sel)) {
      stop("Theme `", theme, "` is a git submodule. However, I cannot locate the URL.")
    }
    url <- settings[sel][[1]]
    if (grepl("url[ =]+http", url)) {
      url <- strsplit(url, split = "/")[[1]]
      uname <- url[[length(url) - 1]]
      repo <- url[[length(url)]]
      if (grepl("@", repo)) {
        tmp <- strsplit(repo, "@")[[1]]
        repo <- tmp[[1]]
        branch <- tmp[[2]]
      } else {
        sel <- grepl("^[^a-zA-Z0-9]branch[ =]", settings)
        if (any(sel)) {
          branch <- settings[sel][[1]]
          branch <- strsplit(branch, "=")[[1]][[2]]
          branch <- sub("[ \t]+", "", branch)
        } else {
          branch <- "main"
        }
      }
    } else if (grepl("url[ =]+git@", url)) {
      url <- strsplit(url, split = ":")[[1]]
      url <- unlist(strsplit(url, split = "/"))
      uname <- url[[length(url) - 1]]
      repo <- url[[length(url)]]
      if (grepl("@", repo)) {
        tmp <- strsplit(repo, "@")[[1]]
        repo <- tmp[[1]]
        branch <- tmp[[2]]
      } else {
        sel <- grepl("^[^a-zA-Z0-9]branch[ =]", settings)
        if (any(sel)) {
          branch <- settings[sel][[1]]
          branch <- strsplit(branch, "=")[[1]][[2]]
          branch <- sub("[ \t]+", "", branch)
        } else {
          branch <- "main"
        }
      }
      if (endsWith(repo, ".git")) {
        repo <- sub("\\.git$", "", repo)
      }
    } else {
      stop("Theme `", theme, "` is a git submodule. However, I cannot parse the URL:\n", url)
    }

    re <- use_template(path = path, user = uname, theme = "",
                       repo = repo, branch = branch, ...)
    return(re)
    # https://github.com/dipterix/rave-pipelines/archive/578c8644b2b67623b7efd138a1e5340fc068e725.zip
    # url <- sprintf("https://github.com/%s/%s/archive/%s.zip", uname, repo, branch)
    # unlink(tempzip)
    # if (branch %in% c("main", "master")) {
    #   tryCatch({
    #     utils::download.file(url, destfile = tempzip, cacheOK = FALSE)
    #   }, error = function(e) {
    #     if (branch == "main") {
    #       branch <- "master"
    #     } else {
    #       branch <- "main"
    #     }
    #     url <- sprintf("https://github.com/%s/%s/archive/%s1.zip", uname, repo, branch)
    #     utils::download.file(url, destfile = tempzip, cacheOK = FALSE)
    #   })
    #
    # } else {
    #   utils::download.file(url, destfile = tempzip, cacheOK = FALSE)
    # }
    #
    # unlink(temppath, recursive = TRUE)
    # utils::unzip(tempzip, exdir = temppath)


  }


  file.copy(fs, path, overwrite = TRUE, recursive = TRUE, copy.date = TRUE)

  # Add RStudio start-up script
  # dput(deparse(quote({
  #   try({
  #     shidashi::template_settings$set(root_path = normalizePath("."))
  #   }, silent = TRUE)
  # })))

  writeLines(
    c(
      "{",
      "    try({",
      "        shidashi::template_settings$set(root_path = normalizePath(\".\"))",
      "    }, silent = TRUE)",
      "}"
    ),
    con = file.path(path, ".Rprofile")
  )

  writeLines(c(
    "library(shidashi)\n",
    "# Render this project",
    sprintf("shidashi::template_settings$set(root_path = '%s')\n", normalizePath(path)),
    "# Render project",
    "shidashi::render(host = '127.0.0.1', port = 8310L)"
  ), con = file.path(path, "start.R"))

  invisible(normalizePath(path))
}

#' @title Configure template options that are shared across the sessions
#' @param ... key-value pair to set options
#' @param name character, key of the value
#' @param default default value if the key is missing
#' @return \code{template_settings_get} returns the values represented by the
#' corresponding keys, or the default value if key is missing.
#' @details The settings is designed to store static key-value pairs that
#' are shared across the sessions. The most important key is
#' \code{"root_path"}, which should be a path pointing to the template
#' folder.
#' @examples
#'
#' # Get current website root path
#'
#' template_root()
#'
#' @export
template_settings <- local({
  map <- fastmap::fastmap()
  list(
    get = function(name, default = NULL) {
      map$get(name, missing = default)
    },
    set = function(...) {
      map$mset(...)
    },
    list = function() {
      map$as_list()
    }
  )
})

#' @rdname template_settings
#' @export
template_settings_set <- template_settings$set

#' @rdname template_settings
#' @export
template_settings_get <- template_settings$get

#' @rdname template_settings
#' @export
template_root <- function() {
  path <- template_settings$get(
    name = "root_path",
    default = NULL
  )
  if (!length(path)) {
    if (template_settings$get("dev.debug", FALSE)) {
      path <- "./inst/builtin-templates/bslib-bare/"
    } else {
      path <- file.path(R_user_dir("shidashi", which = "data"),
                        c("bslib-bare", "AdminLTE3", "AdminLTE3-bare"))
      path <- path[file.exists(path)]
      if (!length(path)) {
        path <- file.path(R_user_dir("shidashi", which = "data"), "bslib-bare")
        create_barebone_bslib(path)
      } else {
        path <- path[[1]]
      }
    }

  }
  normalizePath(path, mustWork = FALSE)
}

test_that("skill_wrapper creates a closure with correct class", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  wrapper <- skill_wrapper(skill_dir)
  expect_s3_class(wrapper, "shidashi_skill_wrapper")
  expect_true(is.function(wrapper))
})

test_that("skill_wrapper errors on missing SKILL.md", {
  tmp <- tempfile("empty_skill")
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  expect_error(skill_wrapper(tmp), "SKILL.md not found")
})

test_that("wrapper() returns an ellmer::ToolDef", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  wrapper <- skill_wrapper(skill_dir)
  tool_def <- wrapper()
  expect_true(inherits(tool_def, "ellmer::ToolDef"))
  expect_equal(tool_def@name, "greet")
})

test_that("action='readme' returns SKILL.md body with script listing", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  tool_def <- skill_wrapper(skill_dir)()
  result <- tool_def(action = "readme")
  expect_type(result, "character")
  expect_match(result, "Instructions")
  expect_match(result, "greet\\.R")
})

test_that("soft gate augments errors before readme", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  tool_def <- skill_wrapper(skill_dir)()

  # Requesting a nonexistent reference triggers error + readme hint
  err <- tryCatch(
    tool_def(action = "script", file_name = "nonexistent.R"),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "action='readme' first", fixed = TRUE)
})

test_that("soft gate allows successful calls before readme", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  tool_def <- skill_wrapper(skill_dir)()

  # Valid script call should succeed even without readme

  result <- tool_def(action = "script", file_name = "greet.R",
                     args = list("TestUser"))
  expect_match(result, "Hello, TestUser!")
})

test_that("after readme, errors are not augmented", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  tool_def <- skill_wrapper(skill_dir)()
  tool_def(action = "readme")  # unlock

  err <- tryCatch(
    tool_def(action = "script", file_name = "nonexistent.R"),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "Script not found")
  expect_false(grepl("action='readme' first", err, fixed = TRUE))
})

test_that("action='script' runs greet.R via processx", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  tool_def <- skill_wrapper(skill_dir)()
  tool_def(action = "readme")  # unlock
  result <- tool_def(action = "script", file_name = "greet.R",
                     args = list("Copilot"))
  expect_match(result, "Hello, Copilot!")
  expect_match(result, "Exit code: 0", fixed = TRUE)
})

test_that("each wrapper() call gets independent gate state", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  wrapper <- skill_wrapper(skill_dir)
  tool1 <- wrapper()
  tool2 <- wrapper()

  # Unlock tool1

  tool1(action = "readme")

  # tool2 should still have gate active (error augmented on failure)
  err <- tryCatch(
    tool2(action = "script", file_name = "nonexistent.R"),
    error = function(e) conditionMessage(e)
  )
  expect_match(err, "action='readme' first", fixed = TRUE)

  # tool1 gate is unlocked — error not augmented
  err1 <- tryCatch(
    tool1(action = "script", file_name = "nonexistent.R"),
    error = function(e) conditionMessage(e)
  )
  expect_false(grepl("action='readme' first", err1, fixed = TRUE))
})

# Format a cost value with appropriate decimal places
format_cost <- function(cost) {
  if (is.na(cost) || is.null(cost)) return("$?")
  if (cost < 0.01) {
    sprintf("$%.4f", cost)
  } else if (cost < 0.10) {
    sprintf("$%.3f", cost)
  } else {
    sprintf("$%.2f", cost)
  }
}

format_token <- function(cost) {
  if (is.na(cost) || is.null(cost)) return("0")
  # 1G = 1000M = 1e6K = 1e9
  base <- 1
  unit <- ""
  if (cost >= 1e10) {
    base <- 1e9
    unit <- "G"
  } else if (cost >= 1e7) {
    base <- 1e6
    unit <- "M"
  } else if (cost >= 1e4) {
    base <- 1e3
    unit <- "K"
  }
  sprintf("%s%s", format(cost / base, big.mark = ",", digits = 4), unit)
}

#' Chat-bot UI panel
#'
#' @description
#' Returns the UI elements for the AI chat panel: a header bar with
#' a \dQuote{New conversation} button and the \pkg{shinychat} widget.
#'
#' When \pkg{shinychat} is not installed or the chat-bot is disabled via
#' \code{options(shidashi.chatbot = FALSE)}, returns an empty
#' \code{tagList()}.
#'
#' Typically called inside \code{shiny::renderUI} by
#' \code{\link{chatbot_server}()} to fill a \code{\link{module_drawer}}.
#' Can also be placed anywhere in module UI directly.
#'
#' @param id character; the Shiny input/output namespace for the chat
#'   widget. Default \code{"shidashi-chatbot"}.
#' @return A \code{shiny::tagList} containing the chat UI or empty.
#' @keywords internal
chatbot_ui <- function(id, modes = NULL, default_mode = NULL) {
  if (!isTRUE(getOption("shidashi.chatbot", TRUE))) {
    return(shiny::tagList("AI is disabled"))
  }
  if (!requireNamespace("shinychat", quietly = TRUE)) {
    return(shiny::tagList(
      "Package `shinychat` must be installed to enable this feature"
    ))
  }
  conv_select_id <- paste0(id, "-conv_select")
  new_conv_id    <- paste0(id, "-new_conversation")
  mode_select_id <- paste0(id, "-mode_select")
  confirm_policy_id <- paste0(id, "-confirm_policy")
  status_id      <- paste0(id, "-status")

  # Build mode selector if modes are defined
  mode_ui <- NULL
  if (length(modes)) {
    # Use mode names only (no descriptions)
    mode_choices <- vapply(modes, function(m) m$name, character(1))
    names(mode_choices) <- vapply(modes, function(m) {
      sprintf("%s - %s", m$name, m$description %||% "")
    }, character(1))
    selected <- if (length(default_mode)) default_mode else mode_choices[[1]]
    mode_ui <- shiny::selectInput(
      mode_select_id,
      label = NULL,
      choices = mode_choices,
      selected = selected,
      width = "100%"
    )
  }

  # Confirmation policy selector
  confirm_policy_ui <- shiny::selectInput(
    confirm_policy_id,
    label = NULL,
    choices = c(
      "Auto-allow" = "auto_allow",
      "Ask before changes" = "ask",
      "Auto-reject" = "auto_reject"
    ),
    selected = "auto_allow",
    width = "100%"
  )

  # Conversation selector
  conv_ui <- shiny::div(
    class = "shidashi-chatbot-conv-select flex-grow-1",
    shiny::selectInput(
      conv_select_id,
      label = NULL,
      choices = c("New conversation" = "1"),
      width = "100%"
    )
  )

  # New conversation button
  new_conv_ui <- shiny::actionLink(
    new_conv_id, label = NULL,
    icon = shiny::icon("plus"),
    title = "New conversation",
    class = "shidashi-chatbot-new-conv shidashi-chatbot-ops-btn btn btn-sm btn-outline-secondary"
  )

  # Copy conversation button (plain button triggers JS directly)
  copy_conv_ui <- shiny::tags$button(
    type = "button",
    class = "shidashi-chatbot-copy-conv shidashi-chatbot-ops-btn btn btn-sm btn-outline-secondary",
    title = "Copy conversation to clipboard",
    `data-shidashi-action` = "copy-conversation",
    `data-shidashi-chat-id` = id,
    shiny::icon("copy")
  )

  shiny::tagList(
    # Chat UI - stop button will be injected dynamically via JS
    shinychat::chat_ui(id, fill = TRUE),
    # Control bar: all selectors in one compact row (dropup)
    shiny::div(
      class = "shidashi-chatbot-controls d-flex align-items-center gap-1 px-1 py-1",
      shiny::div(
        style = "flex: 3;",
        mode_ui
      ),
      shiny::div(
        style = "flex: 3;",
        confirm_policy_ui
      ),
      shiny::div(
        style = "flex: 5;",
        conv_ui
      ),
      shiny::div(
        class = "d-flex gap-1",
        new_conv_ui,
        copy_conv_ui
      )
    ),
    # Status bar: model name, token counts, estimated cost
    shiny::tags$footer(
      id = status_id,
      class = "shidashi-chatbot-status d-flex align-items-center gap-2 small font-monospace px-2 py-1",
      shiny::span(
        id = paste0(status_id, "-model"),
        class = "shidashi-chatbot-status-model text-truncate"
      ),
      shiny::span(class = "ms-auto"),
      shiny::span(
        id = paste0(status_id, "-tokens-input"),
        class = "shidashi-chatbot-status-counter",
        title = "Input tokens",
        "\u2191 0"  # up arrow
      ),
      shiny::span(
        id = paste0(status_id, "-tokens-output"),
        class = "shidashi-chatbot-status-counter",
        title = "Output tokens",
        "\u2193 0"  # down arrow
      ),
      shiny::span(
        id = paste0(status_id, "-cost"),
        class = "shidashi-chatbot-status-counter",
        title = "Estimated cost",
        "$0"
      )
    )
  )
}


# chatbot_drawer_ui() has been removed.
# The drawer shell is now provided by module_drawer() in drawer.R.
# Drawer content is rendered via shiny::renderUI in the module server.


# ---- Server component ----

#' Chat-bot server logic (per-module)
#'
#' @description
#' Sets up the chat server for a single module.
#' Creates its own \code{\link[ellmer]{Chat}} object, binds tools from
#' the \verb{MCP} session registry, and manages per-module conversation
#' history.
#'
#' The function renders \code{\link{chatbot_ui}()} into the drawer's
#' \code{uiOutput} via \code{shiny::renderUI}, and opens the drawer
#' when a \code{"button.click"} shidashi-event with
#' \code{type = "open_drawer"} is received.
#' Any element with \code{data-shidashi-action="shidashi-button"}
#' \code{data-shidashi-type="open_drawer"} can trigger the drawer.
#'
#' This function is injected into each module's server function by
#' \code{modules.R} when \code{agents.yaml} has \code{enabled: yes}.
#' It is called inside \code{shiny::moduleServer()}, so \code{session}
#' is module-scoped.  \pkg{shinychat} operations use the scoped
#' \code{session}; only drawer and event operations use
#' \code{session$rootScope()}.
#'
#' @param input,output,session Standard Shiny server arguments
#'   (typically module-scoped when inside \code{moduleServer}).
#' @param id character; must match the \code{id} used in
#'   \code{chatbot_ui()}. Default \code{"shidashi-chatbot"}.
#' @param drawer_id character; the output ID of the drawer's
#'   \code{uiOutput} placeholder (from \code{\link{module_drawer}()}).
#'   Default \code{"shidashi_drawer"}.
#' @param agent_conf list; parsed \code{agents.yaml} content. Used
#'   for the system prompt and tool configuration.
#' @return Called for side effects (sets up observers). Returns
#'   \code{invisible(NULL)}.
#' @keywords internal
chatbot_server <- function(input, output, session,
                           id = "shidashi-chatbot",
                           drawer_id = "shidashi_drawer",
                           agent_conf = NULL) {
  # Guard: disabled or missing deps
  if (!isTRUE(getOption("shidashi.chatbot", TRUE))) {
    return(invisible(NULL))
  }
  if (!requireNamespace("shinychat", quietly = TRUE)) {
    return(invisible(NULL))
  }
  if (!requireNamespace("ellmer", quietly = TRUE)) {
    return(invisible(NULL))
  }

  # Register session for event bus (theme, button events, etc.)
  shidashi::register_session(session)

  # Module ID from the calling moduleServer namespace
  module_id <- session$ns(NULL)
  if (!length(module_id) || !nzchar(module_id)) {
    module_id <- "__root__"
  }

  # Handle user input from shinychat (scoped session)
  user_input_id <- paste0(id, "_user_input")

  # Conversation selector change
  conv_select_id <- paste0(id, "-conv_select")

  # New conversation button
  new_conv_id <- paste0(id, "-new_conversation")

  # Mode selector change
  mode_select_id <- paste0(id, "-mode_select")

  # Confirmation policy selector
  confirm_policy_id <- paste0(id, "-confirm_policy")

  # Stop button
  stop_btn_id <- paste0(id, "-stop")

  # ---- Local state (fastmap for non-reactive mutable state) ----
  local_data <- fastmap::fastmap()
  local_data$set("chat_token", NULL)
  local_data$set("is_streaming", FALSE)

  # Generate a new chat token (invalidates any in-flight operations)
  new_chat_token <- function() {
    token <- paste0(Sys.time(), "-", rand_string())
    local_data$set("chat_token", token)
    token
  }

  # Check if a token is still valid (matches current)
  is_token_valid <- function(token) {
    isTRUE(token == local_data$get("chat_token"))
  }

  # ---- Mode state ----
  agent_conf <- as.list(agent_conf)
  agent_modes <- agent_conf$modes %||% "None"
  agent_default_mode <- agent_conf$parameters$default_mode %||% agent_modes[[1]]
  globals_set_agent_mode(module_id = module_id, mode = agent_default_mode)
  current_mode <- shiny::reactiveVal(agent_default_mode)

  # ---- Render chatbot UI into the drawer's uiOutput ----
  output[[drawer_id]] <- shiny::renderUI({
    chatbot_ui(id = session$ns(id),
               modes = agent_modes,
               default_mode = agent_default_mode)
  })

  # ---- Per-module Chat object (created lazily) ----

  local_chat <- NULL

  ensure_chat <- function() {
    if (!is.null(local_chat)) return(local_chat)

    # Generate initial token for this chat session
    new_chat_token()

    tryCatch(
      {
        chat <- globals_new_chat(module_id = module_id, session = session)

        chat$on_tool_request(function(request) {
          if (!is_token_valid(local_data$get("callback_token"))) {
            ellmer::tool_reject(reason = paste(
              "Callback token has changed.",
              "User has terminated current conversation.",
              "You should NOT make any tool calls and stop immediately."
            ))
          }
        })

        # Capture current token for tool result callback
        chat$on_tool_result(function(result) {
          # Check if chat was stopped/reset - silently drop if so
          if (!is_token_valid(local_data$get("callback_token"))) {
            return(invisible(NULL))
          }
          if (length(result@error)) {
            try({
              shinychat::chat_append(id = id, session = session, result)
            }, silent = TRUE)
            return()
          }

          if (is.list(result@value)) {
            value_list <- result@value
            unwrap <- FALSE
          } else {
            value_list <- list(result@value)
            unwrap <- TRUE
          }

          value_list <- lapply(value_list, function(value) {
            if (
              S7::S7_inherits(value, ellmer::ContentImage)
            ) {
              img_type <- value@type
              img_data <- value@data
              value <- sprintf(
                "<img src='data:%s;base64,%s' style='max-width:100%%' />",
                img_type, img_data)
              class(value) <- "html"
            }
            value
          })

          if (unwrap) {
            result@value <- value_list[[1]]
          } else {
            result@value <- value_list
          }

          # Try to append the entire result
          try({
            shinychat::chat_append(id = id, session = session, result)
            return()
          }, silent = TRUE)

          # Ultimate fallback
          lapply(value_list, function(value) {
            tryCatch({
              if (inherits(value, "html")) {
                shinychat::chat_append(id = id, session = session, value)
              } else {
                result@value <- value
                shinychat::chat_append(id = id, session = session, result)
              }
            }, error = function(e) {
              # print(result)
              shinychat::chat_append(
                id = id,
                session = session,
                paste0(
                  "<p><span style='font-size: x-small; font-style: italic;'>",
                  "(Tool request result omitted)</span></p>"
                )
              )
            })
          })
          return()
        })

        # Send provider/model info to the status bar
        provider <- chat$get_provider()
        session$sendCustomMessage(
          "shidashi.update_chat_status",
          list(
            id     = session$ns(paste0(id, "-status-model")),
            text   = sprintf("%s/%s", provider@name, provider@model),
            status = "ready"
          )
        )

        # Initialize stop button in the chat input area
        session$sendCustomMessage(
          "shidashi.init_chat_stop_button",
          list(
            chat_id = session$ns(id),
            stop_id = session$ns(stop_btn_id)
          )
        )

        # Initialize code copy buttons for pre blocks
        session$sendCustomMessage(
          "shidashi.init_chat_code_copy",
          list(chat_id = session$ns(id))
        )

        # Share chat object
        local_chat <<- chat
        chat
      },
      error = function(e) {
        shinychat::chat_append(
          id = id,
          session = session,
          response = sprintf(
            "[shidashi] Failed to create chat: \n\n```\n%s\n```\n",
            paste(conditionMessage(e), collapse = "\n")
          ),
          role = "assistant"
        )
        NULL
      }
    )

  }

  get_tokens <- function() {
    res <- list(input = 0L, output = 0L, cached = 0L, cost = NA_real_)
    if (!inherits(local_chat, "Chat")) { return(res) }
    tokens <- local_chat$get_tokens()
    if (!length(tokens)) { return(res) }

    res$output <- sum(c(tokens$output, 0), na.rm = TRUE)
    res$input <- sum(c(tokens$input, 0), na.rm = TRUE)
    if (length(tokens$cached_input)) {
      res$cached <- tokens$cached_input[[length(tokens$cached_input)]]
    }

    res$cost <- tryCatch(
      local_chat$get_cost(),
      error = function(e) {
        NA_real_
      }
    )
    res
  }

  update_chat_status <- function(status = "ready") {
    # Update stop button visibility
    session$sendCustomMessage(
      "shidashi.toggle_stop_button",
      list(
        id = session$ns(stop_btn_id),
        visible = (status == "recalculating")
      )
    )
    if (!inherits(local_chat, "Chat")) { return() }
    tokens <- get_tokens()
    ns_prefix <- session$ns(paste0(id, "-status"))

    # Input tokens
    cached_note <- ""
    if (tokens$cached > 0L) {
      cached_note <- sprintf(" (%s cached)", format_token(tokens$cached))
    }
    session$sendCustomMessage(
      "shidashi.update_chat_status",
      list(
        id     = paste0(ns_prefix, "-tokens-input"),
        text   = paste0("\u2191 ", format_token(tokens$input + tokens$cached)),
        title  = paste0("Input tokens", cached_note),
        status = status
      )
    )
    # Output tokens
    session$sendCustomMessage(
      "shidashi.update_chat_status",
      list(
        id     = paste0(ns_prefix, "-tokens-output"),
        text   = paste0("\u2193 ", format_token(tokens$output)),
        title  = "Output tokens",
        status = status
      )
    )
    # Cost
    session$sendCustomMessage(
      "shidashi.update_chat_status",
      list(
        id     = paste0(ns_prefix, "-cost"),
        text   = format_cost(tokens$cost),
        title  = if (is.na(tokens$cost)) "Token pricing unknown" else "Estimated cost",
        status = if (is.na(tokens$cost)) "unknown" else status
      )
    )
  }

  # ---- Conversation-dropdown helpers ----
  # Push current conversation list into the selectInput dropdown
  update_conv_dropdown <- function() {
    entry <- globals_get_conversation_entry(module_id = module_id)
    choices <- seq_along(entry$conversations)
    names(choices) <- vapply(entry$conversations, function(c) {
      paste(c$title %||% "New conversation", collapse = " ")
    }, character(1))
    shiny::updateSelectInput(
      session  = session,
      inputId  = paste0(id, "-conv_select"),
      choices  = choices,
      selected = as.character(entry$active_idx)
    )
  }

  # ---- ExtendedTask for streaming chat ----
  chat_task <- shiny::ExtendedTask$new(coro::async(function(user_msg) {
    ensure_chat()
    if (is.null(local_chat)) {
      return(NULL)
    }

    # Capture token at request start for validation in callbacks
    request_token <- local_data$get("chat_token")
    local_data$set("callback_token", request_token)
    local_data$set("is_streaming", TRUE)

    update_chat_status(status = "recalculating")

    stream <- local_chat$stream_async(
      user_msg,
      tool_mode = "sequential",
      stream = "text"
    )

    tryCatch(
      {
        user_stopped <- FALSE
        while (TRUE) {
          if (is_token_valid(request_token)) {
            content <- coro::await(stream())
          } else {
            stream(close = TRUE)
            user_stopped <- TRUE
            content <- coro::exhausted()
          }

          if (user_stopped) {
            shinychat::chat_append(
              id,
              response = "\n\n*[Generation stopped by user]*",
              role = "assistant",
              session = session
            )
          } else if (!coro::is_exhausted(content)) {
            shinychat::chat_append_message(
              id = id,
              msg = list(role = "assistant", content = content),
              chunk = "end",
              operation = "append"
            )
          } else {
            # Make sure to end the message to reenable the chatbox
            shinychat::chat_append_message(
              id = id,
              msg = list(role = "assistant", content = ""),
              chunk = "end",
              operation = "append"
            )
          }

          if (coro::is_exhausted(content)) {
            request_token <- "N/A"
            local_data$set("callback_token", request_token)
            local_data$set("is_streaming", FALSE)
            break
          }
        }
      },
      error = function(e) {
        shinychat::chat_append(
          id,
          response = sprintf(
            "Error %s",
            paste(cli::ansi_strip(conditionMessage(e)), collapse = " ")
          ),
          role = "assistant",
          session = session
        )
      }
    ) # tryCatch

    # Runs after the stream is fully consumed -> tokens are available
    globals_save_conversation(module_id = module_id, chat = local_chat)
    update_conv_dropdown()
    update_chat_status()

  }))

  # ---- Observers ----

  # On user prompt
  shiny::bindEvent(
    shiny::observe({
      user_msg <- paste(input[[user_input_id]], collapse = "")
      user_msg <- trimws(user_msg)
      if (!nzchar(user_msg)) { return() }
      chat_task$invoke(user_msg)
    }),
    input[[user_input_id]],
    ignoreNULL = TRUE,
    ignoreInit = TRUE
  )

  # On switching conversation ID
  shiny::bindEvent(
    shiny::observe({

      selected <- as.integer(input[[conv_select_id]])
      if (length(selected) != 1 || is.na(selected)) return()

      entry <- globals_get_conversation_entry(module_id = module_id)
      if (
        selected < 1L ||
        selected > length(entry$conversations) ||
        isTRUE(entry$active_idx == selected)) {

        # no need to change
        return()
      }

      # Invalidate any pending operations from previous conversation
      new_chat_token()

      # Save current conversation before switching
      globals_save_conversation(module_id = module_id, chat = local_chat)

      # Switch conversation
      entry$active_idx <- selected
      globals_set_conversation_entry(entry = entry, module_id = module_id)

      # restore chat
      ensure_chat()
      if (is.null(local_chat)) return()

      # Load the selected conversation's turns into local_chat
      conv <- entry$conversations[[selected]]
      if (length(conv$turns)) {
        local_chat$set_turns(conv$turns)
      } else {
        local_chat$set_turns(list())
      }

      # Update the chat widget
      shinychat::chat_clear(id = id, session = session)

      # as of 0.3.0, chat_restore does not work as expected and throws
      # S7 error. Manually restore text instead
      # shinychat::chat_restore
      lapply(
        local_chat$get_turns(include_system_prompt = FALSE),
        function(turn) {
          tryCatch(
            {
              shinychat::chat_append(
                id = id,
                response = turn@text,
                role = turn@role,
                session = session
              )
            },
            error = function(e) {
              # pass
            }
          )
        }
      )
      return()

    }),
    input[[conv_select_id]],
    ignoreNULL = TRUE,
    ignoreInit = TRUE
  )

  # On starting new conversation
  shiny::bindEvent(
    shiny::observe({

      # Invalidate any pending operations from previous conversation
      new_chat_token()

      # Save current conversation before starting a new one
      globals_save_conversation(module_id = module_id, chat = local_chat)

      chat <- ensure_chat()
      if (is.null(chat)) return()

      # Already saved, hence save_first = FALSE
      globals_new_conversation(module_id = module_id,
                               chat = chat,
                               save_first = FALSE)

      # Clear the UI
      shinychat::chat_clear(id = id, session = session)
      update_conv_dropdown()

    }),
    input[[new_conv_id]],
    ignoreNULL = TRUE,
    ignoreInit = TRUE
  )

  # On stop button click
  shiny::bindEvent(
    shiny::observe({
      if (!isTRUE(local_data$get("is_streaming"))) return()

      # Invalidate current token - this causes all in-flight operations
      # to be silently dropped when they check their captured token
      new_chat_token()
      local_data$set("is_streaming", FALSE)

      # Save whatever we have and update UI
      globals_save_conversation(module_id = module_id, chat = local_chat)
      update_conv_dropdown()
      update_chat_status(status = "ready")
    }),
    input[[stop_btn_id]],
    ignoreNULL = TRUE,
    ignoreInit = TRUE
  )

  # On changing permission mode
  shiny::bindEvent(
    shiny::observe({
      new_mode <- input[[mode_select_id]]
      if (length(new_mode) != 1 || !nzchar(new_mode)) return()
      current_mode(new_mode)

      # Sync mode into globals so MCP handler can read it
      globals_set_agent_mode(module_id = module_id, mode = new_mode)

      if (!inherits(local_chat, "Chat")) { return() }

      globals_bind_chat_tools(chat = local_chat,
                              module_id = module_id,
                              session = session)

    }),
    input[[mode_select_id]],
    ignoreNULL = TRUE,
    ignoreInit = TRUE
  )

  # On changing confirmation policy
  shiny::bindEvent(
    shiny::observe({
      policy <- input[[confirm_policy_id]]
      if (!length(policy) || !nzchar(policy)) return()
      globals_set_confirmation_policy(module_id = module_id, policy = policy)
    }),
    input[[confirm_policy_id]],
    ignoreNULL = TRUE,
    ignoreInit = TRUE
  )

  # Save conversation on session end
  shiny::onSessionEnded(fun = function() {
    globals_save_conversation(module_id = module_id, chat = local_chat)
  }, session = session)

  invisible(NULL)
}

# ---- Skill wrapper for Phase 4 skills ----
#
# Turns an Anthropic-compliant skill directory (SKILL.md + optional
# scripts/ + reference files) into a single MCP tool via closure.
# Mirrors the mcp_wrapper() pattern from Phase 3 but with progressive
# disclosure and a gate mechanism.

#' Wrap a Skill Directory as an \verb{MCP} Tool Generator
#'
#' @description
#' Creates a closure that produces an
#' \code{ellmer::tool} dispatching on an \code{action} enumerator:
#' \describe{
#'   \item{\code{readme}}{Returns the full \verb{SKILL.md} instructions. Must be
#'     called first to unlock other actions.}
#'   \item{\code{reference}}{Returns content from a reference file in the
#'     skill directory (gated behind \code{readme}).}
#'   \item{\code{script}}{Executes a script in the \code{scripts/}
#'     subdirectory via \code{processx::run()} (gated behind \code{readme}).}
#' }
#'
#' @param skill_path Path to the skill directory containing \code{SKILL.md}.
#'   Can be absolute or relative to the project root.
#'
#' @return A function with class \code{c("shidashi_skill_wrapper", "function")}
#'   that returns an \code{ellmer::ToolDef} object.
#'
#' @details
#'   The returned tool enforces a soft gate: calling \code{reference} or
#'   \code{script} before \code{readme} is allowed, but if the call
#'   errors the message is augmented with a condensed summary (~200
#'   tokens) instructing the AI to read the full instructions first.
#'   This minimizes token waste (the summary is only sent on failure).
#'
#'   The gate state is per-instance: each call to the wrapper produces
#'   a closure with an independent \code{readme_unlocked} flag.
#'
#' @examples
#' skill_dir <- system.file(
#'   "builtin-templates/bslib-bare/agents/skills/greet",
#'   package = "shidashi"
#' )
#' wrapper  <- skill_wrapper(skill_dir)
#' tool_def <- wrapper()
#' cat(tool_def(action = "readme"))
#'
#' @export
skill_wrapper <- function(skill_path) {

  # Validate skill directory at definition time
  skill_md_path <- file.path(skill_path, "SKILL.md")
  if (!file.exists(skill_md_path)) {
    stop("SKILL.md not found in: ", skill_path)
  }

  # Parse once at definition time (immutable metadata)
  parsed <- parse_skill_md(skill_md_path)
  ref_files <- discover_references(parsed$skill_dir)
  script_files <- discover_scripts(parsed$skill_dir)

  # Per Anthropic spec, the canonical skill name is the folder name
  canonical_name <- basename(normalizePath(skill_path))

  # Pre-build condensed summary for gate errors (use folder name)
  parsed$name <- canonical_name
  condensed <- build_condensed_summary(parsed, ref_files, script_files)

  # Determine available actions
  available_actions <- "readme"
  if (length(ref_files)) {
    available_actions <- c(available_actions, "reference")
  }
  if (length(script_files)) {
    available_actions <- c(available_actions, "script")
  }

  # Build short description for tools/list (Tier 1: ~30 tokens)
  tool_description <- paste0(
    parsed$description,
    " [Skill: call action='readme' first",
    if (length(script_files)) {
      paste0("; scripts: ", paste(script_files, collapse = ", "))
    },
    "]"
  )

  # Tool name derived from folder name (used as @name on the ToolDef)
  tool_name <- canonical_name

  # Build the argument schema — always include all params so formals match
  # (ellmer::tool requires arguments names to match function formals)
  has_references <- length(ref_files) > 0
  has_scripts <- length(script_files) > 0

  ref_desc <- if (has_references) {
    paste0("Available: ", paste(ref_files, collapse = ", "))
  } else {
    "No reference files available for this skill"
  }
  script_desc <- if (has_scripts) {
    paste0("Available: ", paste(script_files, collapse = ", "))
  } else {
    "No scripts available for this skill"
  }

  arg_list <- list(
    action = ellmer::type_enum(
      values = available_actions,
      description = paste0(
        "Action to perform. IMPORTANT: You must call 'readme' first before ",
        "any other action. Available: ",
        paste(available_actions, collapse = ", ")
      )
    ),
    file_name = ellmer::type_string(
      description = paste0(
        "File name for action='reference' or action='script'. ",
        "References: ", ref_desc, ". Scripts: ", script_desc
      ),
      required = FALSE
    ),
    pattern = ellmer::type_string(
      description = "For action='reference': optional grep pattern to filter lines.", # nolint: line_length_linter.
      required = FALSE
    ),
    line_start = ellmer::type_integer(
      description = "For action='reference': start line (1-based). Default: 1.",
      required = FALSE
    ),
    n_lines = ellmer::type_integer(
      description = "For action='reference': max lines to return. Default: 200.", # nolint: line_length_linter.
      required = FALSE
    ),
    args = ellmer::type_array(
      items = ellmer::type_string(),
      description = "For action='script': CLI arguments to pass to the script.",
      required = FALSE
    ),
    envs = ellmer::type_array(
      items = ellmer::type_string(),
      description = "For action='script': environment variables as KEY=VALUE strings.", # nolint: line_length_linter.
      required = FALSE
    )
  )

  # Build the generator function (closure factory)
  structure(
    function() {

      # Per-instance gate state
      readme_unlocked <- FALSE
      skill_dir <- parsed$skill_dir

      # The tool function that dispatches on action
      tool_fn <- function(action, file_name = NULL, pattern = NULL,
                          line_start = NULL, n_lines = NULL,
                          args = NULL, envs = NULL) {

        # ---- Wrap dispatch: augment errors when readme not yet read ----
        result <- tryCatch(
          {
            # ---- Dispatch ----
            switch(
          action,
          "readme" = {
            readme_unlocked <<- TRUE
            # Build runtime context
            info_parts <- character()
            info_parts <- c(info_parts, parsed$body)

            if (length(ref_files)) {
              info_parts <- c(info_parts, "",
                "## Available reference files",
                paste("-", ref_files)
              )
            }
            if (length(script_files)) {
              info_parts <- c(info_parts, "",
                "## Available scripts",
                paste("-", script_files)
              )
            }

            paste(info_parts, collapse = "\n")
          },

          "reference" = {
            if (!length(file_name) || !nzchar(file_name)) {
              stop("file_name is required for action='reference'. ",
                   "Available: ", paste(ref_files, collapse = ", "),
                   call. = FALSE)
            }
            # Fuzzy match: case-insensitive and supports "references/file"
            matched_file <- fuzzy_match_reference(file_name, ref_files)
            if (is.null(matched_file)) {
              stop("Reference file not found: ", file_name,
                   "\nAvailable: ", paste(ref_files, collapse = ", "),
                   call. = FALSE)
            }
            file_name <- matched_file

            ref_path <- file.path(skill_dir, file_name)
            all_lines <- readLines(ref_path, warn = FALSE)

            # Apply pattern filter if given
            if (length(pattern) && nzchar(pattern)) {
              matched <- grep(pattern, all_lines)
              if (!length(matched)) {
                return(paste0("No lines matching pattern '", pattern,
                              "' in ", file_name,
                              " (", length(all_lines), " total lines)"))
              }
              all_lines <- all_lines[matched]
            }

            # Paginate
            start <- max(1L, as.integer(line_start %||% 1L))
            max_n <- as.integer(n_lines %||% 200L)
            end <- min(length(all_lines), start + max_n - 1L)

            result_lines <- all_lines[start:end]
            header <- sprintf(
              "## %s (lines %d-%d of %d)\n",
              file_name, start, end, length(all_lines)
            )
            paste0(header, paste(result_lines, collapse = "\n"))
          },

          "script" = {
            if (!length(file_name) || !nzchar(file_name)) {
              stop("file_name is required for action='script'. ",
                   "Available: ", paste(script_files, collapse = ", "),
                   call. = FALSE)
            }
            if (!file_name %in% script_files) {
              stop("Script not found: ", file_name,
                   "\nAvailable: ", paste(script_files, collapse = ", "),
                   call. = FALSE)
            }

            # Parse envs from KEY=VALUE strings to named character vector
            env_vec <- character()
            if (is.character(envs) && length(envs)) {
              # Split on first '=' only
              parts <- regmatches(envs, regexpr("=", envs), invert = TRUE)
              valid <- vapply(parts, length, integer(1L)) == 2L
              if (any(valid)) {
                keys <- vapply(parts[valid], `[[`, character(1L), 1L)
                vals <- vapply(parts[valid], `[[`, character(1L), 2L)
                env_vec <- structure(vals, names = keys)
              }
            } else if (is.list(envs) && length(envs)) {
              env_vec <- vapply(envs, as.character, character(1L))
            }

            result <- run_skill_script(
              skill_dir = skill_dir,
              file_name = file_name,
              args = as.character(args %||% character()),
              envs = env_vec,
              timeout_seconds = 60
            )

            # Format output
            parts <- character()
            if (nzchar(result$stdout)) {
              parts <- c(parts, "## stdout", result$stdout)
            }
            if (nzchar(result$stderr)) {
              parts <- c(parts, "## stderr", result$stderr)
            }
            parts <- c(parts, paste0("\nExit code: ", result$status))
            if (isTRUE(result$timeout)) {
              parts <- c(parts, "WARNING: Script timed out after 60 seconds.")
            }
            paste(parts, collapse = "\n")
          },

          stop("Unknown action: ", action,
               "\nAvailable: ", paste(available_actions, collapse = ", "),
               call. = FALSE)
            )
          },
          error = function(e) {
            if (!readme_unlocked && action != "readme") {
              stop(
                conditionMessage(e),
                "\n\n---\nYou need to read the skill instructions with ",
                "action='readme' first.",
                "\n\n", condensed,
                call. = FALSE
              )
            }
            stop(e)
          }
        )
        result
      }

      # Return an ellmer::tool with the skill's metadata
      ellmer::tool(
        fun         = tool_fn,
        name        = tool_name,
        description = tool_description,
        arguments   = arg_list
      )
    },
    class = c("shidashi_skill_wrapper", "function")
  )
}

# ---- ellmer helpers: Content → MCP content serialization ----
#
# Bridges ellmer Content objects to MCP JSON-RPC content items.
#
# The MCP protocol defines its own content format:
#   - text:     {"type": "text", "text": "..."}
#   - image:    {"type": "image", "data": "base64...", "mimeType": "image/png"}
#   - resource: {"type": "resource", "resource": {"uri": "...", "mimeType": "..."}}
#
# This file provides:
#   - ProviderAny: fallback S7 Provider subclass producing MCP-format content
#   - as_json methods for ProviderAny + ellmer Content types
#   - get_mcp_provider(req): detect or default to ProviderAny
#   - content_to_mcp(ret, provider): convert tool return values to MCP content

# --- External generic reference ---

ellmer_as_json <- S7::new_external_generic("ellmer", "as_json", c("provider", "x"))

# --- Fallback MCP provider ---
# Used when we cannot determine the actual LLM provider from the request.
# Produces MCP-standard content format from as_json dispatch.

ProviderAny <- S7::new_class(
  name = "ProviderAny",
  parent = ellmer::Provider
)

# --- as_json methods for ProviderAny ---

# ContentText → MCP text content
S7::method(ellmer_as_json, list(ProviderAny, ellmer::ContentText)) <- function(provider, x, ...) {
  list(type = "text", text = x@text)
}

# ContentImageInline → MCP image content
S7::method(ellmer_as_json, list(ProviderAny, ellmer::ContentImageInline)) <- function(provider, x, ...) {
  list(type = "image", data = x@data, mimeType = x@type)
}

# ContentImageRemote → MCP resource content (URL reference)
S7::method(ellmer_as_json, list(ProviderAny, ellmer::ContentImageRemote)) <- function(provider, x, ...) {
  list(type = "resource", resource = list(uri = x@url, mimeType = "image/jpeg"))
}

# Generic Content fallback: extract S7 props
S7::method(ellmer_as_json, list(ProviderAny, ellmer::Content)) <- function(provider, x, ...) {
  nms <- S7::prop_names(x)
  nms <- nms[!nms %in% "request"]
  res <- S7::props(x, nms)
  # Wrap as text with JSON representation
  list(
    type = "text",
    text = as.character(
      jsonlite::toJSON(res, auto_unbox = TRUE, null = "null", force = TRUE)
    )
  )
}

# character → text content
S7::method(ellmer_as_json, list(ProviderAny, S7::class_character)) <- function(provider, x, ...) {
  list(type = "text", text = paste(x, collapse = "\n"))
}

# any → JSON-serialized text content
S7::method(ellmer_as_json, list(ProviderAny, S7::class_any)) <- function(provider, x, ...) {
  list(
    type = "text",
    text = as.character(
      jsonlite::toJSON(x, auto_unbox = TRUE, null = "null", force = TRUE)
    )
  )
}

# --- Provider detection ---

#' Detect or create an ellmer Provider for MCP content serialization
#'
#' Examines the HTTP request headers (e.g. \code{User-Agent}) to try to
#' identify the upstream LLM provider.  Falls back to \code{ProviderAny}
#' which produces MCP-standard content format.
#'
#' @param req The Rook request environment, or \code{NULL}.
#' @return An \code{ellmer::Provider} instance.
#' @keywords internal
#' @noRd
get_mcp_provider <- function(req = NULL) {
  # TODO: detect actual provider from request headers
  # e.g. req$HTTP_USER_AGENT might contain "anthropic", "openai", etc.
  # For now, always return the MCP-format fallback provider.
  ProviderAny(name = "mcp", model = "unknown", base_url = "http://localhost")
}

# --- Content → MCP result conversion ---

#' Convert a tool return value to MCP content format
#'
#' Handles ellmer \code{Content} objects, \code{ContentToolResult},
#' lists of \code{Content}, plain strings, and arbitrary R objects.
#' Returns a list with \code{content} (array of MCP content items) and
#' \code{isError} (logical).
#'
#' @param ret The raw return value from a tool function.
#' @param provider An \code{ellmer::Provider} for content serialization.
#'   When \code{NULL}, \code{get_mcp_provider()} is used.
#' @return A list with \code{content} and \code{isError}.
#' @keywords internal
#' @noRd
content_to_mcp <- function(ret, provider = NULL) {
  if (is.null(provider)) {
    provider <- get_mcp_provider()
  }
  as_json_fn <- asNamespace("ellmer")[["as_json"]]

  # Unwrap ContentToolResult
  if (S7::S7_inherits(ret, ellmer::ContentToolResult)) {
    if (!is.null(ret@error)) {
      err_msg <- if (inherits(ret@error, "condition")) {
        conditionMessage(ret@error)
      } else {
        as.character(ret@error)
      }
      return(list(
        content = list(list(type = "text", text = paste("Error:", err_msg))),
        isError = TRUE
      ))
    }
    ret <- ret@value
  }

  # Single Content object
  if (S7::S7_inherits(ret, ellmer::Content)) {
    return(list(
      content = list(as_json_fn(provider, ret)),
      isError = FALSE
    ))
  }

  # List of Content objects
  if (is.list(ret) && length(ret) > 0) {
    is_content <- vapply(
      ret,
      function(x) S7::S7_inherits(x, ellmer::Content),
      logical(1)
    )
    if (all(is_content)) {
      return(list(
        content = lapply(ret, function(x) as_json_fn(provider, x)),
        isError = FALSE
      ))
    }
  }

  # NULL
  if (is.null(ret)) {
    return(list(
      content = list(list(type = "text", text = "<empty results>")),
      isError = FALSE
    ))
  }

  # Character
  if (is.character(ret)) {
    return(list(
      content = list(list(type = "text", text = paste(ret, collapse = "\n"))),
      isError = FALSE
    ))
  }

  # Other: JSON serialize and wrap in text content
  list(
    content = list(as_json_fn(provider, ret)),
    isError = FALSE
  )
}

#' @name adminlte
#' @title Generates 'AdminLTE' theme-related 'HTML' tags
#' @description These functions should be called in 'HTML' templates.
#' Please see vignettes for details.
#' @param root_path the root path of the website project; see
#' \code{\link{template_settings}}
#' @param settings_file the settings file containing the module information
#' @param shared_id a shared identification by session to synchronize the
#' inputs; assigned internally.
#' @return 'HTML' tags
#' @export
adminlte_ui <- function(root_path = template_root()) {
  function(req) {

    tryCatch({
      resource <- load_module(root_path = root_path, request = req)
      env <- new.env(parent = resource$environment)

      if (resource$has_module) {
        # load module UI
        template_path <- resource$module$template_path
      } else {
        if (!length(resource$module$id)) {
          # load overall UI
          template_path <- resource$template_path
        } else {
          # 404
          template_path <- file.path(root_path, "views", "404.html")
          if (!file.exists(template_path)) {
            return("Page not found (404)")
          }
        }
      }

      `@args` <- as.list(resource$environment, all.names = TRUE)
      `@args`$filename <- template_path
      call <- as.call(c(list(quote(shiny::htmlTemplate)), `@args`))
      return(eval(call, envir = env))
    }, error = function(e) {
      module_template <- file.path(root_path, "views", "500.html")
      error <- shiny::pre(
        style = "word-wrap: break-word; white-space: break-spaces;",
        paste(
          sep = "\n",
          "Error message:",
          e$message,
          "Traceback:",
          paste(utils::capture.output({
            traceback(e)
          }), collapse = "\n")
        )
      )
      if (file.exists(module_template)) {

        return(shiny::htmlTemplate(module_template, error = error, req = req))
      } else {
        return(paste("Internal error: <br/>", error))
      }
    })

  }

}


#' @rdname adminlte
#' @export
adminlte_sidebar <- function(root_path = template_root(),
                             settings_file = "modules.yaml",
                             shared_id = rand_string(26)) {
  settings <- yaml::read_yaml(file.path(root_path, settings_file))
  # settings <- yaml::read_yaml('modules.yaml')

  divider <- settings$divider
  if (length(divider)) {
    divider <- data.frame(
      name = names(divider),
      order = sapply(divider, function(x) {
        if (isTRUE(is.numeric(x$order))) {
          x$order
        } else {
          NA
        }
      })
    )
    divider <- divider[!is.na(divider$order), ]
  }
  if (!length(divider) || !nrow(divider)) {
    divider <- data.frame(name = "END", order = Inf)
  }

  groups <- settings$groups
  if (length(groups)) {
    groups <- groups[names(groups) != ""]
  }
  group_icons <- sapply(groups, function(x) { ifelse(length(x$icon) == 1, x$icon, "") })
  group_badge <- sapply(groups, function(x) { ifelse(length(x$badge) == 1, x$badge, "") })
  group_order <- sapply(groups, function(x) { ifelse(length(x$order) == 1, x$order, NA) })
  group_open <- sapply(groups, function(x) { isTRUE(x$open) })
  groups <- names(groups)

  modules <- settings$modules
  modules_ids <- names(settings$modules)

  groups <- unique(groups)
  group_level <- factor(groups, levels = groups, ordered = TRUE)

  module_tbl <- do.call("rbind", lapply(modules_ids, function(mid) {
    x <- modules[[mid]]
    if (isTRUE(x$hidden)) {
      return(NULL)
    }
    y <- x[!names(x) %in% c("order", "group", "label", "icon", "badge", "module", "hidden")]
    y$module <- mid
    y$shared_id <- shared_id
    url <- httr2::url_modify("https://dipterix.org/?module=", query = y)
    order <- x$order
    if (!length(order) || is.na(order)) {
      order <- 9999L
    }

    if (length(x$group) == 1 && x$group %in% group_level) {
      x$group <- group_level[group_level == x$group][[1]]
      renderOrder <- group_order[group_level == x$group][[1]] + order / 10000
    } else {
      x$group <- NA
      renderOrder <- order
    }


    data.frame(
      id = mid,
      order = order,
      renderOrder = renderOrder,
      group = x$group,
      label = ifelse(length(x$label) == 1, x$label, "No Label"),
      icon = ifelse(length(x$icon) == 1, x$icon, ""),
      badge = ifelse(length(x$badge) == 1, x$badge, ""),
      url = gsub("^[^\\?]+", "", url),
      stringsAsFactors = FALSE
    )
  }))

  if (nrow(module_tbl)) {
    max_order <- max(c(module_tbl$renderOrder, 10000), na.rm = TRUE) + 1
    # group could be NA, resulting in warning
    suppressWarnings({
      module_tbl <- module_tbl[
        order(module_tbl$renderOrder),
      ]
    })
  }

  side_bar <- list()
  ignore_group <- NULL
  last_order <- -Inf

  for (i in seq_len(nrow(module_tbl))) {
    x <- module_tbl[i, ]

    current_order <- x$renderOrder

    divide_item <- divider[divider$order <= current_order &
                             divider$order > last_order, ]
    if (nrow(divide_item)) {
      for (j in seq_len(nrow(divide_item))) {
        tmp <- divide_item[j, ]
        side_bar[[length(side_bar) + 1]] <- shiny::tags$li(
          class = "nav-header nav-divider",
          shiny::span(
            tmp$name
          )
        )
      }
    }
    last_order <- current_order

    if (is.na(x$group)) {
      item <- menu_item(text = x$label, icon = x$icon, href = x$url, badge = x$badge)
      side_bar[[length(side_bar) + 1]] <- item
    } else if (!as.character(x$group) %in% ignore_group) {

      # add group
      group <- x$group
      ignore_group <- c(ignore_group, as.character(group))
      sub <- module_tbl[!is.na(module_tbl$group) & module_tbl$group == group, ]
      if (nrow(sub)) {

        menu <- lapply(seq_len(nrow(sub)), function(ii) {
          x <- sub[ii, ]
          menu_item(text = x$label, icon = x$icon, href = x$url, badge = x$badge)
        })
        sel <- which(groups %in% group)[[1]]
        item <- menu_item_dropdown(text = group, .list = menu,
                                   icon = group_icons[sel],
                                   badge = group_badge[sel],
                                   active = group_open[sel])
        side_bar[[length(side_bar) + 1]] <- item
      }

    }
  }

  shiny::tagList(side_bar)

  # shiny::tagList(
  #   lapply(seq_along(groups), function(ii) {
  #     group <- groups[[ii]]
  #     sub <- module_tbl[!is.na(module_tbl$group) & module_tbl$group == group, ]
  #     if (!nrow(sub)) { return(NULL) }
  #     menu <- lapply(seq_len(nrow(sub)), function(ii) {
  #       x <- sub[ii, ]
  #       menu_item(text = x$label, icon = x$icon, href = x$url, badge = x$badge)
  #     })
  #     menu_item_dropdown(text = group, .list = menu, icon = group_icons[[ii]], badge = group_badge[[ii]])
  #   }),
  #   local({
  #     sub <- module_tbl[is.na(module_tbl$group), ]
  #     if (!nrow(sub)) { return(NULL) }
  #     lapply(seq_len(nrow(sub)), function(ii) {
  #       x <- sub[ii, ]
  #       menu_item(text = x$label, icon = x$icon, href = x$url, badge = x$badge)
  #     })
  #   })
  # )
}

#' @title Wrapper of shiny progress that can run without shiny
#' @param title the title of the progress
#' @param outputId the element id of \code{\link{progressOutput}}, or
#' \code{NULL} to use the default shiny progress
#' @param max max steps of the procedure
#' @param ... passed to initialization method of \code{\link[shiny]{Progress}}
#' @param quiet whether the progress needs to be quiet
#' @param session shiny session domain
#' @param shiny_auto_close whether to close the progress once function exits
#' @param log alternative log function
#' @return a list of functions that controls the progress
#' @examples
#'
#' {
#'   progress <- shiny_progress("Procedure A", max = 10)
#'   for(i in 1:10) {
#'     progress$inc(sprintf("Step %s", i))
#'     Sys.sleep(0.1)
#'   }
#'   progress$close()
#'
#' }
#'
#' if(interactive()) {
#'   library(shiny)
#'
#'   ui <- fluidPage(
#'     fluidRow(
#'       column(12, actionButton("click", "Click me"))
#'     )
#'   )
#'
#'   server <- function(input, output, session) {
#'     observeEvent(input$click, {
#'       progress <- shiny_progress("Procedure B", max = 10,
#'                                  shiny_auto_close = TRUE)
#'       for(i in 1:10) {
#'         progress$inc(sprintf("Step %s", i))
#'         Sys.sleep(0.1)
#'       }
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
#'
#' @export
shiny_progress <- function(
  title, max = 1, ..., quiet = FALSE,
  session = shiny::getDefaultReactiveDomain(), shiny_auto_close = FALSE,
  log = NULL, outputId = NULL) {
  if (missing(title) || is.null(title)) {
    title <- ""
  }
  if (length(title) > 1) {
    title <- paste(title, collapse = "")
  }
  if (inherits(session, c("ShinySession", "session_proxy",
                          "R6"))) {
    within_shiny <- TRUE
  } else {
    within_shiny <- FALSE
  }
  current <- 0
  closed <- FALSE
  get_value <- function() {
    current
  }
  is_closed <- function() {
    closed
  }
  logger <- function(..., .quiet = quiet, level = "DEFAULT",
                     bullet = "play") {
    if (!.quiet) {
      if (is.function(log)) {
        log(...)
      } else {
        s <- paste(..., collapse = "", sep = "")
        nz <- nchar(s, allowNA = TRUE, keepNA = TRUE)
        w <- getOption("width", 80L)
        s <- paste0(s, paste(rep(" ", w - nz %% w), collapse = ""))
        message("\r", s, appendLF = identical(bullet, "stop"))
      }
    }
  }
  if (quiet || !within_shiny) {
    progress <- NULL
    logger(sprintf("[%s]: initializing...", title), level = "DEFAULT",
           bullet = "play")
    inc <- function(detail, message = NULL, amount = 1, ...) {
      stopifnot(!closed)
      quiet <- c(list(...)[["quiet"]], quiet)[[1]]
      if (!is.null(message) && length(message) == 1) {
        title <<- message
      }
      current <<- amount + current
      logger(sprintf("[%s]: %s (%d out of %d)", title,
                     detail, current, max), level = "DEFAULT", bullet = "arrow_right",
             .quiet = quiet)
    }
    close <- function(message = "Finished") {
      closed <<- TRUE
      logger(message, level = "DEFAULT", bullet = "stop")
    }
    reset <- function(detail = "", message = "", value = 0) {
      title <<- message
      current <<- value
    }
  } else if (length(outputId)) {
    progress <- NULL
    inc <- function(detail, message = NULL, amount = 1, ...) {
      current <<- current + amount
      if (length(message)) { title <<- message[[1]] }
      session$sendCustomMessage("shidashi.set_progress", list(
        outputId = session$ns(outputId),
        value = current,
        max = max,
        description = paste(title, detail, sep = " - ")[[1]]
      ))
    }
    close <- function(message = "Finished") {
      closed <<- TRUE
      session$sendCustomMessage("shidashi.set_progress", list(
        outputId = outputId,
        value = max,
        max = max,
        description = message
      ))
    }
    reset <- function(detail = "", message = "", value = 0) {
      if (length(message)) {
        title <<- message[[1]]
      } else {
        title <<- ""
      }
      if (length(detail)) {
        detail <- paste(title, detail, sep = " - ")[[1]]
      } else {
        detail <- message
      }
      current <<- value
      session$sendCustomMessage("shidashi.set_progress", list(
        outputId = outputId,
        value = current,
        max = max,
        description = detail
      ))
    }
  } else {
    progress <- shiny::Progress$new(session = session, max = max,
                                    ...)
    inc <- function(detail, message = NULL, amount = 1, ...) {
      if (!is.null(message) && length(message) == 1) {
        title <<- message
      }
      progress$inc(detail = detail, message = title, amount = amount)
    }
    close <- function(message = "Finished") {
      if (!closed) {
        progress$close()
        closed <<- TRUE
      }
    }
    reset <- function(detail = "", message = "", value = 0) {
      title <<- message
      current <<- value
      progress$set(value = value, message = title, detail = detail)
    }
    if (shiny_auto_close) {
      parent_frame <- parent.frame()
      do.call(on.exit, list(substitute(close()), add = TRUE),
              envir = parent_frame)
    }
    inc(detail = "Initializing...", amount = 0)
  }
  return(list(.progress = progress, inc = inc, close = close,
              reset = reset, get_value = get_value, is_closed = is_closed))
}

#' Progress bar in shiny dashboard
#' @description For detailed usage, see demo application by
#' running \code{render()}.
#' @param outputId the element id of the progress
#' @param expr R expression that should return a named list of \code{value} and
#' \code{description}
#' @param env where to evaluate \code{expr}
#' @param quoted whether \code{expr} is quoted
#' @param outputArgs a list of other parameters in \code{progressOutput}
#' @param ... extra elements on the top of the progress bar
#' @param description descriptive message below the progress bar
#' @param width width of the progress
#' @param class progress class, default is \code{"bg-primary"}
#' @param value initial value, ranging from 0 to 100; default is 0
#' @param size size of the progress bar; choices are \code{"md"}, \code{"sm"},
#' \code{"xs"}
#' @return \code{progressOutput} returns 'HTML' tags containing progress bars
#' that can be rendered later via \code{\link{shiny_progress}} or
#' \code{renderProgress}. \code{renderProgress} returns shiny render functions
#' internally.
#' @examples
#'
#' library(shiny)
#' library(shidashi)
#' progressOutput("sales_report_prog1",
#'                description = "6 days left!",
#'                "Add Products to Cart",
#'                span(class="float-end", "123/150"),
#'                value = 123/150 * 100)
#'
#' # server function
#' server <- function(input, output, session, ...) {
#'   output$sales_report_prog1 <- renderProgress({
#'     return(list(
#'       value = 140 / 150 * 100,
#'       description = "5 days left!"
#'     ))
#'   })
#' }
#'
#' @export
progressOutput <- function(
  outputId, ..., description = "Initializing",
  width = "100%", class = "bg-primary",
  value = 0, size = c("md", "sm", "xs")
) {

  if (value < 0) {
    value <- 0
  } else if (value > 100) {
    value <- 100L
  }
  size <- match.arg(size)

  shiny::div(
    class = "shidashi-progress-output progress-group",
    id = outputId,
    style = sprintf("width: %s;", width),
    ...,
    shiny::div(
      class = sprintf("progress progress-%s", size),
      shiny::div(
        class = combine_class("progress-bar", class),
        style = sprintf("width: %.0f%%", value),
      )
    ),
    shiny::span(
      class = "progress-description progress-message",
      description
    ),
    shiny::span(
      class = "progress-description progress-error"
    )
  )

}

#' @rdname progressOutput
#' @export
renderProgress <- function(expr,
                           env = parent.frame(),
                           quoted = FALSE,
                           outputArgs = list()) {

  func <- shiny::installExprFunction(expr, "func", env, quoted, label = "renderProgress")
  shiny::createRenderFunction(func, function(value, session, name, ...) {
    if (is.list(value)) {
      description <- value$description
      value <- value$value
    } else {
      description <- NULL
    }
    if (!length(value)) {
      value <- 0L
    } else {
      value <- as.integer(value)
    }

    if ( value < 0L ) { value <- 0L }
    if ( value > 100L ) { value <- 100L }
    list(
      value = value,
      description = description
    )

  }, progressOutput, outputArgs)

}

# ---- Standalone viewer for popped-out outputs ----

#' @title Server function for the standalone viewer module
#' @description Internal server function used by the \code{standalone_viewer}
#'   hidden module. Retrieves the render function from the parent module session
#'   and assigns it to the viewer's output.
#' @param input,output,session Shiny module server arguments
#' @param ... ignored
#' @keywords internal
#' @export
server_standalone_viewer <- function(input, output, session, ...) {

  # The viewer's root scope (same root as all modules in this app)
  root_session <- session$rootScope()

  # Parse query parameters from the URL
  query <- shiny::parseQueryString(
    shiny::isolate(session$clientData$url_search)
  )
  url_output_id <- query$outputId  # namespaced, e.g. "demo-iris_plot"
  url_token     <- query$token     # root session token

  if (!length(url_output_id) || !nzchar(url_output_id) ||
      !length(url_token)     || !nzchar(url_token)) {
    output$viewer_content <- shiny::renderUI({
      shiny::h3("Missing outputId or token in URL.")
    })
    return(invisible())
  }

  # Look up the parent session's entry by token
  entry <- get_session_entry(url_token)
  if (is.null(entry)) {
    output$viewer_content <- shiny::renderUI({
      shiny::h3("Session not found. The parent session may have closed.")
    })
    return(invisible())
  }

  # entry$shiny_session is the module's session proxy (namespaced)
  module_session <- entry$shiny_session
  module_id <- module_session$ns(NULL)

  # Get the root scope of the module session
  module_root <- module_session$rootScope()

  # Build the namespaced output ID
  ns2 <- shiny::NS(module_id)

  # The URL's outputId is already namespaced (e.g. "demo-iris_plot").
  # Strip the namespace prefix to get the bare outputId
  bare_id <- url_output_id
  ns_prefix <- paste0(module_id, "-")
  if (nzchar(module_id) && startsWith(url_output_id, ns_prefix)) {
    bare_id <- substring(url_output_id, nchar(ns_prefix) + 1L)
  }

  full_outputId <- ns2(bare_id)

  # Retrieve the render function already assigned to the output
  render_function <- module_root$getOutput(full_outputId)

  if (!is.function(render_function)) {
    output$viewer_content <- shiny::renderUI({
      shiny::h3(sprintf(
        "Cannot find render function for output: %s", bare_id
      ))
    })
    return(invisible())
  }

  # Get the output UI function (e.g. plotOutput, verbatimTextOutput)
  ui_function <- attr(render_function, "outputFunc")

  # Get output options from the stored renderers
  output_opts <- list()
  renderer <- NULL
  if (!is.null(entry$output_renderers)) {
    renderer <- entry$output_renderers$get(bare_id)
    if (is.list(renderer)) {
      output_opts <- renderer$output_opts
    }
  }

  local_data <- fastmap::fastmap()

  # Use the module's reactive domain so reactive expressions in the
  # render function can access module inputs/reactives
  shiny::withReactiveDomain(module_root, {

    # Assign the wrapper UI to the viewer module's output (namespaced)
    output$viewer_content <- shiny::renderUI({
      output_args <- c(list(full_outputId), as.list(output_opts))
      output_args_fullvp <- output_args
      output_args_fullvp$width <- "100vw"
      output_args_fullvp$height <- "100vh"

      tryCatch(
        do.call(ui_function, output_args_fullvp),
        error = function(e) {
          do.call(ui_function, output_args)
        }
      )
    })

    # Assign the original render function at root scope
    # (matches the namespaced ID embedded in the UI element)
    root_session$output[[full_outputId]] <- render_function

    # stream_viz outputs are handled automatically: the render expression
    # captures session from its lexical scope, so stream_file_id() computes
    # the correct token-qualified path to the parent's binary file.  The
    # streaming flag in the widget's x list triggers startStreaming() in
    # renderValue on the JS side — no custom message needed.

  })

  # Forward inputs from viewer to the parent module's session.
  # Use the standalone viewer's own domain so the observer fires
  # in the same flush cycle as the input update (avoids cross-session
  # timing issues with event-type inputs).
  shiny::observe({
    inputs <- shiny::reactiveValuesToList(root_session$input)
    nms <- names(inputs)
    nms <- nms[startsWith(nms, ns2("")) & !startsWith(nms, "@")]
    if (!length(nms)) {
      return()
    }
    for (nm in nms) {
      sig <- digest::digest(inputs[[nm]])
      if (!identical(sig, local_data$get(nm))) {
        module_root$sendCustomMessage(
          "shidashi.set_shiny_input",
          list(
            inputId = nm,
            value = inputs[[nm]],
            priority = "event"
          )
        )
        local_data$set(nm, sig)
      }
    }
  }, domain = root_session, autoDestroy = TRUE)

  invisible()
}

#' Create an \pkg{ellmer} Chat object for the chat-bot
#'
#' @description
#' Factory function that creates an \code{ellmer::Chat} object based on
#' the configured provider.  Reads from
#' \code{options(shidashi.chat_provider)}, \code{shidashi.chat_model},
#' and \code{shidashi.chat_base_url}. These arguments are passed to
#' \code{\link[ellmer]{chat}}.
#'
#' @param system_prompt character; the system prompt.  Defaults to
#'   \code{getOption("shidashi.chat_system_prompt")}.
#' @param provider character; provider name or provider name with models.
#'   Defaults to \code{getOption("shidashi.chat_provider", "anthropic")}.
#' @param base_url character or \code{NULL}; base URL for
#'   API-compatible providers.
#' @return An \code{ellmer::Chat} R6 object (tools not yet bound).
#' @keywords internal
init_chat <- function(
    system_prompt = getOption("shidashi.chat_system_prompt", NULL),
    provider = getOption("shidashi.chat_provider", "anthropic"),
    # model = getOption("shidashi.chat_model", NULL),
    base_url = getOption("shidashi.chat_base_url", NULL)
) {
  if (!requireNamespace("ellmer", quietly = TRUE)) {
    stop("Package 'ellmer' is required for the chatbot feature.")
  }

  if (!length(system_prompt)) {
    system_prompt <- paste(
      "You are an R Shiny expert. You have access to the shiny",
      "application via provided tools."
    )
  }

  provider <- tolower(provider)

  # Build args common to most providers
  args <- list(system_prompt = system_prompt,
               name = provider,
               echo = "output")
  # if (length(model) == 1L && nzchar(model)) {
  #   args$model <- model
  # }
  if (length(base_url) == 1L && nzchar(base_url)) {
    args$base_url <- base_url
  }
  do.call(ellmer::chat, args)
}



#' Check if a tool is enabled for a given mode
#' @param tool an ellmer::ToolDef with shidashi annotations
#' @param mode character scalar: the current mode name
#' @return logical
#' @keywords internal
#' @noRd
is_tool_enabled_for_mode <- function(tool, mode) {
  enabled <- tool@annotations$shidashi_enabled
  if (is.null(enabled)) return(FALSE)
  if (isTRUE(enabled)) return(TRUE)
  if (isFALSE(enabled)) return(FALSE)
  # enabled is a character vector (or list) of mode names
  mode %in% as.character(unlist(enabled))
}

#' Check if a skill script is enabled for a given mode
#' @param skill_scripts list of per-script override configs from agents.yaml
#' @param script_name character scalar: the script filename
#' @param mode character scalar: the current mode name
#' @return logical
#' @keywords internal
#' @noRd
is_script_enabled_for_mode <- function(skill_scripts, script_name, mode) {
  if (!length(skill_scripts)) return(TRUE)
  for (sc in skill_scripts) {
    if (identical(sc$name, script_name)) {
      enabled <- sc$enabled
      if (is.null(enabled)) return(TRUE)
      if (isTRUE(enabled)) return(TRUE)
      if (isFALSE(enabled)) return(FALSE)
      return(mode %in% as.character(unlist(enabled)))
    }
  }
  TRUE
}

#' Get the category list for a specific skill script
#' @keywords internal
#' @noRd
get_script_category <- function(skill_scripts, script_name) {
  if (!length(skill_scripts)) return(character(0))
  for (sc in skill_scripts) {
    if (identical(sc$name, script_name)) {
      return(as.character(unlist(sc$category)))
    }
  }
  character(0)
}

load_agent_conf <- function(root_path, module_id) {
  root_path <- normalizePath(root_path, mustWork = TRUE)
  module_root <- file.path(root_path, "modules", module_id)
  agent_conf_path <- file.path(module_root, "agents.yaml")
  # read agent.yaml
  if (file.exists(agent_conf_path)) {
    agent_conf <- yaml::read_yaml(agent_conf_path)
    agent_conf$parameters <- as.list(agent_conf$parameters)
    if (!length(agent_conf$parameters$system_prompt)) {
      agent_conf$parameters$system_prompt <- paste(
        "You are an R shiny expert. You have access to the shiny",
        "application via provided tools."
      )
    }
    # Top-level enabled flag: default TRUE when agents.yaml exists
    if (is.null(agent_conf$enabled)) {
      agent_conf$enabled <- TRUE
    } else {
      agent_conf$enabled <- isTRUE(agent_conf$enabled)
    }
  } else {
    # agents.yaml missing → agents disabled for this module
    agent_conf <- list(
      enabled = FALSE,
      tools = list(
        # list(
        #   name = "hello_world",
        #   category = list("exploratory"),
        #   enabled = TRUE
        # ),
        # list(
        #   name = "get_shiny_input_values",
        #   category = list("exploratory"),
        #   enabled = TRUE
        # )
      ),
      skills = list(),
      parameters = list(
        system_prompt = paste(
          "You are a helpful assistant."
        )
      )
    )
  }
  agent_conf
}

compile_tools_and_scripts <- function(root_path, module_id, env) {
  root_path <- normalizePath(root_path, mustWork = TRUE)
  agent_conf <- load_agent_conf(root_path = root_path, module_id = module_id)
  # ---- Register MCP tools ----

  tool_names <- unlist(lapply(agent_conf$tools, "[[", "name"))
  names(agent_conf$tools) <- tool_names

  skill_names <- unlist(lapply(agent_conf$skills, "[[", "name"))
  if (length(agent_conf$skills)) {
    names(agent_conf$skills) <- skill_names
  }

  vnames <- ls(env, all.names = TRUE)
  tools <- lapply(vnames, function(vname) {
    value <- env[[vname]]
    if (!is.function(value)) { return() }
    if (inherits(value, "ellmer::ToolDef")) { return(value) }
    if (inherits(value, "shidashi_mcp_wrapper")) { return(value) }
    return(NULL)
  })

  tools <- drop_null(tools)

  # ---- Discover skill directories (Phase 4) ----
  # Per Anthropic spec, skill name == folder name. Direct lookup,
  # no iteration needed. Missing folders are silently dropped.
  skill_wrappers <- list()
  if (length(skill_names)) {
    root_skills_dir <- file.path(root_path, "agents", "skills")
    for (sname in skill_names) {
      skill_dir <- file.path(root_skills_dir, sname)
      if (file.exists(file.path(skill_dir, "SKILL.md"))) {
        skill_wrappers[[sname]] <- skill_wrapper(skill_dir)
      }
    }
  }

  # Phase 7: mode-getter closure for tool annotations
  get_current_mode <- function() {
    globals_get_agent_mode(module_id = module_id)
  }

  # create a tool-generating function
  tool_gen_fun <- function(session) {

    tool_map <- fastmap::fastmap()
    lapply(tools, function(tool) {
      toolset <- list()
      if (inherits(tool, "ellmer::ToolDef")) {
        toolset <- list(tool)
      } else {
        # generator
        toolset <- tool(session = session)
        if (inherits(toolset, "ellmer::ToolDef")) {
          toolset <- list(toolset)
        }
      }

      lapply(toolset, function(tool) {
        if (!tool@name %in% tool_names) {
          return()
        }
        tool_conf <- agent_conf$tools[[tool@name]]
        tool@annotations$shidashi_type <- "tool"
        tool@annotations$shidashi_enabled <- tool_conf$enabled
        category <- as.character(unlist(tool_conf$category))
        tool@annotations$shidashi_category <- category
        tool@annotations$shidashi_module_id <- module_id
        old_name <- tool@name
        tool@name <- sprintf("tool__%s", tool@name)

        tool_map$set(tool@name, wrap_tools_with_permissions(tool = tool, session = session))
        return()
      })
    })

    # ---- Process skill wrappers (Phase 4) ----
    lapply(names(skill_wrappers), function(sname) {
      wrapper <- skill_wrappers[[sname]]
      skill_tool <- tryCatch(
        wrapper(),
        error = function(e) {
          warning("Failed to create skill tool '", sname, "': ",
                  conditionMessage(e))
          NULL
        }
      )
      if (!inherits(skill_tool, "ellmer::ToolDef")) {
        return()
      }
      skill_conf <- agent_conf$skills[[sname]]
      skill_tool@annotations$shidashi_type <- "skill"
      skill_tool@annotations$shidashi_enabled <- skill_conf$enabled
      skill_tool@annotations$shidashi_category <- c("skill", as.character(skill_conf$category))
      skill_tool@annotations$shidashi_module_id <- module_id
      skill_tool@annotations$shidashi_skill_scripts <- structure(
        as.list(skill_conf$scripts),
        names = vapply(skill_conf$scripts, function(x) {
          x[["name"]]
        }, FUN.VALUE = "")
      )
      skill_tool@name <- sprintf("skill__%s", sname)
      tool_map$set(skill_tool@name, wrap_tools_with_permissions(tool = skill_tool, session = session))
    })

    tool_map

  }

  tool_gen_fun
}


wrap_tools_with_permissions <- function(tool, session) {

  module_id <- tool@annotations$shidashi_module_id
  tool_name <- tool@name

  # tool or skill
  shidashi_type <- tool@annotations$shidashi_type

  # skill, destructive, ...
  shidashi_category <- tool@annotations$shidashi_category

  # permissions
  shidashi_permission <- tool@annotations$shidashi_enabled
  skill_scripts_permission <- as.list(tool@annotations$shidashi_skill_scripts)

  original_fn <- S7::S7_data(tool)

  wrapper_fn <- function(...) {

    agent_mode <- globals_get_agent_mode(module_id = module_id)

    if (identical(agent_mode, "None")) {
      # Agent mode is None
      stop("Agent mode is [None]. All tools & skills are disabled")
    }

    if (is.null(shidashi_permission)) {
      stop("This tool is disabled under current agent permission mode.")
    }

    if (
      !isTRUE(shidashi_permission) &&
        !agent_mode %in% as.character(unlist(shidashi_permission))
    ) {
      stop(
        "This tool is only enabled under the following agent modes: ",
        paste(as.character(unlist(shidashi_permission)), collapse = ", ")
      )
    }

    cl <- match.call()
    arg_exprs <- as.list(cl)[-1L]
    caller <- parent.frame()
    args <- lapply(arg_exprs, eval, envir = caller)

    # Extract and strip _intent before forwarding to original function
    intent <- args[["_intent"]]
    args[["_intent"]] <- NULL

    category <- shidashi_category

    # For skill scripts
    if (length(skill_scripts_permission) > 0 && identical(shidashi_type, "skill") && identical(args$action, "script")) {
      file_name <- args$file_name
      if (length(file_name) == 1 && nzchar(file_name)) {
        script_permission <- as.list(skill_scripts_permission[[file_name]])
        if (
          length(script_permission) > 0 &&
          !isTRUE(script_permission$enabled)
        ) {
          if (isFALSE(script_permission$enabled) ||
              !isTRUE(agent_mode %in% script_permission$enabled)) {
            stop("While skill is permitted, this specific script is disabled under current agent permission mode.")
          }
        }
      }
    }

    # Determine if this specific call is destructive
    needs_confirm <- any(c("destructive", "needs_confirmation") %in% category)

    if (!needs_confirm) {
      return(do.call(original_fn, args))
    }

    # Check confirmation policy
    policy <- globals_get_confirmation_policy(
      module_id = module_id,
      missing = "auto_allow"
    )

    if (identical(policy, "auto_allow")) {
      # Auto-allow: execute without prompting
      return(do.call(original_fn, args))
    }

    if (identical(policy, "auto_reject")) {
      stop(
        "Tool '", tool_name, "' is rejected by policy. ",
        "User needs to change the confirmation policy to 'Auto-allow' ",
        "in the app.",
        call. = FALSE
      )
    }

    # policy == "ask": Ask user for confirmation
    # Extract short tool name (strip type prefix like "tool__" or "skill__")
    short_name <- sub("^(tool|skill)__", "", tool_name)

    confirm_result <- mcp_tool_ask_user(
      arguments = list(
        message = "Do you want to proceed?",
        choices = c("Proceed", "Stop and revise"),
        allow_freeform = FALSE,
        tool_name = short_name,
        intent = if (length(intent) == 1 && nzchar(intent)) intent
      ),
      shiny_session = session
    )
    # mcp_tool_ask_user returns a promise when session is active,
    # or a plain list when session is NULL/closed
    confirm_promise <- if (inherits(confirm_result, "promise")) {
      confirm_result
    } else {
      promises::promise_resolve(confirm_result)
    }

    promises::then(confirm_promise, function(res) {
      user_answer <- res$content[[1]]$text
      if (!identical(user_answer, "Proceed")) {
        stop(
          "User declined destructive action on '", tool_name,
          "'. User response: ", user_answer,
          call. = FALSE
        )
      }
      do.call(original_fn, args)
    })

  }

  # Match formals so ellmer validation stays happy, then inject _intent
  fmls <- formals(original_fn)
  fmls[["_intent"]] <- ""
  formals(wrapper_fn) <- fmls
  S7::S7_data(tool) <- wrapper_fn

  # Inject _intent into the tool's argument schema so the LLM sees it
  tool@arguments@properties[["_intent"]] <- ellmer::type_string(
    "Brief explanation of why you are calling this tool.",
    required = FALSE
  )

  tool
}

#' Create an ellmer::ToolDef for the ask_user built-in
#'
#' Returns a ToolDef that the LLM can invoke to ask the user a question.
#' Uses the browser modal when a Shiny session is available, falls back
#' to the R console when \code{interactive()}, or rejects.
#' @param shiny_session The Shiny session (or \code{NULL})
#' @return An \code{ellmer::ToolDef}
#' @keywords internal
#' @noRd
make_ask_user_tool <- function(shiny_session) {
  ellmer::tool(
    fun = function(message, choices = NULL, allow_freeform = TRUE) {
      result <- mcp_tool_ask_user(
        arguments = list(
          message = message,
          choices = choices,
          allow_freeform = allow_freeform
        ),
        shiny_session = shiny_session
      )
      if (promises::is.promise(result)) {
        promises::then(result, function(res) res$content[[1L]]$text)
      } else {
        result$content[[1L]]$text
      }
    },
    name = "ask_user",
    description = paste(
      "Ask the user a question via a modal dialog (when a Shiny",
      "session is available) or the R console. Use this when you",
      "need the user to make a choice, confirm an action, or",
      "provide free-form input. Returns the user's response text."
    ),
    arguments = list(
      message = ellmer::type_string(
        "The question or message to show the user."
      ),
      choices = ellmer::type_array(
        "Optional predefined choices shown as buttons (e.g. ['Yes', 'No']).",
        items = ellmer::type_string(),
        required = FALSE
      ),
      allow_freeform = ellmer::type_boolean(
        "Whether to show a free-form text area in addition to choices. Default TRUE.",
        required = FALSE
      )
    )
  )
}

# ---- Skill script execution for Phase 4 skills ----
#
# Resolves interpreters and executes scripts via processx::run().
# Interpreter resolution: config.yaml -> interpreters -> built-in defaults.

#' Execute a skill script via processx
#'
#' Runs a script in the skill's \code{scripts/} directory. The interpreter
#' is resolved from:
#' \enumerate{
#'   \item Template-level \code{config.yaml -> interpreters} (when a Shiny
#'     session is active; see \code{template_root()}).
#'   \item Built-in defaults: \code{.R} -> \code{Rscript} (portable),
#'     \code{.sh} -> \code{/bin/sh} (non-Windows only),
#'     \code{.sh.bat} -> \code{cmd.exe /c} on Windows or \code{/bin/sh}
#'     elsewhere (dual-mode scripts that work on both platforms).
#' }
#' All other extensions require explicit \code{config.yaml} configuration.
#'
#' @param skill_dir Absolute path to the skill directory.
#' @param file_name Script filename (relative to \code{scripts/}).
#' @param args Character vector of CLI arguments.
#' @param envs Named character vector of environment variables.
#' @param timeout_seconds Timeout in seconds (default 60).
#' @return A list with \code{stdout}, \code{stderr}, \code{status} (exit code),
#'   and \code{timeout} (logical).
#' @keywords internal
#' @noRd
run_skill_script <- function(skill_dir, file_name, args = character(),
                             envs = character(), timeout_seconds = 60) {
  if (!requireNamespace("processx", quietly = TRUE)) {
    stop("Package 'processx' is required to run skill scripts. ",
         "Try other approaches or stop, and ",
         "tell user to install `processx` by themselves.")
  }

  script_path <- file.path(skill_dir, "scripts", file_name)
  if (!file.exists(script_path)) {
    stop("Script not found: ", file_name,
         "\nAvailable scripts: ",
         paste(discover_scripts(skill_dir), collapse = ", "))
  }
  script_path <- normalizePath(script_path, mustWork = TRUE)
  script_path <- gsub("[/|\\\\]+", "/", script_path)

  # ---- Resolve interpreter ----
  # Detect compound extension .sh.bat before simple extension
  is_sh_bat <- grepl("\\.sh\\.bat$", tolower(file_name))
  ext <- if (is_sh_bat) "sh.bat" else tolower(tools::file_ext(file_name))
  interpreter <- NULL

  # 1. Template config.yaml -> interpreters (when Shiny is running)
  session <- tryCatch(shiny::getDefaultReactiveDomain(),
                      error = function(e) NULL)
  if (!is.null(session)) {
    root <- tryCatch(template_root(), error = function(e) NULL)
    if (!is.null(root)) {
      config_path <- file.path(root, "config.yaml")
      if (file.exists(config_path)) {
        config <- tryCatch(yaml::read_yaml(config_path),
                           error = function(e) NULL)
        if (is.list(config) && length(config$interpreters)) {
          val <- config$interpreters[[ext]]
          if (is.character(val) && length(val)) {
            interpreter <- val
          }
        }
      }
    }
  }

  # 2. Built-in defaults
  if (is.null(interpreter)) {
    interpreter <- switch(
      ext,
      "r" = {
        rscript <- file.path(R.home("bin"),
                             if (.Platform$OS.type == "windows") "Rscript.exe"
                             else "Rscript")
        if (!file.exists(rscript)) rscript <- "Rscript"
        rscript
      },
      "sh" = {
        if (.Platform$OS.type == "windows") {
          stop("Shell scripts (.sh) are not supported on Windows. ",
               "Configure an interpreter in config.yaml -> interpreters -> sh, ",
               "or use a .sh.bat dual-mode script instead.")
        }
        "/bin/sh"
      },
      "sh.bat" = {
        if (.Platform$OS.type == "windows") {
          c("cmd.exe", "/c")
        } else {
          "/bin/sh"
        }
      },
      stop("No interpreter configured for extension '.", ext, "'.\n",
           "Add it to config.yaml under `interpreters:\n  ",
           ext, ": <command>`")
    )
  }

  # ---- Build and run command ----
  command <- interpreter[[1L]]
  cmd_args <- c(interpreter[-1L], script_path, as.character(args))

  env_vars <- NULL
  if (length(envs) && !is.null(names(envs))) {
    env_vars <- envs
  }

  result <- processx::run(
    command = command,
    args    = cmd_args,
    env     = env_vars,
    # TODO: check this against skill specs
    wd      = skill_dir,
    timeout = timeout_seconds,
    error_on_status = FALSE
  )

  list(
    stdout  = result$stdout,
    stderr  = result$stderr,
    status  = result$status,
    timeout = result$timeout
  )
}

test_that("run_skill_script executes R scripts portably", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")

  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  result <- run_skill_script(
    skill_dir = skill_dir,
    file_name = "greet.R",
    args = c("World")
  )

  expect_type(result, "list")
  expect_named(result, c("stdout", "stderr", "status", "timeout"),
               ignore.order = TRUE)
  expect_equal(result$status, 0L)
  expect_match(result$stdout, "Hello, World!")
  expect_false(isTRUE(result$timeout))
})

test_that("run_skill_script passes arguments correctly", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")

  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  result <- run_skill_script(
    skill_dir = skill_dir,
    file_name = "greet.R",
    args = c("Platform Test")
  )
  expect_match(result$stdout, "Hello, Platform Test!")
})

test_that("run_skill_script errors on missing script", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  expect_error(
    run_skill_script(skill_dir, "nonexistent.R"),
    "Script not found"
  )
})

test_that("run_skill_script errors on unconfigured extension", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")

  # Create a temp skill with an unsupported script extension

  tmp <- tempfile("skill_ext")
  dir.create(file.path(tmp, "scripts"), recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  writeLines("print('hi')", file.path(tmp, "scripts", "test.xyz"))

  expect_error(
    run_skill_script(tmp, "test.xyz"),
    "No interpreter configured"
  )
})

test_that("R interpreter resolves portably via R.home", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")

  # Create a minimal skill with an R script that prints R.home
  tmp <- tempfile("skill_rhome")
  dir.create(file.path(tmp, "scripts"), recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  writeLines('cat(R.home("bin"), "\\n")', file.path(tmp, "scripts", "check.R"))

  result <- run_skill_script(tmp, "check.R")
  expect_equal(result$status, 0L)
  # The output should match the current R's bin directory
  expect_match(trimws(result$stdout), R.home("bin"), fixed = TRUE)
})

test_that("shell scripts are rejected on Windows", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")
  skip_if(.Platform$OS.type != "windows", "Windows-only test")

  tmp <- tempfile("skill_sh")
  dir.create(file.path(tmp, "scripts"), recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  writeLines("echo hello", file.path(tmp, "scripts", "test.sh"))

  expect_error(
    run_skill_script(tmp, "test.sh"),
    "not supported on Windows"
  )
})

test_that("shell scripts work on non-Windows", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")
  skip_if(.Platform$OS.type == "windows", "non-Windows only test")

  tmp <- tempfile("skill_sh")
  dir.create(file.path(tmp, "scripts"), recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  writeLines('#!/bin/sh\necho "shell ok"', file.path(tmp, "scripts", "test.sh"))

  result <- run_skill_script(tmp, "test.sh")
  expect_equal(result$status, 0L)
  expect_match(result$stdout, "shell ok")
})

test_that("config.yaml interpreters are used when available", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")

  # This test simulates config.yaml loading by directly testing

  # the interpreter resolution path in run_skill_script.
  # Since there's no Shiny session, config.yaml is skipped and
  # built-in defaults are used — confirming the fallback works.

  tmp <- tempfile("skill_cfg")
  dir.create(file.path(tmp, "scripts"), recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  writeLines("cat('from R\\n')", file.path(tmp, "scripts", "hello.R"))

  # Without a Shiny session, config.yaml is ignored; R fallback works
  result <- run_skill_script(tmp, "hello.R")
  expect_equal(result$status, 0L)
  expect_match(result$stdout, "from R")
})

test_that("environment variables are passed to scripts", {
  skip_if(!requireNamespace("processx", quietly = TRUE), "processx not installed")

  tmp <- tempfile("skill_env")
  dir.create(file.path(tmp, "scripts"), recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  writeLines('cat(Sys.getenv("TEST_VAR"), "\\n")',
             file.path(tmp, "scripts", "env_check.R"))

  result <- run_skill_script(
    tmp, "env_check.R",
    envs = c(TEST_VAR = "hello_from_test")
  )
  expect_equal(result$status, 0L)
  expect_match(result$stdout, "hello_from_test")
})

#' @title 'Accordion' items
#' @param title character title to show in the header
#' @param ... body content
#' @param footer footer element, hidden if \code{NULL}
#' @param class the class of the item
#' @param collapsed whether collapsed at the beginning
#' @param parentId parent \code{\link{accordion}} id
#' @param itemId the item id
#' @param style_header,style_body 'CSS' style of item header and body
#' @param root_path see \code{template_root}
#' @return \code{'shiny.tag.list'} 'HTML' tags
#' @seealso \code{\link{accordion}}
#' @export
accordion_item <- function(
  title, ..., footer = NULL,
  class = "", collapsed = TRUE,
  parentId = rand_string(prefix = "accordion-"),
  itemId = rand_string(prefix = "accordion-item-"),
  style_header = NULL, style_body = NULL,
  root_path = template_root()) {

  body <- shiny::tagList(...)

  template_path <- file.path(root_path, "views", "accordion-item.html")

  if (length(footer)) {
    footer <- shiny::div(
      class = "card-footer",
      footer
    )
  } else {
    footer <- ""
  }

  if (length(title) >= 1) {
    data_title <- trimws(as.character(title[[1]])[[1]])
  } else {
    data_title <- ""
  }

  shiny::htmlTemplate(
    template_path,
    document_ = FALSE,
    title = title,
    data_title = data_title,
    body = body,
    class = class,
    parentId = parentId,
    itemId = itemId,
    style_header = style_header,
    style_body = style_body,
    footer = footer,
    collapsed = collapsed
  )
}

#' @name accordion
#' @title Generates an 'accordion' tab-set
#' @description Generates an 'accordion' tab-set that only one tab is
#' expanded at a time. This feature is experimental and has bugs in
#' some situations. Please use it at your own risk.
#' @param ... 'accordion' items, generated by \code{\link{accordion_item}}
#' @param id the element id, must be unique
#' @param class the additional 'HTML' class
#' @param style_header additional 'CSS' styles for header
#' @param style_body additional 'CSS' styles for content body
#' @param env environment to evaluate \code{...}
#' @param extras key-value pairs that overrides the parameters in
#' \code{\link{accordion_item}}
#' @param root_path see \code{\link{template_root}}
#' @return \code{'shiny.tag.list'} 'HTML' tags
#' @seealso \code{\link{accordion_item}}
#'
#' @examples
#'
#'
#' if(interactive()) {
#'
#'   library(shiny)
#'   library(shidashi)
#'
#'   accordion(
#'     id = "input-set",
#'     accordion_item(
#'       title = "Input Group A",
#'       textInput("input_1", "Input 1"),
#'       collapsed = FALSE,
#'       footer = "Anim pariatur cliche reprehenderit dolor brunch."
#'     ),
#'     accordion_item(
#'       title = "Input Group B",
#'       textInput("input_2", "Input 2"),
#'       footer = actionButton("btn1", "OK"),
#'       collapsed = FALSE
#'     )
#'   )
#' }
#'
#' @export
accordion <- function(
  ..., id = rand_string(prefix = "accordion-"),
  class = NULL, style_header = NULL,
  style_body = NULL, env = parent.frame(), extras = list(),
  root_path = template_root()) {

  call <- match.call(expand.dots = FALSE)

  force(root_path)
  parentId <- id

  items <- unname(lapply(call[["..."]], function(item) {
    item[["parentId"]] <- parentId
    item[["root_path"]] <- root_path

    if (!is.null(style_header)) {
      item[["style_header"]] <- style_header
    }
    if (!is.null(style_body)) {
      item[["style_body"]] <- style_body
    }
    eval(item, envir = env)
  }))

  extras <- as.list(extras)
  extras$id <- parentId
  extras$class <- combine_class("card-accordion", class)
  extras <- c(extras, items)

  do.call(shiny::div, extras)

}


#' @rdname accordion
#' @param itemId \code{\link{accordion_item}} id
#' @param item_title \code{\link{accordion_item}} title, if item id is
#' specified, this title will be ignored
#' @param method operation, choices are \code{'expand'} (default),
#' \code{'collapse'}, or \code{'toggle'}
#' @param session \code{shiny} session
#' @export
accordion_operate <- function(
    id, itemId, item_title,
    method = c("expand", "collapse", "toggle"),
    session = shiny::getDefaultReactiveDomain()) {

  method <- match.arg(method)
  if (missing(item_title) && missing(itemId)) {
    stop("accordion_operate: either accordion item title or id must be specified")
  }

  if (missing(itemId)) {
    selector <- sprintf("#%s .card-accordion-header [data-title='%s']", session$ns(id), item_title)
  } else {
    itemId <- session$ns(itemId)
    selector <- sprintf("#%s .card-accordion-item#%s-accordion-header .card-accordion-title", session$ns(id), itemId)
  }

  session$sendCustomMessage("shidashi.accordion", list(
    selector = selector,
    method = method
  ))

}

#' Generates 'HTML' info box
#' @param ... box content
#' @param icon the box icon; default is \code{"envelope"}, can be hidden by
#' specifying \code{NULL}
#' @param class class of the box container
#' @param class_icon class of the icon
#' @param class_content class of the box body
#' @param root_path see \code{\link{template_root}}
#' @return 'HTML' tags
#' @examples
#'
#' library(shiny)
#' library(shidashi)
#'
#' info_box("Message", icon = "cogs")
#'
#' info_box(
#'   icon = "thumbs-up",
#'   span(class = "info-box-text", "Likes"),
#'   span(class = "info-box-number", "12,320"),
#'   class_icon = "bg-red"
#' )
#'
#' info_box("No icons", icon = NULL)
#'
#' @export
info_box <- function(..., icon = "envelope", class = "",
                     class_icon = "bg-info", class_content = "",
                     root_path = template_root()) {
  call <- match.call(expand.dots = TRUE)
  if (length(icon)) {
    icon <- shiny::span(
      class = combine_class("info-box-icon", class_icon),
      as_icon(icon)
    )
  }

  template_path <- file.path(root_path, "views", "info-box.html")

  re <- shiny::htmlTemplate(
    template_path,
    document_ = FALSE,
    icon = icon,
    body = shiny::tagList(...),
    class = combine_class(class),
    class_content = combine_class(class_content)
  )

  set_attr_call(re, call)
}

#' @title An 'HTML' container that can flip
#' @param inputId element 'HTML' id; must be specified if \code{active_on} is
#' not \code{'click'}
#' @param front 'HTML' elements to show in the front
#' @param back 'HTML' elements to show when the box is flipped
#' @param active_on the condition when a box should be flipped; choices are
#' \code{'click'}: flip when double-click on both sides; \code{'click-front'}:
#' only flip when the front face is double-clicked; \code{'manual'}: manually
#' flip in \code{R} code (see \code{{flip(inputId)}} function)
#' @param session shiny session; default is current active domain
#' @param class 'HTML' class
#' @return \code{flip_box} returns 'HTML' tags; \code{flip} should be called
#' from shiny session, and returns nothing
#' @examples
#'
#' # More examples are available in demo
#'
#' library(shiny)
#' library(shidashi)
#'
#' session <- MockShinySession$new()
#'
#' flip_box(front = info_box("Side A"),
#'          back = info_box("Side B"),
#'          inputId = 'flip_box1')
#'
#' flip('flip_box1', session = session)
#'
#' @export
flip_box <- function(front, back, active_on = c("click", "click-front", "manual"), inputId = NULL, class = NULL) {
  call <- match.call()
  active_on <- match.arg(active_on)
  if (active_on != "click" && length(inputId) != 1) {
    stop("`inputId` must be specified if `active_on` is not 'click'")
  }
  set_attr_call(shiny::div(
    class = combine_class("flip-box", class),
    # AdminLTE3
    "data-toggle" = active_on,
    # bs5
    "data-bs-toggle" = active_on,
    id = inputId,
    shiny::div(
      class = "flip-box-inner",
      shiny::div(
        class = "flip-box-back",
        back
      ),
      shiny::div(
        class = "flip-box-front",
        front
      )
    )
  ), call)
}

#' @rdname flip_box
#' @export
flip <- function(inputId, session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.box_flip", list(
    inputId = session$ns(inputId)
  ))
}

create_barebone <- function(path) {
  dir.create(path, showWarnings = FALSE, recursive = TRUE)
  src <- system.file("builtin-templates", "AdminLTE3-bare", package = "shidashi")
  fs <- list.files(src, full.names = TRUE, recursive = FALSE, all.files = FALSE,
                   no.. = TRUE, include.dirs = TRUE)
  # Exclude node_modules, src, and lock files from copy
  fs <- fs[!basename(fs) %in% c("node_modules", "src", "package-lock.json")]
  file.copy(
    from = fs,
    to = path,
    overwrite = TRUE,
    recursive = TRUE,
    copy.date = TRUE
  )

  # /server.R
  {
    writeLines(
      c(
        "library(shiny)",
        "",
        "server <- function(input, output, session) {",
        "",
        "  shidashi::enable_input_broadcast(session)",
        "  shidashi::enable_input_sync(session)",
        "",
        "  # Load and dispatch module server on navigation",
        "  shiny::observeEvent(session$clientData$url_search, {",
        "    req <- list(QUERY_STRING = session$clientData$url_search)",
        "    resource <- shidashi::load_module(request = req)",
        "    if (resource$has_module) {",
        "      module_table <- shidashi::module_info()",
        "      module_table <- module_table[module_table$id %in%",
        "        resource$module$id, ]",
        "      if (nrow(module_table)) {",
        "        group_name <- as.character(module_table$group[[1]])",
        "        if (is.na(group_name)) {",
        "          group_name <- \"<no group>\"",
        "        }",
        "        if (system.file(package = \"logger\") != \"\") {",
        "          logger::log_info(\"Loading - { module_table$label[1] } ({group_name}/{ module_table$id })\")",
        "        }",
        "        shiny::moduleServer(resource$module$id, resource$module$server,",
        "          session = session)",
        "      }",
        "    }",
        "  })",
        "}"
      ), file.path(path, "server.R"))
  }

  dir.create(file.path(path, "R"), showWarnings = FALSE, recursive = TRUE)
  dir.create(file.path(path, "modules", "module_id", "R"), showWarnings = FALSE, recursive = TRUE)

  # /R/common.R
  {
    writeLines(
      c(
        "library(shiny)",
        "page_title <- function(complete = TRUE) {",
        "  if (complete) {",
        "    \"Shiny Dashboard Template - Barebone\"",
        "  } else {",
        "    \"ShiDashi\"",
        "  }",
        "}",
        "page_logo <- function(size = c(\"normal\", \"small\", \"large\")) {",
        "  # Relative path to your logo icon in www/",
        "  \"shidashi/img/icon.png\"",
        "}",
        "page_loader <- function() {",
        "  # if no loader is needed, then return NULL",
        "  shiny::div(",
        "    class = \"preloader flex-column justify-content-center align-items-center\",",
        "    shiny::img(",
        "      class = \"animation__shake\",",
        "      src = page_logo(\"large\"),",
        "      alt = \"Logo\", height=\"60\", width=\"60\"",
        "    )",
        "  )",
        "}",
        "body_class <- function() {",
        "  c(",
        "    #--- Fix the navigation banner ---",
        "    #\"layout-navbar-fixed\",",
        "",
        "    #--- Collapse the sidebar at the beginning ---",
        "    # \"sidebar-collapse\",",
        "",
        "    #--- Let control sidebar open at the beginning ---",
        "    # \"control-sidebar-slide-open\",",
        "",
        "    #--- Fix the sidebar position ---",
        "    \"layout-fixed\",",
        "",
        "    #--- Default behavior when collapsing sidebar",
        "    # \"sidebar-mini\", \"sidebar-mini-md\", \"sidebar-mini-xs\"",
        "",
        "    #--- Hide the navbar-nav-iframe",
        "    \"navbar-iframe-hidden\",",
        "",
        "    #--- Start as dark-mode ---",
        "    \"dark-mode\"",
        "",
        "    #--- Make scrollbar thinner ---",
        "    # \"fancy-scroll-y\"",
        "",
        "  )",
        "}",
        "nav_class <- function() {",
        "  c(",
        "    \"main-header\",",
        "    \"navbar\",",
        "    \"navbar-expand\",",
        "    \"navbar-dark\",",
        "    \"navbar-primary\"",
        "  )",
        "}",
        "",
        "module_breadcrumb <- function() {}"
      ),
      con = file.path(path, "R", "common.R"))
  }

  # /modules/module_id/R/module-ui.R
  {
    writeLines(
      c(
        "library(shiny)",
        "library(shidashi)",
        "",
        "ui <- function() {",
        "  fluidPage(",
        "    fluidRow(",
        "      column(",
        "        width = 12L,",
        "        # remember to add ns, which is given as shiny::NS(\"module_id\")",
        "        plotOutput(ns(\"plot\"))",
        "      )",
        "    )",
        "  )",
        "}",
        "",
        "server_module_id <- function(input, output, session, ...) {",
        "  shidashi::register_session(session)",
        "",
        "  output$plot <- renderPlot({",
        "    theme <- shidashi::get_theme()",
        "    set.seed(1)",
        "    par(",
        "      bg = theme$background, fg = theme$foreground,",
        "      col.main = theme$foreground,",
        "      col.axis = theme$foreground,",
        "      col.lab = theme$foreground",
        "    )",
        "    hist(rnorm(1000))",
        "  })",
        "}"
      ),
      con = file.path(path, "modules", "module_id", "R", "module-ui.R")
    )
  }

  # /modules/module_id/server.R
  {
    writeLines(
      c(
        "library(shiny)",
        "library(shidashi)",
        "",
        "server <- function(input, output, session, ...) {",
        "  server_module_id(input, output, session, ...)",
        "}"
      ),
      con = file.path(path, "modules", "module_id", "server.R")
    )
  }

  # /agents/ - MCP tools and skills
  create_barebone_agents(path)

  # /modules/module_id/agents.yaml
  {
    writeLines(
      c(
        "enabled: yes",
        "tools:",
        "- name: hello_world",
        "  category:",
        "  - exploratory",
        "  enabled: yes",
        "- name: get_shiny_input_values",
        "  category:",
        "  - exploratory",
        "  enabled: yes",
        "skills:",
        "- name: greet",
        "  category:",
        "  - executing",
        "  enabled: yes",
        "parameters:",
        "  system_prompt: You are an R shiny expert. You have access to the shiny",
        "    application via provided tools."
      ),
      con = file.path(path, "modules", "module_id", "agents.yaml")
    )
  }

  invisible()
}


create_barebone_bslib <- function(path) {
  dir.create(path, showWarnings = FALSE, recursive = TRUE)
  src <- system.file("builtin-templates", "bslib-bare", package = "shidashi")
  if (!nchar(src) || !dir.exists(src)) {
    stop("Cannot find bslib-bare template. Please update the `shidashi` package.")
  }
  fs <- list.files(src, full.names = TRUE, recursive = FALSE, all.files = FALSE,
                   no.. = TRUE, include.dirs = TRUE)
  # Exclude node_modules, src, and lock files from copy
  fs <- fs[!basename(fs) %in% c("node_modules", "src", "package-lock.json")]
  file.copy(
    from = fs,
    to = path,
    overwrite = TRUE,
    recursive = TRUE,
    copy.date = TRUE
  )

  # /server.R
  {
    writeLines(
      c(
        "library(shiny)",
        "",
        "server <- function(input, output, session) {",
        "",
        "  shidashi::enable_input_broadcast(session)",
        "  shidashi::enable_input_sync(session)",
        "",
        "  # Load and dispatch module server on navigation",
        "  shiny::observeEvent(session$clientData$url_search, {",
        "    req <- list(QUERY_STRING = session$clientData$url_search)",
        "    resource <- shidashi::load_module(request = req)",
        "    if (resource$has_module) {",
        "      module_table <- shidashi::module_info()",
        "      module_table <- module_table[module_table$id %in%",
        "        resource$module$id, ]",
        "      if (nrow(module_table)) {",
        "        group_name <- as.character(module_table$group[[1]])",
        "        if (is.na(group_name)) {",
        "          group_name <- \"<no group>\"",
        "        }",
        "        if (system.file(package = \"logger\") != \"\") {",
        "          logger::log_info(\"Loading - { module_table$label[1] } ({group_name}/{ module_table$id })\")",
        "        }",
        "        shiny::moduleServer(resource$module$id, resource$module$server,",
        "          session = session)",
        "      }",
        "    }",
        "  })",
        "",
        "  output$drawer_output <- shiny::renderPrint({",
        "    module_data <- shidashi::active_module()",
        "    if (is.null(module_data)) {",
        "      \"No module\"",
        "    } else {",
        "      utils::str(module_data)",
        "    }",
        "  })",
        "",
        "  # Chatbot server runs per-module (injected by modules.R when",
        "  # agents.yaml has enabled: yes).  No chatbot_server() call here.",
        "}"
      ), file.path(path, "server.R"))
  }

  dir.create(file.path(path, "R"), showWarnings = FALSE, recursive = TRUE)
  dir.create(file.path(path, "modules", "module_id", "R"), showWarnings = FALSE, recursive = TRUE)

  # /R/common.R
  {
    writeLines(
      c(
        "library(shiny)",
        "page_title <- function(complete = TRUE) {",
        "  if (complete) {",
        "    \"Shiny Dashboard Template - bslib\"",
        "  } else {",
        "    \"ShiDashi\"",
        "  }",
        "}",
        "",
        "page_logo <- function(size = c(\"normal\", \"small\", \"large\")) {",
        "  \"shidashi/img/icon.png\"",
        "}",
        "page_loader <- function() {",
        "  NULL",
        "}",
        "",
        "body_class <- function() {",
        "  c(",
        "    #--- Start as dark-mode ---",
        "    \"dark-mode\",",
        "",
        "    # drawer has no overlay",
        "    \"shidashi-drawer-no-overlay\"",
        "  )",
        "}",
        "",
        "nav_class <- function() {",
        "  c(",
        "    \"shidashi-header\",",
        "    \"navbar\",",
        "    \"navbar-expand\"",
        "  )",
        "}",
        "",
        "sidebar_class <- function() {",
        "  c(",
        "    #--- Start as dark-mode ---",
        "    \"dark-mode\"",
        "  )",
        "}",
        "",
        "module_breadcrumb <- function() {}",
        ""
      ),
      con = file.path(path, "R", "common.R"))
  }

  # /modules/module_id/R/module-ui.R
  {
    writeLines(
      c(
        "library(shiny)",
        "library(shidashi)",
        "",
        "ui <- function() {",
        "  fluidPage(",
        "    fluidRow(",
        "      column(",
        "        width = 12L,",
        "        # remember to add ns, which is given as shiny::NS(\"module_id\")",
        "        plotOutput(ns(\"plot\"))",
        "      )",
        "    )",
        "  )",
        "}",
        "",
        "server_module_id <- function(input, output, session, ...) {",
        "  shidashi::register_session(session)",
        "",
        "  output$plot <- renderPlot({",
        "    theme <- shidashi::get_theme()",
        "    set.seed(1)",
        "    par(",
        "      bg = theme$background, fg = theme$foreground,",
        "      col.main = theme$foreground,",
        "      col.axis = theme$foreground,",
        "      col.lab = theme$foreground",
        "    )",
        "    hist(rnorm(1000))",
        "  })",
        "}"
      ),
      con = file.path(path, "modules", "module_id", "R", "module-ui.R")
    )
  }

  # /modules/module_id/server.R
  {
    writeLines(
      c(
        "library(shiny)",
        "library(shidashi)",
        "",
        "server <- function(input, output, session, ...) {",
        "  server_module_id(input, output, session, ...)",
        "}"
      ),
      con = file.path(path, "modules", "module_id", "server.R")
    )
  }

  # /agents/ - MCP tools and skills
  create_barebone_agents(path)

  # /modules/module_id/agents.yaml
  {
    writeLines(
      c(
        "enabled: yes",
        "tools:",
        "- name: hello_world",
        "  category:",
        "  - exploratory",
        "  enabled: yes",
        "- name: get_shiny_input_values",
        "  category:",
        "  - exploratory",
        "  enabled: yes",
        "skills:",
        "- name: greet",
        "  category:",
        "  - executing",
        "  enabled: yes",
        "parameters:",
        "  system_prompt: You are an R shiny expert. You have access to the shiny",
        "    application via provided tools."
      ),
      con = file.path(path, "modules", "module_id", "agents.yaml")
    )
  }

  invisible()

}


# Internal helper: create agents/ directory with MCP tools and skills
create_barebone_agents <- function(path) {
  # Create directory structure
  dir.create(file.path(path, "agents", "tools"), showWarnings = FALSE, recursive = TRUE)
  dir.create(file.path(path, "agents", "skills", "greet", "scripts"), showWarnings = FALSE, recursive = TRUE)

  # agents/tools/hello_world.R
  writeLines(
    c(
      "# agents/tools/hello_world.R",
      "#",
      "# Root-level MCP tool: Returns a greeting.",
      "# Used to verify the MCP tunnel works end-to-end.",
      "",
      "hello_world <- shidashi::mcp_wrapper(",
      "  function(session) {",
      "    ellmer::tool(",
      "      fun = function(name = \"World\") {",
      "        paste0(\"Hello, \", name, \"!\")",
      "      },",
      "      name = \"hello_world\",",
      "      description = \"Returns a greeting. Used to verify the MCP tunnel works.\",",
      "      arguments = list(",
      "        name = ellmer::type_string(",
      "          \"Name to greet (default: 'World')\",",
      "          required = FALSE",
      "        )",
      "      )",
      "    )",
      "  }",
      ")"
    ),
    con = file.path(path, "agents", "tools", "hello_world.R")
  )

  # agents/tools/get_shiny_input_values.R
  writeLines(
    c(
      "# agents/tools/get_shiny_input_values.R",
      "#",
      "# Root-level MCP tool: Read Shiny input values from the bound session.",
      "",
      "get_shiny_input_values <- shidashi::mcp_wrapper(",
      "  function(session) {",
      "",
      "    # Capture the live session in closure",
      "    bound_session <- session",
      "",
      "    ellmer::tool(",
      "      fun = function(input_ids = character()) {",
      "        input_ids <- as.character(input_ids[!is.na(input_ids)])",
      "        values <- tryCatch({",
      "          if (is.null(input_ids) || length(input_ids) == 0L) {",
      "            shiny::isolate(shiny::reactiveValuesToList(bound_session$input))",
      "          } else {",
      "            structure(",
      "              names = input_ids,",
      "              lapply(input_ids, function(id) {",
      "                shiny::isolate(bound_session$input[[id]])",
      "              })",
      "            )",
      "          }",
      "        }, error = function(e) {",
      "          stop(\"Error reading inputs: \", conditionMessage(e))",
      "        })",
      "      },",
      "      name = \"get_shiny_input_values\",",
      "      description = \"Read R-Shiny input values from the bound session.\",",
      "      arguments = list(",
      "        input_ids = ellmer::type_array(",
      "          items = ellmer::type_string(),",
      "          description = \"Input IDs to read. If empty or omitted, returns all input values.\",",
      "          required = FALSE",
      "        )",
      "      )",
      "    )",
      "  }",
      ")"
    ),
    con = file.path(path, "agents", "tools", "get_shiny_input_values.R")
  )

  # agents/skills/greet/SKILL.md
  writeLines(
    c(
      "---",
      "name: greet",
      "description: Greets a user by name using a simple R script",
      "---",
      "",
      "## Instructions",
      "",
      "This skill demonstrates the skill system. It runs a short R script",
      "that prints a personalised greeting.",
      "",
      "### Usage",
      "",
      "1. Call `action='script'`, `file_name='greet.R'`, `args=['World']`",
      "2. The script prints: `Hello, World!`",
      "",
      "### Arguments",
      "",
      "- `args[1]`: The name to greet (default: `\"World\"`)"
    ),
    con = file.path(path, "agents", "skills", "greet", "SKILL.md")
  )

  # agents/skills/greet/scripts/greet.R
  writeLines(
    c(
      "#!/usr/bin/env Rscript",
      "# greet.R - Simple greeting script for the demo skill",
      "#",
      "# Usage: Rscript greet.R [name]",
      "# Output: Hello, <name>!",
      "",
      "args <- commandArgs(trailingOnly = TRUE)",
      "name <- if (length(args) >= 1L) args[[1L]] else \"World\"",
      "",
      "cat(sprintf(\"Hello, %s!\\n\", name))"
    ),
    con = file.path(path, "agents", "skills", "greet", "scripts", "greet.R")
  )

  # agents/tool-schema.yaml
  # Pre-listed tool schemas for MCP clients that do not support
  # dynamic deferred tools.
  writeLines(
    c(
      "# agents/tool-schema.yaml",
      "#",
      "# Pre-listed tool schemas for MCP clients that do not support",
      "# dynamic deferred tools. These schemas are advertised in `tools/list`",
      "# even before a Shiny session is bound. Calling the tools still",
      "# requires a bound session.",
      "#",
      "# Skills (agents/skills/) are auto-discovered and do NOT need to be",
      "# listed here.",
      "",
      "tools:",
      "- name: hello_world",
      "  description: \"Returns a greeting. Used to verify the MCP tunnel works.\"",
      "  inputSchema:",
      "    type: object",
      "    properties:",
      "      name:",
      "        type: string",
      "        description: \"Name to greet (default: 'World')\""
    ),
    con = file.path(path, "agents", "tool-schema.yaml")
  )

  invisible()
}

#' @importFrom fastmap fastmap
#' @importFrom digest digest
#' @importFrom formatR tidy_source
#' @importFrom httr2 url_modify
#' @importFrom yaml read_yaml
#' @importFrom jsonlite fromJSON
#' @importFrom ellmer tool
NULL

#' Get Bootstrap 5 dependencies via \pkg{bslib}
#' @description Returns Bootstrap 5 HTML dependencies provided by
#' \pkg{bslib}. Intended to be called from HTML templates so that
#' \code{headContent()} renders Bootstrap 5 CSS and JavaScript.
#' @param ... additional arguments passed to \code{bslib::bs_theme}
#' @return An \code{\link[shiny]{tagList}} containing Bootstrap 5 dependencies
#' @export
bslib_dependency <- function(...) {
  theme <- bslib::bs_theme(version = 5, ...)
  deps <- bslib::bs_theme_dependencies(theme)
  shiny::tagList(deps)
}

rand_string <- function(length = 10, prefix = NULL) {
  paste(c(prefix, sample(c(letters, LETTERS, 0:9), length, replace = TRUE)),
        collapse = "")
}

R_user_dir <- function(package, which = c("data", "config", "cache")) {
  stopifnot(is.character(package), length(package) == 1L)
  which <- match.arg(which)
  home <- normalizePath("~")
  path <- switch(which, data = {
    p <- Sys.getenv("R_USER_DATA_DIR")
    if (!nzchar(p)) {
      p <- Sys.getenv("XDG_DATA_HOME")
      if (!nzchar(p)) {
        if (.Platform$OS.type == "windows") {
          p <- file.path(Sys.getenv("APPDATA"), "R",
                         "data")
        } else if (Sys.info()["sysname"] == "Darwin") {
          p <- file.path(home, "Library", "Application Support",
                         "org.R-project.R")
        } else {
          p <- file.path(home, ".local", "share")
        }
      }
    }
    p
  }, config = {
    p <- Sys.getenv("R_USER_CONFIG_DIR")
    if (!nzchar(p)) {
      p <- Sys.getenv("R_USER_CONFIG_DIR")
      if (!nzchar(p)) {
        p <- Sys.getenv("XDG_CONFIG_HOME")
        if (!nzchar(p)) {
          if (.Platform$OS.type == "windows") {
            p <- file.path(Sys.getenv("APPDATA"), "R",
                           "config")
          } else if (Sys.info()["sysname"] == "Darwin") {
            p <- file.path(home, "Library", "Preferences",
                           "org.R-project.R")
          } else {
            p <- file.path(home, ".config")
          }
        }
      }
    }
    p
  }, cache = {
    p <- Sys.getenv("R_USER_CACHE_DIR")
    if (!nzchar(p)) {
      p <- Sys.getenv("XDG_CACHE_HOME")
      if (!nzchar(p)) {
        if (.Platform$OS.type == "windows") {
          p <- file.path(Sys.getenv("LOCALAPPDATA"),
                         "R", "cache")
        } else if (Sys.info()["sysname"] == "Darwin") {
          p <- file.path(home, "Library", "Caches", "org.R-project.R")
        } else {
          p <- file.path(home, ".cache")
        }
      }
    }
    p
  })
  file.path(path, "R", package)
}

set_attr_call <- function(x, call, collapse = "\n", ...) {
  if (!is.character(call)) {
    call <- deparse(call)
  }
  call <- paste(call, collapse = collapse, ...)
  attr(x, "shidashi.code") <- call
  x
}

combine_class <- function(...) {
  s <- paste(c(...), collapse = " ", sep = " ")
  s <- unlist(strsplit(s, " "))
  s <- unique(s)
  s <- s[!s %in% ""]
  paste(s, collapse = " ")
}
remove_html_class <- function(target, class) {
  if (!length(target)) { return("") }
  s <- unlist(strsplit(target, " "))
  s <- unique(s)
  s <- s[!s %in% c("", class)]
  paste(s, collapse = " ")
}

#' Guess the 'AdminLTE' body class for modules, used internally
#' @param cls the class string of the \code{<body>} tag in \code{'index.html'}
#' @return The proposed class for \code{<body>} tag
#' @export
guess_body_class <- function(cls) {
  if (missing(cls)) {
    cls <- "fancy-scroll-y darm-mode"
  } else {
    cls <- unlist(strsplit(paste(cls, collapse = " "), " "))
    combine_class(cls[startsWith(cls, "fancy-scroll-") | cls %in% "dark-mode"])
  }
}

#' Get \code{R} expression used to generate the 'HTML' tags
#' @description This function only works on the elements generated by this
#' package
#' @param x 'HTML' tags
#' @return Quoted \code{R} expressions that can generate the 'HTML' tags
#'
#' @seealso \code{\link{format_text_r}}
#' @examples
#'
#' x <- info_box("Message")
#' get_construct_string(x)
#'
#' @export
get_construct_string <- function(x) {
  attr(x, "shidashi.code")
}

#' Get re-formatted \code{R} expressions in characters
#' @seealso \code{\link{get_construct_string}}
#' @param expr \code{R} expressions
#' @param quoted whether \code{expr} is quoted
#' @param reformat whether to reformat
#' @param class class of \code{<pre>} tag
#' @param copy_on_click whether to copy to clipboard if user clicks on the
#' code; default is true
#' @param hover mouse hover behavior
#' @param width.cutoff,indent,wrap,args.newline,blank,... passed to
#' \code{\link[formatR]{tidy_source}}
#' @return \code{format_text_r} returns characters,
#' \code{html_highlight_code} returns the 'HTML' tags wrapping expressions
#' in \code{<pre>} tag
#' @examples
#'
#' s <- format_text_r(print(local({a<-1;a+1})))
#' cat(s)
#'
#' x <- info_box("Message", icon = "cogs")
#' s <- format_text_r(get_construct_string(x),
#'                    width.cutoff = 15L, quoted = TRUE)
#' cat(s)
#'
#'
#' @export
format_text_r <- function(expr, quoted = FALSE, reformat = TRUE,
                          width.cutoff = 80L, indent = 2, wrap = TRUE,
                          args.newline = TRUE, blank = FALSE, ...) {
  if (!quoted) {
    expr <- substitute(expr)
  }

  if (length(expr) != 1 ||
      !is.character(expr)) {
    expr <- paste(deparse(expr), collapse = "\n")
  }

  if (reformat) {
    expr <- formatR::tidy_source(
      text = expr,
      output = FALSE,
      width.cutoff = width.cutoff,
      indent = indent,
      wrap = wrap,
      args.newline = args.newline,
      blank = blank,
      ...
    )$text.tidy
  }
  paste(expr, collapse = "\n")
}

#' @rdname format_text_r
#' @export
html_highlight_code <- function(
  expr, class = NULL, quoted = FALSE,
  reformat = TRUE, copy_on_click = TRUE,
  width.cutoff = 80L, indent = 2, wrap = TRUE,
  args.newline = TRUE, blank = FALSE,
  ..., hover = c("overflow-visible-on-hover", "overflow-auto")) {

  hover <- match.arg(hover)
  if (!quoted) {
    expr <- substitute(expr)
  }
  expr <- format_text_r(expr = expr, quoted = TRUE,
                reformat = reformat, width.cutoff = width.cutoff,
                indent = indent, wrap = wrap, args.newline = args.newline,
                blank = blank, ...)

  shiny::HTML(
    sprintf(
      "<pre class='pre-compact no-padding bg-gray-90 %s %s %s' %s><code class='r'>%s</code></pre>",
      hover,
      paste(class, collapse = " "),
      ifelse(copy_on_click, "clipboard-btn shidashi-clipboard-output", ""),
      ifelse(copy_on_click,
             sprintf("data-clipboard-text='%s' role='button' title='Click to copy!'", expr),
             ""),
      expr
    )
  )
}

#' Used by demo project to show the generating code
#' @seealso \code{html_highlight_code}
#' @description Please write your own version. This function is designed for
#' demo-use only.
#' @param x 'HTML' tags generated by this package
#' @param class additional 'HTML' class
#' @param code_only whether to show code only
#' @param as_card whether to wrap results in \code{\link{card}}
#' @param card_title,class_body used by \code{\link{card}} if \code{as_card=TRUE}
#' @param width.cutoff,indent,wrap,args.newline,blank,copy_on_click,... passed
#' to \code{\link{html_highlight_code}}
#' @return 'HTML' tags
#' @export
show_ui_code <- function(
  x, class = NULL, code_only = FALSE,
  as_card = FALSE, card_title = "", class_body = "bg-gray-70",
  width.cutoff = 80L, indent = 2, wrap = TRUE,
  args.newline = TRUE, blank = FALSE, copy_on_click = TRUE,
  ...)
{
  code <- format_text_r(
    get_construct_string(x),
    quoted = TRUE,
    width.cutoff = width.cutoff,
    indent = indent,
    wrap = wrap,
    args.newline = args.newline,
    blank = blank,
    ...
  )

  res <- info_box(
    class = combine_class("no-margin overflow-visible-on-hover", class),
    class_content = "display-block bg-gray-90 no-padding code-display",
    icon = NULL,
    html_highlight_code(code, quoted = TRUE, reformat = FALSE,
                        copy_on_click = copy_on_click)
  )


  if (as_card) {
    res <- card(
      title = card_title, class_body = class_body,
      tools = clipboardOutput(
        clip_text = code,
        as_card_tool = TRUE),
      footer = res,
      class_foot = "display-block bg-gray-90 no-padding code-display fill-width",
      if (code_only) {
        NULL
      } else {
        x
      }
    )
  }
  res
}


truc_string <- function(x, max_char, annot = "(truncated)", side = c("end", "begin", "both"), collapse = "\n") {
  side <- match.arg(side)

  # x <- "asdbkawjhrgfbqwierkfbwlairefb"
  # max_char <- 10
  # side <- "end"
  # annot = "(truncated)"

  if (side == "both" && max_char < nchar(annot) + 10) {
    side <- "end"
  }

  annot <- switch(
    side,
    both = sprintf("|...%s...|", annot),
    end = sprintf("|...%s", annot),
    begin = sprintf("%s...|", annot)
  )

  alen <- nchar(annot)

  max_char <- max(max_char, nchar(annot) + 2)

  x <- trimws(paste(x, collapse = collapse))
  xlen <- nchar(x)
  if (xlen <= max_char) {
    return(x)
  }

  tlen <- max_char - alen

  switch(
    side,
    both = {
      n_pre <- ceiling(tlen / 2)
      n_post <- tlen - n_pre
      sprintf("%s%s%s", substr(x, 1L, n_pre), annot, substr(x, xlen - n_post + 1, xlen))
    },
    end = {
      sprintf("%s%s", substr(x, 1L, tlen), annot)
    },
    begin = {
      sprintf("%s%s", annot, substr(x, xlen - tlen + 1L, xlen))
    }
  )

}


`%||%` <- function(a, b) if (is.null(a)) b else a

drop_null <- function(x) {
  as.list(x[!vapply(x, is.null, FALSE)])
}

#' @name html_class
#' @title Combine, add, or remove 'HTML' classes
#' @description Combine 'HTML' classes to produce nice, clean 'HTML' class
#' string via \code{combine_html_class}, or to remove a class via
#' \code{remove_html_class}
#' @param ... one or more characters, classes to combine; duplicated classes
#' will be removed
#' @param target characters, class list
#' @param class one or more characters, classes to be removed from \code{target}
#' @return A character string of new 'HTML' class
#' @examples
#'
#' # Combine classes "a b c d e"
#' combine_html_class("a", "b  a", c("c", " d", "b"), list("e ", "a"))
#'
#' # Remove class
#' remove_html_class("a b   c  e", c("b", "c "))
#'
#' @export
combine_html_class <- function(...) {
  s <- paste(c(...), collapse = " ", sep = " ")
  s <- unlist(strsplit(s, " "))
  s <- unique(s)
  s <- s[!s %in% ""]
  paste(s, collapse = " ")
}

#' @rdname html_class
#' @export
remove_html_class <- function(target, class) {
  if (!length(target)) {
    return("")
  }
  s <- unlist(strsplit(target, " "))
  s <- unique(s)
  class <- unlist(strsplit(class, " "))
  s <- s[!s %in% c("", class)]
  paste(s, collapse = " ")
}

#' Escape HTML strings
#' @description Escape HTML strings so that they will be displayed
#' 'as-is' in websites.
#' @param s characters
#' @param space whether to also escape white space, default is true.
#' @return An R string
#' @examples
#'
#' html_asis("<a><----> <b>")
#'
#' @export
html_asis <- function(s, space = TRUE) {
  if (space) {
    pattern <- "&|<|>| |\t" # or "&|<|>|'|\"|\r|\n"
    specials <- list(
      "&" = "&amp;",
      "<" = "&lt;",
      ">" = "&gt;",
      # "'" = "&#39;",
      # '"' = "&quot;",
      # "\r" = "&#13;",
      # "\n" = "&#10;",
      " " = "&nbsp;",
      "\t" = "&nbsp;&nbsp;&nbsp;&nbsp;"
    )
  } else {
    pattern <- "&|<|>|"
    specials <- list("&" = "&amp;", "<" = "&lt;", ">" = "&gt;")
  }

  s <- enc2utf8(as.character(s))
  if (any(grepl(pattern, s, useBytes = TRUE))) {
    for (chr in names(specials)) {
      s <- gsub(chr, specials[[chr]], s, fixed = TRUE, useBytes = TRUE)
    }
  }
  Encoding(s) <- "UTF-8"
  shiny::HTML(s)
}

#' Create a badge widget located at card header
#' @param text inner text content of the badge
#' @param class additional 'HTML' class of the badge; for
#' \code{set_card_badge}, this is the class selector of the badge that is
#' to be changed
#' @param ... additional 'HTML' tag attributes
#' @param id the badge 'HTML' ID to be changed, will be enclosed with
#' session namespace \code{session$ns(id)} automatically.
#' @param add_class,remove_class add or remove class
#' @param session shiny session
#' @param event_name name of the event to trigger
#' @examples
#'
#' library(shidashi)
#'
#' # UI: a Bootstrap badge with green background
#' card_badge("Ready", class = "bg-green shidashi-output-status")
#'
#' # server
#' server <- function(input, output, session) {
#'
#'   shiny::observe({
#'
#'     # ... check if the inputs have changed
#'
#'     set_card_badge(
#'       class = "shidashi-output-status",
#'       text = "Refresh needed",
#'       add_class = "bg-warning",
#'       remove_class = "bg-success disabled"
#'     )
#'
#'   })
#'
#' }
#'
#' @export
card_badge <- function(text = NULL, class = NULL, ...) {
  if (!length(text) || nchar(text) == 0) {
    text <- ""
  }
  class <- combine_class("right badge shidashi-card-badge", class)
  shiny::span(class = class, text, ...)
}

#' @rdname card_badge
#' @export
card_recalculate_badge <- function(
    text = "Recalculate needed", class = NULL, event_name = "run_analysis", ...) {
  stopifnot("`event_name` must be non-empty string" = nzchar(event_name))
  card_badge(
    text = text,
    class = combine_class(
      "btn btn-default shidashi-button shidashi-button-recalculate shidashi-output-status bg-yellow",
      class
    ),
    `data-shidashi-type` = event_name,
    `data-shidashi-action` = "shidashi-button",
    `data-shidashi-dynamic` = "true",
    ...
  )
}

#' @rdname card_badge
#' @export
enable_recalculate_badge <- function(
    text = "Recalculate needed", ...) {
  set_card_badge(
    text = text, class = "shidashi-button-recalculate",
    add_class = "btn btn-default shidashi-button bg-yellow shidashi-button-recalculate",
    remove_class = "disabled"
  )
}

#' @rdname card_badge
#' @export
disable_recalculate_badge <- function(
    text = "Up-to-date", ...) {
  set_card_badge(
    text = text, class = "shidashi-button-recalculate",
    remove_class = "btn btn-default shidashi-button bg-yellow",
    add_class = "disabled"
  )
}

#' @rdname card_badge
#' @export
set_card_badge <- function(
    id = NULL, class = NULL,
    text = NULL, add_class = NULL, remove_class = NULL,
    session = shiny::getDefaultReactiveDomain()) {
  if (!length(id) && !length(class)) {
    class <- "shidashi-card-badge"
  }
  if ( is.null(session) ) { return() }

  if (length(id)) {
    selector <- sprintf("#%s.shidashi-card-badge", session$ns(id))
  } else if (identical(class, "shidashi-card-badge")) {
    selector <- ".shidashi-card-badge"
  } else {
    selector <- sprintf(".%s.shidashi-card-badge", class)
  }

  if (length(text)) {
    session$sendCustomMessage(
      "shidashi.set_html",
      message = list(
        selector = selector,
        content = text
      )
    )
  }
  if (length(add_class)) {
    add_class <- trimws(unlist(strsplit(add_class, " ")))
    lapply(add_class, function(cls) {
      session$sendCustomMessage(
        "shidashi.add_class",
        message = list(
          selector = selector,
          class = cls
        )
      )
    })
  }
  if (length(remove_class)) {
    remove_class <- trimws(unlist(strsplit(remove_class, " ")))
    remove_class <- remove_class[!remove_class %in% "shidashi-output-status"]
    lapply(remove_class, function(cls) {
      session$sendCustomMessage(
        "shidashi.remove_class",
        message = list(
          selector = selector,
          class = cls
        )
      )
    })
  }
  invisible()
}

#' @name drawer
#' @title Open, close, or toggle the drawer panel
#' @description Send messages to the client to open, close, or toggle the
#' off-canvas drawer panel on the right side of the dashboard.
#' @param session shiny session
#' @return No value is returned (called for side effect).
#'
#' @examples
#' server <- function(input, output, session) {
#'   # Open the drawer
#'   drawer_open()
#'
#'   # Close the drawer
#'   drawer_close()
#'
#'   # Toggle the drawer
#'   drawer_toggle()
#' }
#'
#' @export
drawer_open <- function(session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.drawer_open", list())
}

#' @rdname drawer
#' @export
drawer_close <- function(session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.drawer_close", list())
}

#' @rdname drawer
#' @export
drawer_toggle <- function(session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.drawer_toggle", list())
}


#' Drawer shell for module templates
#'
#' @description
#' Emits a minimal \code{.shidashi-drawer} container with a
#' \code{\link[shiny]{uiOutput}} placeholder inside, plus the drawer
#' overlay.  The drawer starts empty; module server code fills it
#' dynamically via \code{shiny::renderUI}.
#'
#' Typical usage in a \file{module-ui.html} template:
#' \preformatted{
#'   \{\{ shidashi::module_drawer() \}\}
#' }
#'
#' Then in the module server:
#' \preformatted{
#'   output$shidashi_drawer <- shiny::renderUI(\{
#'     shiny::tagList(
#'       shiny::h5("My settings"),
#'       shiny::p("Custom drawer content here.")
#'     )
#'   \})
#' }
#'
#' The \code{ns()} function from the module's template evaluation
#' environment is used automatically so that the output ID is
#' properly scoped to the module namespace.
#'
#' @param output_id character; the output ID for the
#'   \code{uiOutput} placeholder inside the drawer.
#'   Defaults to \code{"shidashi_drawer"}.
#' @return A \code{shiny::tagList} containing the drawer div and
#'   its overlay.
#' @export
module_drawer <- function(output_id = "shidashi_drawer") {
  # Resolve the module's ns() from the template evaluation env
  ns_func <- tryCatch(
    get("ns", envir = parent.frame()),
    error = function(e) identity
  )

  shiny::tagList(
    shiny::div(
      class = "shidashi-drawer",
      shiny::div(
        class = "shidashi-drawer-close-tab",
        shiny::tags$i(class = "fas fa-xmark")
      ),
      shiny::div(
        class = "shidashi-drawer-content",
        shiny::uiOutput(ns_func(output_id))
      )
    ),
    shiny::div(class = "shidashi-drawer-overlay")
  )
}

#' @title Card-like 'HTML' element
#' @param title the title of the card
#' @param ... the body content of the card
#' @param body_main,body_side used by \code{card2}, the body
#' content of the front and back sides of the card
#' @param footer the footer of the card; will be hidden if
#' \code{footer=NULL}
#' @param tools a list of tools or badges to be displayed at
#' top-right corner, generated by \code{\link{as_badge}} or
#' \code{\link{card_tool}}
#' @param inputId the id of the card
#' @param class the 'HTML' class of the entire card
#' @param class_header the the 'HTML' class of the card header
#' @param class_body the the 'HTML' class of the card body
#' @param class_foot the the 'HTML' class of the card footer
#' @param style_header 'CSS' style of the header
#' @param style_body 'CSS' style of the body
#' @param start_collapsed whether the card starts as collapsed
#' @param resizable whether the card body can be resized vertically;
#' notice that if true, then the default padding for body will be zero
#' @param root_path see \code{\link{template_root}}
#' @param session shiny session domain
#' @param method action to expand, minimize, or remove the cards;
#' choices are \code{"collapse"}, \code{"expand"}, \code{"remove"},
#' \code{"toggle"}, \code{"maximize"}, \code{"minimize"},
#' and \code{"toggleMaximize"}
#' @return 'HTML' tags
#'
#' @examples
#' library(shiny)
#' library(shidashi)
#'
#' # Used for example only
#' ns <- I
#' session <- MockShinySession$new()
#'
#' # -------------- Basic usage -------------
#' card(
#'   title = "Badges", div(
#'     class = "padding-20",
#'     p(
#'       "Add badges to the top-right corder. ",
#'       "Use \"|\" to indicate the badge classes; ",
#'       "for example: \"badge-info\", \"badge-warning\"..."
#'     ),
#'     hr(), p(
#'       "Use `resizable = TRUE` to make card resizable."
#'     )
#'   ),
#'   tools = list(
#'     as_badge("New|badge-info"),
#'     as_badge("3|badge-warning")
#'   ),
#'   class_body = "height-300",
#'   resizable = TRUE
#' )
#'
#' # ---------------- With tools -----------------
#' card(
#'   title = "Default Tools",
#'   plotOutput(
#'     ns("card_defaulttool_plot"),
#'     height = "100%"
#'   ),
#'   tools = list(
#'     card_tool(
#'       widget = "link",
#'       href = "https://github.com/dipterix"
#'     ),
#'     card_tool(widget = "collapse"),
#'     card_tool(widget = "maximize")
#'   ),
#'   class_body = "height-300",
#'   resizable = TRUE
#' )
#'
#' # -------------- Card2 example --------------
#' card2(
#'   title = "Card2 Example", body_main =
#'     plotOutput(
#'     outputId = ns("card2_plot"),
#'     height = "100%"
#'   ),
#'   body_side = fluidRow(
#'     column(
#'       6L, textInput(
#'         ns("card2_plot_title"),
#'         "Plot title"
#'       )
#'     ),
#'     column(
#'       6L, sliderInput(
#'         ns("card2_plot_npts"),
#'         "# of points", min = 1, max = 100,
#'         value = 10, step = 1, round = TRUE
#'       )
#'     )
#'   ),
#'   tools = list(
#'     card_tool(widget = "link",
#'               href = "https://github.com/dipterix"),
#'     card_tool(widget = "collapse"),
#'     card_tool(widget = "maximize")
#'   ),
#'   class_body = "height-300"
#' )
#'
#' @export
card <- function(
  title, ..., footer = NULL, tools = NULL, inputId = NULL,
  class = "", class_header = "", class_body = "", class_foot = "",
  style_header = NULL, style_body = NULL, start_collapsed = FALSE,
  resizable = FALSE, root_path = template_root()) {

  call <- match.call()
  body <- shiny::tagList(...)

  template_path <- file.path(root_path, "views", "card.html")

  if (length(title) >= 1) {
    data_title <- trimws(as.character(title[[1]])[[1]])
  } else {
    data_title <- ""
  }

  if (length(footer)) {
    footer <- shiny::div(
      class = combine_class("card-footer", class_foot),
      footer
    )
  } else {
    footer <- ""
  }

  if (length(tools)) {
    tools <- shiny::div(
      class = "card-tools",
      tools
    )
  } else {
    tools <- ""
  }

  if (length(inputId) == 1) {
    if (grepl("[\"']", inputId)) {
      stop("`card` ID cannot contain quotation marks.")
    }
    card_id <- sprintf(" id=\"%s\"", inputId)
  } else {
    card_id <- ""
  }

  if (resizable) {
    default_class_body <- "height-400 resize-vertical flex-container no-padding"
    if (length(class_body)) {
      class_body <- unlist(strsplit(class_body, " "))
      if (length(class_body)) {
        tmp <- class_body[
          startsWith(class_body, "height-") |
            startsWith(class_body, "min-height-")
        ]
        size <- sapply(strsplit(tmp, "-"), function(x) { x[[length(x)]] })
        if (length(size)) {
          suppressWarnings({
            size <- as.numeric(size)
            size <- size[!is.na(size)]
            if (length(size) && size %% 50 == 0) {
              default_class_body <- "resize-vertical flex-container no-padding"
            }
          })
        }
      }
    }
    class_body <- combine_class(
      default_class_body,
      class_body
    )
    body <- flex_item(
      class = "fill-height fill-width fill-max-width",
      body
    )
  } else {
    class_body <- combine_class(
      "fill-width fill-height",
      class_body
    )
  }

  set_attr_call(shiny::htmlTemplate(
    template_path,
    document_ = FALSE,
    title = title,
    body = body,
    class = class,
    class_header = class_header,
    class_body = class_body,
    style_header = style_header,
    style_body = style_body,
    footer = footer,
    tools = tools,
    card_id = card_id,
    start_collapsed = start_collapsed,
    data_title = data_title
  ), call)
}

#' @rdname card
#' @export
card2 <- function(
  title, body_main, body_side = NULL,
  footer = NULL, tools = NULL, inputId = NULL,
  class = "", class_header = "", class_body = "min-height-400",
  class_foot = "",
  style_header = NULL, style_body = NULL, start_collapsed = FALSE,
  root_path = template_root()) {

  call <- match.call()
  template_path <- file.path(root_path, "views", "card2.html")

  if (length(title) >= 1) {
    data_title <- trimws(as.character(title[[1]])[[1]])
  } else {
    data_title <- ""
  }

  if (length(footer)) {
    footer <- shiny::div(
      class = combine_class("card-footer", class_foot),
      footer
    )
  } else {
    footer <- ""
  }

  if (length(tools)) {
    tools <- shiny::tagList(
      tools
    )
  } else {
    tools <- ""
  }

  if (length(inputId) == 1) {
    if (grepl("[\"']", inputId)) {
      stop("`card` ID cannot contain quotation marks.")
    }
    card_id <- sprintf(" id=\"%s\"", inputId)
  } else {
    card_id <- ""
  }

  set_attr_call(shiny::htmlTemplate(
    template_path,
    document_ = FALSE,
    title = title,
    body_main = body_main,
    body_side = body_side,
    class = class,
    class_header = class_header,
    class_body = class_body,
    style_header = style_header,
    style_body = style_body,
    footer = footer,
    tools = tools,
    card_id = card_id,
    start_collapsed = start_collapsed,
    data_title = data_title
  ), call)
}

#' @rdname card
#' @export
card2_open <- function(inputId, session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage(
    "shidashi.card2widget",
    list(
      selector = sprintf("#%s:not(.direct-chat-contacts-open) .card-tools>.btn-tool.card2-switch", session$ns(inputId))
    )
  )
}

#' @rdname card
#' @export
card2_close <- function(inputId, session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage(
    "shidashi.card2widget",
    list(
      selector = sprintf("#%s.direct-chat-contacts-open .card-tools>.btn-tool.card2-switch", session$ns(inputId))
    )
  )
}

#' @rdname card
#' @export
card2_toggle <- function(inputId, session = shiny::getDefaultReactiveDomain()) {
  # session$sendCustomMessage(
  #   "shidashi.click",
  #   list(
  #     selector = sprintf("#%s .card-tools>.btn-tool.card2-switch", session$ns(inputId))
  #   )
  # )
  session$sendCustomMessage(
    "shidashi.card2widget",
    list(
      selector = sprintf("#%s .card-tools>.btn-tool.card2-switch", session$ns(inputId))
    )
  )
}

#' @rdname card
#' @export
card_operate <- function(
  inputId, title, method, session = shiny::getDefaultReactiveDomain()
) {
  method <- match.arg(
    method, choices = c("collapse", "expand", "remove", "toggle",
                        "maximize", "minimize", "toggleMaximize")
  )
  params <- list(method = method)
  if (missing(inputId)) {
    params$title <- title
  } else {
    params$inputId <- session$ns(inputId)
  }
  session$sendCustomMessage("shidashi.cardwidget", params)
}

# ---- Skill directory parsing for Phase 4 skills ----
#
# Parses Anthropic-compliant SKILL.md files and discovers companion
# reference/script assets in the skill directory.

#' Sanitize a skill name for use as an MCP tool identifier
#'
#' Converts a human-readable skill name to a valid identifier: lowercase,
#' non-alphanumeric characters replaced with underscores, leading/trailing
#' underscores stripped, consecutive underscores collapsed.
#'
#' @param name Character string; the skill name from SKILL.md frontmatter
#'   or directory name.


#' Parse a SKILL.md file
#'
#' Reads the YAML frontmatter and markdown body from an Anthropic-compliant
#' SKILL.md file.
#'
#' @param skill_md_path Absolute path to the SKILL.md file.
#' @return A named list with:
#'   \describe{
#'     \item{name}{Character; skill name (from frontmatter or dir name)}
#'     \item{description}{Character; one-line description from frontmatter}
#'     \item{body}{Character; full markdown body (instructions) after frontmatter}
#'     \item{frontmatter}{List; all parsed frontmatter fields}
#'     \item{skill_dir}{Character; absolute path to the skill directory}
#'   }
#' @keywords internal
#' @noRd
parse_skill_md <- function(skill_md_path) {
  skill_md_path <- normalizePath(skill_md_path, mustWork = TRUE)
  skill_dir <- dirname(skill_md_path)

  lines <- readLines(skill_md_path, warn = FALSE)

  # Parse YAML frontmatter (between --- delimiters)
  frontmatter <- list()
  body_start <- 1L

  if (length(lines) >= 3L && trimws(lines[[1L]]) == "---") {
    # Find closing ---
    end_idx <- which(trimws(lines[-1L]) == "---")
    if (length(end_idx)) {
      end_idx <- end_idx[[1L]] + 1L  # adjust for skipping line 1
      fm_text <- paste(lines[2L:(end_idx - 1L)], collapse = "\n")
      frontmatter <- tryCatch(
        yaml::read_yaml(text = fm_text),
        error = function(e) {
          warning("Failed to parse SKILL.md frontmatter: ",
                  conditionMessage(e))
          list()
        }
      )
      body_start <- end_idx + 1L
    }
  }

  # Body is everything after frontmatter
  body <- ""
  if (body_start <= length(lines)) {
    body <- trimws(paste(lines[body_start:length(lines)], collapse = "\n"))
  }

  # Name: frontmatter > directory name
  name <- frontmatter$name
  if (!length(name) || !nzchar(trimws(name))) {
    name <- basename(skill_dir)
  }

  # Description: frontmatter or first non-empty line of body
  description <- frontmatter$description
  if (!length(description) || !nzchar(trimws(description))) {
    non_empty <- lines[body_start:length(lines)]
    non_empty <- non_empty[nzchar(trimws(non_empty))]
    description <- if (length(non_empty)) {
      # Strip leading markdown heading markers
      sub("^#+\\s*", "", non_empty[[1L]])
    } else {
      paste("Skill:", name)
    }
  }

  list(
    name        = trimws(name),
    description = trimws(description),
    body        = body,
    frontmatter = frontmatter,
    skill_dir   = skill_dir
  )
}


#' Discover reference files in a skill directory
#'
#' Looks for markdown, text, and data files outside of \code{scripts/}.
#'
#' @param skill_dir Absolute path to the skill directory.
#' @return Character vector of file paths relative to skill_dir, or
#'   \code{character(0)} if none found.
#' @keywords internal
#' @noRd
discover_references <- function(skill_dir) {
  all_files <- list.files(
    skill_dir,
    recursive = TRUE,
    full.names = FALSE,
    include.dirs = FALSE,
    no.. = TRUE
  )
  # Exclude SKILL.md itself and scripts/ directory
  all_files <- all_files[!grepl("^SKILL\\.md$", all_files, ignore.case = TRUE)]
  all_files <- all_files[!grepl("^scripts/", all_files, ignore.case = TRUE)]

  # Keep reference-like extensions
  ref_exts <- "\\.(md|txt|csv|json|yaml|yml|html|xml|r|R|py|sql)$"
  all_files[grepl(ref_exts, all_files)]
}


#' Fuzzy match a reference file name
#'
#' Performs case-insensitive matching and supports flexible path formats:
#' \itemize{
#'   \item Full relative path (e.g., "references/doc_a.md")
#'   \item Just filename (e.g., "doc_a.md" or "DOC_A.MD")
#'   \item Alternate folder names ("reference/" or "references/")
#' }
#'
#' @param query The user-supplied file name to match.
#' @param ref_files Character vector of available reference files
#'   (relative paths from skill_dir).
#' @return The matched file path from \code{ref_files}, or \code{NULL}
#'   if no match found.
#' @keywords internal
#' @noRd
fuzzy_match_reference <- function(query, ref_files) {
  if (!length(ref_files) || !length(query) || !nzchar(query)) {
    return(NULL)
  }

  # Normalize query: strip leading "reference/" or "references/" prefix
  query_normalized <- sub("^references?/", "", query, ignore.case = TRUE)

  # Try exact match first (case-insensitive on full path)
  exact_match <- ref_files[tolower(ref_files) == tolower(query)]
  if (length(exact_match)) {
    return(exact_match[[1L]])
  }

  # Try matching without the prefix
  for (ref in ref_files) {
    # Compare full paths case-insensitively
    ref_normalized <- sub("^references?/", "", ref, ignore.case = TRUE)
    if (tolower(ref_normalized) == tolower(query_normalized)) {
      return(ref)
    }
    # Also try matching just the basename
    if (tolower(basename(ref)) == tolower(query_normalized)) {
      return(ref)
    }
  }

  NULL
}


#' Discover executable scripts in a skill directory
#'
#' Looks for files under the \code{scripts/} subdirectory.
#'
#' @param skill_dir Absolute path to the skill directory.
#' @return Character vector of file names relative to \code{scripts/}, or
#'   \code{character(0)} if none found.
#' @keywords internal
#' @noRd
discover_scripts <- function(skill_dir) {
  scripts_dir <- file.path(skill_dir, "scripts")
  if (!dir.exists(scripts_dir)) {
    return(character(0L))
  }
  list.files(
    scripts_dir,
    recursive = FALSE,
    full.names = FALSE,
    include.dirs = FALSE,
    no.. = TRUE
  )
}


#' Build condensed summary for gate error messages
#'
#' Auto-generates a short summary (~150-200 tokens) from skill metadata,
#' suitable for embedding in a gate rejection error.
#'
#' @param parsed Output of \code{parse_skill_md()}.
#' @param ref_files Character vector from \code{discover_references()}.
#' @param script_files Character vector from \code{discover_scripts()}.
#' @return A single character string with the condensed summary.
#' @keywords internal
#' @noRd
build_condensed_summary <- function(parsed, ref_files, script_files) {
  parts <- character()

  parts <- c(parts, paste0("## ", parsed$name))
  parts <- c(parts, parsed$description)
  parts <- c(parts, "")

  if (length(script_files)) {
    parts <- c(parts, paste0(
      "Available scripts: ",
      paste(script_files, collapse = ", ")
    ))
  }

  if (length(ref_files)) {
    parts <- c(parts, paste0(
      "Reference files: ",
      paste(ref_files, collapse = ", ")
    ))
  }

  parts <- c(parts, "")
  parts <- c(parts, "You MUST call action='readme' first to read the full instructions, then retry your intended action.")

  paste(parts, collapse = "\n")
}

test_that("parse_skill_md parses frontmatter and body", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  parsed <- parse_skill_md(file.path(skill_dir, "SKILL.md"))

  expect_type(parsed, "list")
  expect_named(parsed,
    c("name", "description", "body", "frontmatter", "skill_dir"),
    ignore.order = TRUE
  )
  expect_equal(parsed$name, "greet")
  expect_match(parsed$description, "Greets", ignore.case = TRUE)
  expect_match(parsed$body, "Instructions")
  expect_true(dir.exists(parsed$skill_dir))
})

test_that("parse_skill_md handles missing frontmatter", {
  tmp <- tempfile("skill")
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  writeLines(c("# My Skill", "", "Some instructions"), file.path(tmp, "SKILL.md"))

  parsed <- parse_skill_md(file.path(tmp, "SKILL.md"))
  # Name falls back to directory basename
  expect_equal(parsed$name, basename(tmp))
  # Description falls back to first non-empty line
  expect_equal(parsed$description, "My Skill")
  expect_match(parsed$body, "My Skill")
})

test_that("discover_scripts finds scripts/", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  scripts <- discover_scripts(skill_dir)
  expect_true("greet.R" %in% scripts)
})

test_that("discover_scripts returns empty for missing scripts/", {
  tmp <- tempfile("skill")
  dir.create(tmp)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  expect_length(discover_scripts(tmp), 0)
})

test_that("discover_references excludes SKILL.md and scripts/", {
  tmp <- tempfile("skill")
  dir.create(file.path(tmp, "scripts"), recursive = TRUE)
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)

  writeLines("---\nname: test\n---\nBody", file.path(tmp, "SKILL.md"))
  writeLines("ref content", file.path(tmp, "notes.md"))
  writeLines("data", file.path(tmp, "data.csv"))
  writeLines("#!/bin/sh", file.path(tmp, "scripts", "run.sh"))

  refs <- discover_references(tmp)
  expect_true("notes.md" %in% refs)
  expect_true("data.csv" %in% refs)
  expect_false(any(grepl("SKILL.md", refs, ignore.case = TRUE)))
  expect_false(any(grepl("^scripts/", refs)))
})

test_that("build_condensed_summary produces expected structure", {
  skill_dir <- system.file(
    "builtin-templates/bslib-bare/agents/skills/greet",
    package = "shidashi"
  )
  skip_if(!nzchar(skill_dir), "greet skill not installed")

  parsed <- parse_skill_md(file.path(skill_dir, "SKILL.md"))
  scripts <- discover_scripts(skill_dir)
  refs <- discover_references(skill_dir)

  summ <- build_condensed_summary(parsed, refs, scripts)
  expect_type(summ, "character")
  expect_match(summ, "action='readme'", fixed = TRUE)
  expect_match(summ, "greet", fixed = TRUE)
})


test_that("fuzzy_match_reference matches case-insensitively", {
  ref_files <- c("references/doc_A.md", "config.yaml", "notes/data.txt")

  # Exact match

  expect_equal(fuzzy_match_reference("references/doc_A.md", ref_files),
               "references/doc_A.md")

  # Case-insensitive full path
  expect_equal(fuzzy_match_reference("REFERENCES/DOC_A.MD", ref_files),
               "references/doc_A.md")
  expect_equal(fuzzy_match_reference("References/Doc_A.Md", ref_files),
               "references/doc_A.md")

  # Without prefix (just filename)
  expect_equal(fuzzy_match_reference("doc_A.md", ref_files),
               "references/doc_A.md")
  expect_equal(fuzzy_match_reference("DOC_A.MD", ref_files),
               "references/doc_A.md")

  # reference/ vs references/ prefix handling
  expect_equal(fuzzy_match_reference("reference/doc_A.md", ref_files),
               "references/doc_A.md")

  # Top-level file
  expect_equal(fuzzy_match_reference("config.yaml", ref_files),
               "config.yaml")
  expect_equal(fuzzy_match_reference("CONFIG.YAML", ref_files),
               "config.yaml")

  # Nested file by basename
  expect_equal(fuzzy_match_reference("data.txt", ref_files),
               "notes/data.txt")

  # No match returns NULL
  expect_null(fuzzy_match_reference("nonexistent.md", ref_files))
  expect_null(fuzzy_match_reference("", ref_files))
  expect_null(fuzzy_match_reference(NULL, ref_files))
  expect_null(fuzzy_match_reference("doc_A.md", character(0)))
})

#' @title Obtain the module information
#' @param root_path the root path of the website project
#' @param settings_file the settings file containing the module information
#' @param request 'HTTP' request string
#' @param env environment to load module variables into
#' @return A data frame with the following columns that contain the module
#' information:
#' \describe{
#' \item{\code{id}}{module id, folder name}
#' \item{\code{order}}{display order in side-bar}
#' \item{\code{group}}{group menu name if applicable, otherwise \code{NA}}
#' \item{\code{label}}{the readable label to be displayed on the side-bar}
#' \item{\code{icon}}{icon that will be displayed ahead of label, will be
#' passed to \code{\link{as_icon}}}
#' \item{\code{badge}}{badge text that will be displayed
#' following the module label, will be passed to \code{\link{as_badge}}}
#' \item{\code{url}}{the relative 'URL' address of the module.}
#' }
#' @details The module files are stored in \code{modules/} folder in your
#' project. The folder names are the module id. Within each folder,
#' there should be one \code{"server.R"}, \code{R/}, and a
#' \code{"module-ui.html"}.
#'
#' The \code{R/} folder stores R code files that generate variables,
#' which will be available to the other two files. These variables, along
#' with some built-ins, will be used to render \code{"module-ui.html"}.
#' The built-in functions are
#' \describe{
#' \item{ns}{shiny name-space function; should be used to generate the id for
#' inputs and outputs. This strategy avoids conflict id effectively.}
#' \item{.module_id}{a variable of the module id}
#' \item{module_title}{a function that returns the module label}
#' }
#'
#' The \code{"server.R"} has access to all the code in \code{R/} as well.
#' Therefore it is highly recommended that you write each 'UI' component
#' side-by-side with their corresponding server functions and call
#' these server functions in \code{"server.R"}.
#'
#' @examples
#'
#' library(shiny)
#' module_info()
#'
#' # load master module
#' load_module()
#'
#' # load specific module
#' module_data <- load_module(
#'   request = list(QUERY_STRING = "/?module=module_id"))
#' env <- module_data$environment
#'
#' if (interactive()) {
#'
#' # get module title
#' env$module_title()
#'
#' # generate module-specific shiny id
#' env$ns("input1")
#'
#' # generate part of the UI
#' env$ui()
#'
#' }
#'
#' @export
module_info <- function(root_path = template_root(),
                        settings_file = "modules.yaml") {
  settings <- yaml::read_yaml(file.path(root_path, settings_file))
  # settings <- yaml::read_yaml('modules.yaml')
  groups <- names(settings$groups)
  groups <- groups[groups != ""]

  modules <- settings$modules
  modules_ids <- names(settings$modules)

  groups <- unique(groups)
  group_level <- factor(groups, levels = groups, ordered = TRUE)

  module_tbl <- do.call("rbind", lapply(modules_ids, function(mid) {
    x <- modules[[mid]]
    y <- x[!names(x) %in% c("order", "group", "label", "icon", "badge", "module", "hidden")]
    y$module <- mid
    url <- httr2::url_modify("https://dipterix.org/?module=", query = y)
    if (length(x$group) == 1 && x$group %in% group_level) {
      x$group <- group_level[group_level == x$group][[1]]
    } else {
      x$group <- NA
    }
    data.frame(
      id = mid,
      group = x$group,
      label = ifelse(length(x$label) == 1, x$label, ""),
      icon = ifelse(length(x$icon) == 1, x$icon, ""),
      badge = ifelse(length(x$badge) == 1, x$badge, ""),
      url = gsub("^[^\\?]+", "/", url),
      stringsAsFactors = FALSE
    )
  }))
  module_tbl
}


#' @rdname module_info
#' @description \code{current_module} returns the information of the currently
#' running module. It looks up the \code{.module_id} variable in the calling
#' environment (set automatically when a module is loaded), then retrieves
#' the corresponding row from the module table.
#' @param session shiny reactive domain; used to extract the module id from
#' the URL query string when \code{.module_id} is not found.
#' @returns \code{current_module}: a named list with \code{id}, \code{group},
#' \code{label}, \code{icon}, \code{badge}, and \code{url} of the current
#' module, or \code{NULL} if no module is active.
#' @export
current_module <- function(
    session = shiny::getDefaultReactiveDomain(),
    root_path = template_root()) {

  module_id <- NULL

  # 1. Try the calling environment's .module_id (set by load_module)
  env <- parent.frame()
  if (exists(".module_id", envir = env, inherits = TRUE)) {
    module_id <- get(".module_id", envir = env, inherits = TRUE)
  }

  # 2. Fallback: parse the session's URL query string
  if (!length(module_id) && is.environment(session)) {
    query_str <- shiny::isolate(session$clientData$url_search)
    if (length(query_str) == 1L) {
      query_list <- shiny::parseQueryString(query_str)
      module_id <- query_list$module
    }
  }

  if (!length(module_id) || !nzchar(module_id)) {
    return(NULL)
  }

  modules <- module_info(root_path = root_path)
  idx <- which(modules$id == module_id)
  if (!length(idx)) {
    return(NULL)
  }
  as.list(modules[idx[1L], ])
}


#' @rdname module_info
#' @description \code{active_module} returns a \emph{reactive} value with
#' information about the module that is currently visible in the \verb{iframe}
#' tab (or the standalone module if no \verb{iframe} manager is present).
#' Unlike \code{current_module} which is static and always returns the module
#' whose server code is running, \code{active_module} dynamically tracks
#' which module the user is looking at from any context.
#' @details
#' \code{active_module} works by reading the
#' \code{'@shidashi_active_module@'} Shiny input that is set by the
#' JavaScript front-end whenever a module tab is activated.
#' Because it accesses \code{session$rootScope()$input}, the return value
#' is reactive: when called inside an \code{observe} or \code{reactive}
#' context it will re-fire whenever the user switches modules.
#'
#' If the input has not been set yet (e.g. before any module is opened),
#' the function falls back to \code{current_module()}.
#' @returns \code{active_module}: a named list with \code{id}, \code{group},
#' \code{label}, \code{icon}, \code{badge}, and \code{url} of the
#' currently active (visible) module, or \code{NULL} if no module is active.
#' @export
active_module <- function(
    session = shiny::getDefaultReactiveDomain(),
    root_path = template_root()) {

  # Resolve root scope session
  root_session <- session
  if (is.function(session$rootScope)) {
    root_session <- session$rootScope()
  }
  if (is.null(root_session)) {
    root_session <- session
  }

  # Read the @shidashi_active_module@ input (reactive)
  active_input <- root_session$input[["@shidashi_active_module@"]]

  module_id <- NULL
  if (is.list(active_input) && length(active_input$module_id)) {
    module_id <- active_input$module_id
  } else if (is.character(active_input) && length(active_input) == 1L) {
    module_id <- active_input
  }

  if (!length(module_id) || !nzchar(module_id)) {
    # Fallback to current_module when no active module has been reported yet
    return(current_module(session = session, root_path = root_path))
  }

  modules <- module_info(root_path = root_path)
  idx <- which(modules$id == module_id)
  if (!length(idx)) {
    return(NULL)
  }
  as.list(modules[idx[1L], ])
}


#' @rdname module_info
#' @description \code{switch_module} programmatically switches the active
#' module in the dashboard UI.
#' It sends a \code{shidashi.switch_module} message to the JavaScript
#' front-end.  When called from a module running inside an \verb{iframe}, the
#' handler automatically forwards the request to the parent window via
#' \code{postMessage} so that the sidebar highlight, tab bar, and \verb{iframe}
#' all update correctly.
#' @param module_id character string; the target module identifier (must
#'   match an entry in \file{modules.yaml}).
#' @export
switch_module <- function(
    module_id,
    session = shiny::getDefaultReactiveDomain()) {
  if (is.null(session)) {
    stop("`switch_module`: must be called within a Shiny reactive context")
  }
  session$sendCustomMessage(
    "shidashi.switch_module",
    list(module_id = module_id)
  )
  invisible()
}


load_module_resource <- function(root_path = template_root(), module_id = NULL, env = parent.frame()) {
  if (length(module_id) > 1) {
    stop("length of `module_id` must not exceed one.")
  }
  root_path <- normalizePath(root_path, mustWork = TRUE)

  re <- list(
    environment = env,
    has_module = FALSE,
    root_path = root_path,
    template_path = file.path(root_path, "index.html")
  )

  module_info <- list(
    id = module_id,
    server = function(input, output, session, ...) {},
    template_path = NULL
  )

  r_folder <- file.path(root_path, "R")
  if (dir.exists(r_folder)) {
    fs <- list.files(r_folder, pattern = "\\.R$", ignore.case = TRUE,
                     recursive = FALSE, include.dirs = FALSE,
                     no.. = TRUE, all.files = TRUE, full.names = TRUE)
    for (f in fs) {
      if (startsWith(basename(f), "shared-")) {
        source(f, local = env, chdir = TRUE)
      } else {
        source(f, local = env, chdir = FALSE)
      }
    }
  }

  # root_path <- 'inst/builtin-templates/bslib-bare/'
  # module_id <- "demo"
  root_agent_folder <- file.path(root_path, "agents", "tools")
  if (dir.exists(root_agent_folder)) {
    fs <- list.files(root_agent_folder, pattern = "\\.R$", ignore.case = TRUE,
                     recursive = FALSE, include.dirs = FALSE,
                     no.. = TRUE, all.files = TRUE, full.names = TRUE)
    for (f in fs) {
      source(f, local = env, chdir = TRUE)
    }
  }

  if (length(module_id) == 1) {
    module_root <- file.path(root_path, "modules", module_id)
    module_info$template_path <- file.path(module_root, "module-ui.html")

    if (dir.exists(module_root)) {

      re$has_module <- TRUE
      env$ns <- shiny::NS(module_id)
      env$.module_id <- module_id
      env$module_title <- function() {
        modules <- module_info()
        modules$label[modules$id == module_id]
      }
      shared_input_specs <- globals_get_module_input_specs(module_id)
      shared_output_specs <- globals_get_module_output_specs(module_id)
      wrapper_input_registry <- mcp_wrapper_input_output(
        input_specs = shared_input_specs,
        output_specs = shared_output_specs
      )
      env$.register_input <- wrapper_input_registry$input_helpers$register_input_specification
      env$.register_output <- wrapper_input_registry$input_helpers$register_output_specification
      env$.mcp_wrapper_inputs <- wrapper_input_registry$tool_generator
      current_input_table <- wrapper_input_registry$input_helpers$get_input_specification()

      agent_conf <- load_agent_conf(root_path = root_path, module_id = module_id)

      r_folder <- file.path(module_root, "R")
      if (dir.exists(r_folder)) {
        fs <- list.files(r_folder, pattern = "\\.R$", ignore.case = TRUE,
                         recursive = FALSE, include.dirs = FALSE,
                         no.. = TRUE, all.files = TRUE, full.names = TRUE)
        for (f in fs) {
          if (startsWith(basename(f), "shared-")) {
            source(f, local = env, chdir = TRUE)
          } else {
            source(f, local = env, chdir = FALSE)
          }
        }
      }

      new_input_table <- wrapper_input_registry$input_helpers$get_input_specification()
      if (nrow(new_input_table) > nrow(current_input_table)) {
        message("Registered inputs to MCP server:", appendLF = TRUE)
        print(new_input_table[, c("inputId", "update", "writable")])
      }


      # ---- Register MCP tools ----

      env$.mcptools_maker <- compile_tools_and_scripts(root_path = root_path,
                                                       module_id = module_id,
                                                       env = env)
      # Store agent config in module env for chatbot_ui / back_top_button

      module_handler <- file.path(root_path, "modules", module_id, "server.R")
      if (file.exists(module_handler)) {
        # server_env <- new.env(parent = env)
        server_env <- env
        server_source <- source(module_handler, local = server_env)

        server_function <- server_source$value
        if (!is.function(server_function)) {
          server_function <- server_env$server
        }

        if (!is.function(server_function)) {
          stop("Module `", module_id, "` has server.R, but cannot detect server function.")
        }

        # inject to body
        agent_enabled <- isTRUE(agent_conf$enabled)
        default_mode <- agent_conf$parameters$default_mode %||%
                         agent_conf$modes[[1]] %||% "None"
        body(server_function) <- bquote({
          local({
            shidashi <- asNamespace("shidashi")
            shidashi$register_session(session = session)
            registry <- shidashi$globals_session_registry()
            entry <- registry$get(session$token)

            # Build MCP tools
            tools <- .mcptools_maker(session)
            entry$tools$reset()
            entry$tools$mset(.list = tools$as_list())
            registry$set(session$token, entry)

            # Set agent mode early so MCP can filter tools even if
            # the chat drawer is never opened
            shidashi$globals_set_agent_mode(
              module_id = .(module_id),
              mode = .(default_mode)
            )

            # Initialize chatbot server if agents are enabled for this module
            if (.(isTRUE(agent_conf$enabled))) {
              shidashi$chatbot_server(
                input, output, session,
                agent_conf = .(agent_conf)
              )
            } else {
              output$shidashi_drawer <- shiny::renderUI({
                shiny::p("AI agent has been disabled for this module.")
              })
            }


          })

          .(body(server_function))
        })

        module_info$server <- server_function
      }

      # make sure `ns` and `.module_id` are consistent
      env$ns <- shiny::NS(module_id)
      env$.module_id <- module_id

    }
  }

  re$module <- module_info
  re
}

#' @rdname module_info
#' @export
load_module <- function(
  root_path = template_root(),
  request = list(QUERY_STRING = "/"),
  env = parent.frame()) {

  force(env)
  query_str <- request$QUERY_STRING
  if (length(query_str) != 1) {
    query_str <- "/"
  }
  query_list <- shiny::parseQueryString(query_str)
  module_id <- query_list$module
  shared_id <- query_list$shared_id
  shared_id <- tolower(shared_id)
  shared_id <- gsub("[^a-z0-9_]", "", shared_id)
  if (length(shared_id) != 1 || nchar(shared_id) == 0) {
    shared_id <- getOption("shidashi.shared_id",
                          default = rand_string(26))
  }

  env$.request <- request
  env$.shared_id <- shared_id
  load_module_resource(root_path, module_id, env)
}

.shidahi_globals <- local({
  weak_ref <- NULL

  setter <- function(env) {
    if (is.environment(env)) {
      weak_ref <<- rlang::new_weakref(env)
    } else if (is.null(env)) {
      weak_ref <<- NULL
    }
    weak_ref
  }

  getter <- function() {
    if (rlang::is_weakref(weak_ref)) {
      rlang::wref_key(weak_ref)
    } else {
      NULL
    }
  }

  list(
    getter = getter,
    setter = setter
  )
})

set_shidashi_globals <- function(env) {
  .shidahi_globals$setter(env)
}

get_shidashi_globals <- function() {
  .shidahi_globals$getter()
}


#' Initialize a shidashi application
#'
#' @description
#' Creates a fresh environment holding all per-application global stores
#' (session registry, per-module input registries, etc.).
#' This function is designed to be called from a \file{global.R} file
#' generated by \code{\link{render}()}, not for user's call.
#'
#' @param env environment where the application is initialized into.
#'
#' @return Nothing.
#'
#' @examples
#'
#' init_app(env = new.env())
#'
#' @export
init_app <- function(env = parent.frame()) {

  # check if env$.__shidashi_globals__. exists
  global_env <- env$.__shidashi_globals__.
  if (!is.environment(global_env) || environmentIsLocked(global_env)) {
    global_env <- new.env(parent = emptyenv())

    # Stores the env to make sure the environment is not gc'ed
    env$.__shidashi_globals__. <- global_env
  }

  # Persist shiny sessions
  if (!inherits(global_env$session_registry, "fastmap")) {
    global_env$session_registry <- fastmap::fastmap()
  }

  # For MCP to query input and outputs
  if (!inherits(global_env$module_input_registry, "fastmap")) {
    global_env$module_input_registry <- fastmap::fastmap()
  }

  if (!inherits(global_env$module_output_registry, "fastmap")) {
    global_env$module_output_registry <- fastmap::fastmap()
  }

  # Chatbot: per-module conversation history
  # (Chat objects are created per-module in chatbot_server closures)

  # module_id -> list(active_idx, conversations)
  # Each conversation: list(title, turns, last_visited)
  if (!inherits(global_env$module_conversations, "fastmap")) {
    global_env$module_conversations <- fastmap::fastmap()
  }

  # Phase 7: per-module agent mode state
  # module_id -> list(current_mode, modes, default_mode)
  if (!inherits(global_env$module_agent_modes, "fastmap")) {
    global_env$module_agent_modes <- fastmap::fastmap()
  }

  # Per-module confirmation policy for destructive tools
  # module_id -> "auto_allow" | "ask" | "auto_reject"
  if (!inherits(global_env$module_confirmation_policy, "fastmap")) {
    global_env$module_confirmation_policy <- fastmap::fastmap()
  }

  set_shidashi_globals(global_env)

  # Generate a shared_id once so that both UI rendering (load_module)

  # and server-side registration (register_session) use the same value
  # when the root URL has no ?shared_id= parameter.
  # Try to use existing option
  shared_id <- getOption(
    "shidashi.shared_id",
    Sys.getenv("SHIDASHI_SHARED_ID", unset = "")
  )
  if (
    length(shared_id) == 1 &&
      is.character(shared_id) &&
      !is.na(shared_id) &&
      nzchar(shared_id)
  ) {
    shared_id <- tolower(shared_id)
  } else {
    shared_id <- tolower(rand_string(length = 26))
  }
  options(shidashi.shared_id = shared_id)

  invisible()
}


# ---- helpers: the goal is to avoid direct calls to get_shidashi_globals()
# elsewhere
globals_get_module_input_specs <- function(module_id) {
  global_env <- get_shidashi_globals()
  if (!is.environment(global_env)) {
    stop(
      "MCP module input registry is missing: `init_app()` is needed. ",
      "Is shidashi dashboard running at all?"
    )
  }

  if (!global_env$module_input_registry$has(module_id)) {
    global_env$module_input_registry$set(module_id, fastmap::fastmap())
  }

  global_env$module_input_registry$get(module_id)
}

globals_get_module_output_specs <- function(module_id) {
  global_env <- get_shidashi_globals()
  if (!is.environment(global_env)) {
    stop(
      "MCP module output registry is missing: `init_app()` is needed. ",
      "Is shidashi dashboard running at all?"
    )
  }

  if (!global_env$module_output_registry$has(module_id)) {
    global_env$module_output_registry$set(module_id, fastmap::fastmap())
  }

  global_env$module_output_registry$get(module_id)
}


globals_session_registry <- function() {
  global_env <- get_shidashi_globals()
  if (!is.environment(global_env)) {
    stop(
      "Session registry is missing: `init_app()` is needed. ",
      "Is shidashi dashboard running at all?"
    )
  }
  global_env$session_registry
}

# ---------- Session registry helpers (generic, not MCP-specific) ----------

# Register a Shiny session in the session registry: session MUST be namedspaced
# instead of the root scope session, otherwise the namespace/module_id will
# be invalid


#' @name register_session
#' @title Shiny session registration and cross-tab synchronization
#'
#' @param session A Shiny session object or session proxy (created by
#'   \code{\link[shiny]{moduleServer}}). Most functions in this family default
#'   to \code{shiny::getDefaultReactiveDomain()} so callers inside module
#'   server functions do not need to pass it explicitly.
#'   \code{register_session} and \code{unregister_session} require an explicit
#'   value.
#' @param once Logical; if \code{TRUE}, resume the suspended observer for a
#'   single run via \code{$run()} instead of permanently re-activating it with
#'   \code{$resume()}.  Useful for one-shot snapshots without installing a
#'   persistent observer.
#' @param name A single character string identifying a named slot in the
#'   session's handler list (used by \code{get_handler} and
#'   \code{set_handler}).  Three names are reserved for internal use and will
#'   trigger an error if passed to \code{set_handler}:
#'   \code{"event_handler"}, \code{"broadcast_handler"}, and
#'   \code{"input_sync_handler"}.
#' @param handler A Shiny Observer object created by \code{shiny::observe()},
#'   or \code{NULL} to clear the named slot (used by \code{set_handler}).
#'   Passing any other object type raises an error.
#'
#' @return
#' \code{register_session} returns the session token (\code{session$token})
#' invisibly; it is idempotent and safe to call multiple times for the same
#' session.
#'
#' \code{unregister_session} returns \code{NULL} invisibly; it is idempotent.
#'
#' \code{enable_input_broadcast}, \code{disable_input_broadcast},
#' \code{enable_input_sync}, and \code{disable_input_sync} all return
#' invisibly.  They are silent no-ops when the session is already closed.
#'
#' \code{get_handler} returns the named \code{Observer} object stored under
#' \code{name}, or \code{NULL} if the slot is empty or the session is closed.
#'
#' \code{set_handler} returns \code{TRUE} invisibly when it successfully
#' installs the handler, or \code{FALSE} invisibly when the session is already
#' closed.
#'
#' @details
#' \subsection{Session registration}{
#'
#' \strong{\code{register_session()}} — Call once at the top of every Shiny
#' module server function.  It is idempotent: calling it a second time on the
#' same session safely refreshes the session object and URL in the registry
#' entry without re-creating observers.
#'
#' Internally it creates an entry in the application-global session registry
#' (initialized by \code{\link{init_app}}), resolves a \code{shared_id} token
#' shared across browser tabs from the \code{?shared_id=...} URL query string
#' (or generates a random 26-character string when absent), sets up the
#' per-session reactive event bus, and — for named module sessions — sends a
#' \code{shidashi.register_module_token} custom message to bind the module
#' namespace to its session token on the JavaScript side.
#'
#' \strong{\code{unregister_session()}} — Removes the session entry from the
#' registry and destroys all attached observers.  This is called automatically
#' when the session ends via the \code{onSessionEnded} hook installed by
#' \code{register_session()}.  Direct calls are only needed for explicit early
#' cleanup (e.g. in tests).
#'
#' }
#' \subsection{Session-scoped handlers}{
#'
#' Each registered session maintains a named slot list for Shiny
#' \code{Observer} objects called \emph{handlers}.  This provides a lightweight
#' system for attaching module-level life-cycle hooks that are tied to the
#' session's lifetime.
#'
#' \strong{User-defined handlers — \code{get_handler()} / \code{set_handler()}}
#'
#' \code{set_handler(name, handler)} installs \code{handler} under \code{name},
#' first suspending and destroying any \code{Observer} already stored there.
#' Pass \code{handler = NULL} to clear the slot.  Returns \code{FALSE}
#' invisibly if the session is already closed.
#'
#' \code{get_handler(name)} retrieves the stored \code{Observer} (or
#' \code{NULL}).  It auto-registers the session if not yet registered and
#' returns \code{NULL} gracefully if the session is closed.
#'
#' Three handler names are reserved for internal shidashi infrastructure and
#' will raise an error if passed to \code{set_handler}:
#' \code{"event_handler"}, \code{"broadcast_handler"}, and
#' \code{"input_sync_handler"}.
#'
#' \strong{Built-in cross-tab sync handlers}
#'
#' shidashi installs two opt-in \code{Observer} slots in every registered
#' session (both start \emph{suspended}):
#'
#' \describe{
#'   \item{Input broadcast (\code{enable_input_broadcast()} /
#'         \code{disable_input_broadcast()})}{Continuously monitors the
#'         session's \code{input} values and, whenever they change, pushes a
#'         snapshot to the client via \code{shidashi.cache_session_input}.
#'         Other browser tabs sharing the same \code{shared_id} can read this
#'         cached snapshot to restore or compare input state.}
#'   \item{Input sync (\code{enable_input_sync()} /
#'         \code{disable_input_sync()})}{Listens for serialized input maps
#'         broadcast by \emph{other} sessions sharing the same
#'         \code{shared_id} via the root-session \code{@shidashi@} input.
#'         Values differing from the local \code{input} are written into the
#'         session's private \code{inputs} \code{reactiveValues}.  Messages
#'         from the same session are ignored to prevent feedback loops.}
#' }
#'
#' Both observers run at priority \code{-100000} (after all ordinary reactive
#' computations have settled).  Use \code{once = TRUE} to trigger a single
#' cycle without permanently resuming the observer.  The \code{disable_*}
#' variants suspend the observer and are silent no-ops when the session has
#' already ended.
#'
#' }
#' \subsection{Session life-cycle}{
#'
#' \preformatted{
#' init_app()                       # global.R, once per app start
#'     |
#'     v
#' register_session(session)        # top of each module server()
#'     |
#'     v
#' ... reactive code ...
#'     get_handler() / set_handler() # attach user-defined session observers
#'     enable_input_broadcast()      # optional: push inputs to browser cache
#'     enable_input_sync()           # optional: receive peer-tab inputs
#'     |
#'     v
#' session ends -> unregister_session()  # runs automatically
#' }
#'
#' }
#'
#' @examples
#'
#' library(shiny)
#'
#' # --- Basic usage in a module server ---
#' server <- function(input, output, session) {
#'   shidashi::register_session(session)
#'
#'   # opt-in to cross-tab input broadcast (suspended by default)
#'   shidashi::enable_input_broadcast(session)
#'
#'   # opt-in to receive inputs from peer tabs
#'   shidashi::enable_input_sync(session)
#'
#'   # get_theme must be called within a reactive context
#'   output$plot <- renderPlot({
#'     theme <- shidashi::get_theme()
#'     par(bg = theme$background, fg = theme$foreground)
#'     plot(1:10)
#'   })
#' }
#'
#' # --- Named handler: attach a reusable session-scoped observer ---
#' server2 <- function(input, output, session) {
#'   shidashi::register_session(session)
#'
#'   cleanup <- shiny::observe({
#'     # ... module-level teardown logic ...
#'     shidashi::set_handler("my_cleanup", NULL, session)
#'   }, suspended = TRUE, domain = session)
#'
#'   shidashi::set_handler("my_cleanup", cleanup, session)
#'
#'   # retrieve and resume the observer elsewhere in the same session
#'   h <- shidashi::get_handler("my_cleanup", session)
#'   if (!is.null(h)) h$resume()
#' }
#'
#' @export
register_session <- function(session) {
  token <- session$token

  # shiny will ensure this
  # if (is.null(token) || !nzchar(token)) return(invisible(NULL))

  namespace <- session$ns(NULL)

  registry <- globals_session_registry()
  registered <- registry$has(token)
  # Skip if registered
  if (registered && (length(namespace) != 1 || !nzchar(namespace))) {
    return(invisible(token))
  }

  entry <- registry$get(token, list())

  if (identical(entry$shiny_session, session)) {
    # already registered
    return(invisible(token))
  }
  if (!registered || !length(entry)) {
    url <- shiny::isolate(session$clientData$url_search)

    # Auto-resolve shared_id from URL query string
    shared_id <- NULL
    if (length(url) == 1L && nzchar(url)) {
      query_list <- shiny::parseQueryString(url)
      shared_id <- tolower(query_list$shared_id)
      if (!length(shared_id) || grepl("[^a-z0-9_]", shared_id)) {
        shared_id <- NULL
      }
    }
    if (is.null(shared_id)) {
      shared_id <- getOption("shidashi.shared_id",
                            default = tolower(rand_string(length = 26)))
    }

    entry <- list(
      shiny_session      = session,
      shidashi_module_id = NULL,
      mcp_session_ids    = character(),
      namespace          = namespace,
      url                = url,
      registered_at      = Sys.time(),
      tools              = fastmap::fastmap(),
      output_renderers   = fastmap::fastmap(),
      shared_id          = shared_id,
      events             = shiny::reactiveValues(),
      inputs             = shiny::reactiveValues(),
      handlers           = fastmap::fastmap()
    )
    message("Registered session token: ", token)
  } else {
    entry$shiny_session <- session
    entry$url <- shiny::isolate(session$clientData$url_search)
    entry$namespace <- namespace
    registry$set(token, entry)
    return(invisible(token))
  }

  if (!entry$handlers$has("event_handler")) {
    # register observer
    root_session <- session$rootScope()

    entry$handlers$set("event_handler", shiny::bindEvent(
      shiny::observe({
        event <- root_session$input[["@shidashi_event@"]]
        if (!length(event) || !is.list(event)) { return() }
        if (length(event$type) != 1 || is.na(event$type) || !is.character(event$type)) { return() }
        if (!nzchar(event$type)) { return() }
        entry$events[[event$type]] <- event$message
      }, domain = root_session),
      root_session$input[["@shidashi_event@"]],
      ignoreNULL = TRUE, ignoreInit = FALSE
    ))

    session$sendCustomMessage("shidashi.get_theme", list())
  }

  # broadcast_handler — stored in session's registry entry handlers
  if (!entry$handlers$has("broadcast_handler")) {
    entry$handlers$set("broadcast_handler", shiny::observe(
      {
        inputs <- shiny::reactiveValuesToList(session$input)
        nms <- names(inputs)

        sel <- !startsWith(nms, "@")
        nms <- nms[sel]
        if (length(nms) > 0) {
          inputs <- inputs[sel]
          names(inputs) <- session$ns(nms)
        }
        sig <- session$cache$get("shidashi_input_signature", NULL)
        sig2 <- digest::digest(inputs)
        if (!identical(sig2, sig)) {
          session$cache$set("shidashi_input_signature", sig2)
          message <- list(
            shared_id = entry$shared_id,
            private_id = session$token,
            inputs = inputs
          )
          session$sendCustomMessage("shidashi.cache_session_input", message)
        }
      },
      domain = session,
      priority = -100000,
      suspended = TRUE
    ))
  }

  if (!entry$handlers$has("input_sync_handler")) {
    root_session <- session$rootScope()
    entry$handlers$set("input_sync_handler", shiny::bindEvent(
      shiny::observe({
        try(
          silent = TRUE,
          {
            message <- jsonlite::fromJSON(root_session$input[["@shidashi@"]])

            # Sync from other sessions so ignore if the
            # message source if self
            if (identical(message$last_edit, session$token)) {
              return()
            }

            input_names <- shiny::isolate({
              names(root_session$input)
            })

            input_names <- input_names[!startsWith(input_names, "@")]
            input_names <- input_names[input_names %in% names(message$inputs)]
            if (!length(input_names)) {
              return()
            }

            lapply(input_names, function(nm) {
              v <- message$inputs[[nm]]
              v2 <- shiny::isolate(root_session$input[[nm]])
              if (!identical(v, v2)) {
                entry$inputs[[nm]] <- v
              }
            })
          }
        )
      }, suspended = TRUE, domain = root_session, priority = -100000),
      root_session$input[["@shidashi@"]],
      ignoreNULL = TRUE, ignoreInit = TRUE
    ))
  }

  registry$set(token, entry)


  # Send module token to root-level JS so the chatbot can bind
  if (length(namespace) == 1L && nzchar(namespace)) {
    # It does not matter who send out custom messages, can be root session or
    # session proxy: they will be the same to JS
    session$sendCustomMessage(
      "shidashi.register_module_token",
      list(module_id = namespace, token = token)
    )
  }

  # Belt: onSessionEnded cleanup
  session$onSessionEnded(function() {
    unregister_session(session)
    sweep_closed_sessions()
  })

  invisible(token)
}

# Remove a Shiny session from the session registry
# @param session A Shiny session object

#' @rdname register_session
#' @export
unregister_session <- function(session) {
  token <- session$token
  if (is.null(token) || !nzchar(token)) return(invisible(NULL))

  registry <- globals_session_registry()
  if (registry$has(token)) {
    entry <- registry$get(token)
    if (identical(entry$shiny_session, session) || entry$shiny_session$isClosed()) {

      registry$remove(token)

      handler_keys <- entry$handlers$keys()
      lapply(handler_keys, function(handler_key) {
        handler <- entry$handlers$get(handler_key)
        entry$handlers$remove(handler_key)
        # suspend & destroy observers
        if (inherits(handler, "Observer")) {
          tryCatch({
            handler$suspend()
            handler$destroy()
          }, error = function(e) {})
        }
      })

      entry$handlers$reset()

      # also reset other fastmaps
      entry$tools$reset()
      entry$output_renderers$reset()

      message("Unregistered session token: ", token)
    }
  }
  invisible(NULL)
}

# Sweep closed Shiny sessions from the registry
#
# Iterates all registered sessions and removes any where
# session$isClosed() returns TRUE.
# Called defensively on every MCP request.
# @keywords internal
sweep_closed_sessions <- function() {
  registry <- globals_session_registry()
  tokens <- registry$keys()
  lapply(tokens, function(token) {
    entry <- registry$get(token)
    if (!length(entry) || !is.environment(entry$shiny_session)) {
      registry$remove(token)
      return()
    }
    closed <- tryCatch(entry$shiny_session$isClosed(), error = function(e) TRUE)
    if (isTRUE(closed)) {
      registry$remove(token)
    }
  })
  invisible(NULL)
}

# Look up a registry entry by Shiny session token.
# Returns the entry list, or NULL if not found / closed.
get_session_entry <- function(token) {
  if (
    length(token) != 1 || !is.character(token) ||
    is.na(token) || !nzchar(token)
  ) {
    return(NULL)
  }
  registry <- globals_session_registry()
  if (!registry$has(token)) {
    return(NULL)
  }
  entry <- registry$get(token, missing = list())
  if (!is.environment(entry$shiny_session)) {
    registry$remove(token)
    return(NULL)
  }
  closed <- tryCatch(entry$shiny_session$isClosed(), error = function(e) TRUE)
  if (isTRUE(closed)) {
    registry$remove(token)
    return(NULL)
  }
  entry
}

# ---------- Event bus helpers ---------------------------------------------

# Get shared_id for a session from the registry
globals_get_shared_id <- function(session) {
  token <- session$token
  entry <- get_session_entry(token)
  if (is.null(entry)) return(NULL)
  entry$shared_id
}

# Get all live registry entries with the same shared_id
globals_get_sessions_by_shared_id <- function(shared_id) {
  if (length(shared_id) != 1L || !is.character(shared_id) ||
      !nzchar(shared_id)) {
    return(list())
  }
  registry <- globals_session_registry()
  tokens <- registry$keys()
  entries <- lapply(tokens, function(token) {
    entry <- registry$get(token)
    if (!length(entry) || !is.environment(entry$shiny_session)) return(NULL)
    closed <- tryCatch(entry$shiny_session$isClosed(), error = function(e) TRUE)
    if (isTRUE(closed)) return(NULL)
    if (!identical(entry$shared_id, shared_id)) return(NULL)
    entry
  })
  Filter(Negate(is.null), entries)
}

#' Fire or read a session event
#'
#' @description
#' \code{fire_event} sets a reactive event value on the current session.
#' When \code{global = TRUE}, the event is propagated to all sessions that
#' share the same \code{shared_id} (i.e. other browser tabs for the same
#' user).
#'
#' \code{get_event} reads the current value of an event key from the
#' session registry.
#'
#' \code{get_theme} is a convenience wrapper around
#' \code{get_event("theme.changed")} that returns the current dashboard theme.
#'
#' All three functions must be called inside a Shiny server context.
#'
#' @param key a single character string identifying the event type
#' @param value the event payload (any R object)
#' @param session a Shiny session (defaults to the current reactive domain)
#' @param global logical; if \code{TRUE}, the event is broadcast to all
#'   sessions sharing the same \code{shared_id}
#' @param default value to return when the event has not been fired yet
#'   (used by \code{get_event} only)
#' @return \code{fire_event} returns \code{NULL} invisibly.
#'   \code{get_event} returns the last value fired for \code{key}, or
#'   \code{default} if none.
#'   \code{get_theme} returns a named list with three character elements:
#'   \describe{
#'     \item{theme}{Either \code{"light"} or \code{"dark"}.}
#'     \item{foreground}{Hex color string for text / foreground elements.}
#'     \item{background}{Hex color string for the page background.}
#'   }
#'   Before the browser fires its first theme event, the light-theme fallback
#'   \code{list(theme = "light", background = "#FFFFFF", foreground = "#000000")}
#'   is returned.
#'
#' @details
#' \code{get_theme} and \code{get_event} auto-register the session if needed
#' and must be called inside a reactive context
#' (\code{\link[shiny]{observe}}, \code{\link[shiny]{observeEvent}},
#' \code{\link[shiny]{reactive}}, render functions).
#'
#' @examples
#'
#'
#' library(shiny)
#' server <- function(input, output, session) {
#'   # fire an event
#'   shidashi::fire_event("my_event", list(a = 1), session = session)
#'
#'   # read it back
#'   observe({
#'     evt <- shidashi::get_event("my_event", session = session)
#'     print(evt)
#'   })
#'
#'   # get_theme must be called within a reactive context
#'   output$plot <- renderPlot({
#'     theme <- shidashi::get_theme()
#'     par(bg = theme$background, fg = theme$foreground)
#'     plot(1:10)
#'   })
#' }
#'
#'
#' @name fire_event
#' @export
fire_event <- function(
  key,
  value,
  session = shiny::getDefaultReactiveDomain(),
  global = FALSE
) {
  if (is.null(session)) {
    stop("shidashi::fire_event() must be called in shiny server")
  }
  token <- session$token
  entry <- get_session_entry(token)
  if (is.null(entry)) {
    register_session(session)
    entry <- get_session_entry(token)
  }

  entry$events[[key]] <- value

  if (isTRUE(global)) {
    shared_id <- entry$shared_id
    if (length(shared_id) == 1L && nzchar(shared_id)) {
      peers <- globals_get_sessions_by_shared_id(shared_id)
      for (peer in peers) {
        if (!identical(peer$shiny_session$token, token)) {
          peer$events[[key]] <- value
        }
      }
    }
  }

  invisible()
}

#' @rdname fire_event
#' @export
get_event <- function(
  key,
  session = shiny::getDefaultReactiveDomain(),
  default = NULL
) {
  if (is.null(session)) {
    # Not running in shiny server, return default
    return(default)
  }
  token <- session$token
  entry <- get_session_entry(token)
  if (is.null(entry)) {
    register_session(session)
    entry <- get_session_entry(token)
  }
  entry$events[[key]] %||% default
}


globals_get_agent_mode <- function(module_id, missing = "None") {
  globals <- get_shidashi_globals()
  if (!is.environment(globals)) {
    return(missing)
  }
  mode_entry <- globals$module_agent_modes$get(module_id)
  if (is.null(mode_entry)) {
    return(missing)
  }
  mode <- mode_entry$current_mode
  if (length(mode) != 1) {
    return(missing)
  }
  return(mode)
}

globals_set_agent_mode <- function(module_id, mode) {
  globals <- get_shidashi_globals()
  if (!is.environment(globals)) {
    return(FALSE)
  }
  globals$module_agent_modes$set(module_id, list(current_mode = mode))
  return(TRUE)
}

globals_get_confirmation_policy <- function(module_id, missing = "auto_allow") {
  globals <- get_shidashi_globals()
  if (!is.environment(globals)) {
    return(missing)
  }
  policy <- globals$module_confirmation_policy$get(module_id)
  if (!length(policy) || !policy %in% c("auto_allow", "ask", "auto_reject")) {
    return(missing)
  }
  return(policy)
}

globals_set_confirmation_policy <- function(module_id, policy) {
  globals <- get_shidashi_globals()
  if (!is.environment(globals)) {
    return(FALSE)
  }
  if (!policy %in% c("auto_allow", "ask", "auto_reject")) {
    policy <- "auto_allow"
  }
  globals$module_confirmation_policy$set(module_id, policy)
  return(TRUE)
}

# Ensure module_conversations entry exists and return it
globals_get_conversation_entry <- function(module_id) {
  entry <- list(
    module_id = module_id,
    active_idx = 1L,
    conversations = list(
      list(
        title = "New conversation",
        turns = list(),
        mode = "None",
        last_visited = Sys.time()
      )
    )
  )

  globals <- get_shidashi_globals()
  if (!is.environment(globals)) {
    return(entry)
  }
  if (!globals$module_conversations$has(module_id)) {
    globals$module_conversations$set(module_id, entry)
  } else {
    entry <- globals$module_conversations$get(module_id)
  }
  entry
}

globals_set_conversation_entry <- function(entry, module_id = entry$module_id) {
  globals <- get_shidashi_globals()
  if (!is.environment(globals) || length(module_id) != 1) {
    return()
  }

  entry$module_id <- module_id
  entry$active_idx <- entry$active_idx %||% 1L
  entry$conversations <- entry$conversations %||% list(
    list(
      title = "New conversation",
      turns = list(),
      mode = "None",
      last_visited = Sys.time()
    )
  )

  globals$module_conversations$set(module_id, entry)
  return(invisible(entry))
}

conversation_title <- function(turns, max_chars = 50L) {
  if (!length(turns)) return("New conversation")
  txt <- turns[[1]]@text
  if (!isTRUE(max_chars >= 6L)) {
    max_chars <- 6L
  }
  if (nchar(txt) > max_chars) {
    txt <- truc_string(x = txt, annot = "", side = "end", max_char = max_chars, collapse = "")
  }
  return(txt)
}


globals_save_conversation <- function(module_id, chat) {
  if (!inherits(chat, "Chat")) { return() }
  globals <- get_shidashi_globals()
  if (!is.environment(globals) || length(module_id) != 1) {
    return()
  }

  turns <- chat$get_turns()
  if (!length(turns)) { return() }

  entry <- globals_get_conversation_entry(module_id = module_id)
  idx   <- entry$active_idx
  mode  <- globals_get_agent_mode(module_id = module_id)

  entry$module_id <- module_id
  entry$conversations[[idx]] <- list(
    title        = conversation_title(turns),
    turns        = turns,
    mode         = mode,
    last_visited = Sys.time()
  )

  globals_set_conversation_entry(entry = entry)
  return(invisible(entry))
}

globals_new_conversation <- function(module_id, chat, save_first = TRUE) {
  if (!inherits(chat, "Chat")) { return() }
  if (save_first) {
    globals_save_conversation(module_id = module_id, chat = chat)
  }

  # Append a new empty conversation and make it active
  entry <- globals_get_conversation_entry(module_id = module_id)
  entry$conversations[[length(entry$conversations) + 1]] <- list(
    title        = "New conversation",
    turns        = list(),
    mode         = globals_get_agent_mode(module_id = module_id),
    last_visited = Sys.time()
  )
  entry$active_idx <- length(entry$conversations)
  globals_set_conversation_entry(entry = entry, module_id = module_id)

  # Clear the Chat turns and UI
  chat$set_turns(list())
  invisible(chat)
}

# Bind tools from the MCP session registry, filtered by current mode
globals_bind_chat_tools <- function(
    chat, module_id, session = shiny::getDefaultReactiveDomain()) {

  registry <- globals_session_registry()
  mode <- globals_get_agent_mode(module_id = module_id)
  token <- session$token
  entry <- registry$get(token)
  enabled_tools <- list()
  if (entry$tools$size() > 0) {
    enabled_tools <- Filter(function(t) {
      is_tool_enabled_for_mode(t, mode)
    }, entry$tools$as_list())
  }
  # ask_user is always available
  enabled_tools <- c(enabled_tools, list(make_ask_user_tool(session)))
  chat$set_tools(enabled_tools)

  invisible(chat)
}

globals_new_chat <- function(module_id, session = shiny::getDefaultReactiveDomain()) {
  root_path <- template_root()
  module_path <- file.path(root_path, "modules", module_id)
  agent_conf_path <- file.path(module_path, "agents.yaml")

  if (!file.exists(agent_conf_path)) {
    stop("Module [", module_id, "] does not have `agents.yaml`: Agents are disabled.")
  }

  agent_conf <- yaml::read_yaml(agent_conf_path)
  system_prompt <- agent_conf$parameters$system_prompt
  chat <- init_chat(system_prompt = system_prompt)

  agent_default_mode <- agent_conf$parameters$default_mode %||% "None"
  mode_now <- globals_get_agent_mode(module_id = module_id, missing = agent_default_mode)
  globals_set_agent_mode(module_id = module_id, mode = mode_now)

  # restore conversation
  entry <- globals_get_conversation_entry(module_id = module_id)
  idx   <- entry$active_idx
  conv  <- entry$conversations[[idx]]
  if (length(conv$turns)) {
    chat$set_turns(conv$turns)
  }

  if (length(session)) {
    globals_bind_chat_tools(chat = chat,
                            module_id = module_id,
                            session = session)
  }

  chat

}


#' @rdname fire_event
#' @export
get_theme <- function(
  session = shiny::getDefaultReactiveDomain()
) {
  get_event(
    "theme.changed",
    session = session,
    default = list(
      theme = "light",
      background = "#FFFFFF",
      foreground = "#000000"
    )
  )
}

#' @rdname register_session
#' @export
enable_input_broadcast <- function(
  session = shiny::getDefaultReactiveDomain(),
  once = FALSE
) {
  register_session(session)
  entry <- get_session_entry(session$token)
  if (is.null(entry)) return(invisible())

  handler <- entry$handlers$get("broadcast_handler")
  if (inherits(handler, "Observer")) {
    if (once) {
      handler$run()
    } else {
      handler$resume()
    }
  }
  invisible()
}

#' @rdname register_session
#' @export
disable_input_broadcast <- function(
  session = shiny::getDefaultReactiveDomain()
) {
  entry <- get_session_entry(session$token)
  if (!is.null(entry)) {
    handler <- entry$handlers$get("broadcast_handler")
    if (inherits(handler, "Observer")) {
      handler$suspend()
    }
  }
  invisible()
}

#' @rdname register_session
#' @export
enable_input_sync <- function(session = shiny::getDefaultReactiveDomain(),
                              once = FALSE) {
  register_session(session)
  entry <- get_session_entry(session$token)
  if (is.null(entry)) return(invisible())

  handler <- entry$handlers$get("input_sync_handler")

  if (inherits(handler, "Observer")) {
    if (once) {
      handler$run()
    } else {
      handler$resume()
    }
  }
  invisible()
}

#' @rdname register_session
#' @export
disable_input_sync <- function(session = shiny::getDefaultReactiveDomain()) {
  entry <- get_session_entry(session$token)
  if (!is.null(entry)) {
    handler <- entry$handlers$get("input_sync_handler")
    if (inherits(handler, "Observer")) {
      handler$suspend()
    }
  }
  invisible()
}

#' @rdname register_session
#' @export
get_handler <- function(name, session = shiny::getDefaultReactiveDomain()) {
  if (session$isClosed()) {
    return(NULL)
  }
  entry <- get_session_entry(session$token)
  if (is.null(entry)) {
    register_session(session = session)
    entry <- get_session_entry(session$token)
  }
  if (is.null(entry)) return(NULL)
  entry$handlers$get(name)
}

#' @rdname register_session
#' @export
set_handler <- function(name, handler, session = shiny::getDefaultReactiveDomain()) {
  if (name %in% c("event_handler", "broadcast_handler", "input_sync_handler")) {
    stop("Handler name ", sQuote(name), " is reserved. Please pick another name.")
  }
  if (!inherits(handler, "Observer") && !is.null(handler)) {
    stop("shidashi::set_handler - `handler` must be `NULL` or shiny::observe({...})")
  }
  if (session$isClosed()) {
    return(invisible(FALSE))
  }
  # get current handler: `get_handler` will automatically register session
  # so no need to register here
  current_handler <- get_handler(name = name, session = session)
  if (identical(current_handler, handler)) {
    return(invisible(TRUE))
  }

  if (!is.null(current_handler)) {
    tryCatch({
      current_handler$suspend()
      current_handler$destroy()
    }, error = function(e) {})
  }

  entry <- get_session_entry(session$token)

  if (length(entry)) {
    entry$handlers$set(name, handler)
    return(invisible(TRUE))
  }

  # Cannot register session
  return(invisible(FALSE))
}

card_tabset_header <- function(id_tabset, index, title, active = FALSE) {
  shiny::tags$li(
    class = "nav-item nav-tab-header",
    shiny::a(
      class = ifelse(active, "nav-link active", "nav-link"),
      id = sprintf("%s-%s-tab", id_tabset, index),
      # For AdminLTE3
      "data-toggle" = "tab",
      # For bootstrap5
      "data-bs-toggle" = "tab",
      href = sprintf("#%s-%s", id_tabset, index),
      role = "tab",
      "aria-controls" = sprintf("%s-%s", id_tabset, index),
      "aria-selected" = ifelse(active, "true", "false"),
      # "tab-index" = index,
      title
    )
  )
}

card_tabset_content <- function(id_tabset, index, active, ...) {
  shiny::div(
    class = ifelse(active, "tab-pane fade active fill-min-height show position-relative", "tab-pane fade fill-min-height position-relative"),
    id = sprintf("%s-%s", id_tabset, index),
    role = "tabpanel",
    "aria-labelledby" = sprintf("%s-%s-tab", id_tabset, index),
    # "tab-index" = index,
    ...
  )
}

#' @title Generates a set of card panels
#' @description To insert, remove, or active card panels, see
#' \code{\link{card_tabset_operate}}.
#' @param ... 'HTML' tags; each tag will be placed into a card
#' @param names title of the tabs
#' @param inputId the id of the card-set, must start with letters
#' @param title the title of the card-set
#' @param active the title that will be active on load
#' @param tools a list of tools or badges generated by
#' \code{\link{card_tool}} or \code{\link{as_badge}}
#' @param footer the footer element of the card-set
#' @param class the 'HTML' class the of card-set
#' @param class_header,class_body,class_foot additional 'HTML'
#' class the of card header, body, and footer accordingly
#' @return 'HTML' tags
#' @seealso \code{\link{card_tabset_operate}}
#' @examples
#'
#' library(shiny)
#' library(shidashi)
#'
#' # Fake session to operate on card_tabset without shiny
#' session <- MockShinySession$new()
#'
#' card_tabset(
#'   inputId = "card_set",
#'   title = "Cardset with Tools",
#'   `Tab 1` = p("Tab content 1"),
#'   class_body = "height-500",
#'   tools = list(
#'     as_badge(
#'       "New|badge-success"
#'     ),
#'     card_tool(
#'       widget = "collapse"
#'     ),
#'     card_tool(
#'       widget = "maximize"
#'     )
#'   )
#' )
#'
#' card_tabset_insert(
#'   inputId = "card_set",
#'   title = "Tab 2",
#'   p("New content"),
#'   session = session
#' )
#'
#' card_tabset_activate(
#'   inputId = "card_set",
#'   title = "Tab 1",
#'   session = session
#' )
#'
#' card_tabset_remove(
#'   inputId = "card_set",
#'   title = "Tab 2",
#'   session = session
#' )
#'
#'
#'
#' @export
card_tabset <- function(
  ..., inputId = rand_string(prefix = "tabset-"), title = NULL,
  names = NULL, active = NULL, tools = NULL, footer = NULL,
  class = "", class_header = "", class_body = "", class_foot = "") {

  call_ <- match.call()

  if (grepl("^[^a-zA-Z][^a-zA-Z0-9_-]{0,}", inputId)) {
    stop(
      "card_tabset: invalid `inputId`, can only have letters, digits, '-', or '_', and must start with letters."
    )
  }

  tabs <- list(...)
  ntabs <- length(tabs)
  if (!length(names)) {
    names <- names(tabs)
  }
  if (length(names) != ntabs) {
    stop("card_tabset: `names` must have the same length as tab elements")
  }

  if (length(title) >= 1) {
    data_title <- trimws(as.character(title[[1]])[[1]])
  } else {
    data_title <- ""
  }

  if (length(title) == 1) {
    # Compatible with bs5/4 to center the text vertically
    title <- shiny::tags$li(class = "px-3 d-flex align-items-center", shiny::h4(class = "card-title", title))
  }
  if (length(active)) {
    active <- active[[1]]
  } else if (length(names)) {
    active <- names[[1]]
  }

  if (length(tools)) {
    tools <- shiny::tags$li(class = "nav-item ml-auto ms-auto", shiny::div(class = "card-tools", tools))
  }

  if (!is.null(footer)) {
    footer <- shiny::div(class = combine_class("card-footer", class_foot), footer)
  }

  set_attr_call(shiny::div(
    class = sprintf("card card-tabs %s", class),
    `data-title` = data_title,
    shiny::div(
      class = sprintf("card-header p-0 pt-0 %s", class_header),
      shiny::tags$ul(
        class = "nav nav-tabs",
        id = inputId,
        role = "tablist",
        title,
        lapply(seq_len(ntabs), function(ii) {
          title <- names[[ii]]
          card_tabset_header(inputId, ii, title, active = title %in% active)
        }),
        tools
      )
    ),
    shiny::div(
      class = combine_class("card-body", class_body),
      shiny::div(
        class = "tab-content",
        id = sprintf("%sContent", inputId),
        lapply(seq_len(ntabs), function(ii) {
          title <- names[[ii]]
          card_tabset_content(inputId, ii, active = title %in% active, tabs[[ii]])
        })
      )
    ),
    footer
  ), call = call_)
}

#' @name card_tabset_operate
#' @title Add, active, or remove a card within \code{\link{card_tabset}}
#' @param inputId the element id of \code{\link{card_tabset}}
#' @param title the title of the card to insert, activate, or to remove
#' @param ... the content of the card
#' @param active whether to set the card to be active once added
#' @param notify_on_failure whether to show notifications on failure
#' @param session shiny session domain
#' @return These functions execute \code{session$sendCustomMessage} and return
#' whatever value generated by that function; usually nothing.
#' @seealso \code{\link{card_tabset}}
#' @export
card_tabset_insert <- function(inputId, title, ..., active = TRUE,
                            notify_on_failure = TRUE, session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage(
    "shidashi.card_tabset_insert",
    list(
      inputId = session$ns(inputId),
      title = title,
      body = as.character(shiny::tagList(...)),
      active = isTRUE(active),
      notify_on_failure = isTRUE(notify_on_failure)
    )
  )
}

#' @rdname card_tabset_operate
#' @export
card_tabset_remove <- function(inputId, title, notify_on_failure = TRUE, session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage(
    "shidashi.card_tabset_remove",
    list(
      inputId = session$ns(inputId),
      title = title,
      notify_on_failure = isTRUE(notify_on_failure)
    )
  )
}

#' @rdname card_tabset_operate
#' @export
card_tabset_activate <- function(inputId, title, notify_on_failure = TRUE, session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage(
    "shidashi.card_tabset_activate",
    list(
      inputId = session$ns(inputId),
      title = title,
      notify_on_failure = isTRUE(notify_on_failure)
    )
  )
}

#' @title Generates outputs that can be written to clipboards with one click
#' @param outputId the output id
#' @param message tool tip to show when mouse hovers on the element
#' @param clip_text the initial text to copy to clipboards
#' @param class 'HTML' class of the element
#' @param as_card_tool whether to make the output as \code{\link{card_tool}}
#' @param expr expression to evaluate; the results will replace
#' \code{clip_text}
#' @param env environment to evaluate \code{expr}
#' @param quoted whether \code{expr} is quoted
#' @param outputArgs used to replace default arguments of \code{clipboardOutput}
#' @return 'HTML' elements that can write to clip-board once users click on
#' them.
#' @examples
#' clipboardOutput(clip_text = "Hey there")
#'
#' @export
clipboardOutput <- function(
  outputId = rand_string(prefix = "clipboard"), message = "Copy to clipboard",
  clip_text = "", class = NULL, as_card_tool = FALSE) {

  if (as_card_tool) {
    card_tool(
      class = combine_class("clipboard-btn", "shidashi-clipboard-output", class),
      icon = "copy",
      title = message,
      inputId = outputId,
      widget = "custom",
      "data-clipboard-text" = clip_text
    )
  } else {
    shiny::div(
      id = outputId,
      class = "shidashi-clipboard-output",
      shiny::tags$button(
        class = combine_class("clipboard-btn btn btn-default", class),
        "data-clipboard-text" = clip_text,
        role = "button",
        message
      )
    )
  }

}

#' @rdname clipboardOutput
#' @export
renderClipboard <- function(expr, env = parent.frame(), quoted = FALSE, outputArgs = list()) {

  func <- shiny::installExprFunction(expr, "func", env, quoted, label = "renderClipboard")
  shiny::createRenderFunction(func, function(value, session, name, ...) {

    if (!is.character(value)) {
      value <- deparse(value)
    }
    if (length(value) > 1) {
      value <- paste(value, collapse = "\n")
    }
    value

  }, clipboardOutput, outputArgs)

}

#' @name notification
#' @title The 'Bootstrap' notification
#' @param message notification body content, can be 'HTML' tags
#' @param title,subtitle title and subtitle of the notification
#' @param type type of the notification; can be \code{"default"},
#' \code{"info"}, \code{"warning"}, \code{"success"}, \code{"danger"},
#' \code{"white"}, \code{"dark"}
#' @param close whether to allow users to close the notification
#' @param position where the notification should be; choices are
#' \code{"topRight"}, \code{"topLeft"}, \code{"bottomRight"},
#' \code{"bottomLeft"}
#' @param autohide whether to automatically hide the notification
#' @param fixed whether the position should be fixed
#' @param delay integer in millisecond to hide the notification if
#' \code{autohide=TRUE}
#' @param icon the icon of the title
#' @param collapse if \code{message} is a character vector, the collapse string
#' @param session shiny session domain
#' @param class the extra class of the notification, can be used for style
#' purposes, or by \code{clear_notifications} to close specific notification
#' types.
#' @param ... other options; see
#' \url{https://adminlte.io/docs/3.1//javascript/toasts.html#options}
#'
#' @return Both functions should be used in shiny reactive contexts. The
#' messages will be sent to shiny 'JavaScript' interface and nothing will be
#' returned.
#' @examples
#' \dontrun{
#'
#' # the examples must run in shiny reactive context
#'
#' show_notification(
#'   message = "This validation process has finished. You are welcome to proceed.",
#'   autohide = FALSE,
#'   title = "Success!",
#'   subtitle = "type='success'",
#'   type = "success"
#' )
#'
#' show_notification(
#'   message = "This notification has title and subtitle",
#'   autohide = FALSE,
#'   title = "Hi there!",
#'   subtitle = "Welcome!",
#'   icon = "kiwi-bird",
#'   class = "notification-auto"
#' )
#'
#' # only clear notifications with class "notification-auto"
#' clear_notifications("notification-auto")
#'
#' }
NULL

#' @rdname notification
#' @export
show_notification <- function(
  message,
  title = "Notification!",
  subtitle = "",
  type = c("default", "info", "warning", "success", "danger", "white", "dark"),
  close = TRUE,
  position = c("topRight", "topLeft", "bottomRight", "bottomLeft"),
  autohide = TRUE,
  fixed = TRUE,
  delay = 5000,
  icon = NULL,
  collapse = "",
  session = shiny::getDefaultReactiveDomain(),
  class = NULL,
  ...
) {
  type <- match.arg(type)
  position <- match.arg(position)
  delay <- as.integer(delay)
  if (!length(delay) || is.na(delay) || delay <= 0) {
    delay <- 5000L
    close <- TRUE
  }
  if (length(icon)) {
    icon <- as_icon(icon)
    icon <- icon$attribs$class
  }

  message <- paste(message, collapse = "")


  session$sendCustomMessage("shidashi.show_notification", list(
    position = position,
    autohide = !isFALSE(autohide),
    delay = delay,
    icon = icon,
    title = title,
    subtitle = subtitle,
    close = !isFALSE(close),
    body = message,
    fixed = !isFALSE(fixed),
    class = combine_class(sprintf("bg-%s", type), class),
    ...
  ))
  invisible()
}

#' @rdname notification
#' @export
clear_notifications <- function(
  class = NULL,
  session = shiny::getDefaultReactiveDomain()) {

  class <- unique(c(class, "toast"))
  class <- gsub(" ", "", class)
  class <- class[class != ""]
  selector <- paste0(".", class, collapse = "")
  session$sendCustomMessage("shidashi.clear_notification", list(
    selector = selector
  ))
  invisible()
}

#' Convert characters, shiny icons into 'fontawesome' 4
#' @param icon character or \code{\link[shiny]{icon}}
#' @param class icon class; change this when you are using 'fontawesome'
#' professional version. The choices are \code{'fa'} (compatible),
#' \code{'fas'} (strong), \code{'far'} (regular), \code{'fal'} (light),
#' and \code{'fad'} (duo-tone).
#' @return 'HTML' tag
#' @examples
#'
#' if(interactive()) {
#' as_icon("bookmark", class = "far")
#' as_icon("bookmark", class = "fas")
#'
#' # no icon
#' as_icon(NULL)
#' }
#'
#' @export
as_icon <- function(icon = NULL, class = "fas") {
  class <- combine_class(class)

  if (!length(icon) || (is.character(icon) && icon == "")) {
    icon <- ""
  } else {
    if (inherits(icon, "shiny.tag")) {
      icon$attribs$class <- combine_class(icon$attribs$class, class)
    } else {
      icon <- shiny::icon(icon, class = class)
    }
    # remove class fa
    if ( !"fa" %in% class ) {
      icon$attribs$class <- remove_html_class(icon$attribs$class, "fa")
    }
  }


  icon
}

#' @title Generates badge icons
#' @description Usually used along with \code{\link{card}},
#' \code{\link{card2}}, and \code{\link{card_tabset}}. See \code{tools}
#' parameters in these functions accordingly.
#' @param badge characters, \code{"shiny.tag"} object or \code{NULL}
#' @return 'HTML' tags
#' @details When \code{badge} is \code{NULL} or empty, then \code{as_badge}
#' returns empty strings. When \code{badge} is a \code{"shiny.tag"} object,
#' then 'HTML' class \code{'right'} and \code{'badge'} will be appended.
#' When \code{badge} is a string, it should follow the syntax of
#' \code{"message|class"}. The text before \code{"|"} will be the badge
#' message, and the text after the \code{"|"} becomes the class string.
#' @examples
#'
#' # Basic usage
#' as_badge("New")
#'
#' # Add class `bg-red` and `no-padding`
#' as_badge("New|bg-red no-padding")
#'
#'
#' @export
as_badge <- function(badge = NULL) {
  if (!length(badge) || nchar(badge) == 0) {
    badge <- ""
  } else {
    if (inherits(badge, "shiny.tag")) {
      badge$attribs$class <- combine_class(badge$attribs$class, "right badge")
    } else {
      badge <- strsplit(badge, "\\|")[[1]]
      if (length(badge) > 1) {
        badge <- shiny::span(class = paste("right badge", badge[[2]]), badge[[1]])
      } else {
        badge <- shiny::span(class = paste("right badge"), badge[[1]])
      }
    }
  }
  badge
}


menu_item <- function(
  text, href = "#", icon = NULL, active = FALSE, badge = NULL,
  target = "_blank", root_path = template_root()) {

  icon <- as_icon(icon, class = "nav-icon fas")
  badge <- as_badge(badge)
  module <- ""
  if (startsWith(href, "#")) {
    target <- "_self"
  } else if (startsWith(href, "?module=")) {
    query_list <- shiny::parseQueryString(href)
    module <- query_list$module
    module <- gsub(pattern = " ", replacement = "", module)
    if ( grepl("[^a-zA-Z0-9_]", module) ) {
      stop("Function `menu_item`: for `href` with module link (starts with '?module=<ID>'), the module `ID` can only contain letters, digits, and/or '_'.")
    }
  }
  template_path <- file.path(root_path, "views", "menu-item.html")
  # if(!file.exists(template_path)) {
  #   template_path <- system.file('snippets', 'menu-item.html', package = 'shidashi')
  # }
  shiny::htmlTemplate(
    template_path,
    document_ = FALSE,
    text = text, href = shiny::HTML(href), icon = icon, badge = badge, active = active,
    target = target, module = module)
}

menu_item_dropdown <- function(
  text, ..., .list = NULL, icon = NULL, active = FALSE,
  badge = NULL, root_path = template_root()) {

  sub_items <- c(shiny::tagList(...), .list)
  icon <- as_icon(icon, class = "nav-icon fas")
  badge <- as_badge(badge)
  template_path <- file.path(root_path, "views", "menu-item-dropdown.html")
  # if(!file.exists(template_path)) {
  #   template_path <- system.file('snippets', 'menu-item-dropdown.html', package = 'shidashi')
  # }
  shiny::htmlTemplate(
    template_path,
    document_ = FALSE,
    text = text, icon = icon, badge = badge,
    active = active, sub_items = sub_items)
}

#' @title Generate 'HTML' tags with 'flex' layout
#' @param ... for \code{flex_container}, it's elements of \code{flex_item};
#' for \code{flex_item}, \code{...} are shiny 'HTML' tags
#' @param size numerical relative size of the item; will be ignored if
#' \code{flex} is provided
#' @param style the additional 'CSS' style for containers or inner items
#' @param direction,wrap,justify,align_box,align_content 'CSS' styles for
#' 'flex' containers
#' @param order,align,flex CSS' styles for 'flex' items
#' @param class,.class class to add to the elements
#' @return 'HTML' tags
#'
#' @examples
#'
#' x <- flex_container(
#'   style = "position:absolute;height:100vh;top:0;left:0;width:100%",
#'   flex_item(style = 'background-color:black;'),
#'   flex_item(style = 'background-color:red;')
#' )
#' # You can view it via `htmltools::html_print(x)`
#'
#' @export
flex_container <- function(
  ...,
  style = NULL,
  direction = c("row", "column"),
  wrap = c("wrap", "nowrap", "wrap-reverse"),
  justify = c("flex-start", "center", "flex-end", "space-around", "space-between"),
  align_box = c("stretch", "flex-start", "center", "flex-end", "baseline"),
  align_content = c("stretch", "flex-start", "flex-end", "space-between", "space-around", "center")
) {
  call <- match.call(expand.dots = FALSE)
  style1 <- style
  style <- list()

  direction <- match.arg(direction)
  style[["flex-direction"]] <- direction

  wrap <- match.arg(wrap)
  style[["flex-wrap"]] <- wrap

  justify <- match.arg(justify)
  style[["justify-content"]] <- justify

  align_box <- match.arg(align_box)
  style[["align-content"]] <- align_box

  align_content <- match.arg(align_content)
  style[["align-items"]] <- align_content

  style$display <- "flex"
  style <- paste(names(style), as.vector(style), sep = ":", collapse = "; ")
  if (length(style1)) {
    style <- paste0(style, "; ", style1)
  }


  shiny::div(style = style, ...)
}

#' @rdname flex_container
#' @export
flex_item <- function(
  ..., size = 1,
  style = NULL, order = NULL, flex = as.character(size),
  align = c("flex-start", "flex-end", "center"),
  class = NULL,
  .class = "fill-width padding-5"
) {
  l <- list()
  if (length(align) == 1) {
    align <- match.arg(align)
    l[["align-self"]] <- align
  }
  l[["order"]] <- order
  l[["flex"]] <- flex

  style1 <- paste(names(l), as.vector(l), sep = ":", collapse = "; ")
  if (length(style)) {
    style1 <- paste0(style1, "; ", style)
  }

  class <- combine_class(class, .class)

  shiny::div(
    ...,
    style = style1,
    class = class
  )

}

#' @rdname flex_container
#' @export
flex_break <- function(..., class = NULL) {
  class <- combine_class(class, "flex-break")
  shiny::div(class = class, ...)
}

#' 'HTML' code to generate small back-to-top button
#' @description This function is a template function that should be called
#' in 'HTML' templates before closing the \code{"</body>"} tag.
#'
#' When \code{open_drawer = TRUE}, an additional button is rendered
#' that fires a \code{"button.click"} shidashi-event with
#' \code{type = "open_drawer"}.
#' Module server code (or \code{\link{chatbot_server}}) can observe
#' this event via \code{\link{get_event}} and call
#' \code{\link{drawer_open}} / \code{shiny::renderUI} to fill the
#' drawer with content.
#'
#' @param icon the icon for back-to-top button
#' @param title the expanded menu title
#' @param open_drawer logical; whether to include a drawer-toggle
#'   button. Defaults to \code{TRUE} if a
#'   \code{.shidashi-drawer} element will be present in the page
#'   (e.g.\ from \code{\link{module_drawer}()}).
#' @param drawer_icon the icon for the drawer-toggle button;
#'   defaults to \code{"ellipsis"} (three dots). Use e.g.
#'   \code{"robot"} for AI-agent modules.
#' @return 'HTML' tags
#'
#' @examples
#'
#' back_top_button()
#' back_top_button("rocket")
#' back_top_button("rocket", drawer_icon = "robot")
#'
#' @export
back_top_button <- function(icon = "chevron-up", title = "Jump to",
                            open_drawer = TRUE, drawer_icon = "ellipsis") {
  if (!length(title)) {
    title <- NULL
  } else {
    title <- shiny::h6(class = "dropdown-header", title)
  }

  # ---- Optional drawer button (fires shidashi-event, no logic) ----
  if (isTRUE(open_drawer)) {
    drawer_btn <- shiny::a(
      type = "button",
      class = "btn btn-default btn-drawer-toggle",
      href = "#",
      "data-shidashi-action" = "drawer-toggle",
      title = "Open panel",
      as_icon(drawer_icon)
    )
  } else {
    drawer_btn <- NULL
  }

  shiny::div(
    class = "shidashi-back-to-top",
    drawer_btn,
    shiny::div(
      class = "btn-group dropup",
      role = "group",
      shiny::a(
        type = "button",
        class = "btn btn-default btn-go-top border-right-1",
        href = "#",
        as_icon(icon)
      ),
      shiny::tags$button(
        type = "button",
        class = "btn btn-default dropdown-toggle dropdown-toggle-split border-left-1",
        # AdminLTE3
        "data-toggle" = "dropdown",
        # bs5
        "data-bs-toggle" = "dropdown",
        "aria-haspopup" = "false",
        "aria-expanded" = "false",
        shiny::span(class = "sr-only visually-hidden", "Dropdown-Open")
      ),
      shiny::div(class = "dropdown-menu dropdown-menu-end", title)
    )
  )
}



#' @name add-remove-html-class
#' @title Add or remove 'HTML' class from 'RAVE' application
#' @description Only works in template framework provided by 'shidashi' package,
#' see \code{\link[shidashi]{use_template}}
#' @param selector 'CSS' selector
#' @param class class to add or to remove from selected elements
#' @param session shiny session
#' @return No value is returned
#'
#' @examples
#'
#' server <- function(input, output, session) {
#'
#'   # Add class `hidden` to element with ID `elemid`
#'   add_class("#elemid", "hidden")
#'
#'   # Remove class `hidden` from element with class `shiny-input-optional`
#'   remove_class(".shiny-input-optional", "hidden")
#' }
#'
#' @export
add_class <- function(selector, class,
                      session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.add_class", list(
    selector = selector,
    class = class
  ))
}

#' @rdname add-remove-html-class
#' @export
remove_class <- function(selector, class,
                         session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.remove_class", list(
    selector = selector,
    class = class
  ))
}


#' @title Open a URL in a new browser tab
#' @description Sends a message to the client to open the specified URL
#' in a new browser window/tab.
#' @param url character string, the URL to open
#' @param target the \code{window.open} target; default is
#' \code{"_blank"} (new tab)
#' @param session shiny session
#' @return No value is returned (called for side effect).
#' @export
open_url <- function(url, target = "_blank",
                     session = shiny::getDefaultReactiveDomain()) {
  session$sendCustomMessage("shidashi.open_url", list(
    url = url,
    target = target
  ))
}

---
name: greet
description: Greets a user by name using a simple R script
---

## Instructions

This skill demonstrates the skill system. It runs a short R script
that prints a personalised greeting.

### Usage

1. Call `action='script'`, `file_name='greet.R'`, `args=['World']`
2. The script prints: `Hello, World!`

### Arguments

- `args[1]`: The name to greet (default: `"World"`)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream.R
\name{stream_to_js}
\alias{stream_to_js}
\title{Write data to a \pkg{shidashi} stream binary file}
\usage{
stream_to_js(
  abspath,
  data,
  type = c("raw", "json", "int32", "float32", "float64"),
  ...
)
}
\arguments{
\item{abspath}{Character scalar. Absolute path to the target \code{.bin}
file.  Use \code{\link{stream_path}(id)} to obtain a path under the
app's \code{www/stream/} directory.}

\item{data}{The data to serialize.  See Details for how each \code{type}
interprets this argument.}

\item{type}{Character scalar. One of \code{"raw"}, \code{"json"},
\code{"int32"}, \code{"float32"}, or \code{"float64"}.}

\item{...}{Additional named scalar values to embed in the JSON header and
expose to the JavaScript caller via \code{result.header}.}
}
\value{
Invisibly returns \code{abspath}.
}
\description{
Serializes \code{data} together with a JSON header into the binary
envelope expected by the JavaScript \code{window.shidashi.fetchStreamData(id)}
method.
}
\details{
\strong{Wire format}
\preformatted{
  [endianFlag: 1 byte]  [headerLen: uint32 LE]  [header: UTF-8 JSON]  [body]
}
\code{endianFlag} is always \code{0x01} (little-endian).
\code{header} is a JSON object containing at minimum \code{data_type}
plus any extra fields passed via \code{...}.

\strong{Body encoding by type}
\describe{
  \item{\code{"raw"}}{Passed through verbatim; \code{data} must be a
    \code{raw} vector or will be coerced via \code{as.raw()}.}
  \item{\code{"json"}}{\code{data} is serialized with
    \code{jsonlite::toJSON(auto_unbox = TRUE)}.}
  \item{\code{"int32"}}{\code{data} is coerced to integer and written as
    4-byte little-endian signed integers.}
  \item{\code{"float32"}}{\code{data} is coerced to double and written as
    4-byte little-endian IEEE 754 single-precision floats.}
  \item{\code{"float64"}}{\code{data} is coerced to double and written as
    8-byte little-endian IEEE 754 double-precision floats.}
}
}
\seealso{
\code{\link{stream_path}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accordion.R
\name{accordion}
\alias{accordion}
\alias{accordion_operate}
\title{Generates an 'accordion' tab-set}
\usage{
accordion(
  ...,
  id = rand_string(prefix = "accordion-"),
  class = NULL,
  style_header = NULL,
  style_body = NULL,
  env = parent.frame(),
  extras = list(),
  root_path = template_root()
)

accordion_operate(
  id,
  itemId,
  item_title,
  method = c("expand", "collapse", "toggle"),
  session = shiny::getDefaultReactiveDomain()
)
}
\arguments{
\item{...}{'accordion' items, generated by \code{\link{accordion_item}}}

\item{id}{the element id, must be unique}

\item{class}{the additional 'HTML' class}

\item{style_header}{additional 'CSS' styles for header}

\item{style_body}{additional 'CSS' styles for content body}

\item{env}{environment to evaluate \code{...}}

\item{extras}{key-value pairs that overrides the parameters in
\code{\link{accordion_item}}}

\item{root_path}{see \code{\link{template_root}}}

\item{itemId}{\code{\link{accordion_item}} id}

\item{item_title}{\code{\link{accordion_item}} title, if item id is
specified, this title will be ignored}

\item{method}{operation, choices are \code{'expand'} (default),
\code{'collapse'}, or \code{'toggle'}}

\item{session}{\code{shiny} session}
}
\value{
\code{'shiny.tag.list'} 'HTML' tags
}
\description{
Generates an 'accordion' tab-set that only one tab is
expanded at a time. This feature is experimental and has bugs in
some situations. Please use it at your own risk.
}
\examples{


if(interactive()) {

  library(shiny)
  library(shidashi)

  accordion(
    id = "input-set",
    accordion_item(
      title = "Input Group A",
      textInput("input_1", "Input 1"),
      collapsed = FALSE,
      footer = "Anim pariatur cliche reprehenderit dolor brunch."
    ),
    accordion_item(
      title = "Input Group B",
      textInput("input_2", "Input 2"),
      footer = actionButton("btn1", "OK"),
      collapsed = FALSE
    )
  )
}

}
\seealso{
\code{\link{accordion_item}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/widgets.R
\name{open_url}
\alias{open_url}
\title{Open a URL in a new browser tab}
\usage{
open_url(url, target = "_blank", session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{url}{character string, the URL to open}

\item{target}{the \code{window.open} target; default is
\code{"_blank"} (new tab)}

\item{session}{shiny session}
}
\value{
No value is returned (called for side effect).
}
\description{
Sends a message to the client to open the specified URL
in a new browser window/tab.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{include_view}
\alias{include_view}
\title{Template function to include 'snippets' in the view folder}
\usage{
include_view(file, ..., .env = parent.frame(), .root_path = template_root())
}
\arguments{
\item{file}{files in the template \code{views} folder}

\item{...}{ignored}

\item{.env, .root_path}{internally used}
}
\value{
rendered 'HTML' segments
}
\description{
Store the reusing 'HTML' segments in the
\code{views} folder. This function should be used in the
\code{'index.html'} template
}
\examples{
\dontrun{
# in your 'index.html' file
<html>
<header>
{{ shidashi::include_view("header.html") }}
</header>
<body>

</body>
<!-- Before closing html tag -->
{{ shidashi::include_view("footer.html") }}
</html>
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress.R
\name{progressOutput}
\alias{progressOutput}
\alias{renderProgress}
\title{Progress bar in shiny dashboard}
\usage{
progressOutput(
  outputId,
  ...,
  description = "Initializing",
  width = "100\%",
  class = "bg-primary",
  value = 0,
  size = c("md", "sm", "xs")
)

renderProgress(expr, env = parent.frame(), quoted = FALSE, outputArgs = list())
}
\arguments{
\item{outputId}{the element id of the progress}

\item{...}{extra elements on the top of the progress bar}

\item{description}{descriptive message below the progress bar}

\item{width}{width of the progress}

\item{class}{progress class, default is \code{"bg-primary"}}

\item{value}{initial value, ranging from 0 to 100; default is 0}

\item{size}{size of the progress bar; choices are \code{"md"}, \code{"sm"},
\code{"xs"}}

\item{expr}{R expression that should return a named list of \code{value} and
\code{description}}

\item{env}{where to evaluate \code{expr}}

\item{quoted}{whether \code{expr} is quoted}

\item{outputArgs}{a list of other parameters in \code{progressOutput}}
}
\value{
\code{progressOutput} returns 'HTML' tags containing progress bars
that can be rendered later via \code{\link{shiny_progress}} or
\code{renderProgress}. \code{renderProgress} returns shiny render functions
internally.
}
\description{
For detailed usage, see demo application by
running \code{render()}.
}
\examples{

library(shiny)
library(shidashi)
progressOutput("sales_report_prog1",
               description = "6 days left!",
               "Add Products to Cart",
               span(class="float-end", "123/150"),
               value = 123/150 * 100)

# server function
server <- function(input, output, session, ...) {
  output$sales_report_prog1 <- renderProgress({
    return(list(
      value = 140 / 150 * 100,
      description = "5 days left!"
    ))
  })
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card-tool.R
\name{card_tool}
\alias{card_tool}
\title{Generates small icon widgets}
\usage{
card_tool(
  inputId = NULL,
  title = NULL,
  widget = c("maximize", "collapse", "remove", "flip", "refresh", "link", "custom"),
  icon,
  class = "",
  href = "#",
  target = "_blank",
  start_collapsed = FALSE,
  ...
)
}
\arguments{
\item{inputId}{the button id, only necessary when \code{widget}
is \code{"custom"}}

\item{title}{the tip message to show when the mouse cursor hovers
on the icon}

\item{widget}{the icon widget type; choices are \code{"maximize"},
\code{"collapse"}, \code{"remove"}, \code{"flip"},
\code{"refresh"}, \code{"link"}, and \code{"custom"}; see 'Details'}

\item{icon}{icon to use if you are unsatisfied with the default ones}

\item{class}{additional class for the tool icons}

\item{href, target}{used when \code{widget} is \code{"link"}, will
open an external website; default is open a new tab}

\item{start_collapsed}{used when \code{widget} is \code{"collapse"},
whether the card should start collapsed}

\item{...}{passed to the tag as attributes}
}
\value{
'HTML' tags to be included in \code{tools} parameter in
\code{\link{accordion}}, \code{\link{card}}, \code{\link{card2}},
\code{\link{card_tabset}}
}
\description{
The icons cane be displayed at header line within
\code{\link{accordion}}, \code{\link{card}}, \code{\link{card2}},
\code{\link{card_tabset}}. See their examples.
}
\details{
There are 7 \code{widget} types:
\describe{
\item{\code{"maximize"}}{allow the elements to maximize
themselves to full-screen}
\item{\code{"collapse"}}{allow the elements to collapse}
\item{\code{"remove"}}{remove a \code{\link{card}} or
\code{\link{card2}}}
\item{\code{"flip"}}{used together with \code{\link{flip_box}},
to allow card body to flip over}
\item{\code{"refresh"}}{refresh all shiny outputs}
\item{\code{"link"}}{open a hyper-link pointing to external
websites}
\item{\code{"custom"}}{turn the icon into a \code{actionButton}.
in this case, \code{inputId} must be specified.}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\name{init_app}
\alias{init_app}
\title{Initialize a shidashi application}
\usage{
init_app(env = parent.frame())
}
\arguments{
\item{env}{environment where the application is initialized into.}
}
\value{
Nothing.
}
\description{
Creates a fresh environment holding all per-application global stores
(session registry, per-module input registries, etc.).
This function is designed to be called from a \file{global.R} file
generated by \code{\link{render}()}, not for user's call.
}
\examples{

init_app(env = new.env())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drawer.R
\name{module_drawer}
\alias{module_drawer}
\title{Drawer shell for module templates}
\usage{
module_drawer(output_id = "shidashi_drawer")
}
\arguments{
\item{output_id}{character; the output ID for the
\code{uiOutput} placeholder inside the drawer.
Defaults to \code{"shidashi_drawer"}.}
}
\value{
A \code{shiny::tagList} containing the drawer div and
  its overlay.
}
\description{
Emits a minimal \code{.shidashi-drawer} container with a
\code{\link[shiny]{uiOutput}} placeholder inside, plus the drawer
overlay.  The drawer starts empty; module server code fills it
dynamically via \code{shiny::renderUI}.

Typical usage in a \file{module-ui.html} template:
\preformatted{
  \{\{ shidashi::module_drawer() \}\}
}

Then in the module server:
\preformatted{
  output$shidashi_drawer <- shiny::renderUI(\{
    shiny::tagList(
      shiny::h5("My settings"),
      shiny::p("Custom drawer content here.")
    )
  \})
}

The \code{ns()} function from the module's template evaluation
environment is used automatically so that the output ID is
properly scoped to the module namespace.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/widgets.R
\name{back_top_button}
\alias{back_top_button}
\title{'HTML' code to generate small back-to-top button}
\usage{
back_top_button(
  icon = "chevron-up",
  title = "Jump to",
  open_drawer = TRUE,
  drawer_icon = "ellipsis"
)
}
\arguments{
\item{icon}{the icon for back-to-top button}

\item{title}{the expanded menu title}

\item{open_drawer}{logical; whether to include a drawer-toggle
button. Defaults to \code{TRUE} if a
\code{.shidashi-drawer} element will be present in the page
(e.g.\ from \code{\link{module_drawer}()}).}

\item{drawer_icon}{the icon for the drawer-toggle button;
defaults to \code{"ellipsis"} (three dots). Use e.g.
\code{"robot"} for AI-agent modules.}
}
\value{
'HTML' tags
}
\description{
This function is a template function that should be called
in 'HTML' templates before closing the \code{"</body>"} tag.

When \code{open_drawer = TRUE}, an additional button is rendered
that fires a \code{"button.click"} shidashi-event with
\code{type = "open_drawer"}.
Module server code (or \code{\link{chatbot_server}}) can observe
this event via \code{\link{get_event}} and call
\code{\link{drawer_open}} / \code{shiny::renderUI} to fill the
drawer with content.
}
\examples{

back_top_button()
back_top_button("rocket")
back_top_button("rocket", drawer_icon = "robot")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/settings.R
\docType{data}
\name{template_settings}
\alias{template_settings}
\alias{template_settings_set}
\alias{template_settings_get}
\alias{template_root}
\title{Configure template options that are shared across the sessions}
\format{
An object of class \code{list} of length 3.
}
\usage{
template_settings

template_settings_set(...)

template_settings_get(name, default = NULL)

template_root()
}
\arguments{
\item{...}{key-value pair to set options}

\item{name}{character, key of the value}

\item{default}{default value if the key is missing}
}
\value{
\code{template_settings_get} returns the values represented by the
corresponding keys, or the default value if key is missing.
}
\description{
Configure template options that are shared across the sessions
}
\details{
The settings is designed to store static key-value pairs that
are shared across the sessions. The most important key is
\code{"root_path"}, which should be a path pointing to the template
folder.
}
\examples{

# Get current website root path

template_root()

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream.R
\name{stream_path}
\alias{stream_path}
\title{Get the absolute path for a shidashi stream file}
\usage{
stream_path(id, session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{id}{Character scalar. Stream identifier.  Must be a valid file-name
component (no path separators).}

\item{session}{Shiny session object.  Defaults to the currently active
reactive domain.}
}
\value{
Invisible character scalar: absolute path to the \code{.bin} file.
}
\description{
Returns the full path to the binary file that will be served at
\code{stream/{token}_{id}.bin}.  The filename embeds the session token so
that simultaneous Shiny sessions never overwrite each other's data.
Call \code{\link{stream_init}} at least once per session before relying on
this function.
}
\seealso{
\code{\link{stream_init}}, \code{\link{stream_to_js}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{html_class}
\alias{html_class}
\alias{combine_html_class}
\alias{remove_html_class}
\title{Combine, add, or remove 'HTML' classes}
\usage{
combine_html_class(...)

remove_html_class(target, class)
}
\arguments{
\item{...}{one or more characters, classes to combine; duplicated classes
will be removed}

\item{target}{characters, class list}

\item{class}{one or more characters, classes to be removed from \code{target}}
}
\value{
A character string of new 'HTML' class
}
\description{
Combine 'HTML' classes to produce nice, clean 'HTML' class
string via \code{combine_html_class}, or to remove a class via
\code{remove_html_class}
}
\examples{

# Combine classes "a b c d e"
combine_html_class("a", "b  a", c("c", " d", "b"), list("e ", "a"))

# Remove class
remove_html_class("a b   c  e", c("b", "c "))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream.R
\name{stream_init}
\alias{stream_init}
\title{Initialize the shidashi stream directory for a Shiny session}
\usage{
stream_init(session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{session}{Shiny session object.  Defaults to the currently active
reactive domain.}
}
\value{
Invisibly returns the absolute path to the stream directory.
}
\description{
Call once in the server function to set up the directory that
\code{\link{stream_path}} and \code{\link{stream_to_js}} will write to.
When running inside a shidashi template the directory is automatically
resolved from the template root; when running in plain Shiny a temporary
directory is created and registered with
\code{\link[shiny]{addResourcePath}} so that the browser can fetch
\code{stream/{token}_{id}.bin}.  An optional cleanup hook is registered
to remove this session's files when the session ends.
}
\seealso{
\code{\link{stream_path}}, \code{\link{stream_to_js}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modules.R
\name{module_info}
\alias{module_info}
\alias{current_module}
\alias{active_module}
\alias{switch_module}
\alias{load_module}
\title{Obtain the module information}
\usage{
module_info(root_path = template_root(), settings_file = "modules.yaml")

current_module(
  session = shiny::getDefaultReactiveDomain(),
  root_path = template_root()
)

active_module(
  session = shiny::getDefaultReactiveDomain(),
  root_path = template_root()
)

switch_module(module_id, session = shiny::getDefaultReactiveDomain())

load_module(
  root_path = template_root(),
  request = list(QUERY_STRING = "/"),
  env = parent.frame()
)
}
\arguments{
\item{root_path}{the root path of the website project}

\item{settings_file}{the settings file containing the module information}

\item{session}{shiny reactive domain; used to extract the module id from
the URL query string when \code{.module_id} is not found.}

\item{module_id}{character string; the target module identifier (must
match an entry in \file{modules.yaml}).}

\item{request}{'HTTP' request string}

\item{env}{environment to load module variables into}
}
\value{
A data frame with the following columns that contain the module
information:
\describe{
\item{\code{id}}{module id, folder name}
\item{\code{order}}{display order in side-bar}
\item{\code{group}}{group menu name if applicable, otherwise \code{NA}}
\item{\code{label}}{the readable label to be displayed on the side-bar}
\item{\code{icon}}{icon that will be displayed ahead of label, will be
passed to \code{\link{as_icon}}}
\item{\code{badge}}{badge text that will be displayed
following the module label, will be passed to \code{\link{as_badge}}}
\item{\code{url}}{the relative 'URL' address of the module.}
}

\code{current_module}: a named list with \code{id}, \code{group},
\code{label}, \code{icon}, \code{badge}, and \code{url} of the current
module, or \code{NULL} if no module is active.

\code{active_module}: a named list with \code{id}, \code{group},
\code{label}, \code{icon}, \code{badge}, and \code{url} of the
currently active (visible) module, or \code{NULL} if no module is active.
}
\description{
\code{current_module} returns the information of the currently
running module. It looks up the \code{.module_id} variable in the calling
environment (set automatically when a module is loaded), then retrieves
the corresponding row from the module table.

\code{active_module} returns a \emph{reactive} value with
information about the module that is currently visible in the \verb{iframe}
tab (or the standalone module if no \verb{iframe} manager is present).
Unlike \code{current_module} which is static and always returns the module
whose server code is running, \code{active_module} dynamically tracks
which module the user is looking at from any context.

\code{switch_module} programmatically switches the active
module in the dashboard UI.
It sends a \code{shidashi.switch_module} message to the JavaScript
front-end.  When called from a module running inside an \verb{iframe}, the
handler automatically forwards the request to the parent window via
\code{postMessage} so that the sidebar highlight, tab bar, and \verb{iframe}
all update correctly.
}
\details{
The module files are stored in \code{modules/} folder in your
project. The folder names are the module id. Within each folder,
there should be one \code{"server.R"}, \code{R/}, and a
\code{"module-ui.html"}.

The \code{R/} folder stores R code files that generate variables,
which will be available to the other two files. These variables, along
with some built-ins, will be used to render \code{"module-ui.html"}.
The built-in functions are
\describe{
\item{ns}{shiny name-space function; should be used to generate the id for
inputs and outputs. This strategy avoids conflict id effectively.}
\item{.module_id}{a variable of the module id}
\item{module_title}{a function that returns the module label}
}

The \code{"server.R"} has access to all the code in \code{R/} as well.
Therefore it is highly recommended that you write each 'UI' component
side-by-side with their corresponding server functions and call
these server functions in \code{"server.R"}.

\code{active_module} works by reading the
\code{'@shidashi_active_module@'} Shiny input that is set by the
JavaScript front-end whenever a module tab is activated.
Because it accesses \code{session$rootScope()$input}, the return value
is reactive: when called inside an \code{observe} or \code{reactive}
context it will re-fire whenever the user switches modules.

If the input has not been set yet (e.g. before any module is opened),
the function falls back to \code{current_module()}.
}
\examples{

library(shiny)
module_info()

# load master module
load_module()

# load specific module
module_data <- load_module(
  request = list(QUERY_STRING = "/?module=module_id"))
env <- module_data$environment

if (interactive()) {

# get module title
env$module_title()

# generate module-specific shiny id
env$ns("input1")

# generate part of the UI
env$ui()

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{format_text_r}
\alias{format_text_r}
\alias{html_highlight_code}
\title{Get re-formatted \code{R} expressions in characters}
\usage{
format_text_r(
  expr,
  quoted = FALSE,
  reformat = TRUE,
  width.cutoff = 80L,
  indent = 2,
  wrap = TRUE,
  args.newline = TRUE,
  blank = FALSE,
  ...
)

html_highlight_code(
  expr,
  class = NULL,
  quoted = FALSE,
  reformat = TRUE,
  copy_on_click = TRUE,
  width.cutoff = 80L,
  indent = 2,
  wrap = TRUE,
  args.newline = TRUE,
  blank = FALSE,
  ...,
  hover = c("overflow-visible-on-hover", "overflow-auto")
)
}
\arguments{
\item{expr}{\code{R} expressions}

\item{quoted}{whether \code{expr} is quoted}

\item{reformat}{whether to reformat}

\item{width.cutoff, indent, wrap, args.newline, blank, ...}{passed to
\code{\link[formatR]{tidy_source}}}

\item{class}{class of \code{<pre>} tag}

\item{copy_on_click}{whether to copy to clipboard if user clicks on the
code; default is true}

\item{hover}{mouse hover behavior}
}
\value{
\code{format_text_r} returns characters,
\code{html_highlight_code} returns the 'HTML' tags wrapping expressions
in \code{<pre>} tag
}
\description{
Get re-formatted \code{R} expressions in characters
}
\examples{

s <- format_text_r(print(local({a<-1;a+1})))
cat(s)

x <- info_box("Message", icon = "cogs")
s <- format_text_r(get_construct_string(x),
                   width.cutoff = 15L, quoted = TRUE)
cat(s)


}
\seealso{
\code{\link{get_construct_string}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info-box.R
\name{flip_box}
\alias{flip_box}
\alias{flip}
\title{An 'HTML' container that can flip}
\usage{
flip_box(
  front,
  back,
  active_on = c("click", "click-front", "manual"),
  inputId = NULL,
  class = NULL
)

flip(inputId, session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{front}{'HTML' elements to show in the front}

\item{back}{'HTML' elements to show when the box is flipped}

\item{active_on}{the condition when a box should be flipped; choices are
\code{'click'}: flip when double-click on both sides; \code{'click-front'}:
only flip when the front face is double-clicked; \code{'manual'}: manually
flip in \code{R} code (see \code{{flip(inputId)}} function)}

\item{inputId}{element 'HTML' id; must be specified if \code{active_on} is
not \code{'click'}}

\item{class}{'HTML' class}

\item{session}{shiny session; default is current active domain}
}
\value{
\code{flip_box} returns 'HTML' tags; \code{flip} should be called
from shiny session, and returns nothing
}
\description{
An 'HTML' container that can flip
}
\examples{

# More examples are available in demo

library(shiny)
library(shidashi)

session <- MockShinySession$new()

flip_box(front = info_box("Side A"),
         back = info_box("Side B"),
         inputId = 'flip_box1')

flip('flip_box1', session = session)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream.R
\name{read_stream_vis}
\alias{read_stream_vis}
\title{Read a shidashi stream binary file}
\usage{
read_stream_vis(path)
}
\arguments{
\item{path}{Character scalar. Absolute path to a \code{.bin} file
produced by \code{\link{stream_to_js}}.}
}
\value{
A list with components:
  \describe{
    \item{\code{header}}{Named list parsed from the JSON header
      (contains \code{data_type}, \code{signature}, \code{timestamp},
      and any extra fields).}
    \item{\code{data}}{Decoded body: a \code{raw} vector for
      \code{"raw"}, an R object for \code{"json"}, or a numeric/integer
      vector for \code{"int32"}, \code{"float32"}, \code{"float64"}.}
  }
}
\description{
Reads the binary envelope written by \code{\link{stream_to_js}} and
returns the header and decoded body as a list.
}
\seealso{
\code{\link{stream_to_js}}, \code{\link{stream_path}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/widgets.R
\name{add-remove-html-class}
\alias{add-remove-html-class}
\alias{add_class}
\alias{remove_class}
\title{Add or remove 'HTML' class from 'RAVE' application}
\usage{
add_class(selector, class, session = shiny::getDefaultReactiveDomain())

remove_class(selector, class, session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{selector}{'CSS' selector}

\item{class}{class to add or to remove from selected elements}

\item{session}{shiny session}
}
\value{
No value is returned
}
\description{
Only works in template framework provided by 'shidashi' package,
see \code{\link[shidashi]{use_template}}
}
\examples{

server <- function(input, output, session) {

  # Add class `hidden` to element with ID `elemid`
  add_class("#elemid", "hidden")

  # Remove class `hidden` from element with class `shiny-input-optional`
  remove_class(".shiny-input-optional", "hidden")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/progress.R
\name{shiny_progress}
\alias{shiny_progress}
\title{Wrapper of shiny progress that can run without shiny}
\usage{
shiny_progress(
  title,
  max = 1,
  ...,
  quiet = FALSE,
  session = shiny::getDefaultReactiveDomain(),
  shiny_auto_close = FALSE,
  log = NULL,
  outputId = NULL
)
}
\arguments{
\item{title}{the title of the progress}

\item{max}{max steps of the procedure}

\item{...}{passed to initialization method of \code{\link[shiny]{Progress}}}

\item{quiet}{whether the progress needs to be quiet}

\item{session}{shiny session domain}

\item{shiny_auto_close}{whether to close the progress once function exits}

\item{log}{alternative log function}

\item{outputId}{the element id of \code{\link{progressOutput}}, or
\code{NULL} to use the default shiny progress}
}
\value{
a list of functions that controls the progress
}
\description{
Wrapper of shiny progress that can run without shiny
}
\examples{

{
  progress <- shiny_progress("Procedure A", max = 10)
  for(i in 1:10) {
    progress$inc(sprintf("Step \%s", i))
    Sys.sleep(0.1)
  }
  progress$close()

}

if(interactive()) {
  library(shiny)

  ui <- fluidPage(
    fluidRow(
      column(12, actionButton("click", "Click me"))
    )
  )

  server <- function(input, output, session) {
    observeEvent(input$click, {
      progress <- shiny_progress("Procedure B", max = 10,
                                 shiny_auto_close = TRUE)
      for(i in 1:10) {
        progress$inc(sprintf("Step \%s", i))
        Sys.sleep(0.1)
      }
    })
  }

  shinyApp(ui, server)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream-viz.R
\name{updateStreamViz}
\alias{updateStreamViz}
\title{Trigger an in-place update of a streaming widget}
\usage{
updateStreamViz(
  outputId,
  streaming = NULL,
  token = NULL,
  session = shiny::getDefaultReactiveDomain()
)
}
\arguments{
\item{outputId}{Character scalar. The output ID passed to
\code{\link{streamVizOutput}} (and to \code{\link{stream_path}}).}

\item{streaming}{Logical or \code{NULL}.  \code{TRUE} starts active
streaming (the widget polls for new data via \code{requestAnimationFrame});
\code{FALSE} pauses it; \code{NULL} (default) leaves the current
streaming state unchanged.}

\item{token}{Character scalar or \code{NULL}.  When \code{NULL} (default)
the session's own token is used.  Override with a parent session token
when running inside a standalone viewer (pop-out window) so that the
child session fetches the parent's binary stream file.}

\item{session}{Shiny session object. Defaults to the active reactive domain.}
}
\value{
Invisibly \code{NULL}.
}
\description{
Sends a custom Shiny message that causes the browser to re-fetch the binary
stream file and re-render the signal viewer without tearing down and
recreating the widget.  Call this after writing new data with
\code{\link{stream_to_js}}.  The \code{outputId} is used both
to locate the widget as the stream file identifier — it must
match the \code{id} passed to \code{\link{stream_path}}.
}
\seealso{
\code{\link{streamVizOutput}}, \code{\link{stream_to_js}},
  \code{\link{stream_file_id}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chatbot.R
\name{chatbot_server}
\alias{chatbot_server}
\title{Chat-bot server logic (per-module)}
\usage{
chatbot_server(
  input,
  output,
  session,
  id = "shidashi-chatbot",
  drawer_id = "shidashi_drawer",
  agent_conf = NULL
)
}
\arguments{
\item{input, output, session}{Standard Shiny server arguments
(typically module-scoped when inside \code{moduleServer}).}

\item{id}{character; must match the \code{id} used in
\code{chatbot_ui()}. Default \code{"shidashi-chatbot"}.}

\item{drawer_id}{character; the output ID of the drawer's
\code{uiOutput} placeholder (from \code{\link{module_drawer}()}).
Default \code{"shidashi_drawer"}.}

\item{agent_conf}{list; parsed \code{agents.yaml} content. Used
for the system prompt and tool configuration.}
}
\value{
Called for side effects (sets up observers). Returns
  \code{invisible(NULL)}.
}
\description{
Sets up the chat server for a single module.
Creates its own \code{\link[ellmer]{Chat}} object, binds tools from
the \verb{MCP} session registry, and manages per-module conversation
history.

The function renders \code{\link{chatbot_ui}()} into the drawer's
\code{uiOutput} via \code{shiny::renderUI}, and opens the drawer
when a \code{"button.click"} shidashi-event with
\code{type = "open_drawer"} is received.
Any element with \code{data-shidashi-action="shidashi-button"}
\code{data-shidashi-type="open_drawer"} can trigger the drawer.

This function is injected into each module's server function by
\code{modules.R} when \code{agents.yaml} has \code{enabled: yes}.
It is called inside \code{shiny::moduleServer()}, so \code{session}
is module-scoped.  \pkg{shinychat} operations use the scoped
\code{session}; only drawer and event operations use
\code{session$rootScope()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream-viz.R
\name{stream_viz}
\alias{stream_viz}
\title{Create a streaming widget}
\usage{
stream_viz(
  outputId,
  session = shiny::getDefaultReactiveDomain(),
  stream_id = NULL,
  width = NULL,
  height = NULL,
  refresh_rate = 33,
  show_controls = TRUE,
  streaming = FALSE,
  token = NULL
)
}
\arguments{
\item{outputId}{Character scalar.  The output ID matching the
corresponding \code{\link{streamVizOutput}} call.  Used to compute
\code{stream_id} automatically when \code{stream_id} is \code{NULL}.}

\item{session}{Shiny session object.  Defaults to the active reactive
domain.  Together with \code{outputId} determines the token-qualified
stream file identifier.}

\item{stream_id}{Character scalar or \code{NULL}.  When \code{NULL}
(default) the stream identifier is computed as
\code{stream_file_id(outputId, session, token = token)}.  Provide an
explicit value to override for non-Shiny or advanced use.}

\item{width, height}{Widget dimensions (passed to \code{\link[htmlwidgets]{createWidget}}).}

\item{refresh_rate}{Numeric scalar (milliseconds). Polling interval used
when the viewer is in \emph{active streaming} mode (play button pressed).
Lower values give smoother animation but increase CPU/network load.
Default \code{33} (approximately 30 Hz).}

\item{show_controls}{Logical. Whether to display the hover toolbar
(play/pause, zoom, reset, export).  Default \code{TRUE}.}

\item{streaming}{Logical.  Whether to start active streaming immediately
when the widget is rendered.  Default \code{FALSE}; set to \code{TRUE}
to begin polling for new data automatically.  Can also be toggled later
via \code{\link{updateStreamViz}}.}

\item{token}{Character scalar or \code{NULL}.  When \code{NULL} (default)
the session's own token is used.  Override with a parent session token
when running inside a standalone viewer (pop-out window) so that the
child session fetches the parent's binary stream file.}
}
\value{
An \pkg{htmlwidgets} object.
}
\description{
Creates a multi-channel signal viewer widget that fetches its data from
the binary stream file produced by \code{\link{stream_to_js}}.
}
\details{
The \code{outputId} together with \code{session} is used to auto-compute
the token-qualified stream file identifier via
\code{\link{stream_file_id}}.  You can override this by passing an
explicit \code{stream_id}.

Use \code{\link{updateStreamViz}} for in-place updates without recreating
the widget.
}
\seealso{
\code{\link{streamVizOutput}}, \code{\link{renderStreamViz}},
  \code{\link{updateStreamViz}}, \code{\link{stream_file_id}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\name{fire_event}
\alias{fire_event}
\alias{get_event}
\alias{get_theme}
\title{Fire or read a session event}
\usage{
fire_event(
  key,
  value,
  session = shiny::getDefaultReactiveDomain(),
  global = FALSE
)

get_event(key, session = shiny::getDefaultReactiveDomain(), default = NULL)

get_theme(session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{key}{a single character string identifying the event type}

\item{value}{the event payload (any R object)}

\item{session}{a Shiny session (defaults to the current reactive domain)}

\item{global}{logical; if \code{TRUE}, the event is broadcast to all
sessions sharing the same \code{shared_id}}

\item{default}{value to return when the event has not been fired yet
(used by \code{get_event} only)}
}
\value{
\code{fire_event} returns \code{NULL} invisibly.
  \code{get_event} returns the last value fired for \code{key}, or
  \code{default} if none.
  \code{get_theme} returns a named list with three character elements:
  \describe{
    \item{theme}{Either \code{"light"} or \code{"dark"}.}
    \item{foreground}{Hex color string for text / foreground elements.}
    \item{background}{Hex color string for the page background.}
  }
  Before the browser fires its first theme event, the light-theme fallback
  \code{list(theme = "light", background = "#FFFFFF", foreground = "#000000")}
  is returned.
}
\description{
\code{fire_event} sets a reactive event value on the current session.
When \code{global = TRUE}, the event is propagated to all sessions that
share the same \code{shared_id} (i.e. other browser tabs for the same
user).

\code{get_event} reads the current value of an event key from the
session registry.

\code{get_theme} is a convenience wrapper around
\code{get_event("theme.changed")} that returns the current dashboard theme.

All three functions must be called inside a Shiny server context.
}
\details{
\code{get_theme} and \code{get_event} auto-register the session if needed
and must be called inside a reactive context
(\code{\link[shiny]{observe}}, \code{\link[shiny]{observeEvent}},
\code{\link[shiny]{reactive}}, render functions).
}
\examples{


library(shiny)
server <- function(input, output, session) {
  # fire an event
  shidashi::fire_event("my_event", list(a = 1), session = session)

  # read it back
  observe({
    evt <- shidashi::get_event("my_event", session = session)
    print(evt)
  })

  # get_theme must be called within a reactive context
  output$plot <- renderPlot({
    theme <- shidashi::get_theme()
    par(bg = theme$background, fg = theme$foreground)
    plot(1:10)
  })
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clipboard.R
\name{clipboardOutput}
\alias{clipboardOutput}
\alias{renderClipboard}
\title{Generates outputs that can be written to clipboards with one click}
\usage{
clipboardOutput(
  outputId = rand_string(prefix = "clipboard"),
  message = "Copy to clipboard",
  clip_text = "",
  class = NULL,
  as_card_tool = FALSE
)

renderClipboard(
  expr,
  env = parent.frame(),
  quoted = FALSE,
  outputArgs = list()
)
}
\arguments{
\item{outputId}{the output id}

\item{message}{tool tip to show when mouse hovers on the element}

\item{clip_text}{the initial text to copy to clipboards}

\item{class}{'HTML' class of the element}

\item{as_card_tool}{whether to make the output as \code{\link{card_tool}}}

\item{expr}{expression to evaluate; the results will replace
\code{clip_text}}

\item{env}{environment to evaluate \code{expr}}

\item{quoted}{whether \code{expr} is quoted}

\item{outputArgs}{used to replace default arguments of \code{clipboardOutput}}
}
\value{
'HTML' elements that can write to clip-board once users click on
them.
}
\description{
Generates outputs that can be written to clipboards with one click
}
\examples{
clipboardOutput(clip_text = "Hey there")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{guess_body_class}
\alias{guess_body_class}
\title{Guess the 'AdminLTE' body class for modules, used internally}
\usage{
guess_body_class(cls)
}
\arguments{
\item{cls}{the class string of the \code{<body>} tag in \code{'index.html'}}
}
\value{
The proposed class for \code{<body>} tag
}
\description{
Guess the 'AdminLTE' body class for modules, used internally
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\name{register_session}
\alias{register_session}
\alias{unregister_session}
\alias{enable_input_broadcast}
\alias{disable_input_broadcast}
\alias{enable_input_sync}
\alias{disable_input_sync}
\alias{get_handler}
\alias{set_handler}
\title{Shiny session registration and cross-tab synchronization}
\usage{
register_session(session)

unregister_session(session)

enable_input_broadcast(
  session = shiny::getDefaultReactiveDomain(),
  once = FALSE
)

disable_input_broadcast(session = shiny::getDefaultReactiveDomain())

enable_input_sync(session = shiny::getDefaultReactiveDomain(), once = FALSE)

disable_input_sync(session = shiny::getDefaultReactiveDomain())

get_handler(name, session = shiny::getDefaultReactiveDomain())

set_handler(name, handler, session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{session}{A Shiny session object or session proxy (created by
\code{\link[shiny]{moduleServer}}). Most functions in this family default
to \code{shiny::getDefaultReactiveDomain()} so callers inside module
server functions do not need to pass it explicitly.
\code{register_session} and \code{unregister_session} require an explicit
value.}

\item{once}{Logical; if \code{TRUE}, resume the suspended observer for a
single run via \code{$run()} instead of permanently re-activating it with
\code{$resume()}.  Useful for one-shot snapshots without installing a
persistent observer.}

\item{name}{A single character string identifying a named slot in the
session's handler list (used by \code{get_handler} and
\code{set_handler}).  Three names are reserved for internal use and will
trigger an error if passed to \code{set_handler}:
\code{"event_handler"}, \code{"broadcast_handler"}, and
\code{"input_sync_handler"}.}

\item{handler}{A Shiny Observer object created by \code{shiny::observe()},
or \code{NULL} to clear the named slot (used by \code{set_handler}).
Passing any other object type raises an error.}
}
\value{
\code{register_session} returns the session token (\code{session$token})
invisibly; it is idempotent and safe to call multiple times for the same
session.

\code{unregister_session} returns \code{NULL} invisibly; it is idempotent.

\code{enable_input_broadcast}, \code{disable_input_broadcast},
\code{enable_input_sync}, and \code{disable_input_sync} all return
invisibly.  They are silent no-ops when the session is already closed.

\code{get_handler} returns the named \code{Observer} object stored under
\code{name}, or \code{NULL} if the slot is empty or the session is closed.

\code{set_handler} returns \code{TRUE} invisibly when it successfully
installs the handler, or \code{FALSE} invisibly when the session is already
closed.
}
\description{
Shiny session registration and cross-tab synchronization
}
\details{
\subsection{Session registration}{

\strong{\code{register_session()}} — Call once at the top of every Shiny
module server function.  It is idempotent: calling it a second time on the
same session safely refreshes the session object and URL in the registry
entry without re-creating observers.

Internally it creates an entry in the application-global session registry
(initialized by \code{\link{init_app}}), resolves a \code{shared_id} token
shared across browser tabs from the \code{?shared_id=...} URL query string
(or generates a random 26-character string when absent), sets up the
per-session reactive event bus, and — for named module sessions — sends a
\code{shidashi.register_module_token} custom message to bind the module
namespace to its session token on the JavaScript side.

\strong{\code{unregister_session()}} — Removes the session entry from the
registry and destroys all attached observers.  This is called automatically
when the session ends via the \code{onSessionEnded} hook installed by
\code{register_session()}.  Direct calls are only needed for explicit early
cleanup (e.g. in tests).

}
\subsection{Session-scoped handlers}{

Each registered session maintains a named slot list for Shiny
\code{Observer} objects called \emph{handlers}.  This provides a lightweight
system for attaching module-level life-cycle hooks that are tied to the
session's lifetime.

\strong{User-defined handlers — \code{get_handler()} / \code{set_handler()}}

\code{set_handler(name, handler)} installs \code{handler} under \code{name},
first suspending and destroying any \code{Observer} already stored there.
Pass \code{handler = NULL} to clear the slot.  Returns \code{FALSE}
invisibly if the session is already closed.

\code{get_handler(name)} retrieves the stored \code{Observer} (or
\code{NULL}).  It auto-registers the session if not yet registered and
returns \code{NULL} gracefully if the session is closed.

Three handler names are reserved for internal shidashi infrastructure and
will raise an error if passed to \code{set_handler}:
\code{"event_handler"}, \code{"broadcast_handler"}, and
\code{"input_sync_handler"}.

\strong{Built-in cross-tab sync handlers}

shidashi installs two opt-in \code{Observer} slots in every registered
session (both start \emph{suspended}):

\describe{
  \item{Input broadcast (\code{enable_input_broadcast()} /
        \code{disable_input_broadcast()})}{Continuously monitors the
        session's \code{input} values and, whenever they change, pushes a
        snapshot to the client via \code{shidashi.cache_session_input}.
        Other browser tabs sharing the same \code{shared_id} can read this
        cached snapshot to restore or compare input state.}
  \item{Input sync (\code{enable_input_sync()} /
        \code{disable_input_sync()})}{Listens for serialized input maps
        broadcast by \emph{other} sessions sharing the same
        \code{shared_id} via the root-session \code{@shidashi@} input.
        Values differing from the local \code{input} are written into the
        session's private \code{inputs} \code{reactiveValues}.  Messages
        from the same session are ignored to prevent feedback loops.}
}

Both observers run at priority \code{-100000} (after all ordinary reactive
computations have settled).  Use \code{once = TRUE} to trigger a single
cycle without permanently resuming the observer.  The \code{disable_*}
variants suspend the observer and are silent no-ops when the session has
already ended.

}
\subsection{Session life-cycle}{

\preformatted{
init_app()                       # global.R, once per app start
    |
    v
register_session(session)        # top of each module server()
    |
    v
... reactive code ...
    get_handler() / set_handler() # attach user-defined session observers
    enable_input_broadcast()      # optional: push inputs to browser cache
    enable_input_sync()           # optional: receive peer-tab inputs
    |
    v
session ends -> unregister_session()  # runs automatically
}

}
}
\examples{

library(shiny)

# --- Basic usage in a module server ---
server <- function(input, output, session) {
  shidashi::register_session(session)

  # opt-in to cross-tab input broadcast (suspended by default)
  shidashi::enable_input_broadcast(session)

  # opt-in to receive inputs from peer tabs
  shidashi::enable_input_sync(session)

  # get_theme must be called within a reactive context
  output$plot <- renderPlot({
    theme <- shidashi::get_theme()
    par(bg = theme$background, fg = theme$foreground)
    plot(1:10)
  })
}

# --- Named handler: attach a reusable session-scoped observer ---
server2 <- function(input, output, session) {
  shidashi::register_session(session)

  cleanup <- shiny::observe({
    # ... module-level teardown logic ...
    shidashi::set_handler("my_cleanup", NULL, session)
  }, suspended = TRUE, domain = session)

  shidashi::set_handler("my_cleanup", cleanup, session)

  # retrieve and resume the observer elsewhere in the same session
  h <- shidashi::get_handler("my_cleanup", session)
  if (!is.null(h)) h$resume()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream-viz.R
\name{streamVizOutput}
\alias{streamVizOutput}
\title{Output placeholder for a streaming visualization widget}
\usage{
streamVizOutput(outputId, width = "100\%", height = "400px")
}
\arguments{
\item{outputId}{Character scalar. Output ID that matches the corresponding
\code{\link{renderStreamViz}} call.}

\item{width, height}{CSS width and height of the widget container.}
}
\value{
An HTML output element suitable for inclusion in a Shiny UI.
}
\description{
Use in a Shiny UI to reserve a slot for a multi-channel signal viewer
that is driven by binary stream data.
}
\seealso{
\code{\link{renderStreamViz}}, \code{\link{updateStreamViz}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card-tabset.R
\name{card_tabset}
\alias{card_tabset}
\title{Generates a set of card panels}
\usage{
card_tabset(
  ...,
  inputId = rand_string(prefix = "tabset-"),
  title = NULL,
  names = NULL,
  active = NULL,
  tools = NULL,
  footer = NULL,
  class = "",
  class_header = "",
  class_body = "",
  class_foot = ""
)
}
\arguments{
\item{...}{'HTML' tags; each tag will be placed into a card}

\item{inputId}{the id of the card-set, must start with letters}

\item{title}{the title of the card-set}

\item{names}{title of the tabs}

\item{active}{the title that will be active on load}

\item{tools}{a list of tools or badges generated by
\code{\link{card_tool}} or \code{\link{as_badge}}}

\item{footer}{the footer element of the card-set}

\item{class}{the 'HTML' class the of card-set}

\item{class_header, class_body, class_foot}{additional 'HTML'
class the of card header, body, and footer accordingly}
}
\value{
'HTML' tags
}
\description{
To insert, remove, or active card panels, see
\code{\link{card_tabset_operate}}.
}
\examples{

library(shiny)
library(shidashi)

# Fake session to operate on card_tabset without shiny
session <- MockShinySession$new()

card_tabset(
  inputId = "card_set",
  title = "Cardset with Tools",
  `Tab 1` = p("Tab content 1"),
  class_body = "height-500",
  tools = list(
    as_badge(
      "New|badge-success"
    ),
    card_tool(
      widget = "collapse"
    ),
    card_tool(
      widget = "maximize"
    )
  )
)

card_tabset_insert(
  inputId = "card_set",
  title = "Tab 2",
  p("New content"),
  session = session
)

card_tabset_activate(
  inputId = "card_set",
  title = "Tab 1",
  session = session
)

card_tabset_remove(
  inputId = "card_set",
  title = "Tab 2",
  session = session
)



}
\seealso{
\code{\link{card_tabset_operate}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/widgets.R
\name{flex_container}
\alias{flex_container}
\alias{flex_item}
\alias{flex_break}
\title{Generate 'HTML' tags with 'flex' layout}
\usage{
flex_container(
  ...,
  style = NULL,
  direction = c("row", "column"),
  wrap = c("wrap", "nowrap", "wrap-reverse"),
  justify = c("flex-start", "center", "flex-end", "space-around", "space-between"),
  align_box = c("stretch", "flex-start", "center", "flex-end", "baseline"),
  align_content = c("stretch", "flex-start", "flex-end", "space-between", "space-around",
    "center")
)

flex_item(
  ...,
  size = 1,
  style = NULL,
  order = NULL,
  flex = as.character(size),
  align = c("flex-start", "flex-end", "center"),
  class = NULL,
  .class = "fill-width padding-5"
)

flex_break(..., class = NULL)
}
\arguments{
\item{...}{for \code{flex_container}, it's elements of \code{flex_item};
for \code{flex_item}, \code{...} are shiny 'HTML' tags}

\item{style}{the additional 'CSS' style for containers or inner items}

\item{direction, wrap, justify, align_box, align_content}{'CSS' styles for
'flex' containers}

\item{size}{numerical relative size of the item; will be ignored if
\code{flex} is provided}

\item{order, align, flex}{CSS' styles for 'flex' items}

\item{class, .class}{class to add to the elements}
}
\value{
'HTML' tags
}
\description{
Generate 'HTML' tags with 'flex' layout
}
\examples{

x <- flex_container(
  style = "position:absolute;height:100vh;top:0;left:0;width:100\%",
  flex_item(style = 'background-color:black;'),
  flex_item(style = 'background-color:red;')
)
# You can view it via `htmltools::html_print(x)`

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drawer.R
\name{drawer}
\alias{drawer}
\alias{drawer_open}
\alias{drawer_close}
\alias{drawer_toggle}
\title{Open, close, or toggle the drawer panel}
\usage{
drawer_open(session = shiny::getDefaultReactiveDomain())

drawer_close(session = shiny::getDefaultReactiveDomain())

drawer_toggle(session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{session}{shiny session}
}
\value{
No value is returned (called for side effect).
}
\description{
Send messages to the client to open, close, or toggle the
off-canvas drawer panel on the right side of the dashboard.
}
\examples{
server <- function(input, output, session) {
  # Open the drawer
  drawer_open()

  # Close the drawer
  drawer_close()

  # Toggle the drawer
  drawer_toggle()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accordion.R
\name{accordion_item}
\alias{accordion_item}
\title{'Accordion' items}
\usage{
accordion_item(
  title,
  ...,
  footer = NULL,
  class = "",
  collapsed = TRUE,
  parentId = rand_string(prefix = "accordion-"),
  itemId = rand_string(prefix = "accordion-item-"),
  style_header = NULL,
  style_body = NULL,
  root_path = template_root()
)
}
\arguments{
\item{title}{character title to show in the header}

\item{...}{body content}

\item{footer}{footer element, hidden if \code{NULL}}

\item{class}{the class of the item}

\item{collapsed}{whether collapsed at the beginning}

\item{parentId}{parent \code{\link{accordion}} id}

\item{itemId}{the item id}

\item{style_header, style_body}{'CSS' style of item header and body}

\item{root_path}{see \code{template_root}}
}
\value{
\code{'shiny.tag.list'} 'HTML' tags
}
\description{
'Accordion' items
}
\seealso{
\code{\link{accordion}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card-tabset.R
\name{card_tabset_operate}
\alias{card_tabset_operate}
\alias{card_tabset_insert}
\alias{card_tabset_remove}
\alias{card_tabset_activate}
\title{Add, active, or remove a card within \code{\link{card_tabset}}}
\usage{
card_tabset_insert(
  inputId,
  title,
  ...,
  active = TRUE,
  notify_on_failure = TRUE,
  session = shiny::getDefaultReactiveDomain()
)

card_tabset_remove(
  inputId,
  title,
  notify_on_failure = TRUE,
  session = shiny::getDefaultReactiveDomain()
)

card_tabset_activate(
  inputId,
  title,
  notify_on_failure = TRUE,
  session = shiny::getDefaultReactiveDomain()
)
}
\arguments{
\item{inputId}{the element id of \code{\link{card_tabset}}}

\item{title}{the title of the card to insert, activate, or to remove}

\item{...}{the content of the card}

\item{active}{whether to set the card to be active once added}

\item{notify_on_failure}{whether to show notifications on failure}

\item{session}{shiny session domain}
}
\value{
These functions execute \code{session$sendCustomMessage} and return
whatever value generated by that function; usually nothing.
}
\description{
Add, active, or remove a card within \code{\link{card_tabset}}
}
\seealso{
\code{\link{card_tabset}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skill-wrapper.R
\name{skill_wrapper}
\alias{skill_wrapper}
\title{Wrap a Skill Directory as an \verb{MCP} Tool Generator}
\usage{
skill_wrapper(skill_path)
}
\arguments{
\item{skill_path}{Path to the skill directory containing \code{SKILL.md}.
Can be absolute or relative to the project root.}
}
\value{
A function with class \code{c("shidashi_skill_wrapper", "function")}
  that returns an \code{ellmer::ToolDef} object.
}
\description{
Creates a closure that produces an
\code{ellmer::tool} dispatching on an \code{action} enumerator:
\describe{
  \item{\code{readme}}{Returns the full \verb{SKILL.md} instructions. Must be
    called first to unlock other actions.}
  \item{\code{reference}}{Returns content from a reference file in the
    skill directory (gated behind \code{readme}).}
  \item{\code{script}}{Executes a script in the \code{scripts/}
    subdirectory via \code{processx::run()} (gated behind \code{readme}).}
}
}
\details{
The returned tool enforces a soft gate: calling \code{reference} or
  \code{script} before \code{readme} is allowed, but if the call
  errors the message is augmented with a condensed summary (~200
  tokens) instructing the AI to read the full instructions first.
  This minimizes token waste (the summary is only sent on failure).

  The gate state is per-instance: each call to the wrapper produces
  a closure with an independent \code{readme_unlocked} flag.
}
\examples{
skill_dir <- system.file(
  "builtin-templates/bslib-bare/agents/skills/greet",
  package = "shidashi"
)
wrapper  <- skill_wrapper(skill_dir)
tool_def <- wrapper()
cat(tool_def(action = "readme"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{reset_output}
\alias{reset_output}
\title{Reset shiny outputs with messages}
\usage{
reset_output(
  outputId,
  message = "This output has been reset",
  session = shiny::getDefaultReactiveDomain()
)
}
\arguments{
\item{outputId}{output ID}

\item{message}{output message}

\item{session}{shiny reactive domain}
}
\value{
No value
}
\description{
Forces outdated output to reset and show a silent message.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream-viz.R
\name{renderStreamViz}
\alias{renderStreamViz}
\title{Render a streaming widget}
\usage{
renderStreamViz(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{expr}{An R expression that returns a \code{\link{stream_viz}} widget.}

\item{env}{Environment in which to evaluate \code{expr}.}

\item{quoted}{Logical. Whether \code{expr} is already quoted.}
}
\value{
A server-side render function for use with Shiny.
}
\description{
Server-side render function for \code{\link{streamVizOutput}}.
The expression should evaluate to a \code{\link{stream_viz}} object.
}
\seealso{
\code{\link{streamVizOutput}}, \code{\link{updateStreamViz}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/menu-item.R
\name{as_icon}
\alias{as_icon}
\title{Convert characters, shiny icons into 'fontawesome' 4}
\usage{
as_icon(icon = NULL, class = "fas")
}
\arguments{
\item{icon}{character or \code{\link[shiny]{icon}}}

\item{class}{icon class; change this when you are using 'fontawesome'
professional version. The choices are \code{'fa'} (compatible),
\code{'fas'} (strong), \code{'far'} (regular), \code{'fal'} (light),
and \code{'fad'} (duo-tone).}
}
\value{
'HTML' tag
}
\description{
Convert characters, shiny icons into 'fontawesome' 4
}
\examples{

if(interactive()) {
as_icon("bookmark", class = "far")
as_icon("bookmark", class = "fas")

# no icon
as_icon(NULL)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card-badge.R
\name{card_badge}
\alias{card_badge}
\alias{card_recalculate_badge}
\alias{enable_recalculate_badge}
\alias{disable_recalculate_badge}
\alias{set_card_badge}
\title{Create a badge widget located at card header}
\usage{
card_badge(text = NULL, class = NULL, ...)

card_recalculate_badge(
  text = "Recalculate needed",
  class = NULL,
  event_name = "run_analysis",
  ...
)

enable_recalculate_badge(text = "Recalculate needed", ...)

disable_recalculate_badge(text = "Up-to-date", ...)

set_card_badge(
  id = NULL,
  class = NULL,
  text = NULL,
  add_class = NULL,
  remove_class = NULL,
  session = shiny::getDefaultReactiveDomain()
)
}
\arguments{
\item{text}{inner text content of the badge}

\item{class}{additional 'HTML' class of the badge; for
\code{set_card_badge}, this is the class selector of the badge that is
to be changed}

\item{...}{additional 'HTML' tag attributes}

\item{event_name}{name of the event to trigger}

\item{id}{the badge 'HTML' ID to be changed, will be enclosed with
session namespace \code{session$ns(id)} automatically.}

\item{add_class, remove_class}{add or remove class}

\item{session}{shiny session}
}
\description{
Create a badge widget located at card header
}
\examples{

library(shidashi)

# UI: a Bootstrap badge with green background
card_badge("Ready", class = "bg-green shidashi-output-status")

# server
server <- function(input, output, session) {

  shiny::observe({

    # ... check if the inputs have changed

    set_card_badge(
      class = "shidashi-output-status",
      text = "Refresh needed",
      add_class = "bg-warning",
      remove_class = "bg-success disabled"
    )

  })

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{html_asis}
\alias{html_asis}
\title{Escape HTML strings}
\usage{
html_asis(s, space = TRUE)
}
\arguments{
\item{s}{characters}

\item{space}{whether to also escape white space, default is true.}
}
\value{
An R string
}
\description{
Escape HTML strings so that they will be displayed
'as-is' in websites.
}
\examples{

html_asis("<a><----> <b>")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chatbot.R
\name{chatbot_ui}
\alias{chatbot_ui}
\title{Chat-bot UI panel}
\usage{
chatbot_ui(id, modes = NULL, default_mode = NULL)
}
\arguments{
\item{id}{character; the Shiny input/output namespace for the chat
widget. Default \code{"shidashi-chatbot"}.}
}
\value{
A \code{shiny::tagList} containing the chat UI or empty.
}
\description{
Returns the UI elements for the AI chat panel: a header bar with
a \dQuote{New conversation} button and the \pkg{shinychat} widget.

When \pkg{shinychat} is not installed or the chat-bot is disabled via
\code{options(shidashi.chatbot = FALSE)}, returns an empty
\code{tagList()}.

Typically called inside \code{shiny::renderUI} by
\code{\link{chatbot_server}()} to fill a \code{\link{module_drawer}}.
Can also be placed anywhere in module UI directly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mcp-wrapper.R
\name{register_io}
\alias{register_io}
\alias{register_input}
\alias{register_output}
\title{Register Shiny Inputs and Outputs for \verb{MCP} Access}
\usage{
register_input(
  expr,
  inputId,
  update,
  description = "",
  writable = TRUE,
  quoted = FALSE,
  env = parent.frame()
)

register_output(
  expr,
  outputId,
  description = "",
  quoted = FALSE,
  env = parent.frame(),
  ...,
  output_opts = list(),
  download_function = NULL,
  download_type = c("image", "htmlwidget", "threeBrain", "no-download", "data",
    "stream_viz"),
  extension = NULL,
  session = shiny::getDefaultReactiveDomain()
)
}
\arguments{
\item{expr}{For \code{register_input}: a call expression that creates
a \code{shiny} input widget.
For \code{register_output}: a render function call such as
\code{renderPlot(\{...\})}.}

\item{inputId}{character string.  The \code{shiny} input ID
(without the module namespace prefix).}

\item{update}{character string.  The fully qualified update function,
e.g. \code{"shiny::updateTextInput"}.  Field mappings such as
\code{"shiny::updateSelectInput(value=selected)"} override the
default argument names passed to the update function.}

\item{description}{character string.  A human-readable description of
the input or output purpose, exposed to \verb{LLM} agents via \verb{MCP} tools.}

\item{writable}{logical (default \code{TRUE}).  Whether the \verb{MCP}
update tool is allowed to change this input.}

\item{quoted}{logical (default \code{FALSE}).  If \code{TRUE},
\code{expr} is treated as already quoted; otherwise it is captured
with \code{substitute()}.}

\item{env}{the environment in which to evaluate \code{expr}.}

\item{outputId}{character string.  The \code{shiny} output ID
(without the module namespace prefix).}

\item{...}{reserved for future use.}

\item{output_opts}{a named list of extra options for the output
(e.g. width, height defaults).}

\item{download_function}{a custom download handler function.  When
\code{download_type = "data"}, this function receives the file path
and writes the download content.}

\item{download_type}{character string.  One of \code{"image"},
\code{"threeBrain"}, \code{"data"}, or \code{"no-download"}.}

\item{extension}{character vector of allowed file extension for
download, or \code{NULL}.}

\item{session}{the \code{shiny} session object.  For
\code{register_output}, defaults to
\code{shiny::getDefaultReactiveDomain()}.}
}
\value{
\code{register_input} returns the evaluated UI element.
  \code{register_output} is called for its side effects (assigning
  the render function and registering widgets) and returns \code{NULL}
  invisibly.
}
\description{
Register \code{shiny} inputs and outputs for \verb{MCP} (Model Context
Protocol) agent access.

\code{register_input()} wraps a \code{shiny} input constructor to
register metadata.  It evaluates \code{expr} and returns the UI element.

\code{register_output()} is a server-side function that registers a
render function call (e.g. \code{renderPlot(\{...\})}), assigns it to
\code{session$output}, registers the \verb{MCP} output spec, and sets
up download-widget handlers.  The UI overlay icons are injected
entirely by \verb{JS}.
}
\examples{
\dontrun{
# inside a shidashi module UI function:
ns <- shiny::NS("demo")

register_input(
  expr = shiny::sliderInput(
    inputId = ns("threshold"),
    label = "Threshold",
    min = 0, max = 1, value = 0.5
  ),
  inputId = "threshold",
  update = "shiny::updateSliderInput",
  description = "Filter threshold for the plot"
)

# inside a shidashi module server function:
register_output(
  expr = renderPlot(\{ plot(iris) \}),
  outputId = "my_plot",
  description = "Scatter plot of iris data",
  download_type = "image"
)
}
}
\seealso{
\code{\link{init_app}}, \code{\link{mcp_wrapper}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mcp-wrapper.R
\name{mcp_wrapper}
\alias{mcp_wrapper}
\title{Wrap an \verb{MCP} Tool Generator Function}
\usage{
mcp_wrapper(generator)
}
\arguments{
\item{generator}{A function that accepts a `session` parameter and returns
either a single tool object or a list/vector of such objects; see
\code{\link[ellmer]{tool}}.}
}
\value{
A wrapped function with class \code{'shidashi_mcp_wrapper'} that:
  - Accepts a `session` parameter
  - Calls the generator function with the session
  - Normalizes the output to a list
  - Filters to keep only valid tool objects
  - Returns a list of tool objects (possibly empty)
}
\description{
Creates a wrapper around a generator function to ensure it returns valid
Model Context Protocol (\verb{MCP}) tool definitions. The wrapper validates
input and filters output to contain only \code{'ellmer::ToolDef'} objects.
}
\details{
The wrapper performs the following validations:
  - Ensures `generator` is a function
  - Checks that `generator` accepts a `session` parameter

  The returned function automatically handles both single tool definitions
  and lists of tools, providing a consistent interface for \verb{MCP} tool
  registration.
}
\examples{
# Define a generator function that returns tool definitions
my_tool_generator <- function(session) {
  # Define MCP tools using ellmer package

  tool_rnorm <- tool(
    function(n, mean = 0, sd = 1) {
      shiny::updateNumericInput(session, "rnorm", value = rnorm)
    },
    description = "Draw numbers from a random normal distribution",
    arguments = list(
      n = type_integer("The number of observations. Must be positive"),
      mean = type_number("The mean value of the distribution."),
      sd = type_number("The standard deviation of the distribution.")
    )
  )

  # or `list(tool_rnorm)`
  tool_rnorm
}

# Wrap the generator
wrapped_generator <- mcp_wrapper(my_tool_generator)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream-viz.R
\name{stream_file_id}
\alias{stream_file_id}
\title{Build the token-qualified stream file identifier}
\usage{
stream_file_id(id, session = shiny::getDefaultReactiveDomain(), token = NULL)
}
\arguments{
\item{id}{Character scalar. The base stream identifier (no path separators).}

\item{session}{Shiny session object. Defaults to the active reactive domain.}

\item{token}{Character scalar or \code{NULL}.  When \code{NULL} (default)
the session's own token is used.  Override with a parent session token
so that a standalone viewer (child session) can address the parent's
binary stream file.}
}
\value{
Character scalar: \code{"{token}_{id}"}.
}
\description{
Combines \code{session$token} with \code{id} to produce the string used
both by \code{\link{stream_path}} (as the filename stem) and by the browser
as the URL path component under \code{stream/}.
}
\seealso{
\code{\link{stream_path}}, \code{\link{stream_viz}},
  \code{\link{updateStreamViz}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{bslib_dependency}
\alias{bslib_dependency}
\title{Get Bootstrap 5 dependencies via \pkg{bslib}}
\usage{
bslib_dependency(...)
}
\arguments{
\item{...}{additional arguments passed to \code{bslib::bs_theme}}
}
\value{
An \code{\link[shiny]{tagList}} containing Bootstrap 5 dependencies
}
\description{
Returns Bootstrap 5 HTML dependencies provided by
\pkg{bslib}. Intended to be called from HTML templates so that
\code{headContent()} renders Bootstrap 5 CSS and JavaScript.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/notification.R
\name{notification}
\alias{notification}
\alias{show_notification}
\alias{clear_notifications}
\title{The 'Bootstrap' notification}
\usage{
show_notification(
  message,
  title = "Notification!",
  subtitle = "",
  type = c("default", "info", "warning", "success", "danger", "white", "dark"),
  close = TRUE,
  position = c("topRight", "topLeft", "bottomRight", "bottomLeft"),
  autohide = TRUE,
  fixed = TRUE,
  delay = 5000,
  icon = NULL,
  collapse = "",
  session = shiny::getDefaultReactiveDomain(),
  class = NULL,
  ...
)

clear_notifications(class = NULL, session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{message}{notification body content, can be 'HTML' tags}

\item{title, subtitle}{title and subtitle of the notification}

\item{type}{type of the notification; can be \code{"default"},
\code{"info"}, \code{"warning"}, \code{"success"}, \code{"danger"},
\code{"white"}, \code{"dark"}}

\item{close}{whether to allow users to close the notification}

\item{position}{where the notification should be; choices are
\code{"topRight"}, \code{"topLeft"}, \code{"bottomRight"},
\code{"bottomLeft"}}

\item{autohide}{whether to automatically hide the notification}

\item{fixed}{whether the position should be fixed}

\item{delay}{integer in millisecond to hide the notification if
\code{autohide=TRUE}}

\item{icon}{the icon of the title}

\item{collapse}{if \code{message} is a character vector, the collapse string}

\item{session}{shiny session domain}

\item{class}{the extra class of the notification, can be used for style
purposes, or by \code{clear_notifications} to close specific notification
types.}

\item{...}{other options; see
\url{https://adminlte.io/docs/3.1//javascript/toasts.html#options}}
}
\value{
Both functions should be used in shiny reactive contexts. The
messages will be sent to shiny 'JavaScript' interface and nothing will be
returned.
}
\description{
The 'Bootstrap' notification
}
\examples{
\dontrun{

# the examples must run in shiny reactive context

show_notification(
  message = "This validation process has finished. You are welcome to proceed.",
  autohide = FALSE,
  title = "Success!",
  subtitle = "type='success'",
  type = "success"
)

show_notification(
  message = "This notification has title and subtitle",
  autohide = FALSE,
  title = "Hi there!",
  subtitle = "Welcome!",
  icon = "kiwi-bird",
  class = "notification-auto"
)

# only clear notifications with class "notification-auto"
clear_notifications("notification-auto")

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/standalone-viewer.R
\name{server_standalone_viewer}
\alias{server_standalone_viewer}
\title{Server function for the standalone viewer module}
\usage{
server_standalone_viewer(input, output, session, ...)
}
\arguments{
\item{input, output, session}{Shiny module server arguments}

\item{...}{ignored}
}
\description{
Internal server function used by the \code{standalone_viewer}
  hidden module. Retrieves the render function from the parent module session
  and assigns it to the viewer's output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{get_construct_string}
\alias{get_construct_string}
\title{Get \code{R} expression used to generate the 'HTML' tags}
\usage{
get_construct_string(x)
}
\arguments{
\item{x}{'HTML' tags}
}
\value{
Quoted \code{R} expressions that can generate the 'HTML' tags
}
\description{
This function only works on the elements generated by this
package
}
\examples{

x <- info_box("Message")
get_construct_string(x)

}
\seealso{
\code{\link{format_text_r}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/card.R
\name{card}
\alias{card}
\alias{card2}
\alias{card2_open}
\alias{card2_close}
\alias{card2_toggle}
\alias{card_operate}
\title{Card-like 'HTML' element}
\usage{
card(
  title,
  ...,
  footer = NULL,
  tools = NULL,
  inputId = NULL,
  class = "",
  class_header = "",
  class_body = "",
  class_foot = "",
  style_header = NULL,
  style_body = NULL,
  start_collapsed = FALSE,
  resizable = FALSE,
  root_path = template_root()
)

card2(
  title,
  body_main,
  body_side = NULL,
  footer = NULL,
  tools = NULL,
  inputId = NULL,
  class = "",
  class_header = "",
  class_body = "min-height-400",
  class_foot = "",
  style_header = NULL,
  style_body = NULL,
  start_collapsed = FALSE,
  root_path = template_root()
)

card2_open(inputId, session = shiny::getDefaultReactiveDomain())

card2_close(inputId, session = shiny::getDefaultReactiveDomain())

card2_toggle(inputId, session = shiny::getDefaultReactiveDomain())

card_operate(
  inputId,
  title,
  method,
  session = shiny::getDefaultReactiveDomain()
)
}
\arguments{
\item{title}{the title of the card}

\item{...}{the body content of the card}

\item{footer}{the footer of the card; will be hidden if
\code{footer=NULL}}

\item{tools}{a list of tools or badges to be displayed at
top-right corner, generated by \code{\link{as_badge}} or
\code{\link{card_tool}}}

\item{inputId}{the id of the card}

\item{class}{the 'HTML' class of the entire card}

\item{class_header}{the the 'HTML' class of the card header}

\item{class_body}{the the 'HTML' class of the card body}

\item{class_foot}{the the 'HTML' class of the card footer}

\item{style_header}{'CSS' style of the header}

\item{style_body}{'CSS' style of the body}

\item{start_collapsed}{whether the card starts as collapsed}

\item{resizable}{whether the card body can be resized vertically;
notice that if true, then the default padding for body will be zero}

\item{root_path}{see \code{\link{template_root}}}

\item{body_main, body_side}{used by \code{card2}, the body
content of the front and back sides of the card}

\item{session}{shiny session domain}

\item{method}{action to expand, minimize, or remove the cards;
choices are \code{"collapse"}, \code{"expand"}, \code{"remove"},
\code{"toggle"}, \code{"maximize"}, \code{"minimize"},
and \code{"toggleMaximize"}}
}
\value{
'HTML' tags
}
\description{
Card-like 'HTML' element
}
\examples{
library(shiny)
library(shidashi)

# Used for example only
ns <- I
session <- MockShinySession$new()

# -------------- Basic usage -------------
card(
  title = "Badges", div(
    class = "padding-20",
    p(
      "Add badges to the top-right corder. ",
      "Use \"|\" to indicate the badge classes; ",
      "for example: \"badge-info\", \"badge-warning\"..."
    ),
    hr(), p(
      "Use `resizable = TRUE` to make card resizable."
    )
  ),
  tools = list(
    as_badge("New|badge-info"),
    as_badge("3|badge-warning")
  ),
  class_body = "height-300",
  resizable = TRUE
)

# ---------------- With tools -----------------
card(
  title = "Default Tools",
  plotOutput(
    ns("card_defaulttool_plot"),
    height = "100\%"
  ),
  tools = list(
    card_tool(
      widget = "link",
      href = "https://github.com/dipterix"
    ),
    card_tool(widget = "collapse"),
    card_tool(widget = "maximize")
  ),
  class_body = "height-300",
  resizable = TRUE
)

# -------------- Card2 example --------------
card2(
  title = "Card2 Example", body_main =
    plotOutput(
    outputId = ns("card2_plot"),
    height = "100\%"
  ),
  body_side = fluidRow(
    column(
      6L, textInput(
        ns("card2_plot_title"),
        "Plot title"
      )
    ),
    column(
      6L, sliderInput(
        ns("card2_plot_npts"),
        "# of points", min = 1, max = 100,
        value = 10, step = 1, round = TRUE
      )
    )
  ),
  tools = list(
    card_tool(widget = "link",
              href = "https://github.com/dipterix"),
    card_tool(widget = "collapse"),
    card_tool(widget = "maximize")
  ),
  class_body = "height-300"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/info-box.R
\name{info_box}
\alias{info_box}
\title{Generates 'HTML' info box}
\usage{
info_box(
  ...,
  icon = "envelope",
  class = "",
  class_icon = "bg-info",
  class_content = "",
  root_path = template_root()
)
}
\arguments{
\item{...}{box content}

\item{icon}{the box icon; default is \code{"envelope"}, can be hidden by
specifying \code{NULL}}

\item{class}{class of the box container}

\item{class_icon}{class of the icon}

\item{class_content}{class of the box body}

\item{root_path}{see \code{\link{template_root}}}
}
\value{
'HTML' tags
}
\description{
Generates 'HTML' info box
}
\examples{

library(shiny)
library(shidashi)

info_box("Message", icon = "cogs")

info_box(
  icon = "thumbs-up",
  span(class = "info-box-text", "Likes"),
  span(class = "info-box-number", "12,320"),
  class_icon = "bg-red"
)

info_box("No icons", icon = NULL)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat-helpers.R
\name{init_chat}
\alias{init_chat}
\title{Create an \pkg{ellmer} Chat object for the chat-bot}
\usage{
init_chat(
  system_prompt = getOption("shidashi.chat_system_prompt", NULL),
  provider = getOption("shidashi.chat_provider", "anthropic"),
  base_url = getOption("shidashi.chat_base_url", NULL)
)
}
\arguments{
\item{system_prompt}{character; the system prompt.  Defaults to
\code{getOption("shidashi.chat_system_prompt")}.}

\item{provider}{character; provider name or provider name with models.
Defaults to \code{getOption("shidashi.chat_provider", "anthropic")}.}

\item{base_url}{character or \code{NULL}; base URL for
API-compatible providers.}
}
\value{
An \code{ellmer::Chat} R6 object (tools not yet bound).
}
\description{
Factory function that creates an \code{ellmer::Chat} object based on
the configured provider.  Reads from
\code{options(shidashi.chat_provider)}, \code{shidashi.chat_model},
and \code{shidashi.chat_base_url}. These arguments are passed to
\code{\link[ellmer]{chat}}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rstudio.R
\name{use_template}
\alias{use_template}
\title{Download 'shidashi' templates from 'Github'}
\usage{
use_template(
  path,
  user = "dipterix",
  theme = "bslib",
  repo = "shidashi-templates",
  branch = "main",
  ...
)
}
\arguments{
\item{path}{the path to create 'shidashi' project}

\item{user}{'Github' user name}

\item{theme}{the theme to download}

\item{repo}{repository if the name is other than \code{'shidashi-templates'}}

\item{branch}{branch name if other than \code{'main'} or \code{'master'}}

\item{...}{ignored}
}
\value{
the target project path
}
\description{
Download 'shidashi' templates from 'Github'
}
\details{
To publish a 'shidashi' template, create a 'Github' repository
called \code{'shidashi-templates'}, or fork the
\href{https://github.com/dipterix/shidashi-templates}{built-in templates}.
The \code{theme} is the sub-folder of the template repository.

An easy way to use a template in your project is through the 'RStudio'
project widget. In the 'RStudio' navigation bar, go to "File" menu,
click on the "New Project..." button, select the "Create a new project"
option, and find the item that creates 'shidashi' templates. Use the
widget to set up template directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{show_ui_code}
\alias{show_ui_code}
\title{Used by demo project to show the generating code}
\usage{
show_ui_code(
  x,
  class = NULL,
  code_only = FALSE,
  as_card = FALSE,
  card_title = "",
  class_body = "bg-gray-70",
  width.cutoff = 80L,
  indent = 2,
  wrap = TRUE,
  args.newline = TRUE,
  blank = FALSE,
  copy_on_click = TRUE,
  ...
)
}
\arguments{
\item{x}{'HTML' tags generated by this package}

\item{class}{additional 'HTML' class}

\item{code_only}{whether to show code only}

\item{as_card}{whether to wrap results in \code{\link{card}}}

\item{card_title, class_body}{used by \code{\link{card}} if \code{as_card=TRUE}}

\item{width.cutoff, indent, wrap, args.newline, blank, copy_on_click, ...}{passed
to \code{\link{html_highlight_code}}}
}
\value{
'HTML' tags
}
\description{
Please write your own version. This function is designed for
demo-use only.
}
\seealso{
\code{html_highlight_code}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render.R
\name{render}
\alias{render}
\title{Render a 'shidashi' project}
\usage{
render(
  root_path = template_root(),
  ...,
  prelaunch = NULL,
  prelaunch_quoted = FALSE,
  launch_browser = TRUE,
  as_job = TRUE,
  test_mode = getOption("shiny.testmode", FALSE)
)
}
\arguments{
\item{root_path}{the project path, default is the demo folder from
\code{template_root()}}

\item{...}{additional parameters passed to \code{\link[shiny]{runApp}},
such as \code{host}, \code{port}}

\item{prelaunch}{expression to execute before launching the session; the
expression will execute in a brand new session}

\item{prelaunch_quoted}{whether the expression is quoted; default is false}

\item{launch_browser}{whether to launch browser; default is \code{TRUE}}

\item{as_job}{whether to run as 'RStudio' jobs; this options is only
available when 'RStudio' is available}

\item{test_mode}{whether to test the project; this options is helpful when
you want to debug the project without relaunching shiny applications}
}
\value{
This functions runs a 'shiny' application, and returns the job id
if 'RStudio' is available.
}
\description{
Render a 'shidashi' project
}
\examples{

template_root()

if(interactive()) {
  render()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ui-adminlte.R
\name{adminlte}
\alias{adminlte}
\alias{adminlte_ui}
\alias{adminlte_sidebar}
\title{Generates 'AdminLTE' theme-related 'HTML' tags}
\usage{
adminlte_ui(root_path = template_root())

adminlte_sidebar(
  root_path = template_root(),
  settings_file = "modules.yaml",
  shared_id = rand_string(26)
)
}
\arguments{
\item{root_path}{the root path of the website project; see
\code{\link{template_settings}}}

\item{settings_file}{the settings file containing the module information}

\item{shared_id}{a shared identification by session to synchronize the
inputs; assigned internally.}
}
\value{
'HTML' tags
}
\description{
These functions should be called in 'HTML' templates.
Please see vignettes for details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/menu-item.R
\name{as_badge}
\alias{as_badge}
\title{Generates badge icons}
\usage{
as_badge(badge = NULL)
}
\arguments{
\item{badge}{characters, \code{"shiny.tag"} object or \code{NULL}}
}
\value{
'HTML' tags
}
\description{
Usually used along with \code{\link{card}},
\code{\link{card2}}, and \code{\link{card_tabset}}. See \code{tools}
parameters in these functions accordingly.
}
\details{
When \code{badge} is \code{NULL} or empty, then \code{as_badge}
returns empty strings. When \code{badge} is a \code{"shiny.tag"} object,
then 'HTML' class \code{'right'} and \code{'badge'} will be appended.
When \code{badge} is a string, it should follow the syntax of
\code{"message|class"}. The text before \code{"|"} will be the badge
message, and the text after the \code{"|"} becomes the class string.
}
\examples{

# Basic usage
as_badge("New")

# Add class `bg-red` and `no-padding`
as_badge("New|bg-red no-padding")


}
